#!/bin/sh
# The next line is executed by /bin/sh, but not tcl \
exec tclsh $0 ${1+"$@"}

# Generate a ladder graph 3col problem with <int> levels, i.e., 2*<int> nodes.
#
# Written by Wolfgang Faber.

proc gen_level {l x} {

    # calculate the node indices:
    #
    #  level 1          1 o---o 2
    #                     |   |
    #                     .....
    #                     |   |
    #  level i-1  (i-1)*2 o---o (i-1)*2+1
    #                     |   |
    #  level i        i*2 o---o i*2+1
    #                     |   |
    #  level i+1  (i+1)*2 o---o (i+1)*2+1

    set left [expr $l * 2]
    set right [expr $l * 2 + 1]
    set upperleft [expr ($l - 1) * 2]
    set upperright [expr ($l - 1) * 2 + 1]
    set lowerleft [expr ($l + 1) * 2]
    set lowerright [expr ($l + 1) * 2 + 1]

    # for the left node, we generate three rules (one for each color)
    # each of them means "node i has color j, if the connected nodes
    # do not have color j and node i is not colored differently"
    foreach {c1 c2 c3} {1 2 3 2 3 1 3 1 2} {
        puts -nonewline "v${left}n${c1} :- "
        if {$l != 1} {
            puts -nonewline "not v${upperleft}n${c1}, "
        }
        puts -nonewline "not v${right}n${c1}, "
        if {$l != $x} {
            puts -nonewline "not v${lowerleft}n${c1}, "
        }
        puts "not v${left}n${c2}, not v${left}n${c3}."
    }

    # similar for the right node
    foreach {c1 c2 c3} {1 2 3 2 3 1 3 1 2} {
        puts -nonewline "v${right}n${c1} :- "
        if {$l != 1} {
            puts -nonewline "not v${upperright}n${c1}, "
        }
        puts -nonewline "not v${left}n${c1}, "
        if {$l != $x} {
            puts -nonewline "not v${lowerright}n${c1}, "
        }
        puts "not v${right}n${c2}, not v${right}n${c3}."
    }
}

proc gen {x} {

    if {$x < 1} {
        return
    }

    puts "h2 :- not h."
    puts "h2 :- not h2."

    for {set i 1} {$i <= $x} {incr i} {
        puts "h :- not v[expr 2*$i]n1, not v[expr 2*$i]n2, not v[expr 2*$i]n3."
        puts "h :- not v[expr 2*${i}+1]n1, not v[expr 2*${i}+1]n2, not v[expr 2*${i}+1]n3."
    }

    for {set i 1} {$i <= $x} {incr i} {
        gen_level $i $x
    }

# uncomment the following to obtain the unnecessary rules of Niemela's
# original examples
#    puts "s :- h2."
#    for {set i 1} {$i <= $x} {incr i} {
#        foreach j {1 2 3} {
#            puts "s :- v[expr ${i} * 2]n${j}."
#        }
#        foreach j {1 2 3} {
#            puts "s :- v[expr ${i} * 2 + 1]n${j}."
#        }
#    }
}

if {$argc > 1 || [scan $argv "%d%s" dummy1 dummy2] != 1} then {
    puts "Usage: ladder <int>"
} else {
    gen $argv
}

