%{
/*
QBF2DLV - Translator from Cadoli's QBF format to DLP. Since Cadoli solves
QBF instances for   forall X exists Y CNF
and this translation is designed for  exists X forall Y DNF
the QBF is valid iff the result of the DLP is BRAVELY FALSE !

GenQbf and QbfEvaluate can be downloaded from
http://www.dis.uniroma1.it/~cadoli/projects/QBF/

Proposed Makefile:
***
qbf2dlv:
        flex -oqbf2dlv.C qbf2dlv.l
        g++ -o qbf2dlv qbf2dlv.C

example: qbf2dlv
        rm -f example1.qbf example1.dl
        -GenQbf 2 5 5 500 8 F2 > example1.qbf
        qbf2dlv < example1.qbf > example1.dl
        -time QbfEvaluate example1.qbf
        time ../dl -silent -FB example1.dl

***
Attention: Both Cadoli's GenQbf and QbfEvaluate return error code 1 after
successful execution, therefore the '-' in the makefile.
*/


#include<iostream.h>
#include<vector.h>
#include<stdlib.h>
#include<assert.h>

#define LEX_NUMBER 1
#define LEX_ENDOFLINE 2
%}

%%

P\n2\n		{ /* that's a Cadoli 2QBF header */ } 
-?[0-9]+	{ return LEX_NUMBER; }
[\t ]+		{}
[\n]+		{ return LEX_ENDOFLINE; }
.		{ cout << "Error: Wrong file format!" << endl; exit(1); }

%%

int yywrap(void)
{
   return 1;
}


// end of lex-based "parser"


// this program complies to the initial transformation in the paper
// Eiter et al. Th. Comp. Sc. 189 (1997) 129-177


typedef vector<int> QBF_CONJUNCTION;


int main(int argc, char **argv)
{
   if((argc > 2) || ((argc == 2) && (strcmp(argv[1], "-smodels") != 0)))
   {
      cerr << "USAGE: " << argv[0] << " [-smodels]" << endl;
      exit(1);
   }

   // true: dlv false: smodels
   bool dlv_mode = true;

   if((argc == 2) && (strcmp(argv[1], "-smodels") == 0))
   {
      dlv_mode = false;
   }

   // 2QBF in DNF
   vector<QBF_CONJUNCTION> qbf;

   // here we store which variable indexes appeared in the file
   vector<bool> var_idx;

   // this parses the first two lines of the header (P\n2\n) as well as the
   // following integer
   int retval = yylex();

   assert(retval == LEX_NUMBER);
   int max_var_cnt = atoi(yytext);
   retval = yylex();
   assert(retval == LEX_ENDOFLINE);

   retval = yylex();
   assert(retval == LEX_NUMBER);
   int num_clauses = atoi(yytext);
   retval = yylex();
   assert(retval == LEX_ENDOFLINE);

   retval = yylex();
   assert(retval == LEX_NUMBER);
   int num_atoms = atoi(yytext);
   retval = yylex();
   assert(retval == LEX_ENDOFLINE);

   // now the header of a Cadoli-style QBF file is read. The clauses follow.

   // this code ignores the number of atoms in the clauses as specified in the
   // header. The clauses may have differing lengths.
   for(unsigned i = 0; i < num_clauses; i++)
   {
      qbf.push_back(QBF_CONJUNCTION());
      int j;
      while((j = yylex()) == LEX_NUMBER)
      {
         while(var_idx.size() < abs(atoi(yytext)))
         {
            var_idx.push_back(false);
         }
         var_idx[abs(atoi(yytext))-1] = true;

         qbf.back().push_back(atoi(yytext));
      }
   }
   // the input file is now completely read.


   // first come a number of Xs which are existentially quantified
   unsigned num_X = 0;
   // next come a number of Ys which are all-quantified
   unsigned num_Y = 0;


   // here we find out how many X and Y variables there are. This is not
   // explicitly stated in a Cadoli-style file.
   for(unsigned i = 0; i < var_idx.size(); i++)
   {
      if(var_idx[i])
      {
         if(i < max_var_cnt)
            {
            if(i+1 > num_X) num_X = i+1;
            //num_X++;
            }
         else
            {
            if(i+1-max_var_cnt > num_Y) num_Y = i+1-max_var_cnt;
            //num_Y++;
            }
      }
   }
   cout << "% |X| = " << num_X << "   |Y| = " << num_Y << endl;


   //assert(max_var_cnt == ((num_X > num_Y) ? num_X : num_Y));

   char not_prefix = 'n';

   for(unsigned i = 1; i <= num_X; i++)
   {
      if(dlv_mode)
         cout << "x" << i << " v " << not_prefix << "x" << i << "." << endl;
      else // smodels_mode
         cout << "{x" << i << ", " << not_prefix << "x" << i << "}." << endl;
   }

   for(unsigned i = max_var_cnt + 1; i <= max_var_cnt + num_Y; i++)
   {
      if(dlv_mode)
         cout << "y" << i << " v " << not_prefix << "y" << i << "." << endl;
      else // smodels_mode
         cout << "{y" << i << ", " << not_prefix << "y" << i << "}." << endl;
// these rules are superfluous and make dlv choke
//      cout << "w :- y" << i << ", " << not_prefix << "y" << i << "." << endl;
      cout << "y" << i << " :- w." << endl;
      cout << not_prefix << "y" << i << " :- w." << endl;
   }

   for(vector<QBF_CONJUNCTION>::const_iterator j = qbf.begin();
       j != qbf.end(); j++)
   {
      cout << "w :- ";

      for(QBF_CONJUNCTION::const_iterator k = j->begin();
         k != j->end(); k++)
      {
         if(k != j->begin())
            cout << ", ";

         char v;

         if(abs(*k) <= num_X)
            v = 'x';
         else
            v = 'y';

         if(*k > 0)
            cout << v << abs(*k);
         else
            cout << not_prefix << v << abs(*k);
      }
      cout << "." << endl;
   }

   if(dlv_mode)
   {
      // cout << ":- not w." << endl;
      // this requires that brave reasoning is turned on (dl -FB)
      cout << "w ?" << endl;
   }
   else // smodels_mode
      cout << "w :- not w." << endl;

   return 0;
}


