/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/*
 * File:   Aggregate.h
 * Author: Carmine Dodaro
 *
 * Created on 5 May 2011, 11.29
 */

#ifndef _AGGREGATE_H
#define	_AGGREGATE_H

#include <cassert>
#include <cstring>
#include "Constants.h"

class Literal;

class Aggregate
{
    friend ostream &operator<<( ostream &, const Aggregate & );
public:

    Aggregate( AtomId atom, unsigned int minGuard, unsigned int maxGuard, unsigned int size, AggregateType type );
    
    inline ~Aggregate();

    inline AtomId getId() const;
    
    inline unsigned int getMaxGuard() const;    

    inline unsigned int getMinGuard() const;
    
    inline void atomSetFalse( unsigned int position );

    inline void atomSetTrue( unsigned int position );    
    
//    inline Atom* getAtomPointer( unsigned int position );
    
    inline Literal& getLiteral( unsigned int position );    
    
    inline unsigned int getSize() const;

    inline unsigned int getLiteralValue( unsigned int position ) const;

    inline unsigned int getCurrentValue() const;

    inline unsigned int& getCurrentValueAddress();
    
    inline unsigned int getMaxPossibleValue() const;

    inline unsigned int& getMaxPossibleValueAddress();
    
    inline void addLiteralInTheAggregateSet( Literal& literal, int value );
    
    inline Literal& getLeftWatchLiteral();
    
    inline Literal& getRightWatchLiteral();
    
    inline unsigned int getLeftWatch();
    
    inline unsigned int getRightWatch();
    
    inline unsigned int& getLeftWatchAddress();
    
    inline unsigned int& getRightWatchAddress();
    
    inline unsigned int getLeftWatchValue();
    
    inline unsigned int getRightWatchValue();
    
    inline void setMinGuard( unsigned int minGuard );
    
    inline void updateLeftWatch();
    
    inline void updateRightWatch();
    
private:

    inline Aggregate( const Aggregate & aggregate );        

    AtomId id;

    unsigned int maxGuard;
    
    unsigned int minGuard;

    AggregateType type;
    
    //An array of atoms appearing in the aggregate.
    vector< Literal > literalsInTheAggregate;
    
    //An array of atoms values appearing in the aggregate. For example, the i-th atom, atoms[ i ], has a value atomsValues[ i ].
    vector< int > literalsValues;

    //The current sum/count of the aggregate.
    unsigned int currentValue;

    //The max sum/count of the aggregate.
    unsigned int maxPossibleValue;
    
    unsigned int leftWatch;
    
    unsigned int rightWatch;
};

Aggregate::Aggregate(
    const Aggregate& )
{
    assert( 0 );
}

Aggregate::~Aggregate()
{
}

AtomId
Aggregate::getId() const
{
    return id;
}

unsigned int
Aggregate::getMaxGuard() const
{
    return maxGuard;
}

unsigned int
Aggregate::getMinGuard() const
{
    return minGuard;
}

void
Aggregate::atomSetFalse(
    unsigned int position )
{
    assert( position < getSize() );
    maxPossibleValue -= literalsValues[ position ];
}

void
Aggregate::atomSetTrue(
    unsigned int position )
{
    assert( position < getSize() );
    currentValue += literalsValues[ position ];
}

//Atom*
//Aggregate::getAtomPointer( 
//    unsigned int position )
//{
//    assert( position < getSize() );
//    return literalsInTheAggregate[ position ];
//}

Literal&
Aggregate::getLiteral( 
    unsigned int position )
{
    assert( position < getSize() );
    return literalsInTheAggregate[ position ];
}

unsigned int
Aggregate::getSize() const
{
    return literalsInTheAggregate.size();
}

unsigned int
Aggregate::getLiteralValue( 
    unsigned int position ) const
{
    assert( position < getSize() );
    return literalsValues[ position ];
}

unsigned int
Aggregate::getCurrentValue() const
{
    return currentValue;
}

unsigned int&
Aggregate::getCurrentValueAddress()
{
    return currentValue;
}

unsigned int
Aggregate::getMaxPossibleValue() const
{
    return maxPossibleValue;
}

unsigned int&
Aggregate::getMaxPossibleValueAddress()
{
    return maxPossibleValue;
}

void
Aggregate::addLiteralInTheAggregateSet(
    Literal& literal, 
    int value )
{
    literalsInTheAggregate.push_back( literal );
    literalsValues.push_back( value );
    
    maxPossibleValue += value;
}

Literal&
Aggregate::getLeftWatchLiteral()
{
    assert( leftWatch < literalsInTheAggregate.size() );
    return literalsInTheAggregate[ leftWatch ];
}
    
Literal&
Aggregate::getRightWatchLiteral()
{
    assert( rightWatch < literalsInTheAggregate.size() );
    return literalsInTheAggregate[ rightWatch ];
}
    
unsigned int 
Aggregate::getLeftWatch()
{
    return leftWatch;
}
    
unsigned int 
Aggregate::getRightWatch()
{
    return rightWatch;
}
    
unsigned int 
Aggregate::getLeftWatchValue()
{
    assert( leftWatch < literalsValues.size() );
    return literalsValues[ leftWatch ];
}
    
unsigned int 
Aggregate::getRightWatchValue()
{
    assert( rightWatch < literalsValues.size() );
    return literalsValues[ rightWatch ];
}

unsigned int& 
Aggregate::getLeftWatchAddress()
{
    return leftWatch;
}
    
unsigned int& 
Aggregate::getRightWatchAddress()
{
    return rightWatch;
}

void 
Aggregate::updateLeftWatch()
{
    assert( leftWatch < rightWatch );
    ++leftWatch;
}
    
void 
Aggregate::updateRightWatch()
{
    assert( rightWatch > leftWatch && rightWatch != MAXVALUE );
    --rightWatch;
}

void 
Aggregate::setMinGuard( 
    unsigned int mG )
{
    minGuard = mG;
}

#endif /* _AGGREGATE_H */
