/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   Atom.h
 * Author: Carmine Dodaro
 *
 * Created on 5 October 2010, 17.53
 */
#ifndef _ATOM_H
#define	_ATOM_H

#include <cassert>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include "Aggregate.h"
#include "Constants.h"
#include "options.h"
#include "unordered_map.h"
#include "RuleContainer.h"
#include "Rule.h"
#include "WatchedList.h"
#include "ProgramBuilder.h"

using namespace std;

class Atom;
class ModelGenerator;
class WaspRule;
class ChoiceRule;

typedef void ( ModelGenerator::*functionPointer )( Atom& );

/**
 * This class stores all information about an atom.
 */
class Atom
{
    friend ostream &operator<<( ostream &, const Atom & );
    friend class ModelGenerator;    
    
public:    

    unsigned int countFoundingRules;

    /**
     * This constructor can be used to create an atom with its name.
     * @param atomName The atom name.
     */
    Atom( AtomId id, string atomName );
    
    inline ~Atom();
    
    inline void addInvolvedRules( const InvolvedPointer id, const InvolvedType type );

    /**
     * It adds a waspRule where atom appears in the body with the negative polarity.
     *
     * @param rule The waspRule.
     */
    inline void addNegativeBodyWaspRule( WaspRule* rule );
    
    /**
     * It adds a waspRule where atom appears in the head.
     *
     * @param rule The waspRule.
     */
    inline void addNegativeHeadWaspRule( WaspRule* rule );

    /**
     * It adds a waspRule where atom appears in the body with the positive polarity.
     *
     * @param rule The waspRule.
     */
    inline void addPositiveWaspRule( WaspRule* rule );
        
   /**
    * It adds a choice rule where atom appears in the body with the negative polarity.
    *
    * @param rule The choice rule.
    */
    inline void addNegativeBodyChoiceRule( ChoiceRule* rule );

   /**
    * It adds a choice rule where atom appears in the body with the positive polarity.
    *
    * @param rule The choice rule.
    */
    inline void addPositiveBodyChoiceRule( ChoiceRule* rule );
    
   /**
    * It adds a choice rule where atom appears in the head.
    *
    * @param rule The choice rule.
    */
    inline void addHeadChoiceRule( ChoiceRule* rule );
    
    
    /**
     * Returns if the atom appears only in one head.
     * 
     * @return true if the atom appears in one head, false otherwise.
     */    
    inline bool appearsInOneHead() const;

    /**
     * Returns if there is at least one rule can support it.
     *
     * @return true if atom can be supported, false otherwise.
     */
    inline bool canBeSupported() const;

    /**
     * It returns a string, which represent the atom name.
     *
     * @return atom name.
     */
    inline string getAtomName() const;

    /**
     * Returns the component where atom appears
     *
     * @return the component Id.
     */
    inline ComponentId getComponent() const;    

    /**
     * Returns id of the atom.
     *
     * @return the atom Id.
     */
    inline AtomId getId() const;
    
    inline InvolvedPointer getIdOfDerivation() const;
    
    inline const RuleContainer< WaspRule* >& getNegativeBodyWaspRules() const;
    
    inline const RuleContainer< WaspRule* >& getNegativeHeadWaspRules() const;
    
    inline const RuleContainer< WaspRule* >& getPositiveWaspRules() const;    
    
    inline virtual unsigned int getNegativeOccurrencesAggregates() const;    
    
    inline virtual unsigned int getPositiveOccurrencesAggregates() const;
    
    inline const RuleContainer< ChoiceRule* >& getNegativeBodyChoiceRules() const; 

    inline const RuleContainer< ChoiceRule* >& getPositiveBodyChoiceRules() const;

    inline const RuleContainer< ChoiceRule* >& getHeadChoiceRules() const;
    
    virtual functionPointer getPropagationFunction();
    
    /**
     * It returns the level where atom is derived true or false.
     *
     * @return the level where atom truthvalue is derived.
     */
    inline const Level getRecursionLevel() const;

    inline const RuleContainer< WaspRule* >& getRuleAtomInHead() const;    

    inline InvolvedType getTypeOfDerivation() const;    

    inline virtual bool isAggregate() const;
    
    /**
     * Returns if this atom is a choice.
     *
     * @return true if this atom is a choice, false otherwise.
     */
    inline bool isChoice() const;

    bool isFalse() const;   
    
    bool isFalsePropagated() const;    
    
    /**
     * Returns if this atom is hidden.
     *
     * @return true if this atom is hidden, false otherwise.
     */
    inline bool isHidden() const;

    /**
     * Returns if this atom is yet supported.
     *
     * @return true if atom is yet supported, false otherwise.
     */
    inline bool isSupported() const;
    
    bool isTrue() const;
    
    bool isTruePropagated() const;        
    
    inline bool isTrueSupported() const;
    
    bool isUndefined() const;

    /**
     * Returns if top watch and bottom watch are on the same position.
     *
     * @return true if top watch and bottom watch are on the same position, false otherwise.
     */
    inline bool onlyOneSupportingRuleLeft() const;
    
    /**
     * It sets the atom name.
     *
     * @param atomName The string associated to atom.
     */
    inline void setAtomName( string atomName );

    /**
     * It sets that the atom is a choice.
     *
     */
    inline void setChoice();

    /**
     * It sets the component where the atom appears.
     *
     * @param component The component id.
     */
    inline void setComponent ( ComponentId component );

    /**
     * It sets the backjumping level where atom is derived true or false.
     *
     * @param level The level.
     */
    inline void setRecursionLevel ( Level level );

     /**
     * Set this atom supported.
     *
     */
    inline void setSupported();
    
    inline void setTrueSupported();
    
    /**
     * Support counter is decreased.
     *
     */
    inline void updateSupportCounter();

    inline bool operator==( const Atom& ) const;
    inline bool operator!=( const Atom& ) const;
    
    inline unsigned int getRulePosition( Rule* );    
    
    inline unsigned int& getChoiceAddress();
    
    inline const RuleContainer< WaspRule* >& getBinaryNegativeWaspRules() const;
    inline const RuleContainer< WaspRule* >& getBinaryPositiveWaspRules() const;
    
    inline void addBinaryWasprule( WaspRule* waspRule, bool positive );
    
    inline WatchedList< WaspRule* >::iterator addWatchedPositiveWaspRule( WaspRule* );
    
    inline WatchedList< WaspRule* >::iterator addWatchedNegativeWaspRule( WaspRule* );
    
    inline WatchedList< WaspRule* >::iterator addWatchedPositiveConstraint( WaspRule* );
    
    inline WatchedList< WaspRule* >::iterator addWatchedNegativeConstraint( WaspRule* );
    
    inline WatchedList< WaspRule* >& getWatchedPositiveWaspRules();
    
    inline WatchedList< WaspRule* >& getWatchedNegativeWaspRules();
    
    inline WatchedList< WaspRule* >& getWatchedPositiveConstraints();
    
    inline WatchedList< WaspRule* >& getWatchedNegativeConstraints();
    
    inline void eraseWatchedPositiveConstraint( WatchedList< WaspRule* >::iterator );
    
    inline void eraseWatchedNegativeConstraint( WatchedList< WaspRule* >::iterator );
    
    inline bool isFounded() const;
    
    inline void setFounded();
    
    inline void setFounded( bool );
    
    inline unsigned int& getFoundedAddress();
    
    inline Rule* getSourcePointer();
    
    inline void setSourcePointer( Rule* sourcePointer );
    
private:

    /**
     * Copy constructor is disabled.
     */
    Atom( const Atom & atom );
    
    /**
     * Atom name is stored in this variable.
     */
    string atomName;

    /**
     * If atoms is a choice this variable is equal to 1.
     */
    unsigned int choice;

    /**
     * Component id where this atom appears.
     */
    ComponentId component;
    
    AtomId id;

    ReasonPointer idOfDerivation;
    
    bool isAtomTrueSupported;
    
    /**
     * The recursion level where atom is derived true or false on backjumping.
     */
    Level recursionLevel;

    RuleContainer< WaspRule* > negativeBodyWaspRules;
    
    RuleContainer< WaspRule* > negativeHeadWaspRules;    
  
    RuleContainer< WaspRule* > positiveWaspRules;
        
    WatchedList< WaspRule* > watchedPositiveWaspRules;
    
    WatchedList< WaspRule* > watchedNegativeWaspRules;
    
    WatchedList< WaspRule* > watchedPositiveConstraints;
    
    WatchedList< WaspRule* > watchedNegativeConstraints;
    
    RuleContainer< ChoiceRule* > negativeBodyChoiceRules;
    
    RuleContainer< ChoiceRule* > positiveBodyChoiceRules;
    
    RuleContainer< ChoiceRule* > headChoiceRules;
    
    unsigned int supportCounter;
    
    unsigned int truthValue;    
    
    unsigned int typeOfDerivation;
    
    Rule* sourcePointer;
    
    unsigned int founded;
    
    RuleContainer< WaspRule* > binaryPositiveWaspRules;
    RuleContainer< WaspRule* > binaryNegativeWaspRules;
    
};

Atom::~Atom()
{
}

void
Atom::addInvolvedRules(
    const InvolvedPointer id,
    const InvolvedType type )
{
    this->idOfDerivation = id;
    this->typeOfDerivation = type;
}

void 
Atom::addNegativeBodyWaspRule( 
    WaspRule* rule )
{
    negativeBodyWaspRules.push_back( rule );
}

void
Atom::addNegativeHeadWaspRule(
    WaspRule* rule )
{
    supportCounter++;
    negativeHeadWaspRules.push_back( rule );
}

void
Atom::addPositiveWaspRule(
    WaspRule* rule )
{
    positiveWaspRules.push_back( rule );
}

bool
Atom::appearsInOneHead() const
{
    return ( negativeHeadWaspRules.size() == 1 );
}

bool
Atom::canBeSupported() const
{
    return( supportCounter > 0 );
}

string
Atom::getAtomName() const
{
    if( isHidden() )
    {
        stringstream s;        
        if( isAggregate() )
            s << "#aggr" << id;
        else
            s << "#hidden" << id;
        return s.str();
    }
    else
        return atomName;
}

ComponentId
Atom::getComponent() const
{
    return component;
}

AtomId
Atom::getId() const
{
    return id;
}

InvolvedPointer
Atom::getIdOfDerivation() const
{
    return idOfDerivation;
}

const RuleContainer< WaspRule* >&
Atom::getNegativeBodyWaspRules() const
{
    return negativeBodyWaspRules;
}

const RuleContainer< WaspRule* >&
Atom::getNegativeHeadWaspRules() const
{
    return negativeHeadWaspRules;
}

unsigned int 
Atom::getNegativeOccurrencesAggregates() const
{
    return 0;
}

unsigned int 
Atom::getPositiveOccurrencesAggregates() const
{
    return 0;
}

const RuleContainer< WaspRule* >&
Atom::getPositiveWaspRules() const
{
    return positiveWaspRules;
}

const Level
Atom::getRecursionLevel() const
{
    return recursionLevel;
}

InvolvedType
Atom::getTypeOfDerivation() const
{
    return typeOfDerivation;
}

bool 
Atom::isAggregate() const
{
    return false;
}

bool
Atom::isChoice() const
{
    return choice == 1;
}

bool
Atom::isHidden() const
{
    return atomName == "";
}

bool
Atom::isSupported() const
{
    return( supportCounter == MAXVALUE );
}

bool
Atom::isTrueSupported() const
{
    return isAtomTrueSupported;
}

bool
Atom::onlyOneSupportingRuleLeft() const
{
    return( supportCounter == 1 );
}

void
Atom::setAtomName(
    string atomName )
{
    this->atomName = atomName;
}

void
Atom::setChoice()
{
    this->choice = 1;
}

void
Atom::setComponent(
    ComponentId component )
{
    this->component = component;
}

void
Atom::setRecursionLevel(
    Level level )
{
    this->recursionLevel = level;
}

void
Atom::setSupported()
{
    supportCounter = MAXVALUE;
}

void
Atom::setTrueSupported()
{
    isAtomTrueSupported = true;
}

void
Atom::updateSupportCounter()
{
    assert( supportCounter <= ( negativeHeadWaspRules.size() + headChoiceRules.size() ) );
    supportCounter--;
}

bool
Atom::operator==(
    const Atom& atom ) const
{
    return( id == atom.id );
}

bool
Atom::operator!=(
    const Atom& atom ) const
{
    return( id != atom.id );
}

unsigned int 
Atom::getRulePosition( 
    Rule* r )
{   
    if( isTrueSupported() )
        return 0;
    
    assert( r != NULL );
    if( r->isWaspRule() )
    {
        WaspRule* w = ( WaspRule* ) r;
        return negativeHeadWaspRules.getPosition( w );
    }
    else
    {
        ChoiceRule* c = ( ChoiceRule* ) r;
        return negativeHeadWaspRules.size() + headChoiceRules.getPosition( c );
    }
}

unsigned int& 
Atom::getChoiceAddress()
{
    return choice;
}

const RuleContainer< WaspRule* >& 
Atom::getBinaryNegativeWaspRules() const
{
    return binaryNegativeWaspRules;
}

const RuleContainer< WaspRule* >& 
Atom::getBinaryPositiveWaspRules() const
{
    return binaryPositiveWaspRules;
}

void 
Atom::addBinaryWasprule( 
    WaspRule* waspRule, 
    bool positive )
{
    if( positive )
        binaryPositiveWaspRules.push_back( waspRule );
    else
        binaryNegativeWaspRules.push_back( waspRule );
}

void 
Atom::addNegativeBodyChoiceRule( 
    ChoiceRule* rule )
{
    negativeBodyChoiceRules.push_back( rule );
}

void 
Atom::addPositiveBodyChoiceRule( 
    ChoiceRule* rule )
{
    positiveBodyChoiceRules.push_back( rule );
}
    
void 
Atom::addHeadChoiceRule( 
    ChoiceRule* rule )
{
    supportCounter++;
    headChoiceRules.push_back( rule );
}

const RuleContainer< ChoiceRule* >& 
Atom::getNegativeBodyChoiceRules() const
{
    return negativeBodyChoiceRules;
}

const RuleContainer< ChoiceRule* >& 
Atom::getPositiveBodyChoiceRules() const
{
    return positiveBodyChoiceRules;
}

const RuleContainer< ChoiceRule* >& 
Atom::getHeadChoiceRules() const
{
    return headChoiceRules;
}

WatchedList< WaspRule* >::iterator 
Atom::addWatchedPositiveWaspRule( 
    WaspRule* w )
{
    return watchedPositiveWaspRules.add( w );
}
    
WatchedList< WaspRule* >::iterator 
Atom::addWatchedNegativeWaspRule( 
    WaspRule* w )
{
    return watchedNegativeWaspRules.add( w );
}

WatchedList< WaspRule* >::iterator 
Atom::addWatchedPositiveConstraint( 
    WaspRule* w )
{
    return watchedPositiveConstraints.add( w );
}

WatchedList< WaspRule* >::iterator 
Atom::addWatchedNegativeConstraint( 
    WaspRule* w )
{
    return watchedNegativeConstraints.add( w );
}

WatchedList< WaspRule* >& 
Atom::getWatchedPositiveWaspRules()
{
    return watchedPositiveWaspRules;
}
    
WatchedList< WaspRule* >& 
Atom::getWatchedNegativeWaspRules()
{
    return watchedNegativeWaspRules;
}

WatchedList< WaspRule* >& 
Atom::getWatchedPositiveConstraints()
{
    return watchedPositiveConstraints;
}

WatchedList< WaspRule* >&
Atom::getWatchedNegativeConstraints()
{
    return watchedNegativeConstraints;
}

void 
Atom::eraseWatchedPositiveConstraint( 
    WatchedList< WaspRule* >::iterator it )
{
    watchedPositiveConstraints.erase( it );
}
    
void 
Atom::eraseWatchedNegativeConstraint( 
    WatchedList< WaspRule* >::iterator it )
{
    watchedNegativeConstraints.erase( it );
}

unsigned int&
Atom::getFoundedAddress()
{
    return founded;
}

bool
Atom::isFounded() const
{
    return founded;
}
    
void 
Atom::setFounded()
{
    founded = 1;
}

void 
Atom::setFounded( 
    bool b )
{
    assert( !b );
    founded = 0;
}

Rule* 
Atom::getSourcePointer()
{
    return sourcePointer;
}
    
void
Atom::setSourcePointer( 
    Rule* source )
{
    sourcePointer = source;
}

#endif	/* _ATOM_H */
