/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   ChoiceRule.h
 * Author: Carmine Dodaro
 *
 * Created on 1 February 2012, 12.27
 */
#ifndef _CHOICERULE_H
#define	_CHOICERULE_H

#include <iostream>
#include <cassert>
#include "Constants.h"
#include "Literal.h"
#include "Rule.h"
using namespace std;

class Atom;

class ChoiceRule : public Rule
{
    friend class ModelGenerator;
    friend ostream& operator<<( ostream& o, ChoiceRule& c )
    {   
        unsigned int headAtomsSize = c.getNumberOfAtomsInTheHead();
        unsigned int bodyLiteralsSize = c.getNumberOfLiteralsInTheBody();
        
        assert( headAtomsSize > 0 );
        Literal l = c.bodyLiterals[ 0 ];
        l.setPositive();
        o << "{ " << l;

        for( unsigned int i = 1; i < headAtomsSize; i++ )
        {
            Literal l = c.bodyLiterals[ i ];
            l.setPositive();
            o << " ; " << l;
        }
        
        o << "}";
        
        if( bodyLiteralsSize > 0 )
        {
            o << " :- ";
            
            o << c.getBodyLiteral( 0 );
            
            for( unsigned int i = 1; i < bodyLiteralsSize; i++ )
            {
                o << ", " << c.getBodyLiteral( i );
            }
        }
        o << ".";
        
        return o;
    }
    
    public:
        inline ChoiceRule( unsigned int id, unsigned int bodyLiteralsSize, unsigned int headAtoms );
        inline ~ChoiceRule();
        
        inline void addHeadAtom( Atom* atom );
        inline void addBodyLiteral( Literal& literal );
        
        inline unsigned int getNumberOfAtomsInTheHead() const;
        inline unsigned int getNumberOfLiteralsInTheBody() const;
        
        inline unsigned int getId() const;
        inline void setId( RuleId id );
        
        inline unsigned int getIdFirstFalseLiteral() const;
        inline void setIdFirstFalseLiteral( unsigned int i );
        
        inline bool isBodyUndefined() const;
        inline bool isBodyTrue() const;
        inline bool isBodyFalse() const;
        inline bool canSupport() const;
        
        inline void updateCounter();
        
        inline Atom* getHeadAtom( unsigned int position );
        inline Literal& getBodyLiteral( unsigned int position );
        
        inline bool isActiveForAtom() const;
        
        virtual bool isRuleActiveForAtom( AtomId atomId ) const;
        virtual bool isWaspRule() const;
        virtual void print();
        virtual unsigned int getSize() const;
        virtual Literal& getRuleBodyLiteral( unsigned int position );
        
    private:
        //The id of the choice rule.
        unsigned int idChoiceRule;
        
        //A vector containing the Atoms in the choice rule.
        vector< Atom* > headAtoms;
        
        //A vector containing literals in the body of the choice rule.
        vector< Literal > bodyLiterals;
        
       /*
        * A counter is used to count the number of undefined literals. 
        * It is updating when a literal becomes true.
        */
        unsigned int counter;
        
        //Id of the first false literal.
        unsigned int idFirstFalseLiteral;
        
        //Copy constructor. Disabled by default.
        inline ChoiceRule( const ChoiceRule& );        
};

ChoiceRule::ChoiceRule(
    unsigned int id,
    unsigned int bodyLiteralSize, 
    unsigned int choiceRuleAtomsSize ) : 
    counter( bodyLiteralSize ), idChoiceRule( id ), idFirstFalseLiteral( 0 )
{
    bodyLiterals.reserve( bodyLiteralSize );
    headAtoms.reserve( choiceRuleAtomsSize );
}

ChoiceRule::ChoiceRule(
    const ChoiceRule& )
{
    assert( 0 );
}

ChoiceRule::~ChoiceRule()
{
}

void
ChoiceRule::addBodyLiteral(
    Literal& literal )
{
    bodyLiterals.push_back( literal );
}

void
ChoiceRule::addHeadAtom(
    Atom* atom )
{
    assert( atom != NULL );
    headAtoms.push_back( atom );
}

unsigned int 
ChoiceRule::getNumberOfAtomsInTheHead() const
{
    return headAtoms.size();
}

unsigned int
ChoiceRule::getNumberOfLiteralsInTheBody() const
{
    return bodyLiterals.size();
}

unsigned int
ChoiceRule::getIdFirstFalseLiteral() const
{
    return idFirstFalseLiteral;
}

void
ChoiceRule::setIdFirstFalseLiteral(
    unsigned int i )
{
    idFirstFalseLiteral = i;
}

Atom*
ChoiceRule::getHeadAtom(
    unsigned int i )
{
    assert( i < headAtoms.size() );
    return headAtoms[ i ];
}

Literal&
ChoiceRule::getBodyLiteral(
    unsigned int i )
{
    assert( i < getNumberOfLiteralsInTheBody() );
    return bodyLiterals[ i ];
}

RuleId
ChoiceRule::getId() const
{
    return idChoiceRule;
}

bool 
ChoiceRule::isBodyUndefined() const
{
    return ( idFirstFalseLiteral == 0 && counter != 0 );
}

bool
ChoiceRule::isBodyFalse() const
{
    return idFirstFalseLiteral != 0;
}

bool 
ChoiceRule::canSupport() const
{
    return !isBodyFalse();
}

bool 
ChoiceRule::isBodyTrue() const
{
    return counter == 0;
}

void
ChoiceRule::setId(
    RuleId id )
{
    idChoiceRule = id;
}

void
ChoiceRule::updateCounter()
{
    assert( counter > 0 );
    counter--;
}

bool
ChoiceRule::isActiveForAtom() const
{
    return !isBodyFalse();
}

#endif	/* _CHOICERULE_H */