/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   Component.h
 * Author: Carmine Dodaro
 *
 * Created on 25 October 2010, 13.50
 */

#ifndef _COMPONENT_H
#define	_COMPONENT_H

#include <vector>
#include <cassert>
#include <iostream>
using namespace std;

#include "Constants.h"
#include "stack.h"
#include "unordered_map.h"

class Component {

public:
    
    inline Component();    

    inline ~Component();

    /**
     * This function add an atom in the component.
     * @param atomId the atom is getting to be added.
     */
    inline void addAtomInComponent( AtomId atomId );

    inline AtomId getAtom( unsigned int i );
    
    /**
     * This function return the vector of atoms that are in this component.
     * @return a vector of AtomId.
     */
    inline vector< AtomId >& getAtomsInComponent();
    
    /**
     * This function return if this component is cyclic.
     * @return true if the component is cyclic, false otherwise.
     */
    inline bool isCyclic() const;

    /**
     * This function return if this component is head cycle free.
     * @return true if the component is head cycle free, false otherwise.
     */
    inline bool isHeadCycleFree() const;

    #ifndef TRACE_OFF
    void printComponentForDebug()
    {
        if( size() == 0 )
        {
            cerr << "Empty component.";
            return;
        }
        
        cerr << "Component: " << atomsInComponent[ 0 ];
                
        for( int i = 1; i < size(); i++ )
        {
            cerr << ", " << atomsInComponent[ i ];
        }
        cerr << "." << endl;
    }
    
    #endif
    
    /**
     * This function sets that current component is cyclic.
     */
    inline void setCyclic();

    /**
     * This function sets that current component is head cycle free.
     */
    inline void setHeadCycleFree();

    inline unsigned int size();
    
private:

    inline Component( const Component& orig );
    
    /**
     * This fields contain the atoms in this component.
     */
    vector< AtomId > atomsInComponent;

    /**
     * This field is true if the component is head cycle free.
     */
    bool headCycleFree;

    /**
     * This field is true if the component is cyclic.
     */
    bool cyclic;

};

Component::Component()
{
    headCycleFree = false;
    cyclic = false;
}

Component::Component(
    const Component& )
{
    assert( 0 );
}

Component::~Component()
{
}

void
Component::addAtomInComponent(
    AtomId atomId )
{
    atomsInComponent.push_back( atomId );
}

AtomId
Component::getAtom(
    unsigned int i )
{
    assert( i < atomsInComponent.size() );
    return atomsInComponent[ i ];
}

vector< AtomId >&
Component::getAtomsInComponent()
{
    return atomsInComponent;
}

bool
Component::isCyclic() const
{
    return cyclic;
}

bool
Component::isHeadCycleFree() const
{
    return headCycleFree;
}

void
Component::setHeadCycleFree()
{
    headCycleFree = true;
}

void
Component::setCyclic()
{
    cyclic = true;
}

unsigned int
Component::size()
{
    return atomsInComponent.size();
}

#endif	/* _COMPONENT_H */