/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "DependencyGraph.h"

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/strong_components.hpp>

/**
 * The dependency graph.
 */
static boost::adjacency_list <> g;
static unsigned numOfInstances = 0;

DependencyGraph::DependencyGraph()
{
#ifndef NDEBUG
    // Check that DependencyGraph is instantiated just one time!
    // This for ensuring that variable g is used exactly by one instance.
    // If more than one instance is needed, this should be changed, maybe by 
    // introducing some vector of adjacency_list.
    assert( numOfInstances == 0 );
    numOfInstances++;
#endif
}

DependencyGraph::DependencyGraph(
    const DependencyGraph& )
{
    assert( 0 );
}

DependencyGraph::~DependencyGraph()
{
    assert( numOfInstances == 1 );
    numOfInstances--;
    g.clear();
    assert( g.vertex_set().empty() );
    delete [] components;
}

void
DependencyGraph::clear()
{
    g.clear();
    assert( g.vertex_set().empty() );
    componentSize = 0;
    delete [] components;
    selfLoopVertex.clear();
}

void
DependencyGraph::addEdge(
    unsigned int v1,
    unsigned int v2 )
{
    boost::add_edge( v1, v2, g );
    if( v1 == v2 )
        selfLoopVertex.insert( v1 );    
}

Component*
DependencyGraph::componentsStrongConnected()
{
    vector< int > component( boost::num_vertices( g ) ), discover_time( boost::num_vertices( g ) );
    vector< boost::default_color_type > color( boost::num_vertices( g ) );
    vector< boost::graph_traits< boost::adjacency_list< > >::vertex_descriptor > root( boost::num_vertices( g ) );
    int num = boost::strong_components( g, &component[ 0 ] , boost::root_map( &root[ 0 ] ).color_map( &color[ 0 ] ).discover_time_map( &discover_time[ 0 ] ) );

    componentSize = num;
    components = new Component[ num ];

    for( vector< int >::size_type i = 0; i != component.size(); ++i )
    {
        components[ component[ i ] ].addAtomInComponent( i );
    }
    
    return components;
}

bool
DependencyGraph::existSelfLoop(
    unsigned int node )
{
    return ( selfLoopVertex.find( node ) != selfLoopVertex.end() );
}

void
DependencyGraph::printAdjacencyList()
{
   boost::graph_traits < boost::adjacency_list <> >::vertex_iterator firstIndex, lastIndex;
   boost::graph_traits < boost::adjacency_list <> >::adjacency_iterator adjacencyFirstIndex, adjacencyLastIndex;
   boost::property_map < boost::adjacency_list <>, boost::vertex_index_t >::type index_map = get( boost::vertex_index, g );

   for( boost::tie( firstIndex , lastIndex ) = boost::vertices( g ); firstIndex != lastIndex; ++firstIndex )
   {
       int index = get(index_map, *firstIndex);
       cout << index << ": ";

       boost::tie( adjacencyFirstIndex, adjacencyLastIndex ) = boost::adjacent_vertices( *firstIndex, g );
       if( adjacencyFirstIndex == adjacencyLastIndex )
           cout << "NO CONNECTION." << endl;
       else
       {
           for(; adjacencyFirstIndex != adjacencyLastIndex; ++adjacencyFirstIndex )
           {
               int newIndex = get( index_map, *adjacencyFirstIndex );
               cout << newIndex;
               if ( boost::next( adjacencyFirstIndex ) != adjacencyLastIndex )
                   cout << ", ";
           }
           cout << endl;
       }
   }
}