/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "GringoNumericFormat.h"
#include "Constants.h"
#include "options.h"

#include <cstdlib>
#include <cstring>
#include <fstream>
#include <algorithm>

GringoNumericFormat::~GringoNumericFormat()
{    
}

void
GringoNumericFormat::parse()
{
    this->parse( cin );
}

void
GringoNumericFormat::parse(
    istream& input )
{
    preprocessing( input );
    postprocessing();
}

void
GringoNumericFormat::preprocessing( 
    istream& input )
{
    bool loop = true;
    
    while( loop )
    {
        unsigned int type;
        input >> type;
    
        switch( type )
        {
            case GRINGO_NORMAL_RULE_ID:
                readNormalRule( input );
                break;
                
            case GRINGO_CHOICE_RULE_ID:
                readChoiceRule( input );
                break;
                
            case GRINGO_CONSTRAINT_RULE_ID:
                readConstraintRule( input );
                break;
                
            case GRINGO_DISJUNCTIVE_RULE_ID:
                readDisjunctiveRule( input );
                break;

            case GRINGO_LINE_SEPARATOR:
                loop = false;
                break;
        }
    }
    
    readAtomsTable( input );
    readTrueAtoms( input );
    readFalseAtoms( input );
    readErrorNumber( input );
}

void
GringoNumericFormat::readNormalRule(
    istream& input )
{
    GringoRule* gRule = new GringoRule();
    
    unsigned int headAtom;
    
    input >> headAtom;
    
    assert( headAtom > 0 );
    headAtom--;
    
    checkMax( headAtom );
    
    if( headAtom > 0 )
        gRule->pushHeadAtom( headAtom );
    
    readBody( input, gRule );
}

void
GringoNumericFormat::readChoiceRule(
    istream& input )
{
    GringoRule* gRule = new GringoRule();
    
    unsigned int size;
    input >> size;
    
    unsigned int i = 0;
    
    gRule->reserveHeadsSize( size );    
    
    while( i < size )
    {    
        unsigned int headAtom;

        input >> headAtom;

        assert( headAtom > 1 );
        
        headAtom--;
        checkMax( headAtom );
        
        gRule->pushHeadAtom( headAtom );
        
        i++;
    }
    
    gRule->setChoiceRule();
    
    readBody( input, gRule );
}

void
GringoNumericFormat::readBody( 
    istream& input, 
    GringoRule* gRule )
{
    unsigned int size;
    unsigned int negativeLiteralsSize;
    
    input >> size;
    input >> negativeLiteralsSize;
    
    gRule->reserveNegativeLiteralsSize( negativeLiteralsSize );
    
    unsigned int i = 0;    
    while( i < negativeLiteralsSize )
    {
        unsigned int negativeLiteral;
        input >> negativeLiteral;
        
        assert( negativeLiteral > 1 );
        negativeLiteral--;
        
        checkMax( negativeLiteral );
        
        gRule->pushNegativeBodyLiteral( negativeLiteral );
        i++;
    }
    
    gRule->reservePositiveLiteralsSize( size - negativeLiteralsSize );
    
    while( i < size )
    {
        unsigned int positiveLiteral;
        input >> positiveLiteral;
        
        assert( positiveLiteral > 1 );
        positiveLiteral--;
        
        checkMax( positiveLiteral );
        
        gRule->pushPositiveBodyLiteral( positiveLiteral );
        i++;
    }
    
    if( !gRule->isChoiceRule() )
    {
        if( gRule->size() == 1 )
        {
            if( gRule->headSize() == 1 )
            {
                trueSupportedAtoms.push_back( gRule->getHeadAtoms()[ 0 ] );
            }
            else if( gRule->positiveLiteralsSize() == 1 )
            {
                falseAtoms.push_back( gRule->getPositiveBodyLiterals()[ 0 ] );
            }
            else
            {
                assert( gRule->negativeLiteralsSize() == 1 );
                trueUnsupportedAtoms.push_back( gRule->getNegativeBodyLiterals()[ 0 ] );
            }

            delete gRule;

            return;
        }
        else if( gRule->size() == 0 )
        {        
            delete gRule;

            if( wasp::options.competitionOutput )
                cout << NOMODEL_COMPETITION_OUTPUT << endl;
            else if( !wasp::options.silentMode )
            {
                cout << WASP_STRING << endl;
                cout << NOMODEL << endl;
            }
            exit( 0 );
        }
    }
    
    if( !gRule->haveToDelete() )
        gringoRules.push_back( gRule );
}

void
GringoNumericFormat::readDisjunctiveRule(
    istream& input )
{
    GringoRule* gRule = new GringoRule();
    
    unsigned int size;
    input >> size;
    
    unsigned int i = 0;
    
    gRule->reserveHeadsSize( size );    
    
    while( i < size )
    {    
        unsigned int headAtom;

        input >> headAtom;

        assert( headAtom > 1 );
        
        headAtom--;
        checkMax( headAtom );
        
        gRule->pushHeadAtom( headAtom );
        
        i++;
    }
    
    readBody( input, gRule );    
}

bool compare( GringoAtom* i, GringoAtom* j )
{
    return( i ->getId() < j->getId() );
}

void 
GringoNumericFormat::readAtomsTable(
    istream& input )
{
    GringoAtom* falseAtom = new GringoAtom( 0, "" );
    gringoAtoms.push_back( falseAtom );
    
    unsigned int nextAtom;    
    input >> nextAtom;
    
    while( nextAtom != 0 )
    {
        nextAtom--;
        
        string name;
        input >> name;
        
        GringoAtom* atom = new GringoAtom( nextAtom, name );
        gringoAtoms.push_back( atom );
        
        input >> nextAtom;
    }    
    
    gringoAtoms.sort( compare );
}

void
GringoNumericFormat::readTrueAtoms(
    istream& input )
{
    char b;
    
    input >> b;
    input >> b;
    
    unsigned int nextAtom;
    input >> nextAtom;
    
    while( nextAtom != 0 )
    {
        nextAtom--;
        
        assert( nextAtom > 0 );
        trueSupportedAtoms.push_back( nextAtom );
        
        input >> nextAtom;
    }
}

void
GringoNumericFormat::readFalseAtoms(
    istream& input )
{
    char b;
    
    input >> b;
    input >> b;
    
    unsigned int nextAtom;
    input >> nextAtom;
    
    while( nextAtom != 0 )
    {
        nextAtom--;
        
        if( nextAtom != 0 )
            falseAtoms.push_back( nextAtom );
        
        input >> nextAtom;
    }
}

void
GringoNumericFormat::readErrorNumber(
    istream& input )
{
    unsigned int errorNumber;
    input >> errorNumber;
    
    if( errorNumber != 1 )
    {
        cout << ERRORPARSING << endl;
        exit( 0 );
    }
    else
    {
    }
}

void 
GringoNumericFormat::readConstraintRule( 
    istream & input )
{
    unsigned int id;
    input >> id;
    
    assert( id > 1 );
    id--;
    
    checkMax( id );
    
    unsigned int literalsSize;
    input >> literalsSize;
    
    unsigned int negLiteralsSize;
    input >> negLiteralsSize;
    
    if( negLiteralsSize > 0 )
    {
        cerr << GRINGO_NEG_LIT_NOT_SUPPORTED << endl;
        exit( 0 );
    }
    
    unsigned int bound;
    input >> bound;
    
    GringoAggregate* gAggregate = new GringoAggregate( id, bound, literalsSize );
    
    unsigned int i = 0;
    while( i < literalsSize )
    {    
        unsigned int atom;

        input >> atom;

        assert( atom > 1 );
        
        atom--;
        checkMax( atom );
        
        gAggregate->addAtom( atom );
        
        atomAggregateTypeSet[ atom ] = GRINGO_ATOM_IN_AGGREGATE_VALUE;
        
        i++;
    }
    
    atomAggregateTypeSet[ id ] = GRINGO_AGGREGATE_GREATER_VALUE;
    
    gringoAggregates.push_back( gAggregate );
}

void
GringoNumericFormat::addAtom(
    unsigned int id, 
    const char* name )
{
    switch( atomAggregateTypeSet[ id ] )
    {
        case GRINGO_ATOM_IN_AGGREGATE_VALUE:
            programBuilder->addAtomInAnAggregateSet( id, name );
            break;
            
        case GRINGO_AGGREGATE_GREATER_VALUE:            
            programBuilder->addAggregateAtomGreater( id, name );
            break;
            
        default:
            programBuilder->addAtom( id, name );
            break;
    }
}

void
GringoNumericFormat::postprocessing()
{    
    programBuilder->startProgram();
 
    unsigned int id;    
    unsigned int oldId = 0;
    
    for( List< GringoAtom* >::iterator it = gringoAtoms.begin(); it != gringoAtoms.end(); ++it )
    {
        id = ( *it )->getId();
        
        while( id > oldId )
        {
            //programBuilder->addAtom( oldId, "" );            
            addAtom( oldId, "" );
            oldId++;
        }
        
//        programBuilder->addAtom( id, ( *it )->getName() );
        addAtom( id, ( *it )->getName().c_str() );
        
        oldId = id + 1;
        
        assert( *it != NULL );
        delete ( *it );
    }
    
    while( max >= oldId )
    {
//        programBuilder->addAtom( oldId, "" );
        addAtom( oldId, "" );
        oldId++;
    }
    
    gringoAtoms.clear();    
    
    for( unsigned int i = 0; i < gringoRules.size(); i++ )
    {
        GringoRule* g = gringoRules[ i ];
        
        if( g->isChoiceRule() )
        {
            ChoiceRule* c = programBuilder->startChoiceRule( g->bodySize(), g->headSize() );
            
            vector< unsigned int >& headAtoms = g->getHeadAtoms();        
            for( unsigned int j = 0; j < headAtoms.size(); j++ )
            {
                programBuilder->addHeadAtomInChoiceRule( c, headAtoms[ j ] );
            }

            vector< unsigned int >& positiveLiterals = g->getPositiveBodyLiterals();
            for( unsigned int j = 0; j < positiveLiterals.size(); j++ )
            {
                programBuilder->addBodyLiteralInChoiceRule( c, positiveLiterals[ j ], true );
            }

            vector< unsigned int >& negativeLiterals = g->getNegativeBodyLiterals();
            for( unsigned int j = 0; j < negativeLiterals.size(); j++ )
            {
                programBuilder->addBodyLiteralInChoiceRule( c, negativeLiterals[ j ], false );
            }

            programBuilder->endChoiceRule( c );
        }
        else
        {
            WaspRule* w = programBuilder->startWaspRule( g->size(), g->headSize() );
        
            vector< unsigned int >& headAtoms = g->getHeadAtoms();        
            for( unsigned int j = 0; j < headAtoms.size(); j++ )
            {
                programBuilder->addHeadAtom( w, headAtoms[ j ] );
            }

            if( !headAtoms.empty() )
            {
                vector< unsigned int >& positiveLiterals = g->getPositiveBodyLiterals();
                for( unsigned int j = 0; j < positiveLiterals.size(); j++ )
                {
                    programBuilder->addBodyLiteral( w, positiveLiterals[ j ], true );
                }

                vector< unsigned int >& negativeLiterals = g->getNegativeBodyLiterals();
                for( unsigned int j = 0; j < negativeLiterals.size(); j++ )
                {
                    programBuilder->addBodyLiteral( w, negativeLiterals[ j ], false );
                }
            }
            else
            {
                vector< unsigned int >& positiveLiterals = g->getPositiveBodyLiterals();
                for( unsigned int j = 0; j < positiveLiterals.size(); j++ )
                {
                    programBuilder->addBodyLiteralInConstraint( w, positiveLiterals[ j ], true );
                }

                vector< unsigned int >& negativeLiterals = g->getNegativeBodyLiterals();
                for( unsigned int j = 0; j < negativeLiterals.size(); j++ )
                {
                    programBuilder->addBodyLiteralInConstraint( w, negativeLiterals[ j ], false );
                }
            }

            programBuilder->endWaspRule( w );
        }
        delete g;
    }
    
    gringoRules.clear();
    
    for( unsigned int i = 0; i < gringoAggregates.size(); i++ )
    {
        GringoAggregate* g = gringoAggregates[ i ];
        Aggregate* aggregate = programBuilder->startAggregateCount( g->getId(), g->getBound(), MAXVALUE, g->getAtomsSize() );
        
        vector< unsigned int >& atomsInTheAggregateSet = g->getAtoms();
        for( unsigned int j = 0; j < atomsInTheAggregateSet.size(); j++ )
        {
            unsigned int atomInTheAggregateSet = atomsInTheAggregateSet[ j ];
            programBuilder->addAtomInTheAggregate( atomInTheAggregateSet, 1, aggregate, j );
        }
        
        programBuilder->endAggregateCount( aggregate );
    }
    
    for( unsigned int i = 0; i < trueSupportedAtoms.size(); i++ )
    {
        programBuilder->addTrueSupportedAtom( trueSupportedAtoms[ i ] );
    }
    
    for( unsigned int i = 0; i < trueUnsupportedAtoms.size(); i++ )
    {
        programBuilder->addTrueUnsupportedAtom( trueUnsupportedAtoms[ i ] );
    }
    
    for( unsigned int i = 0; i < falseAtoms.size(); i++ )
    {
        programBuilder->addFalseAtom( falseAtoms[ i ] );
    }
    
    programBuilder->endProgram();
}

void
GringoNumericFormat::checkMax(
    unsigned int value )
{
    if( value > max )
        max = value;
}
