/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   GringoNumericFormat.h
 * Author: Carmine Dodaro
 *
 * Created on 7 January 2013, 17.31
 */


#ifndef _GRINGONUMERICFORMAT_H
#define	_GRINGONUMERICFORMAT_H

#include <vector>
#include "stl/List.h"
#include <unordered_set>
#include <iostream>
using namespace std;

#include "unordered_map.h"
#include "ProgramBuilder.h"

class GringoRule
{
    public:
        
        GringoRule() : toDelete( false ), choice( false )
        {
        }
        
        void pushHeadAtom( 
            unsigned int headAtom )
        {
            headAtomsUnorderedSet.insert( headAtom );
            headAtomsPosition[ headAtom ] = headAtoms.size();
            headAtoms.push_back( headAtom );
        }

        void pushPositiveBodyLiteral( 
            unsigned int literal )
        {
            if( headAtomsUnorderedSet.find( literal ) != headAtomsUnorderedSet.end() )
                toDelete = true;
            positiveBodyLiterals.push_back( literal );
        }
        
        void pushNegativeBodyLiteral( 
            unsigned int literal )
        {
            if( headAtomsUnorderedSet.find( literal ) != headAtomsUnorderedSet.end() )
            {
                unsigned int pos = headAtomsPosition[ literal ];
                headAtoms.erase( headAtoms.begin() + pos );
            }
            
            negativeBodyLiterals.push_back( literal );
        }
        
        unsigned int size()
        {
            return ( headAtoms.size() + positiveBodyLiterals.size() + negativeBodyLiterals.size() );
        }
        
        unsigned int headSize()
        {
            return headAtoms.size();
        }
        
        unsigned int bodySize()
        {
            return positiveBodyLiterals.size() + negativeBodyLiterals.size();
        }
        
        unsigned int positiveLiteralsSize()
        {
            return positiveBodyLiterals.size();
        }
        
        unsigned int negativeLiteralsSize()
        {
            return negativeBodyLiterals.size();
        }
        
        vector< unsigned int >& getHeadAtoms()
        {
            return headAtoms;
        }
        
        vector< unsigned int >& getPositiveBodyLiterals()
        {
            return positiveBodyLiterals;
        }
        
        vector< unsigned int >& getNegativeBodyLiterals()
        {
            return negativeBodyLiterals;
        }
        
        void reservePositiveLiteralsSize( unsigned int size )
        {
            positiveBodyLiterals.reserve( size );
        }
        
        void reserveNegativeLiteralsSize( unsigned int size )
        {
            negativeBodyLiterals.reserve( size );
        }
        
        void reserveHeadsSize( unsigned int size )
        {
            headAtoms.reserve( size );
        }
        
        bool haveToDelete()
        {
            return toDelete;
        }
        
        bool setChoiceRule()
        {
            choice = true;
        }
        
        bool isChoiceRule()
        {
            return choice;
        }

    private:
        unordered_set< unsigned int > headAtomsUnorderedSet;
        unordered_map< unsigned int, unsigned int > headAtomsPosition;
        vector< unsigned int > headAtoms;
        vector< unsigned int > positiveBodyLiterals;
        vector< unsigned int > negativeBodyLiterals;
        bool toDelete;
        bool choice;
};

class GringoAtom
{
    public:
        GringoAtom( unsigned int i, string n ) : id( i ), name( n ){}
        
        unsigned int getId()
        { 
            return id; 
        }
        
        string getName()
        {
            return name;
        }
        
    private:
        unsigned int id;
        string name;
};

class GringoAggregate
{
    public:
        GringoAggregate( unsigned int i, unsigned int b, unsigned int s ) : id( i ), bound( b )
        {
            atoms.reserve( s );
        }
        
        void addAtom( unsigned int a )
        {
            atoms.push_back( a );
        }
        
        vector< unsigned int >& getAtoms()
        {
            return atoms;
        }
        
        unsigned int getBound()
        {
            return bound;
        }
        
        unsigned int getId()
        {
            return id;
        }
        
        unsigned int getAtomsSize()
        {
            return atoms.size();
        }
        
    private:
        unsigned int id;
        unsigned int bound;
        vector< unsigned int > atoms;
};

class GringoNumericFormat
{

public:

    inline GringoNumericFormat( ProgramBuilder* p );
    ~GringoNumericFormat();
    
    /**
     * This function read instruction from standard input and
     * build the program.
     */
    void parse();

    /**
     * This function read instruction from input and
     * build the program.
     *
     * @param input The istream input.
     */
    void parse( istream& input );
    
private:

    void addAtom( unsigned int, const char* s );
    
    void preprocessing( istream& input );
    
    void readAtomsTable( istream& input );
    void readNormalRule( istream & input );
    void readDisjunctiveRule( istream & input );
    void readChoiceRule( istream & input );
    void readConstraintRule( istream & input );
    
    void readBody( istream& input, GringoRule* gRule );    
    
    void readTrueAtoms( istream& input );
    void readFalseAtoms( istream& input );
    
    void readErrorNumber( istream& input );
    
    void postprocessing();
        
    void checkMax( unsigned int value );
    
    ProgramBuilder* programBuilder;
    
    vector< GringoRule* > gringoRules;
    List< GringoAtom* > gringoAtoms;
    vector< GringoAggregate* > gringoAggregates;
    vector< unsigned int > trueSupportedAtoms;
    vector< unsigned int > trueUnsupportedAtoms;
    vector< unsigned int > falseAtoms;
    unsigned int max;
    unordered_map< AtomId, GringoAggregateType > atomAggregateTypeSet;
};

GringoNumericFormat::GringoNumericFormat(
    ProgramBuilder* p ) : programBuilder( p ), max( 0 )
{
}

#endif	/* _GRINGONUMERICFORMAT_H */
