/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   Interpretation.h
 * Author: Carmine Dodaro
 *
 * Created on 11 October 2010, 18.35
 */

#ifndef _INTERPRETATION_H
#define	_INTERPRETATION_H

#include "Constants.h"
#include "Atom.h"
#include "Literal.h"
#include <cassert>
#include <string.h>

class Interpretation
{
    friend class ModelGenerator;
public:

    /**
     * This constructor create the atoms array.
     * @param size The size of atoms array.
     */
    Interpretation();    
    
    inline ~Interpretation();
    
    bool operator==( const Interpretation& interpretation ) const;

    /**
     * Returns the truth value of an atom.
     *
     * @param atomId the id of Atom.
     * @return the truth value of atom.
     */
    inline TruthValue getAtomTruthValue( Atom& atom ) const;

    inline TruthValue getAtomTruthValue( AtomId atomId ) const;

    /**
     * Returns the left watch value.
     *
     * @return the first undefined atom.
     */
    inline const unsigned int getLeftWatch() const;

    /**
     * Returns the left watch truth value.
     *
     * @return Undefined if there is at least an undefined atom, true or false otherwise.
     */
    inline const TruthValue getLeftWatchTruthValue() const;

    /**
     * Returns the right watch value.
     *
     * @return the last undefined atom.
     */
    inline const unsigned int getRightWatch() const;

    /**
     * Returns the right watch truth value.
     *
     * @return Undefined if there is at least an undefined atom, true or false otherwise.
     */
    inline const TruthValue getRightWatchTruthValue() const;

    /**
     * Returns the size of the interpretation.
     *
     * @return the size of the interpretation.
     */
    inline const unsigned int getSize() const;

    void init( unsigned int size );

    /*
     * This function update left watch.
     */
    inline void updateLeftWatch();

    /*
     * This function update right watch.
     */
    inline void updateRightWatch();

    /**
     * Set the truth value of an atom.
     *
     * @param atomId the id of Atom.
     * @param truthValue the truth value.
     */
    inline void setAtomTruthValue( AtomId atomId, TruthValue truthValue );

    inline bool isFalse( const Atom& atom ) const;

    inline bool isFalsePropagated( const Atom& atom ) const;

    inline bool isFalseUndefined( const Atom& atom ) const;

    inline bool isTrue( const Atom& atom ) const;

    inline bool isTruePropagated( const Atom& atom ) const;

    inline bool isTrueUndefined( const Atom& atom ) const;

    inline bool isUndefined( const Atom& atom ) const;

    inline bool isUndefinedNotTrue( const Atom& atom ) const;

    inline bool isUndefinedNotFalse( const Atom& atom ) const;

    inline bool isFalse( AtomId atom ) const;

    inline bool isFalsePropagated( AtomId atom ) const;

    inline bool isFalseUndefined( AtomId atom ) const;

    inline bool isTrue( AtomId atom ) const;

    inline bool isTruePropagated( AtomId atom ) const;

    inline bool isTrueUndefined( AtomId atom ) const;

    inline bool isUndefined( AtomId atom ) const;

    inline bool isUndefinedNotTrue( AtomId atom ) const;

    inline bool isUndefinedNotFalse( AtomId atom ) const;

    inline bool isFalse( Literal& literal ) const;

    inline bool isFalsePropagated( Literal& literal ) const;

    inline bool isFalseUndefined( Literal& literal ) const;

    inline bool isTrue( Literal& literal ) const;

    inline bool isTruePropagated( Literal& literal ) const;

    inline bool isTrueUndefined( Literal& literal ) const;

    inline bool isUndefined( Literal& literal ) const;

    inline bool isUndefinedNotTrue( Literal& literal ) const;

    inline bool isUndefinedNotFalse( Literal& literal ) const;

    inline bool thereAreUndefinedAtoms() const;
    
private:

    /**
     * Truth values of atoms are stored in a dinamic array.
     */
    TruthValue *atoms;

    /**
     * Size of atoms.
     */
    unsigned int size;

    /*
     * This watch contains the first undefined in the interpretation.
     */
    unsigned int leftWatch;

    /*
     * This watch contains the last undefined in the interpretation.
     */
    unsigned int rightWatch;

    bool checkUndefined() const;
};

Interpretation::~Interpretation()
{
    delete [] atoms;
}

TruthValue
Interpretation::getAtomTruthValue(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return atoms[ atomId ];
}

TruthValue
Interpretation::getAtomTruthValue(
    Atom& atom ) const
{
    return getAtomTruthValue( atom.getId() );
}

const unsigned int
Interpretation::getLeftWatch() const
{
    return leftWatch;
}

const TruthValue
Interpretation::getLeftWatchTruthValue() const
{
    assert( 0 );
//    assert( leftWatch > 0 && leftWatch < size );
//    return this->getAtomTruthValue( leftWatch );
}

const unsigned int
Interpretation::getRightWatch() const
{
    return rightWatch;
}

const TruthValue
Interpretation::getRightWatchTruthValue() const
{
    assert( 0 );
//    assert( rightWatch > 0 && rightWatch < size );
//    return this->getAtomTruthValue( rightWatch );
}

const unsigned int
Interpretation::getSize() const
{
    return size;
}

void
Interpretation::setAtomTruthValue(
    AtomId atomId,
    TruthValue truthValue )
{
    assert( truthValue != UNDEFINED );
    assert( ( atoms[ atomId ] == UNDEFINED ) || ( atoms[ atomId ] == TRUE_UNDEFINED ) || ( atoms[ atomId ] == FALSE_UNDEFINED ) );

    atoms[ atomId ] = truthValue;
}

void
Interpretation::updateLeftWatch()
{
    assert( atoms[ leftWatch ] != UNDEFINED );
    leftWatch++;
    assert( leftWatch <= rightWatch );
}

void
Interpretation::updateRightWatch()
{
    assert( atoms[ rightWatch ] != UNDEFINED );
    rightWatch--;
    assert( rightWatch >= leftWatch );
}

bool
Interpretation::isFalse(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
//    return ( ( atoms[ atomId ] & FALSE ) == FALSE );
    return ( atoms[ atomId ] & FALSE );
//    return( isFalsePropagated( atomId ) || isFalseUndefined ( atomId ) );
}

bool
Interpretation::isFalsePropagated(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( ( atoms[ atomId ] & INTERPRETATION_MASK ) == FALSE );
}

bool
Interpretation::isFalseUndefined(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( ( atoms[ atomId ] & INTERPRETATION_MASK ) == FALSE_UNDEFINED );
}

bool
Interpretation::isTrue(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
//    return ( ( atoms[ atomId ] & TRUE ) == TRUE );
    return ( atoms[ atomId ] & TRUE );
//    return ( isTruePropagated( atomId ) || isTrueUndefined( atomId ) );
}

bool
Interpretation::isTruePropagated(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( ( atoms[ atomId ] & INTERPRETATION_MASK ) == TRUE );
}

bool
Interpretation::isTrueUndefined(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( ( atoms[ atomId ] & INTERPRETATION_MASK ) == TRUE_UNDEFINED );
}

bool
Interpretation::isUndefined(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( atoms[ atomId ] & 1 );
}

bool
Interpretation::isUndefinedNotTrue(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( ( atoms[ atomId ] & TRUE_UNDEFINED ) == UNDEFINED );
}

bool
Interpretation::isUndefinedNotFalse(
    AtomId atomId ) const
{
    assert( atomId > 0 && atomId < size );
    return ( ( atoms[ atomId ] & FALSE_UNDEFINED ) == UNDEFINED );
}

bool
Interpretation::isFalse(
    const Atom& atom ) const
{
    return isFalse( atom.getId() );
}

bool
Interpretation::isFalsePropagated(
    const Atom& atom ) const
{
    return isFalsePropagated( atom.getId() );
}

bool
Interpretation::isFalseUndefined(
    const Atom& atom ) const
{
    return isFalseUndefined( atom.getId() );
}

bool
Interpretation::isTrue(
    const Atom& atom ) const
{
    return isTrue( atom.getId() );
}

bool
Interpretation::isTruePropagated(
    const Atom& atom ) const
{
    return isTruePropagated( atom.getId() );
}

bool
Interpretation::isTrueUndefined(
    const Atom& atom ) const
{
    return isTrueUndefined( atom.getId() );
}

bool
Interpretation::isUndefined(
    const Atom& atom ) const
{
    return isUndefined( atom.getId() );
}

bool
Interpretation::isUndefinedNotTrue(
    const Atom& atom ) const
{
    return isUndefinedNotTrue( atom.getId() );
}

bool
Interpretation::isUndefinedNotFalse(
    const Atom& atom ) const
{
    return isUndefinedNotFalse( atom.getId() );
}

bool
Interpretation::isFalse(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isFalse( id );
    else
        return isTrue( id );
}

bool
Interpretation::isFalsePropagated(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isFalsePropagated( id );
    else
        return isTruePropagated( id );
}

bool
Interpretation::isFalseUndefined(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isFalseUndefined( id );
    else
        return isTrueUndefined( id );
}

bool
Interpretation::isTrue(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isTrue( id );
    else
        return isFalse( id );
}

bool
Interpretation::isTruePropagated(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isTruePropagated( id );
    else
        return isFalsePropagated( id );
}

bool
Interpretation::isTrueUndefined(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isTrueUndefined( id );
    else
        return isFalseUndefined( id );
}

bool
Interpretation::isUndefined(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    return isUndefined( id );
}

bool
Interpretation::isUndefinedNotTrue(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isUndefinedNotTrue( id );
    else
        return isUndefinedNotFalse( id );
}

bool
Interpretation::isUndefinedNotFalse(
    Literal& literal ) const
{
    AtomId id = literal.getAtomId();
    if( literal.isPositive() )
        return isUndefinedNotFalse( id );
    else
        return isUndefinedNotTrue( id );
}

bool
Interpretation::thereAreUndefinedAtoms() const
{
    assert( isUndefined( leftWatch ) || leftWatch == rightWatch );
    assert( isUndefined( leftWatch ) ? checkUndefined() : !checkUndefined() );
    
//    if( isUndefined( leftWatch ) )
//        assert( checkUndefined() );
//    else
//        assert( !checkUndefined() );
    return isUndefined( leftWatch ) ;
}

#endif	/* _INTERPRETATION_H */
