/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/*
 * File:   Literal.h
 * Author: Carmine Dodaro
 *
 * Created on 29 Aprile 2012, 17.44
 */

#ifndef _LITERAL_H
#define	_LITERAL_H

#include <cassert>
using namespace std;

#include "Constants.h"
#include "RuleContainer.h"

#include "Atom.h"
//#include "WaspRule.h"
#include "WatchedList.h"

class Atom;
class WaspRule;

class Literal
{
    friend ostream &operator<<( ostream &, const Literal & );

public:

    // Instantiate a positive literal
    inline Literal( Atom* = NULL );
    
    // Instantiate a negative literal
    inline Literal( Atom* , bool );
    
    inline Literal( const Literal& );
    
    inline ~Literal();

    inline bool isPositive() const;

    inline Atom* getAtom();
    inline const Atom* getAtom() const;
    
    inline void setAtom( Atom* );

    inline void setPositive();
    inline void setNegative();
    
    inline int getId() const;
    inline AtomId getAtomId() const;

    inline bool isFalsePropagated() const;
    inline bool isTruePropagated() const;
    inline bool isTrue() const;
    inline bool isFalse() const;
    inline bool isUndefined() const;
    
    inline const RuleContainer< WaspRule* >& getBinaryWaspRules() const;
    inline const RuleContainer< WaspRule* >& getBinaryWaspRulesReverse() const;

    inline bool operator==( const Literal& ) const;
    
    inline WatchedList< WaspRule* >::iterator addWatchedWaspRule( WaspRule* );
    inline WatchedList< WaspRule* >::iterator addWatchedConstraint( WaspRule* );
    inline void eraseWatchedConstraint( WatchedList< WaspRule* >::iterator it );
    
private:
    inline void setPositiveAtom( Atom* );
    inline void setNegativeAtom( Atom* );

    Atom* atom;
    
};

Literal::~Literal()
{

}

Literal::Literal(
    const Literal& l )
{
    this->atom = l.atom; 
}

bool
Literal::isPositive() const
{
    return ( long long ) atom & 1;
}

Atom*
Literal::getAtom()
{
    return ( Atom* ) ( ( long long ) atom & ( ~1 ) );
}

const Atom*
Literal::getAtom() const
{
    return ( Atom* ) ( ( long long ) atom & ( ~1 ) );
}

int
Literal::getId() const
{
    return isPositive() ? getAtomId() : -getAtomId();
}

AtomId
Literal::getAtomId() const
{
    return getAtom()->getId();
}

bool
Literal::operator==(
    const Literal& literal ) const
{
    return ( atom == literal.atom );
}

Literal::Literal(
    Atom* a ) : atom( ( Atom* ) ( ( long long ) a | 1 ) )
{
    assert( isPositive() );
    assert( getAtom() == a );
}

Literal::Literal(
    Atom* a,
    bool ) : atom( a )
{
    assert( !isPositive() );
    assert( getAtom() == a );
}

void
Literal::setAtom(
    Atom* a )
{
    atom = a;
}

void
Literal::setPositive()
{
    atom = ( Atom* ) ( ( long long ) atom | 1 );
}

void
Literal::setNegative()
{
    atom = ( Atom* ) ( ( long long ) atom | 0 );
}

void
Literal::setPositiveAtom(
    Atom* a )
{
    atom = ( Atom* ) ( ( long long ) a | 1 );
}

void
Literal::setNegativeAtom(
    Atom* a )
{
    atom = a;
}

bool
Literal::isFalsePropagated() const
{
    return isPositive() ? getAtom()->isFalsePropagated() : getAtom()->isTruePropagated();
}

bool
Literal::isTruePropagated() const
{
    return isPositive() ? getAtom()->isTruePropagated() : getAtom()->isFalsePropagated();
}

bool
Literal::isFalse() const
{
    return isPositive() ? getAtom()->isFalse() : getAtom()->isTrue();
}

bool
Literal::isTrue() const
{
    return isPositive() ? getAtom()->isTrue() : getAtom()->isFalse();
}

bool
Literal::isUndefined() const
{    
    return getAtom()->isUndefined();
}

const RuleContainer< WaspRule* >&
Literal::getBinaryWaspRules() const
{
    return isPositive() ? getAtom()->getBinaryPositiveWaspRules() : getAtom()->getBinaryNegativeWaspRules();
}

const RuleContainer< WaspRule* >&
Literal::getBinaryWaspRulesReverse() const
{
    return isPositive() ? getAtom()->getBinaryNegativeWaspRules() : getAtom()->getBinaryPositiveWaspRules();
}

WatchedList< WaspRule* >::iterator 
Literal::addWatchedWaspRule( 
    WaspRule* w )
{
    if( isPositive() )
        return getAtom()->addWatchedPositiveWaspRule( w );
    else
        return getAtom()->addWatchedNegativeWaspRule( w );
}
    
WatchedList< WaspRule* >::iterator 
Literal::addWatchedConstraint( 
    WaspRule* w )
{
    if( isPositive() )
        return getAtom()->addWatchedPositiveConstraint( w );
    else
        return getAtom()->addWatchedNegativeConstraint( w );
}

void 
Literal::eraseWatchedConstraint( 
    WatchedList< WaspRule* >::iterator it )
{
    if( isPositive() )
        getAtom()->eraseWatchedPositiveConstraint( it );
    else
        getAtom()->eraseWatchedNegativeConstraint( it );
}

#endif	/* _LITERAL_H */
