/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/*
 * ModelChecker.cpp
 *
 *  Created on: 02/dic/2010
 *      Author: marco
 */

#include <vector>

#include "ModelChecker.h"

#include "trace.h"
#include "options.h"
#include "timer.h"

void
ModelChecker::initializeUnfoundedSet(
    Interpretation& I )
{
    trace( mc, 2, "Initializing the set.\n" );

    unfoundedSet.clear();

    //This code copy the interpretation
    for( unsigned v = 1; v < I.getSize(); v++ )
    {
		Atom& a = program->getAtom( v );
        if( I.isTruePropagated( v ) && !a.isAggregate() && !a.isTrueSupported() )
        {
            unfoundedSet.insert( v );
        }
    }
}

/*
 * This function implement the R operator used to find the maximal unfounded set.
 * The fix point of R tells us if a given interpretation is a stable model
 * (for head cycle free programs)
 *
 * This is the R operator
 * For each atoms a that belongs to unfoundedSet it is still included in unfoundedSet if
 *
 *      - for each rule r in which a belongs to the head of r at least one of the
 *        following conditions holds:
 *              - the body of r contains something that is in unfoundedSet
 *               or in the complement of I
 *              - the head of r contains b != a such that b belongs to I
 *
 * @return true if the unfoundedSet changed, false otherwise. 
 * 
 * */
bool
ModelChecker::operatorR(
    Interpretation& )//I )
{
    trace( mc, 2, "Applying R operator.\n" );

    statistics( incrementOperatorRCalls() );
    
    bool unfoundedSetHasChanged = false;

    //For each atom a
    for( unsigned int it = 0; it < unfoundedSet.size(); it++ )
    {
        unsigned int i = unfoundedSet[ it ];
        Atom& a = program->getAtom( i );

        trace( mc, 2, "Checking atom %s.\n", program->atomToString( i ).c_str() );

        //These are the heads of the rules to which a belongs
        const RuleContainer< WaspRule* >& rules = a.getNegativeHeadWaspRules();

        unsigned int rulesSize = rules.size();
        //Perform the check for each rule
        for( unsigned int rh = 0; rh != rulesSize; rh++ )
        {
            WaspRule& r = *( rules[ rh ] );

            trace( mc, 3, "Considering rule %s", program->waspRuleToString( r ).c_str() );

            if( r.isSatisfiedCannotSupport() )
            {
                trace( mc, 3, "The rule is satisfied and cannot support the atom: skip it.\n" );
                continue;
            }
            
            bool ruleSupport = true;
            unsigned int bodySize = r.getNumberOfLiteralsInTheBody();
            for( unsigned int at = 0; at < bodySize; at++ )
            {
                int val = r.getBodyLiteral( at ).getId();
                if( val > 0 && unfoundedSet.contains( val ) )
                {
                    trace( mc, 3, "Found an atom in the positive body which is also in the unfounded set: skip this rule.\n" );
                    ruleSupport = false;
                    
                    atomRulesRecursiveMap[ val ].insert( rules[ rh ] );                    
                    break;
                }
            }
            
            if( ruleSupport )
            {
                trace( mc, 2, "The atom %s does not belong to the set anymore.\n", program->atomToString( i ).c_str() );
                unfoundedSet.erase( i );
                unfoundedSetHasChanged = true;
                break;
            }
        }                
    }//for each atom a
    
    return unfoundedSetHasChanged;
}

void
ModelChecker::fixPointoperatorR(
    Interpretation& I )
{
    startTimer( polynomialOperator );    
    trace( mc, 1, "Applying fixpoint of the R operator.\n" );

    statistics( incrementFixOperatorRCalls() );
    //First of all initialize the unfounded set to I for the polynomial check
    initializeUnfoundedSet( I );

    #ifndef TRACE_OFF
    unsigned count = 0;
    trace( mc, 2, "Iteration: %d.\n", count );
    #endif
    
    //Then calculate the fixpoint of R
    while( operatorR( I ) )
    {
        #ifndef TRACE_OFF
        count++;
        trace( mc, 2, "Iteration: %d.\n", count );
        #endif
    }

    trace( mc, 1, "The fixpoint of R contains %u atoms.\n", static_cast< unsigned >( unfoundedSet.size() ) );
    stopTimer( polynomialOperator );
}

bool
ModelChecker::isStable( 
    Interpretation& I )
{
    //Perform the polynomial check
    fixPointoperatorR( I );

    if( unfoundedSet.size() == 0 ) //No unfoundend set founded, the model is stable
    {
        trace( mc, 1, "The fixpoint of R is empty: The model is stable.\n" );
        return true;
    }
    else //Found that unfounded set is not empty
    {
        if( program->isHeadCycleFree() ) //if the program is head-cycle-free model in not stable
        {
            trace( mc, 1, "The program is head-cycle-free and the fixpoint of R is not empty: The model is not stable.\n" );
            return false;
        }
        else
        {
            trace( mc, 1, "The program is non-head-cycle-free and the fixpoint of R is not empty: Checking the model via SAT.\n" );
            return !performSATCheck( I );
        }
    }
}

bool
ModelChecker::performSATCheck(
    Interpretation& I )
{
    //Declaration of the solver
    Solver minisat;

    //  Controls which polarity the decision heuristic chooses
    if( options.polarity == 1 )
        minisat.polarity_mode = Solver::polarity_true;
    else if( options.polarity == 2 )
        minisat.polarity_mode = Solver::polarity_rnd;
    else
        minisat.polarity_mode = Solver::polarity_false;

    trace( mc, 1, "Building sat formula.\n" );
    //Build SAT formula and check if resulting formula is empty
    if( build_simple_minisat_instance( minisat, I ) )
    {
        startTimer( satSolverInvocation );
        trace( mc, 1, "The formula is non empty. Going to solve it.\n" );
        minisat.solve();
        stopTimer( satSolverInvocation );    
    }
    else
    {
        trace( mc, 1, "The formula is empty. All atoms in the fixpoint of R are an unfounded set: We pick the first.\n" );
        //If the formula is empty each atoms in unfoundedSet is an
        //unfounded set, just pick the first!
        //AtomId unf = *unfoundedSet.begin();
        AtomId unf = unfoundedSet[ 0 ];
        unfoundedSet.clear();
        unfoundedSet.insert( unf );

        return true;
    }

    const bool isSat = minisat.okay();

    if( isSat )
    {
        trace( mc, 1, "The formula is SAT: the model is unstable.\n" );
        trace( mc, 1, "The unfounded set is: " );
        unsigned int size = program->getNumberOfAtoms();
        for( unsigned int i = 1; i < size; i++ )
        {
            // Take care: the indices of SATZ internal variables are
            // one less than what SATZ takes as input, thus -1 below.
            if( containedInUnfoundedSet( i ) && minisat.model[ i - 1 ] == l_False )
            {
                unfoundedSet.erase( i );
            }
            else if( containedInUnfoundedSet( i ) && minisat.model[ i - 1 ] == l_True )
            {
                trace( mc, 1, "%d", i );
            }
        }
    }
    else
    {
        trace( mc, 1, "The formula is unsatisfiable: the model is stable.\n" );
    }
    return isSat;
}

/*
 *
 * This function is used by addTransformedRule.
 *
 * A rule R is relevant if:
 *     - the body of R is not false and
 *     - the head of R intersect I+ contains atoms from Rw(I+) and
 *     - R is not positive recursive
 */
bool
ModelChecker::isRelevantForTransformation(
    WaspRule& r,
    Interpretation& I )
{
    trace( mc, 3, "Checking the body.\n" );

    if( r.isConstraint() || r.isBodyFalse() )
        return false;    
    
    unsigned int headSize = r.getNumberOfAtomsInTheHead();
    unsigned int bodySize = r.getNumberOfLiteralsInTheBody();
    
//    if( headSize > 0 )
//    {
//        for( unsigned int i = 0; i < bodySize; i++ )
//        {
//            Literal& l = r.getBodyLiteral( i );
//            if( I.isFalsePropagated( l ) )
//            {
//                trace( mc, 3, "The body is false: the rule is irrelevant.\n" );            
//                return false;
//            }
//        }
//    }
//    else
//    {
//        return false;
//    }

    assert( r.getIdFirstFalseLiteral() != 0 );
    assert( I.isTruePropagated( r.getIdFirstFalseLiteral() ) );
    if( !containedInUnfoundedSet( r.getIdFirstFalseLiteral() ) )
        return false;
    
    if( r.getIdSecondFalseLiteral() != 0 )
    {
        assert( I.isTruePropagated( r.getIdSecondFalseLiteral() ) );
        if( !containedInUnfoundedSet( r.getIdSecondFalseLiteral() ) )
            return false;
        trace( mc, 3, "Checking the head.\n" );
        for( unsigned int i = 0; i < headSize; i++ )
        {
            Atom& a = *( r.getHeadAtom( i ) );
            AtomId id = a.getId();
            if( ( !containedInUnfoundedSet( id ) ) && I.isTruePropagated( id ) )
            {
                trace( mc, 3, "The head contains a true atom that is not in the unfounded set.\n" );
                return false;
            }
            
            unordered_set< WaspRule * >& recursiveRules = atomRulesRecursiveMap[ id ];
            if( recursiveRules.find( &r ) != recursiveRules.end()  )
                return false;

//            for( unsigned int j = 0; j < bodySize; j++ )
//            {
//                Literal& lit = r.getBodyLiteral( j );
//                if( lit.isPositive() )
//                {
//                    if( a == *( lit.getAtom() ) )
//                        return false;
//                }
//            }
        }
    }
    else
    {
        unordered_set< WaspRule * >& recursiveRules = atomRulesRecursiveMap[ r.getIdFirstFalseLiteral() ];
        if( recursiveRules.find( &r ) != recursiveRules.end()  )
            return false;
        
//        for( unsigned int j = 0; j < bodySize; j++ )
//        {
//            Literal& lit = r.getBodyLiteral( j );
//            if( lit.isPositive() )
//            {
//                if( r.getIdFirstFalseLiteral() == lit.getAtomId() )
//                    return false;
//            }
//        }
    }    
    return true;   
}

bool
ModelChecker::build_simple_minisat_instance(
    Solver& minisat,
    Interpretation& I )
{
    startTimer( buildFormula );
    vec< Lit > veclits;

    trace( mc, 3, "Adding the disjunction containing the fixpoint of R.\n" );
    trace( mc, 3, "It is made up of the atoms:" );

    RuleContainer< WaspRule* > rules;
    unordered_set< WaspRule* > inserted;
    
    /*
    * First of all, the big disjunction that force the unfounded set to
    * be non empty.
    */
//    for( set< AtomId >::iterator it = unfoundedSet.begin(); it != unfoundedSet.end(); ++it )
    for( unsigned int it = 0; it < unfoundedSet.size(); it++ )
    {
        trace( mc, 3, " %s", program->atomToString( unfoundedSet[ it ] ).c_str() );
        addPositiveInVecLits( minisat, veclits, unfoundedSet[ it ] );
        
        Atom& a = program->getAtom( unfoundedSet[ it ] );
        for( unsigned int j = 0; j < a.getNegativeHeadWaspRules().size(); j++ )
        {
            WaspRule* w = a.getNegativeHeadWaspRules()[ j ];
            
            if( inserted.insert( w ).second && isRelevantForTransformation( *w, I ) )
                rules.push_back( w );
        }
    }
    trace( mc, 3, ".\n" );

    minisat.addClause( veclits );
    veclits.clear();
    
    //Then we try to generate the formula
//    RuleContainer< WaspRule* >& rules = program->getWaspRules();
    
    bool transformationEmpty = true;

    unsigned int rulesSize = rules.size();
    //Iterate over rules
    for( unsigned int it = 0; it < rulesSize; it++ )
    {
        WaspRule& rule = *( rules[ it ] );
        trace( mc, 3, "Checking the rule %s", program->waspRuleToString( rule ).c_str() );

        trace( mc, 3, "The rule is relevant for transformation: constructing the clause.\n" );
        trace( mc, 3, "The clause is made up of literals:" );

        unsigned int headSize = rule.getNumberOfAtomsInTheHead();
        for( unsigned i = 0; i < headSize; i++ )
        {
            int val = rule.getHeadAtom( i )->getId();
            /*
                * This condition was modified in order to deal with partial interpretations.
                */
            //if( !I.isFalsePropagated( val ) )
            if( I.isTruePropagated( val ) )
            {
                addNegativeInVecLits( minisat, veclits, val );
                trace( mc, 3, " not %s", program->atomToString( val ).c_str() );
            }
        }

        unsigned int bodySize = rule.getNumberOfLiteralsInTheBody();
        for( unsigned i = 0; i < bodySize; i++ )
        {
            int val = rule.getBodyLiteral( i ).getId();
            if( val > 0 && unfoundedSet.contains( val ) )//unfoundedSet.find( val ) != unfoundedSet.end() ) //Only if body literal is contained in unfoundedSet
            {
                trace( mc, 3, " %s", program->atomToString( val ).c_str() );
                addPositiveInVecLits( minisat, veclits, val );
            }
        }
        trace( mc, 3, ".\n" );
        trace( mc, 3, "Pushing clause in the formula.\n" );
        minisat.addClause( veclits );
        veclits.clear();

        transformationEmpty = false;
    }
    stopTimer( buildFormula );
    return !transformationEmpty;
}

/*
 * This function converts an array of int in a set of minisat clause;
 */
void
ModelChecker::readClause(
    int* in,
    unsigned size,
    Solver& S,
    vec< Lit >& lits )
{
    int var;
    lits.clear();
    for( unsigned i = 0; i < size; i++ )
    {
        if( in[ i ] == 0 )
            break;
        var = abs( in[ i ] ) - 1;
        while ( var >= S.nVars() )
            S.newVar();
        lits.push( ( in[ i ] > 0 ) ? Lit( var ) : ~Lit( var ) );
    }
}

void
ModelChecker::addPositiveInVecLits(
    Solver& S,
    vec< Lit >& lits,
    unsigned int value )
{
    if( value != 0 )
    {
        unsigned int var = value - 1;
        while ( var >= S.nVars() )
            S.newVar();
        lits.push( Lit( var ) );
    }
}

void
ModelChecker::addNegativeInVecLits(
    Solver& S,
    vec< Lit >& lits,
    unsigned int value )
{
    if( value != 0 )
    {
        unsigned int var = value - 1;
        while ( var >= S.nVars() )
            S.newVar();
        lits.push( ~Lit( var ) );
    }
}

/*
 * A set of atoms X is unfounded with respect to an interpretation I iff for each rule
 * whose head contains at least one atom from X, the three following condition holds:
 *
 *  - a positive body literal is in X.
 *  - the body is false w.r.t. I.
 *  - a head atom which is not in X is true in I.
 */
bool
ModelChecker::isUnfoundedSet(
    Interpretation& I )
{
    trace( mc, 3, "Checking if the previous unfounded set is still unfounded after backtracking.\n" );

    for( unsigned int it = 0; it < unfoundedSet.size(); it++ )
    {
        AtomId unfounded = unfoundedSet[ it ];
        if( !I.isTruePropagated( unfounded ) )
            unfoundedSet.erase( unfounded );
    }
    
    if( unfoundedSet.size() == 0 )
        return false;
    
    bool positiveBodyCond = false, falseBodyCondition = false, multipleTrueInHeadCondition = false;

    //Iterate over the unfounded set
//    for( set< AtomId >::iterator it = unfoundedSet.begin(); it != unfoundedSet.end(); it++ )
    for( unsigned int it = 0; it < unfoundedSet.size(); it++ )
    {
        AtomId unfounded = unfoundedSet[ it ];
//        if( !I.isTruePropagated( unfounded ) )
//            return false;
        
        const RuleContainer< WaspRule* >& rules = program->getAtom( unfounded ).getNegativeHeadWaspRules();
        unsigned int rulesSize = rules.size();
        
        for( unsigned int rh = 0; rh != rulesSize; rh++ )
        {
            positiveBodyCond = false, falseBodyCondition = false, multipleTrueInHeadCondition = false;
            WaspRule& r = *( rules[ rh ] );

            trace( mc, 3 , "Considering rule %s.\n", program->waspRuleToString( r ).c_str() );

            unsigned int bodySize = r.getNumberOfLiteralsInTheBody();
            for( int at = 0; at != bodySize; at++ )
            {
                int val = r.getBodyLiteral( at ).getId();
                if( val > 0 && unfoundedSet.contains( val ) )//unfoundedSet.find( val ) != unfoundedSet.end() )
                {
                    positiveBodyCond = true;
                    break;
                }
            }

            //If the first condition does not hold check the other two
            if( !positiveBodyCond )
            {
                for( int at = 0; at != bodySize; at++ )
                {
                    if( I.isFalsePropagated( r.getBodyLiteral( at ) ) )
                    {
                        falseBodyCondition = true;
                        break;
                    }
                }

                if( !falseBodyCondition )
                {                    
                    unsigned int headSize = r.getNumberOfAtomsInTheHead();
                    if( headSize > 1 )
                    {
                        for( unsigned int h2 = 0; h2 < headSize; h2++ )
                        {
                            AtomId atomId = r.getHeadAtom( h2 )->getId();
                            if( I.isTruePropagated( atomId ) && !unfoundedSet.contains( atomId ) )// unfoundedSet.find( atomId ) == unfoundedSet.end() )
                            {
                                multipleTrueInHeadCondition = true;
                                break;
                            }
                        }
                    }
                }//Third condition

            }//second condition

            if( !falseBodyCondition && !positiveBodyCond && !multipleTrueInHeadCondition )
                return false;
        }//for each rule with it in the head
    }
    return true;
}
