/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/*
 * ModelChecker.h
 *
 *  Created on: 02/dic/2010
 *      Author: marco
 */

#ifndef MODELCHECKER_H_
#define MODELCHECKER_H_

#include "Interpretation.h"
#include "Program.h"
#include <iostream>
#include <set>
using namespace std;

#include "minisat/Solver.h"
#include "UnorderedSet.h"

using namespace wasp;

class ModelChecker {
    
    public:

        inline ModelChecker( Program& );

        inline ~ModelChecker();

       /**
        * This function perform the fixpoint of the R operator.
        *
        * @param I the interpretation.
        * 
        */
        void fixPointoperatorR( Interpretation& I );

       /**
        * This function takes a model as input and determines if it is stable.
        *
        * @param I the model.
        * 
        * @return a boolean that is true if the model is stable, false otherwise.
        */
        bool isStable( Interpretation& I );

       /**
        * @return the last unfounded set calculated.
        */
        inline const UnorderedSet< AtomId >& getUnfoundedSet() const;

       /**
        * 
        * Returns the size of the unfounded set.
        * 
        * @return the size of the last unfounded set calculated.
        */
        inline const unsigned int getUnfoundedSetSize() const;

       /**
        *
        * This function checks whether the previous unfoundedSet is still unfounded w.r.t. I.
        * 
        * @paramn I an interpretation.
        * 
        * @return true if unfoundedSet is an unfounded set w.r.t. I, false otherwise.
        */
        bool isUnfoundedSet( Interpretation& I );

    private:

        Program* program;

       /**
        * A set containing the unfounded atoms.
        */
        UnorderedSet< AtomId > unfoundedSet;

        unordered_map< AtomId, unordered_set< WaspRule* > > atomRulesRecursiveMap;
        
       /**
        * Check if an atom A is contained in unfoundedSet.
        *
        * @param atomId the id of the atom A.
        * 
        * @return true if A is contained in the unfoundedSet, false otherwise.
        */
        inline bool containedInUnfoundedSet( AtomId atomId );

       /**  
        * This is used for determine the max number of variable that will be
        * placed in the SAT formula. As it indicate the larger index of a literal
        * seen in unfoundedSet, it is an upper bound for the number of variable.
        */
//        unsigned int maxIndexOfVariableSeen;

       /**
        * This function performs the R operator.
        *
        * @return true if the unfoundedSet has been modified, false otherwise.
        */
        bool operatorR( Interpretation& I );

       /**
        *
        * Given an interpretation I, initialize the truth values of the unfounded set
        * basing on I. Two possible (TRUE/FALSE) values are assigned to literals.
        * 
        */
        void initializeUnfoundedSet( Interpretation& I );

       /**
        * This function transform the relevant part of the program into a SAT
        * Formula, than perform a call to minisat to determine if the formula is
        * satisfiable. If it is, the input interpretation, which is included in
        * the formula, is an unfounded set. Input interpretation considered is the
        * result of the fixPoint of the R operator.
        *
        * @return bool that tells if the formula is satisfiable.
        */
        bool performSATCheck( Interpretation& I );

       /**
        * This function converts an array of int in a set of minisat clause.
        * 
        * @param in the array of int.
        * @param size the size of the array.
        * @param S the SAT solver.
        * @param lits the vector of Literals.
        */
        void readClause( int* in, unsigned size, Solver& S, vec< Lit >& lits );

       /**
        * This function builds the SAT formula.
        * 
        * @param minisat the SAT solver.
        * @param I the interpretation.
        * 
        * 
        */
        bool build_simple_minisat_instance( Solver& minisat, Interpretation& I );

       /**
        * This function determines if a rule R will become a clause.
        * 
        * @param rule the rule R.
        * @param I the interpretation.
        * 
        * @return true if R will become a clause, false otherwise.
        */
        bool isRelevantForTransformation( WaspRule& rule, Interpretation& I );
        
        void addPositiveInVecLits( Solver& S, vec< Lit >& lits, unsigned int value );
        
        void addNegativeInVecLits( Solver& S, vec< Lit >& lits, unsigned int value );
};

ModelChecker::ModelChecker(
    Program& p )
{
    program = &p;
}

ModelChecker::~ModelChecker()
{
}

const UnorderedSet< AtomId >&
ModelChecker::getUnfoundedSet() const
{
    return unfoundedSet;
}

const unsigned int
ModelChecker::getUnfoundedSetSize() const
{
    return unfoundedSet.size();
}

bool
ModelChecker::containedInUnfoundedSet(
    AtomId a )
{
    return unfoundedSet.contains( a );
}

#endif /* MODELCHECKER_H_ */