/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "ModelGenerator.h"
#include "AtomUT.h"
#include "minisat/Sort.h"
#include "Util.h"
#include "AggregateAtomGreater.h"
#include "AggregateAtomDoubleGuards.h"
#include "heuristics/FirstUndefinedHeuristic.h"
#include "heuristics/BerkminHeuristic.h"
#include "heuristics/BerkminHeuristicTwo.h"
#include "heuristics/BerkminHeuristicTwoLight.h"
#include "heuristics/BerkminHeuristicMoms.h"
#include "heuristics/RandomHeuristic.h"
#include "heuristics/PrelookaheadHeuristic.h"
#include "heuristics/PrelookaheadHeuristicLight.h"
#include "heuristics/MixedHeuristic.h"
#include "heuristics/BerkminHeuristicBalancedMO.h"
#include "heuristics/BerkminRandomHeuristic.h"
#include "heuristics/MixedHeuristicPrelookahead.h"
#include "heuristics/MixedHeuristicPrelookaheadRestart.h"
#include "LearnedWaspRuleFromModel.h"
#include "heuristics/BerkminHeuristicLimitedActivity.h"
#include "heuristics/BerkminHeuristicWeakConstraints.h"
#include "outputBuilders/CompetitionOutputBuilder.h"
#include "outputBuilders/QueryCompetitionOutputBuilder.h"
#include "outputBuilders/QueryOutputBuilder.h"
#include "outputBuilders/SilentOutputBuilder.h"
#include "outputBuilders/WaspOutputBuilder.h"
#include "outputBuilders/ThirdCompetitionOutputBuilder.h"
#include <cmath>
#include <unordered_set>
#include <vector>
#include <set>
#include <bits/stl_stack.h>
#include <stack>
#include <bits/stl_set.h>
#include <bits/stl_vector.h>

extern int COMPLETED;
extern int EXIT_CODE;

List< RuleId > uselessConstraints;
unordered_set< RuleId > uselessConstraintsSet;

Interpretation* ModelGenerator::currentInterpretation = NULL;

ModelGenerator::ModelGenerator(
    int heuristicId )
    : stackInvolvedForGus( reasonsForGus )
{
    startTimer( startup );
    InputHandler in( &program );
    in.build();
    stopTimer( startup );
    init( heuristicId );
}

ModelGenerator::ModelGenerator(
    istream& input )
    : stackInvolvedForGus( reasonsForGus )
{    
    startTimer( startup );
    InputHandler in( &program );
    in.build( input );
    stopTimer( startup );
    init();
}

ModelGenerator::~ModelGenerator()
{
    if( mc )
    {
        delete mc;
        mc = NULL;
    }
    
    if( decisionHeuristic )
    {
        delete decisionHeuristic;
        decisionHeuristic = NULL;
    }
    
    if( outputBuilder )
    {
        delete outputBuilder;
        outputBuilder = NULL;
    }
}

int
computeRestartNumber( 
    int i )
{
    assert( i > 0 );
    static unsigned int a[] = { 1,3,7,15,31,63,127,255,511,1023, 2047, 4095, 8191, 16383, 32767 };
    //A binary search to compute
    int first = 0;
    int last = 32;
    int mid = ( last + first ) / 2;

    while( first < last )
    {
        if( ( i >> mid ) == 0 )
        {
            last = mid;
        }
        else
        {
            first = mid + 1;
        }
        mid = ( last + first ) / 2;
    }

    if( a[ first - 1 ] == i )
    {
        int pos = pow(2, ( first - 1 ) );
        return pos;
    }
    else
    {
        int newValue = i - pow( 2, ( first - 1 ) ) + 1 ;
        return computeRestartNumber( newValue );
    }
}

void
ModelGenerator::init(
    int heuristicId )
{
    COMPLETED = 1;
    
    switch( heuristicId )
    {
    case BERKMIN_HEURISTIC:
        
        if( program.getAggregateRepresentingWeakConstraints() == NULL )
            decisionHeuristic = new BerkminHeuristic( this, program.getNumberOfAtoms() );
        else
            decisionHeuristic = new BerkminHeuristicWeakConstraints( this, program.getNumberOfAtoms() );
        break;
        
    case BERKMIN_MOMS:
        decisionHeuristic = new BerkminHeuristicMoms( this, program.getNumberOfAtoms() );
        break;

    case BERKMIN_TWO:
        decisionHeuristic = new BerkminHeuristicTwo( this, program.getNumberOfAtoms() );
        break;
        
    case BERKMIN_TWOLIGHT:
        decisionHeuristic = new BerkminHeuristicTwoLight( this, program.getNumberOfAtoms() );
        break;
        
    case FIRST_UNDEFINED_HEURISTIC:
        decisionHeuristic = new FirstUndefinedHeuristic( this );
        break;
        
    case RANDOM_HEURISTIC:
        decisionHeuristic = new RandomHeuristic( this, program.getNumberOfAtoms() );
        break;
        
    case PRELOOKAHEAD_HEURISTIC:
        decisionHeuristic = new PrelookaheadHeuristic( this, program.getNumberOfAtoms() );
        break;
        
    case PRELOOKAHEADLIGHT_HEURISTIC:
        decisionHeuristic = new PrelookaheadHeuristicLight( this, program.getNumberOfAtoms() );
        break;
        
    case MIXED_HEURISTIC:
        decisionHeuristic = new MixedHeuristic( this, new FirstUndefinedHeuristic( this ), new BerkminHeuristic( this, program.getNumberOfAtoms() ) );
        break;
        
    case MIXED_HEURISTIC_PRELOOKAHEAD:
        decisionHeuristic = new MixedHeuristicPrelookahead( this, new FirstUndefinedHeuristic( this ), new BerkminHeuristic( this, program.getNumberOfAtoms() ) );
        break;
        
    case MIXED_HEURISTIC_RESTARTLOOKAHEAD:
        decisionHeuristic = new MixedHeuristicPrelookaheadRestart( this, new FirstUndefinedHeuristic( this ), new BerkminHeuristic( this, program.getNumberOfAtoms() ) );
        break;
        
    case BALANCED_MO:
        decisionHeuristic = new BerkminHeuristicBalancedMO( this, program.getNumberOfAtoms() );
        break;
        
    case BERKMIN_RANDOM:
        decisionHeuristic = new BerkminRandomHeuristic( this, program.getNumberOfAtoms() );
        break;
        
    case BERKMIN_LIMITED_ACTIVITY:
        decisionHeuristic = new BerkminHeuristicLimitedActivity( this, program.getNumberOfAtoms() );
        break;
        
    case LOOKAHEAD_HEURISTIC:
        cerr << ERRORHEURISTIC << endl;
        exit( 0 );
        break;

    case LOOKAHEADBERKMIN_HEURISTIC:
        cerr << ERRORHEURISTIC << endl;
        exit( 0 );
        break;

    default:
        cerr << ERRORHEURISTIC << endl;
        exit( 0 );
    }

//    if( options.doSubsumption )
//    {
//        selfSubsume = &ModelGenerator::selfSubsumption;
//    }
//    else
//    {
//        selfSubsume = &ModelGenerator::selfSubsumptionDoNothing;
//    }

    programSize = program.getWaspRulesSize();
    unsigned int size = program.getNumberOfAtoms();
    interpretation.init( size );
    currentChoice.first = 1;
    currentChoice.second = TRUE;
    mc = new ModelChecker( program );
    models = 0;
    constraintIncrement = 1;
    constraintDecrement = 1/0.999;
    conflictCount = 0;
    conflictCountForRestart = 0;
    countRestart = 1;
    if( options.geometricRestarts )
        nextRestartValue = options.restartsNumbers;
    else
        nextRestartValue = computeRestartNumber( countRestart ) * options.restartsNumbers;
    conflictLiteral.first = 0;
    conflictLiteral.second = UNDEFINED;
    conflictAtomId = 0;
    conflictAtomIdTruthvalue = UNDEFINED;

    conflictImplicantIdType.first = { NULL };
    conflictImplicantIdType.second = 0;
    callComputeGUS = true;

    currentInterpretation = &interpretation;        
    if( options.competitionOutput )
    {
        if( program.hasQuery() )
        {
            outputBuilder = new QueryCompetitionOutputBuilder();
        }
        else
        {
            outputBuilder = new CompetitionOutputBuilder();
        }
    }
    else if( options.thirdCompetitionOutput )
    {
        outputBuilder = new ThirdCompetitionOutputBuilder();
    }        
    else if( program.hasQuery() )
    {
        outputBuilder = new QueryOutputBuilder();        
    }
    else if( options.silentMode )
    {   
        outputBuilder = new SilentOutputBuilder();
    }
    else
    {
        outputBuilder = new WaspOutputBuilder();
    }
}

void
ModelGenerator::mg()
{
    
    if( program.isErrorFromGrounding() )
        return;

    if( program.isMessageFromGrounding() )
    {
        assert( program.getType() == QUERY_BRAVE || program.getType() == QUERY_CAUTIOUS );
        assert( program.getValue() == QUERY_FALSE || program.getValue() == QUERY_TRUE );
        
        if( options.competitionOutput )
        {
            if( program.getValue() == QUERY_FALSE )
            {
                cout << "no." << endl;
                EXIT_CODE = 20;
            }
            else
            {
                cout << "yes." << endl;
                EXIT_CODE = 10;
            }
        }
        else
        {
            if( program.getType() == QUERY_BRAVE )
            {
                if( program.getValue() == QUERY_FALSE )
                    cout << "the query is bravely false." << endl;
                else
                    cout << "the query is bravely true." << endl;
            }
            else
            {
                if( program.getValue() == QUERY_FALSE )
                    cout << "the query is cautiously false." << endl;
                else
                    cout << "the query is cautiously true." << endl;
            }
        }
        return;
    }

    if( options.printProgram )
    {
        program.printProgram();
        return;
    }
    
    statistics( setRulesNumber( program.getWaspRulesSize() ) );
    statistics( setAtomsNumber( program.getNumberOfAtoms() - 1 ) );
    
    if( program.getNumberOfAtoms() == 1 )
    {
        onModelFound();
        return;
    }
    
    if( !options.silentMode )
        cout << WASP_STRING << endl;

    startTimer( initialPropagation );
    bool result = initialPropagation();
    stopTimer( initialPropagation );
    
    if( result )
    {
        #ifndef STATS_OFF
        unsigned int count = 0;
        for( int i = 1; i < program.getNumberOfAtoms(); i++ )
        {
            if( interpretation.isFalsePropagated( program.getAtom( i ) ) || interpretation.isTruePropagated( program.getAtom( i ) ) )
                count++;
        }
        statistics( setInitialPropagationNumber( count ) );
        #endif

        assert( program.getNumberOfAtoms() > 1 );
        
        updateInterpretationWatch();
        if( !interpretation.thereAreUndefinedAtoms() )
        {
            if( checkModel() )
            {
                onModelFound();
                if( program.hasWeakConstraints() )
                {
                    EXIT_CODE = 30;
                    cout << OPTIMUM << endl;
                }
            }
            else
            {
                outputBuilder->onProgramIncoherent();                    
            }
        }
        else
        {
            if( options.simplifyProgram )
            {
                if( program.isHeadCycleFree() )
                {
                    program.reduceProgram();
                    programSize = program.getWaspRulesSize();
                }
            }

            backtracking();
            
            if( models == 0 )
            {
                outputBuilder->onProgramIncoherent();                    
            }
            else
            {
                if( program.hasWeakConstraints() )
                {
                    EXIT_CODE = 30;
                    cout << OPTIMUM << endl;
                }
            }
        }

        if( options.stats )
        {
            statistics( printStatistics() );
        }
    }
    else
    {
        assert( models == 0 );
        outputBuilder->onProgramIncoherent();
        
        if( options.stats )
        {
            statistics( printStatistics() );
        }
    }
}

bool
ModelGenerator::initialPropagation()
{
    trace( mg, 1, "Starting first propagation.\n" );
    trace( mg, 1, "Computing deterministic consequences.\n" );
    computeDeterministicConsequences();
    trace( mg, 1, "Computed deterministic consequences.\n" );    

    const vector< ComponentId >& components = program.getCyclicAndHeadCycleFreeComponents();
    unsigned int componentsSize = components.size();
    
    for( unsigned int i = 0; i < componentsSize; i++ )
    {
        ComponentId componentId = components[ i ];
        Component& component = program.getComponents()[ componentId ];
        
        assert( component.isCyclic() );
        assert( component.isHeadCycleFree() );
        for( unsigned int j = 0; j < component.size(); j++ )
        {
            AtomId atomId = component.getAtom( j );
            if( mapComponentListOfAtomsLostSourcePointer[ componentId ].empty() )
                componentsInTheMap.push_back( componentId );
            
            mapComponentListOfAtomsLostSourcePointer[ componentId ].push_back( program.getAtomPointer( atomId ) );                        
//            atomsLostSourcePointer.push_back( program.getAtomPointer( atomId ) );
        }
//        assert( program.getComponents()[ componentId ].mustBeAddedToGUSVector() );        
//        program.getComponents()[ componentId ].setMustBeAddedToGUSVector( false );
//        componentsToComputeGus.push_back( componentId );
    }
    
    if( !propagationStack.empty() && !propagate() )
        return false;
    else if( propagationStack.empty() )
    {
        computeGUS();
        if( !propagationStack.empty() && !propagate() )
            return false;
    }
        
    if( !decisionHeuristic->onInit() )
        return false;

    trace( mg, 1, "Ending first propagation.\n" );

    return true;
}

void
ModelGenerator::backtracking()
{    
    unsigned int maxModels = options.numModels;

    #ifndef STATS_OFF
    unsigned int loopCount = 0;
    #endif

    while( true )
    {
        assert( models < maxModels || program.hasWeakConstraints() );
        updateInterpretationWatch();
        assert( interpretation.thereAreUndefinedAtoms() || !backtrackingAtomStack.empty() );
        
        /*
        static unsigned int PROVA = 0;
        static time_t PROVA_TIME = time( 0 );
        
        if( ++PROVA > options.statsTimer ) {
            cerr << "PROVA END!" << endl;
            return;
        }
        else if( ++PROVA % 10000 == 0 )
        {
            cout << PROVA << " " << program.getLearnedWaspRules().size() <<  " " << ( time( 0 ) - PROVA_TIME ) << endl;
        }
        //*/
        
        #ifndef STATS_OFF
        ++loopCount;
        if( loopCount == options.statsTimer )
        {
            if( options.stats )
            {
                statistics( printStatistics() );
            }
            
            if( options.statsExit )
            {
                return;
            }
            loopCount = 0;
        }
        #endif

        int aux;
        if( propagationStack.empty() )
        {
            makeAChoice();
            assert( program.getAtom( currentChoice.first ).getTypeOfDerivation() == TYPEDETERMINISTIC );
            aux = propagate( currentChoice );        
        }
        else
        {
            aux = propagate() ? 1 : -1;
        }
           
        switch( aux )
        {   
        case -1:
        {
            statistics( incrementCalculateFirstUipInvocationsNumber() );
            if( backtrackingStack.getCurrentLevel() == 0 )
            {
                trace( mg, 1, "Conflict at level 0. Stop computation.\n" );
                return;
            }

            AtomId firstUip = calculateFirstUIP();               
            assert( firstUip != 0 );
            unsigned int levelToBacktrack;
            LearnedWaspRule* learnedWaspRulePointer = learnRule( firstUip, levelToBacktrack );
            assert( learnedWaspRulePointer != NULL );
            
            if( levelToBacktrack == FOUND_INCOHERENCE )
            {
                trace( mg, 1, "Learned an empty constraint. Stop computation.\n" );
                return;
            }

            decrementConstraintActivity();
            ++conflictCount;
            ++conflictCountForRestart;
            
            if( learnedWaspRulePointer->size() == 1 )
            {
                conflictCountForRestart = 0;
                countRestart++;
                statistics( incrementRestartNumber() );
                
                if( options.geometricRestarts )
                    nextRestartValue *= 1.5;
                else
                    nextRestartValue = computeRestartNumber( countRestart ) * options.restartsNumbers;
            }
            
            if( conflictCount == 100 )
            {
                startTimer( divideCountersTimer );                
                conflictCount = 0;
                decisionHeuristic->onReachedANumberOfConflicts();
                stopTimer( divideCountersTimer );
            }
            if( conflictCountForRestart == nextRestartValue && options.heuristic != LOOKAHEAD_HEURISTIC && options.heuristic != LOOKAHEADBERKMIN_HEURISTIC )
            {
                conflictCountForRestart = 0;
                countRestart++;
                statistics( incrementRestartNumber() );
                if( !restart() )
                {
                    trace( mg, 1, "Found a conflict after the restart. Stop computation.\n" );
                    return;
                }
                break;
            }
            
            if( options.deleteMode == DELETE_WHEN_LEARN_A_NEW_CONSTRAINT )
                onLearningAWaspRule();
            else if( options.deleteMode == RESTART_BASED_POLICY )
                onLearningAWaspRuleRestartBasedPolicy();
            
            unroll( levelToBacktrack );

            unsigned levelChoice;
            do
            {
                levelChoice = backtrackingLevelAtomStack.back();
                backtrackingAtomStack.pop_back();
                backtrackingLevelAtomStack.pop_back();         
            } while( levelToBacktrack < levelChoice );

            if( !backtrackingAtomStack.empty() )
            {
                currentChoice.first = backtrackingAtomStack.back().first;                
                currentChoice.second = backtrackingAtomStack.back().second;                
            }

            LearnedWaspRule& learnedWaspRule = *learnedWaspRulePointer;
            assert( interpretation.isUndefined( learnedWaspRule.getFirstWatchValue() ) );
            
            unsigned int value;
            if( learnedWaspRule.size() > 1 )
            {
                assert( interpretation.isTruePropagated( learnedWaspRule.getSecondWatchValue() ) );
                value = updateSecondWatch( learnedWaspRule );
            }
            else
            {
                value = learnedWaspRule.getFirstWatch();
            }

//            Literal* undefinedLiteral = learnedWaspRule.onlyOneUndefined();

            assert( value != MAXVALUE );

            Literal& literal = learnedWaspRule.getLiteral( value );
            assert( literal == learnedWaspRule.getFirstWatchValue() );

            Atom& a = *( literal.getAtom() );
            statistics( incrementCountUnit() );
            
            addInvolvedRules( a, { learnedWaspRulePointer }, TYPEALL );
            trace( mg, 1, "Inserting UIP: %s in stack with truth value %s.\n", a.getAtomName().c_str(), literal.isPositive() ? "false" : "true" );
            pushAtomInStack( a, literal.isPositive() ? FALSE : TRUE );                                

//            addAddressInBackTrackingStack( learnedWaspRule.getLockedAddress() );
            backtrackingStack.push( &( learnedWaspRule.getLockedAddress() ) );
            learnedWaspRule.setLocked();
            
            assert( !propagationStack.empty() );            
            updateInterpretationWatch();
            break;
        }

        case 1:
            if( options.deleteMode == DELETE_WHEN_MODEL_IS_FOUND )
                onFindingAModel();

            updateInterpretationWatch();
            if( !interpretation.thereAreUndefinedAtoms() )
            {
                if( checkModel() )
                {
                    statistics( incrementNumModels() );
                    ++models;
                    this->printInterpretation();
                    
                    cout << options.modelSeparator;
                        
                    // All models have been found. Stop the backtracking now!
                    // TODO: Maybe we want to clean stacks and similar stuff.
                    // TODO: Fix when there are weak constraints.
                    if( models >= maxModels && !program.hasWeakConstraints() )
                    {
                        trace( mg, 1, "All models have been found.\n" );
                        return;
                    }      
                    
                    if( backtrackingAtomStack.empty() )
                        return;
                    
                    if( !program.hasWeakConstraints() )
                        learnRuleModelFound();
                    else
                    {
                        if( !updateAggregateRepresentingWeakConstraints() )
                            return;
                    }
                }
                else
                {
                    trace( mg, 1, "Model Checker Failure.\n" );
                    statistics( incrementNumCheckModelFailures() );
                    
                    if( backtrackingAtomStack.empty() )
                        return;
                    
                    if( !learnRuleModelCheckerFailure() )
                        return;
                }                
            }
            break;
            
        default:
            assert( 0 );
        }
    }
}

void
updateMax( 
    unsigned int atomRecursionLevel,
    unsigned int position,
    unsigned int& maxLevel, 
    unsigned int& maxPosition, 
    unsigned int& secondMaxLevel, 
    unsigned int& secondMaxPosition )
{
    if( atomRecursionLevel > maxLevel )
    {
        secondMaxLevel = maxLevel;
        secondMaxPosition = maxPosition;

        maxLevel = atomRecursionLevel;
        maxPosition = position;
    }
    else if( atomRecursionLevel > secondMaxLevel )
    {
        secondMaxLevel = atomRecursionLevel;
        secondMaxPosition = position;
    }
}

bool
ModelGenerator::learnRuleModelCheckerFailure()
{
    trace( mg, 1, "Learning WaspRule from after a model checker failure.\n" );
    bool changedSomething;
	do
	{
		changedSomething = false;
		const UnorderedSet< AtomId >& unfoundedSet = mc->getUnfoundedSet();        
        
		vector< pair< AtomId, TruthValue > > loopFormula;
		unordered_set< AtomId > inserted;
		
		unsigned int maxPosition = 0;
		unsigned int maxLevel = 0;
		    
		unsigned int secondMaxPosition = 0;
		unsigned int secondMaxLevel = 0;    
		
		for( unsigned int it = 0; it < unfoundedSet.size(); it++ )
		{
		    AtomId atomInTheUnfoundedSet = unfoundedSet[ it ];
		    Atom& a = program.getAtom( atomInTheUnfoundedSet );		    
            
            trace( mg, 2, "Processing atom %s which is in the unfounded set.\n", a.getAtomName().c_str() );
            
            if( inserted.insert( atomInTheUnfoundedSet ).second )
		    {
		        if( a.getRecursionLevel() != 0 )
		        {
                    trace( mg, 3, "Inserting atom %s in the loop formula.\n", a.getAtomName().c_str() );
		            updateMax( a.getRecursionLevel(), loopFormula.size(), maxLevel, maxPosition, secondMaxLevel, secondMaxPosition );
		            loopFormula.push_back( pair< AtomId, TruthValue >( atomInTheUnfoundedSet, TRUE ) );
		        }
		    }		    
            
		    const RuleContainer< WaspRule* >& headWaspRules = a.getNegativeHeadWaspRules();
		    unsigned int headWaspRulesSize = headWaspRules.size();        
		    
		    for( unsigned int i = 0; i < headWaspRulesSize; i++ )
		    {
		        assert( headWaspRules[ i ] != NULL );
		        
		        bool skipRule = false;
		        WaspRule& w = *( headWaspRules[ i ] );
                
                trace( mg, 3, "Processing WaspRule: %s.\n", program.waspRuleToString( w ).c_str() );
		        
		        for( unsigned int j = 0; j < w.getNumberOfLiteralsInTheBody(); j++ )
		        {
		            Literal& literal = w.getBodyLiteral( j );
		            AtomId atomId = literal.getAtomId();
		            if( literal.isPositive() && unfoundedSet.contains( atomId ) )//unfoundedSet.find( atomId ) != unfoundedSet.end() )
		            {
                        trace( mg, 3, "Skipped WaspRule.\n" );
		                skipRule = true;
		                break;                    
		            }
		        }
		        
		        if( !skipRule )
		        {
		            bool found = false;                    
		            
		            for( unsigned int j = 0; j < w.size() && !found; j++ )
		            {
		                Literal& literal = w.getLiteral( j );
		                
		                assert( literal.getAtom() != NULL );
		                Atom& atom = *( literal.getAtom() );
		                
		                AtomId atomId = atom.getId();
		                //if( unfoundedSet.find( atomId ) == unfoundedSet.end() )
		                if( !unfoundedSet.contains( atomId ) )
		                {
		                    if( interpretation.isFalse( literal ) )
		                    {
		                        if( inserted.insert( atomId ).second && atom.getRecursionLevel() != 0 )
		                        {
		                            TruthValue truth = literal.isPositive() ? FALSE : TRUE;
		                            updateMax( atom.getRecursionLevel(), loopFormula.size(), maxLevel, maxPosition, secondMaxLevel, secondMaxPosition );
		                            loopFormula.push_back( pair< AtomId, TruthValue >( atomId, truth ) );
                                    trace( mg, 3, "Inserting atom %s in the loop formula.\n", atom.getAtomName().c_str() );
		                        }
		                        found = true;
		                    }
		                }
		            }
		            assert( found );
		        }
		    }
		}
        
		if( loopFormula.empty() )
        {
            trace( mg, 1, "Loop formula is empty: return.\n");
		    return false;
        }
		
		assert( maxPosition < loopFormula.size() );
		swap( loopFormula.front(), loopFormula[ maxPosition ] );
		
		if( secondMaxPosition == 0 )
		    secondMaxPosition = maxPosition;
		assert( secondMaxPosition < loopFormula.size() );    
		swap( loopFormula.back(), loopFormula[ secondMaxPosition ] );    
		
		LearnedWaspRule* waspRule = insertConstraintFromLoopFormula( loopFormula );    
		assert( waspRule != NULL );
		LearnedWaspRule& learnedWaspRule = *( waspRule );
        
		assert( learnedWaspRule.size() != 0 );
		
		unsigned int levelToBacktrack;
		
		if( learnedWaspRule.size() == 1 )
		    levelToBacktrack = 1;
		else if( maxLevel == secondMaxLevel )
		    levelToBacktrack = maxLevel;
		else
		    levelToBacktrack = secondMaxLevel + 1;
		
        trace( mg, 1, "Unroll until level %d.\n", levelToBacktrack );
		unroll( levelToBacktrack );    
		
		unsigned levelChoice;
		do
		{
		    levelChoice = backtrackingLevelAtomStack.back();
		    backtrackingAtomStack.pop_back();
		    backtrackingLevelAtomStack.pop_back();         
		} while( levelToBacktrack < levelChoice );

		if( !backtrackingAtomStack.empty() )
		{
		    currentChoice.first = backtrackingAtomStack.back().first;                
		    currentChoice.second = backtrackingAtomStack.back().second;                
		}		
		
		assert( interpretation.isUndefined( learnedWaspRule.getFirstWatchValue() ) );
		
        unsigned int value = NOTHING_TO_DO;
		if( learnedWaspRule.size() > 1 )
		{
		    if( interpretation.isTruePropagated( learnedWaspRule.getSecondWatchValue() ) )
		    {
		        value = updateSecondWatch( learnedWaspRule );
                
                assert( value != WATCH_CHANGED );
		    }
		}
        else
        {
            value = learnedWaspRule.getFirstWatch();
        }

//		Literal* undefinedLiteral = learnedWaspRule.onlyOneUndefined();

		if( value != NOTHING_TO_DO )
		{
		    Literal& literal = learnedWaspRule.getLiteral( value );
		    assert( literal == learnedWaspRule.getFirstWatchValue() );

		    Atom& a = *( literal.getAtom() );
		    statistics( incrementCountUnit() );
		    addInvolvedRules( a, { waspRule }, TYPEALL );

		    trace( mg, 1, "Unit propagation from loop formula: %s in stack with truth value %s.\n", a.getAtomName().c_str(), literal.isPositive() ? "false" : "true" );
		    pushAtomInStack( a, literal.isPositive() ? FALSE : TRUE );

//            addAddressInBackTrackingStack( learnedWaspRule.getLockedAddress() );
            backtrackingStack.push( &( learnedWaspRule.getLockedAddress() ) );
            learnedWaspRule.setLocked();
            
		    assert( !propagationStack.empty() );
		    updateInterpretationWatch();
		}
		else
		{
			if( !options.partialMCoff )
			{        
                if( !backtrackingAtomStack.empty() && !mc->isStable( interpretation ) )
                //if( !backtrackingAtomStack.empty() && mc->isUnfoundedSet( interpretation ) )
				{
                    trace( mg, 1, "Partial model checking.\n" );
                    statistics( incrementNumberOfPartialModelChecks() );
					changedSomething = true;
				}				
			}
		}
	} while( changedSomething );

    return true;
}

LearnedWaspRule*
ModelGenerator::insertConstraintFromLoopFormula(
    vector< pair< AtomId, TruthValue > >& loopFormula )
{        
    unsigned int size = loopFormula.size();

    assert( size > 0 );
    assert( size < options.maxLearnedConstraintSize );
    
    statistics( incrementNumberOfLearnedWaspRules() );
    statistics( incrementSizeOfLearnedWaspRules( size ) );

    unsigned int learnedWaspRulesSize = program.getLearnedWaspRules().size();

    LearnedWaspRule* learnedWaspRule = new LearnedWaspRule( loopFormula.size() );

    for( unsigned int i = 0; i < size; i++ )
    {
        AtomId atom = loopFormula[ i ].first;
        TruthValue truth = loopFormula[ i ].second;
        
        assert( atom > 0 && atom < program.getNumberOfAtoms() );

        Atom* atomPointer = program.getAtomPointer( atom );        
        assert( atomPointer != NULL );
        Atom& a = *atomPointer;
        
        decisionHeuristic->onLearnedWaspRule( atom, truth == TRUE );
        if( truth == TRUE )
        {            
            learnedWaspRule->addLiteral( Literal( atomPointer ) );
        }
        else
        {
            assert( truth == FALSE );
            learnedWaspRule->addLiteral( Literal( atomPointer, true ) );
        }
    }
    
    WatchedList< WaspRule* >::iterator it1 = learnedWaspRule->getFirstWatchValue().addWatchedConstraint( learnedWaspRule );
    learnedWaspRule->setIteratorFirstWatch( it1 );
    WatchedList< WaspRule* >::iterator it2 = learnedWaspRule->getSecondWatchValue().addWatchedConstraint( learnedWaspRule );
    learnedWaspRule->setIteratorSecondWatch( it2 );
    
    //Add the learned constraint.
    program.addLearnedWaspRule( learnedWaspRule );
    
    //WARNING: We need this in order to avoid the deletion of new constraints.
    backtrackingStack.push( &( learnedWaspRule->getLockedAddress() ) );
    learnedWaspRule->setLocked();

    incrementLearnedWaspRuleActivity( learnedWaspRule );       
    trace( mg, 1, "Learning waspRule from model checker failure.\n" );
    trace( mg, 1, "%s", program.waspRuleToString( *learnedWaspRule ).c_str() );
    return learnedWaspRule;
}

bool
ModelGenerator::updateAggregateRepresentingWeakConstraints()
{
    assert( program.hasWeakConstraints() );
    Aggregate* aggregateRepresentingWeakConstraintsPointer = program.getAggregateRepresentingWeakConstraints();
    assert( aggregateRepresentingWeakConstraintsPointer != NULL );
    Aggregate& aggregateRepresentingWeakConstraints = *aggregateRepresentingWeakConstraintsPointer;
    
    aggregateRepresentingWeakConstraints.setMinGuard( aggregateRepresentingWeakConstraints.getCurrentValue() );
    
    do
    {
        unroll( backtrackingStack.getCurrentLevel() );
        
        assert( !backtrackingAtomStack.empty() );
        backtrackingAtomStack.pop_back();
        backtrackingLevelAtomStack.pop_back();
    } while( aggregateRepresentingWeakConstraints.getCurrentValue() >= aggregateRepresentingWeakConstraints.getMinGuard() && !backtrackingAtomStack.empty() );
    
    while( ( aggregateRepresentingWeakConstraints.getRightWatchValue() + aggregateRepresentingWeakConstraints.getCurrentValue() ) >= aggregateRepresentingWeakConstraints.getMinGuard() && !backtrackingAtomStack.empty() )
    {
        unroll( backtrackingStack.getCurrentLevel() );
        backtrackingAtomStack.pop_back();
        backtrackingLevelAtomStack.pop_back(); 
    }
    
//    if( backtrackingAtomStack.empty() )
//        return false;
    
    if( !backtrackingAtomStack.empty() )
    {
        currentChoice.first = backtrackingAtomStack.back().first;                
        currentChoice.second = backtrackingAtomStack.back().second;                
    }
    
    assert( ( backtrackingAtomStack.empty() && backtrackingStack.getCurrentLevel() == 0 ) || backtrackingLevelAtomStack.back() == backtrackingStack.getCurrentLevel() );    
    
    if( aggregateRepresentingWeakConstraints.getCurrentValue() >= aggregateRepresentingWeakConstraints.getMinGuard() )
        return false;
        
    aggregateIsFalse( aggregateRepresentingWeakConstraints );    
    updateInterpretationWatch();
    
    return true;
}

void
ModelGenerator::learnRuleModelFound()
{
    vector< AtomId > choicesVector;
    choicesVector.reserve( backtrackingAtomStack.size() );

    unsigned int btSize = backtrackingAtomStack.size() - 1;

    choicesVector.push_back( backtrackingAtomStack.back().first );
    for( int i = 0; i < btSize; i++ )
    {
        choicesVector.push_back( backtrackingAtomStack[ i ].first );
    }

    assert( program.getAtom( choicesVector[ 0 ] ).getRecursionLevel() == backtrackingStack.getCurrentLevel() );
    assert( backtrackingStack.getCurrentLevel() == 1 || program.getAtom( choicesVector.back() ).getRecursionLevel() == backtrackingStack.getCurrentLevel() - 1 );

    trace( mg, 1, "Learning waspRule from model.\n" );
    LearnedWaspRuleFromModel* learnedWaspRulePointer = insertConstraintFromModel( choicesVector );

    assert( learnedWaspRulePointer != NULL );

    unroll( backtrackingStack.getCurrentLevel() );

    backtrackingAtomStack.pop_back();
    backtrackingLevelAtomStack.pop_back();

    if( !backtrackingAtomStack.empty() )
    {
        currentChoice.first = backtrackingAtomStack.back().first;
        currentChoice.second = backtrackingAtomStack.back().second;
    }

    LearnedWaspRuleFromModel& learnedWaspRule = *learnedWaspRulePointer;
    assert( interpretation.isUndefined( learnedWaspRule.getFirstWatchValue() ) );

    unsigned int value;
    if( learnedWaspRule.size() > 1 )
    {
        assert( interpretation.isTruePropagated( learnedWaspRule.getSecondWatchValue() ) );
        value = updateSecondWatch( learnedWaspRule );
    }
    else
    {
        value = learnedWaspRule.getFirstWatch();
    }

    assert( learnedWaspRule.isLocked() );

    assert( value != MAXVALUE );

    Literal& literal = learnedWaspRule.getLiteral( value );
    assert( literal == learnedWaspRule.getFirstWatchValue() );

    Atom& a = *( literal.getAtom() );
    statistics( incrementCountUnit() );
    addInvolvedRules( a, { learnedWaspRulePointer }, TYPEALL );
    trace( mg, 1, "Inserting from model: %s in stack with truth value %s.\n", a.getAtomName().c_str(), literal.isPositive() ? "false" : "true" );
    pushAtomInStack( a, literal.isPositive() ? FALSE : TRUE );

    assert( !propagationStack.empty() );
    updateInterpretationWatch();    
}

void
ModelGenerator::makeAChoice()
{
    setCurrentChoiceAccordingToAHeuristic();
    assert( currentChoice.second == TRUE || currentChoice.second == FALSE );
    assert( interpretation.isUndefined( currentChoice.first ) );

    nextLevel();

    statistics( addReachedLevel( backtrackingStack.getCurrentLevel() ) );
    statistics( setMaxBacktrackingLevel( backtrackingStack.getCurrentLevel() ) );

    backtrackingAtomStack.push_back( currentChoice );
    backtrackingLevelAtomStack.push_back( backtrackingStack.getCurrentLevel() );

    Atom& atom = program.getAtom( currentChoice.first );

    backtrackingStack.push( &( atom.choice ) );
    atom.setChoice();

    assert( atom.getTypeOfDerivation() == 0 );
    addInvolvedRules( atom, { NULL }, TYPEDETERMINISTIC );                

    trace( mg, 1, "Choosing first undefined %s with truthvalue %s at level %d.\n",
            atom.getAtomName().c_str(),
            currentChoice.second == TRUE ? "true" : "false",
            backtrackingStack.getCurrentLevel() );

    statistics( incrementNumberOfChoices() );
}

void
ModelGenerator::printInterpretationWithUndefined()
{
    vector< Atom* >& atoms = program.getAtoms();

    unsigned int size = program.getNumberOfAtoms();
    cout << "true: {";
    int i = 1;
    for( ; i < size; i++ )
    {
        Atom& atom = *( atoms[ i ] );
        if( interpretation.isTruePropagated( atom ) && !atom.isHidden() )
        {
            cout << atom;
            break;
        }
    }
    i++;
    for( ; i < size; i++ )
    {
        Atom& atom = *( atoms[ i ] );
        if( interpretation.isTruePropagated( atom ) && !atom.isHidden() )
        {
            cout << ", " << atom;
        }
    }
    cout << "}"  << endl;

    cout << "undefined: {";
    i = 1;
    for( ; i < size; i++ )
    {
        Atom& atom = *( atoms[ i ] );
        if( interpretation.isUndefined( atom ) && !atom.isHidden() )
        {
            cout << atom;
            break;
        }
    }
    i++;
    for( ; i < size; i++ )
    {
        Atom& atom = *( atoms[ i ] );
        if( interpretation.isUndefined( atom ) && !atom.isHidden() )
        {
                cout << ", " << atom;
        }
    }
    cout << "}"  << endl;
}

bool
ModelGenerator::propagate()
{   
    assert( !this->propagationStack.empty() );
    startTimer( propagation );
    statistics( incrementNumPropagations() );
    trace( mg, 1, "Starting propagation.\n" );
    trace( mg, 1, "Starting to pop from the stack.\n" );
    
    while( true )
    {
        statistics( incrementNumPropagations() );
        assert( !this->propagationStack.empty() || conflictLiteral.first != 0 );

        //IMPORTANT: THIS MUST BE BEFORE THE OTHER STUFF. SOMETIMES, THE PROPAGATION STACK IS EMPTY, BUT CONFLICT LITERAL IS DIFFERENT FROM 0.
        //THIS HAPPENS WHEN AN ATOM IS DERIVED FALSE IN THE COMPUTEGUS FUNCTION.
        if( conflictLiteral.first != 0 )
        {
            Atom& atom = program.getAtom( conflictLiteral.first );
            trace( mg, 1, "Found inconsistency on atom %s.\n", atom.getAtomName().c_str() );
            propagationStack = stack< AtomTruthValue >();
            
            if( conflictLiteral.second != UNDEFINED )
            {
                backtrackingStack.push( &( interpretation.atoms[ conflictLiteral.first ] ) );
                interpretation.setAtomTruthValue( conflictLiteral.first, conflictLiteral.second );
                
                conflictAtomIdTruthvalue = conflictLiteral.second;
            }
            else
            {
                assert ( interpretation.isFalsePropagated( atom ) || interpretation.isTruePropagated( atom ) );
                if( interpretation.isFalsePropagated( atom ) )
                    conflictAtomIdTruthvalue = FALSE; 
                else
                    conflictAtomIdTruthvalue = TRUE;
            }

            conflictAtomId = conflictLiteral.first;            
            conflictLiteral.first = 0;
            conflictLiteral.second = UNDEFINED;

            stopTimer( propagation );
            return false;
        }

        assert( !this->propagationStack.empty() );
        AtomTruthValue& stackContent = propagationStack.top();
        AtomId atomId = stackContent.first;
        TruthValue truthValue = stackContent.second;

        assert( atomId > 0 && atomId < program.getNumberOfAtoms() );
        Atom& atom = program.getAtom( atomId );

        trace( mg, 1, "Propagating atom %s with id %d and truthvalue %s.\n", atom.getAtomName().c_str(), atomId, truthValue == TRUE ? "true" : "false" );
        
        propagationStack.pop();
        assert( atom.getRecursionLevel() == MAXVALUE );

        backtrackingStack.push( &( atom.recursionLevel ) );
        atom.setRecursionLevel( backtrackingStack.getCurrentLevel() );

        assert( interpretation.isUndefined( atom ) );
        backtrackingStack.push( &( interpretation.atoms[ atomId ] ) );
        interpretation.setAtomTruthValue( atomId, truthValue );

        assert( interpretation.isTruePropagated( atom ) || interpretation.isFalsePropagated( atom ) );

        propagateAtom( atom );

        trace( mg, 1, "Propagated atom %s.\n", atom.getAtomName().c_str() );        
        
        if( !this->propagationStack.empty() )
            continue;
        trace( mg, 1, "Computing GUS.\n" );

        if( callComputeGUS )
            computeGUS();
        if( this->propagationStack.empty() && conflictLiteral.first == 0 )
            break;
    }
    trace( mg, 1, "Ending propagation.\n" );
    stopTimer( propagation );
    assert( conflictLiteral.first == 0 );
    assert( conflictLiteral.second == UNDEFINED );    
    return true;
}

void
ModelGenerator::computeDeterministicConsequences()
{
    const vector< Atom* >& trueSupportedAtoms = program.getTrueSupportedAtoms();
    const vector< Atom* >& trueUnsupportedAtoms = program.getTrueUnsupportedAtoms();
    const vector< Atom* >& falseAtoms = program.getFalseAtoms();
    
    unsigned int trueSupportedSize = trueSupportedAtoms.size();
    
    for( unsigned int i = 0; i < trueSupportedSize; i++ )
    {
        Atom& a = *( trueSupportedAtoms[ i ] );
        a.setSupported();
        if( !interpretation.isTrue( a ) )
        {
            addInvolvedRules( a, { NULL }, TYPEDETERMINISTIC );
            this->pushAtomInStack( a, TRUE );
        }
    }
    
    unsigned int trueUnsupportedSize = trueUnsupportedAtoms.size();
    
    for( unsigned int i = 0; i < trueUnsupportedSize; i++ )
    {
        Atom& a = *( trueUnsupportedAtoms[ i ] );
        if( !interpretation.isTrue( a ) )
        {
            addInvolvedRules( a, { NULL }, TYPEDETERMINISTIC );
            this->pushAtomInStack( a, TRUE );
        }
    }
    
    unsigned int falseSize = falseAtoms.size();
    
    for( unsigned int i = 0; i < falseSize; i++ )
    {
        Atom& a = *( falseAtoms[ i ] );
        if( !interpretation.isFalse( a ) )
        {
            addInvolvedRules( a, { NULL }, TYPEDETERMINISTIC );
            this->pushAtomInStack( a, FALSE );
        }
    }
    
    vector< Atom* >& atoms = program.getAtoms();

    unsigned int size = atoms.size();
    for( unsigned int i = 1; i < size; i++ )
    {
        Atom& atom = *( atoms[ i ] );
        if( atom.isAggregate() )
        {
            AggregateAtomGreater& aggregateAtom = *( static_cast< AggregateAtomGreater* >( &atom ) );
            Aggregate& aggregate = *( aggregateAtom.getAggregate() );
            if( aggregate.getMaxPossibleValue() < aggregate.getMinGuard() )
            {
                addInvolvedRules( atom, { NULL }, TYPEDETERMINISTIC );
                this->pushAtomInStack( atom, FALSE );
            }
            else if( aggregate.getCurrentValue() >= aggregate.getMinGuard() )
            {
                addInvolvedRules( atom, { NULL }, TYPEDETERMINISTIC );
                this->pushAtomInStack( atom, TRUE );
            }
        }
        else if( !atom.canBeSupported() && !interpretation.isFalse( atom ) )//&& !atom.isAggregate() )
        {
            addInvolvedRules( atom, { NULL }, TYPEDETERMINISTIC );
            this->pushAtomInStack( atom, FALSE );
        }
    }
    
    unsigned int choiceRulesSize = program.getChoiceRulesSize();
    for( unsigned i = 0; i < choiceRulesSize; i++ )
    {
        ChoiceRule& choiceRule = *( program.getChoiceRule( i ) );
        if( choiceRule.isBodyTrue() )
        {
            unsigned int headAtomsSize = choiceRule.getNumberOfAtomsInTheHead();
            for( unsigned int j = 0; j < headAtomsSize; j++ )
            {
                assert( choiceRule.getHeadAtom( j ) != NULL );
                choiceRule.getHeadAtom( j )->setSupported();
            }
        }
    }
}

void
ModelGenerator::haveToCalculateGUS(
    Atom& atom,
    Rule* rule )
{
    if( !interpretation.isFalsePropagated( atom ) )
    {
        if( atom.getSourcePointer() == rule )
        {
            trace( us, 2, "Atom %s lost its source pointer.\n", atom.getAtomName().c_str() );
            backtrackingStack.push( &( atom.getFoundedAddress() ) );
            atom.setFounded( false );
            
            assert( !atom.isFounded() );
//            atomsLostSourcePointer.push_back( &atom );
            ComponentId componentId = atom.getComponent();            
            if( mapComponentListOfAtomsLostSourcePointer[ componentId ].empty() )
                componentsInTheMap.push_back( componentId );
            
            mapComponentListOfAtomsLostSourcePointer[ componentId ].push_back( &atom );
        }    
    }
}

unsigned int
ModelGenerator::estimateShortWaspRulesPropagation(
    const Literal& l ) const
{
    int id = l.getId();
    stack< Literal* > next;
    const RuleContainer< WaspRule* >& binaryWaspRules = l.getBinaryWaspRules();
    
    unsigned int count = 0;
    unsigned int size = binaryWaspRules.size();    
    
    for( unsigned int i = 0; i < size; i++ )
    {
        assert( binaryWaspRules[ i ] != NULL );
        WaspRule& w = *( binaryWaspRules[ i ] );
        
        if( w.isUndefined() )
        {
            count++;
            
            if( w.getFirstWatchValue().getId() != id )
                next.push( &w.getFirstWatchValue() );
            else
                next.push( &w.getSecondWatchValue() );
        }
    }
    
    unsigned int maxSteps = 5;
    
    unsigned int k = 0;
    while( k < maxSteps && k < next.size() )
    {
        Literal* current = next.top();
        next.pop();

        const RuleContainer< WaspRule* >& binaryWaspRules = current->getBinaryWaspRulesReverse();
        unsigned int size = binaryWaspRules.size();
        for( unsigned int i = 0; i < size; i++ )
        {
            assert( binaryWaspRules[ i ] != NULL );
            WaspRule& w = *( binaryWaspRules[ i ] );

            if( w.isUndefined() )
            {
                count++;

                if( w.getFirstWatchValue().getId() != current->getId() )
                    next.push( &w.getFirstWatchValue() );
                else
                    next.push( &w.getSecondWatchValue() );
            }
        }
        
        k++;
    }
    return count;
}

bool compareAtoms( const pair< AtomId, unsigned int >& i, const pair< AtomId, unsigned int >& j )
{
    return( i.second > j.second );
}

//IT DOESN'T WORK. TO FIX.
//void
//ModelGenerator::lookaheadHeuristic()
//{
//    startTimer( lookahead );
//
//    startTimer( pt );
//    vector< pair< AtomId, unsigned int > > possTrue;
//    computePT( possTrue );
//    stopTimer( pt );
//
//    int min = MAXVALUE;
//    AtomId id = 0;
//
//    vector< AtomUT > sameValuesOfUT;
//
//    startTimer( chosen );
//    for( int z = 0; z < possTrue.size() && z < 5; z++ )
//    {
//        AtomId atomId = possTrue[ z ].first;
//
//        if( !interpretation.isUndefined( atomId ) )
//            continue;
//
//        Atom& a = program.getAtoms()[ atomId ];
//        TruthValue atomTruthValue = TRUE;
//        TruthValue otherAtomTruthValue = FALSE;
//
//        Literal literal;
//        literal.first = atomId;
//        literal.second = atomTruthValue;
//
//        nextLevel();
//
//        backtrackingStack.push( &( a.choice ) );
//        a.setChoice();
//        this->propagationStack.push( literal );
//
//        startTimer( look2 );
//        bool prop = propagate();
//        stopTimer( look2 );
//
//        if( prop )
//        {
//            int ut = 0, ut2 = 0, ut3 = 0, trueAtoms = 0, count = 0, supportingRulesForTrueAtoms = 0;
//            computeUTs( ut, ut2, ut3, trueAtoms );
//
//            computeSatisfiedRules( count, supportingRulesForTrueAtoms );
//
//            unrollOne();
//
//            nextLevel();
//
//            backtrackingStack.push( &( a.choice ) );
//            a.setChoice();
//
//            literal.second = otherAtomTruthValue;
//            this->propagationStack.push( literal );
//
//            startTimer( look2 );
//            bool prop = propagate();
//            stopTimer( look2 );
//
//            if( prop )
//            {
//                int other_ut = 0, other_ut2 = 0, other_ut3 = 0, other_trueAtoms = 0;
//                computeUTs( other_ut, other_ut2, other_ut3, other_trueAtoms );
//
//                unrollOne();
//
//                if( ( ut + other_ut ) < min )
//                {
//                    sameValuesOfUT.clear();
//                    AtomUT atomToAdd;
//                    atomToAdd.id = atomId;
//                    atomToAdd.ut = ut; atomToAdd.ut2 = ut2; atomToAdd.ut3 = ut3;
//                    atomToAdd.other_ut = other_ut; atomToAdd.other_ut2 = other_ut2; atomToAdd.other_ut3 = other_ut3;
//                    atomToAdd.supportingRulesForTrueAtoms = supportingRulesForTrueAtoms;
//                    atomToAdd.satisfiedRules = count;
//
//                    sameValuesOfUT.push_back( atomToAdd );
//
//                    int other_count = 0;
//                    int other_supportingRulesForTrueAtoms = 0;
//
//                    computeSatisfiedRules( other_count, other_supportingRulesForTrueAtoms );
//
//                    atomToAdd.other_supportingRulesForTrueAtoms = other_supportingRulesForTrueAtoms;
//                    atomToAdd.other_satisfiedRules = other_count;
//
//                    id = atomId;
//                    min = ut + other_ut;
//                }
//                else if( ( ut + other_ut ) == min )
//                {
//                    AtomUT atomToAdd;
//                    atomToAdd.id = atomId;
//                    atomToAdd.ut = ut; atomToAdd.ut2 = ut2; atomToAdd.ut3 = ut3;
//                    atomToAdd.other_ut = other_ut; atomToAdd.other_ut2 = other_ut2; atomToAdd.other_ut3 = other_ut3;
//                    atomToAdd.supportingRulesForTrueAtoms = supportingRulesForTrueAtoms;
//                    atomToAdd.satisfiedRules = count;
//
//                    sameValuesOfUT.push_back( atomToAdd );
//
//                    int other_count = 0;
//                    int other_supportingRulesForTrueAtoms = 0;
//
//                    computeSatisfiedRules( other_count, other_supportingRulesForTrueAtoms );
//
//                    atomToAdd.other_supportingRulesForTrueAtoms = other_supportingRulesForTrueAtoms;
//                    atomToAdd.other_satisfiedRules = other_count;
//                }
//            }
//            else
//            {
//                //If proapagation failed we push the original choice like a deterministic consequence.
//                unrollOne();
//                currentChoice.first = atomId;
//                currentChoice.second = TRUEDET;
//                stopTimer( chosen );
//                stopTimer( lookahead );
//                return;
////                this->propagationStack.push( Literal( atomId, TRUE ) );
////                bool prop = propagate();
////                assert( prop );
//            }
//        }
//        else
//        {
//            unrollOne();
//            currentChoice.first = atomId;
//            currentChoice.second = FALSEDET;
//            stopTimer( chosen );
//            stopTimer( lookahead );
//            return;
////            nextLevel();
////            this->propagationStack.push( Literal( atomId, FALSE ) );
////            bool prop = propagate();
////
////            unrollOne();
////            if( prop )
////            {
////                this->propagationStack.push( Literal( atomId, FALSE ) );
////                bool prop = propagate();
////                assert( prop );
////            }
////            else
////            {
////                currentChoice.first = atomId;
////                currentChoice.second = TRUE;
////                stopTimer( chosen );
////                stopTimer( lookahead );
////                return;
////            }
//        }
//    }
//
//    if( id == 0 )
//    {
//        Counter mostOccurrencyMax = 0;
//        bool found = false;
//
//        startTimer( mo );
//        for( int i = interpretation.getLeftWatch(); i <= interpretation.getRightWatch(); i++ )
//        {
//            if( interpretation.isUndefined( i ) )
//            {
//                Atom& a = program.getAtoms()[ i ];
//                if( a.getOccurrences() > mostOccurrencyMax )
//                {
//                    mostOccurrencyMax = a.getOccurrences();
//                    id = i;
//                    found = true;
//                }
//            }
//        }
//        stopTimer( mo );
//
//        if( !found )
//        {
//            //If !found there are no undefined atoms left in the interpretation!
//            currentChoice.first = interpretation.getLeftWatch();
//
//            assert( currentChoice.first != 0 );
//            assert( !interpretation.isUndefined( currentChoice.first ) );
//
//            interpretation.isTrue( currentChoice.first ) ? currentChoice.second = TRUE : currentChoice.second = FALSE;
//            stopTimer( chosen );
//            stopTimer( lookahead );
//            return;
//        }
//
//        assert( id != 0 );
//
//        currentChoice.first = id;
//        program.getAtoms()[ currentChoice.first ].isMoreNegativeOccurrences() ? currentChoice.second = FALSE : currentChoice.second = TRUE;
//
//        assert( interpretation.isUndefined( currentChoice.first ) );
//        assert( currentChoice.first != 0 );
//
//        stopTimer( chosen );
//        stopTimer( lookahead );
//        return;
//    }
//    stopTimer( chosen );
//
//    if( sameValuesOfUT.size() > 1 )
//    {
//        vector< AtomUT > sameValuesOfUT_2;
//        int min = sameValuesOfUT[ 0 ].ut2 + sameValuesOfUT[ 0 ].other_ut2;
//        sameValuesOfUT_2.push_back( sameValuesOfUT[ 0 ] );
//
//        for( int i = 1; i < sameValuesOfUT.size(); i++ )
//        {
//            int values = sameValuesOfUT[ i ].ut2 + sameValuesOfUT[ i ].ut2;
//            if( values < min )
//            {
//                sameValuesOfUT_2.clear();
//                sameValuesOfUT_2.push_back( sameValuesOfUT[ i ] );
//                min = values;
//            }
//            else if( values == min )
//            {
//                sameValuesOfUT_2.push_back( sameValuesOfUT[ i ] );
//            }
//        }
//
//        if( sameValuesOfUT_2.size() > 1 )
//        {
//            vector< AtomUT > sameValuesOfUT_3;
//            int min = sameValuesOfUT_2[ 0 ].ut3 + sameValuesOfUT_2[ 0 ].other_ut3;
//            sameValuesOfUT_3.push_back( sameValuesOfUT_2[ 0 ] );
//
//            for( int i = 1; i < sameValuesOfUT_2.size(); i++ )
//            {
//                int values = sameValuesOfUT_2[ i ].ut3 + sameValuesOfUT_2[ i ].ut3;
//                if( values < min )
//                {
//                    sameValuesOfUT_3.clear();
//                    sameValuesOfUT_3.push_back( sameValuesOfUT_2[ i ] );
//                    min = values;
//                }
//                else if( values == min )
//                {
//                    sameValuesOfUT_3.push_back( sameValuesOfUT_2[ i ] );
//                }
//            }
//
//            if( sameValuesOfUT_3.size() > 1 )
//            {
//                vector< AtomUT > sameValuesOfSAT;
//                int max = sameValuesOfUT_3[ 0 ].satisfiedRules + sameValuesOfUT_3[ 0 ].other_satisfiedRules;
////                int max = program.getAtoms()[ sameValuesOfUT_3[ 0 ].id ].getOccurrences();
//                id = sameValuesOfUT_3[ 0 ].id;
//                sameValuesOfSAT.push_back( sameValuesOfUT_3[ 0 ] );
//
//                for( int it = 0; it < sameValuesOfUT_3.size(); it++ )
//                {
//                    if( ( sameValuesOfUT_3[ it ].satisfiedRules + sameValuesOfUT_3[ it ].other_satisfiedRules ) > max )
//                    {
//                        sameValuesOfSAT.clear();
//                        sameValuesOfSAT.push_back( sameValuesOfUT_3[ it ] );
//                        max = sameValuesOfUT_3[ it ].satisfiedRules + sameValuesOfUT_3[ it ].other_satisfiedRules;
//                    }
//                    else if( ( sameValuesOfUT_3[ it ].satisfiedRules + sameValuesOfUT_3[ it ].other_satisfiedRules ) == max )
//                    {
//                        sameValuesOfSAT.push_back( sameValuesOfUT_3[ it ] );
//                    }
//                }
//
//                if( sameValuesOfSAT.size() > 1 )
//                {
//                    float maxDegreeOfSupport = ( sameValuesOfSAT[ 0 ].supportingRulesForTrueAtoms / sameValuesOfSAT[ 0 ].trueAtoms )
//                                                                                +
//                                               ( sameValuesOfSAT[ 0 ].other_supportingRulesForTrueAtoms / sameValuesOfSAT[ 0 ].other_trueAtoms );
//                    AtomUT atom = sameValuesOfSAT[ 0 ];
//
//                    for( int it = 0; it < sameValuesOfSAT.size(); it++ )
//                    {
//                        float degreeOfSupport = ( sameValuesOfSAT[ it ].supportingRulesForTrueAtoms / sameValuesOfSAT[ it ].trueAtoms )
//                                                                                +
//                                               ( sameValuesOfSAT[ it ].other_supportingRulesForTrueAtoms / sameValuesOfSAT[ it ].other_trueAtoms );
//                        if( degreeOfSupport > maxDegreeOfSupport )
//                        {
//                            maxDegreeOfSupport = maxDegreeOfSupport;
//                            atom = sameValuesOfSAT[ it ];
//                        }
//                    }
//
//                    currentChoice.first = atom.id;
//                    if( atom.ut < atom.other_ut )
//                    {
//                        currentChoice.second = TRUE;
//                    }
//                    else if( atom.ut > atom.other_ut )
//                    {
//                        currentChoice.second = FALSE;
//                    }
//                    else
//                    {
//                        if( atom.ut2 < atom.other_ut2 )
//                        {
//                            currentChoice.second = TRUE;
//                        }
//                        else if( atom.ut2 > atom.other_ut2 )
//                        {
//                            currentChoice.second = FALSE;
//                        }
//                        else
//                        {
//                            if( atom.ut3 < atom.other_ut3 )
//                            {
//                                currentChoice.second = TRUE;
//                            }
//                            else if( atom.ut3 > atom.other_ut3 )
//                            {
//                                currentChoice.second = FALSE;
//                            }
//                            else
//                            {
//                                if( atom.satisfiedRules > atom.satisfiedRules )
//                                {
//                                    currentChoice.second = TRUE;
//                                }
//                                else if( atom.satisfiedRules < atom.satisfiedRules )
//                                {
//                                    currentChoice.second = FALSE;
//                                }
//                                else
//                                {
//                                    float degree1 = ( atom.supportingRulesForTrueAtoms / atom.trueAtoms );
//                                    float degree2 = ( atom.other_supportingRulesForTrueAtoms / atom.other_trueAtoms );
//                                    degree1 > degree2 ? currentChoice.second = TRUE : currentChoice.second = FALSE;
//                                }
//                            }
//                        }
//                    }
//
//                }
//                else
//                {
//                    currentChoice.first = sameValuesOfSAT[ 0 ].id;
//                    if( sameValuesOfSAT[ 0 ].ut < sameValuesOfSAT[ 0 ].other_ut )
//                    {
//                        currentChoice.second = TRUE;
//                    }
//                    else if( sameValuesOfSAT[ 0 ].ut > sameValuesOfSAT[ 0 ].other_ut )
//                    {
//                        currentChoice.second = FALSE;
//                    }
//                    else
//                    {
//                        if( sameValuesOfSAT[ 0 ].ut2 < sameValuesOfSAT[ 0 ].other_ut2 )
//                        {
//                            currentChoice.second = TRUE;
//                        }
//                        else if( sameValuesOfSAT[ 0 ].ut2 > sameValuesOfSAT[ 0 ].other_ut2 )
//                        {
//                            currentChoice.second = FALSE;
//                        }
//                        else
//                        {
//                            if( sameValuesOfSAT[ 0 ].ut3 < sameValuesOfSAT[ 0 ].other_ut3 )
//                            {
//                                currentChoice.second = TRUE;
//                            }
//                            else if( sameValuesOfSAT[ 0 ].ut3 > sameValuesOfSAT[ 0 ].other_ut3 )
//                            {
//                                currentChoice.second = FALSE;
//                            }
//                            else
//                            {
//                                sameValuesOfSAT[ 0 ].satisfiedRules >= sameValuesOfSAT[ 0 ].other_satisfiedRules ? currentChoice.second = TRUE : currentChoice.second = FALSE;
//                            }
//                        }
//                    }
//                }
//            }
//            else
//            {
//                currentChoice.first = sameValuesOfUT_3[ 0 ].id;
//                if( sameValuesOfUT_3[ 0 ].ut < sameValuesOfUT_3[ 0 ].other_ut )
//                {
//                    currentChoice.second = TRUE;
//                }
//                else if( sameValuesOfUT_3[ 0 ].ut > sameValuesOfUT_3[ 0 ].other_ut )
//                {
//                    currentChoice.second = FALSE;
//                }
//                else
//                {
//                    if( sameValuesOfUT_3[ 0 ].ut2 < sameValuesOfUT_3[ 0 ].other_ut2 )
//                    {
//                        currentChoice.second = TRUE;
//                    }
//                    else if( sameValuesOfUT_3[ 0 ].ut2 > sameValuesOfUT_3[ 0 ].other_ut2 )
//                    {
//                        currentChoice.second = FALSE;
//                    }
//                    else
//                    {
//                        sameValuesOfUT_3[ 0 ].ut3 < sameValuesOfUT_3[ 0 ].other_ut3 ? currentChoice.second = TRUE : currentChoice.second = FALSE;
//                    }
//                }
//            }
//        }
//        else
//        {
//            currentChoice.first = sameValuesOfUT_2[ 0 ].id;
//            if( sameValuesOfUT_2[ 0 ].ut < sameValuesOfUT_2[ 0 ].other_ut )
//            {
//                currentChoice.second = TRUE;
//            }
//            else if( sameValuesOfUT_2[ 0 ].ut > sameValuesOfUT_2[ 0 ].other_ut )
//            {
//                currentChoice.second = FALSE;
//            }
//            else
//            {
//                sameValuesOfUT_2[ 0 ].ut2 < sameValuesOfUT_2[ 0 ].other_ut2 ? currentChoice.second = TRUE : currentChoice.second = FALSE;
//            }
//        }
//    }
//    else
//    {
//        currentChoice.first = sameValuesOfUT[ 0 ].id;
//        sameValuesOfUT[ 0 ].ut < sameValuesOfUT[ 0 ].other_ut ? currentChoice.second = TRUE : currentChoice.second = FALSE;
//    }
//
//    assert( interpretation.isUndefined( currentChoice.first ) );
//    assert( currentChoice.first != 0 );
//    stopTimer( lookahead );
//}
//

void
ModelGenerator::calculateAtomNeighbours(
    vector< AtomId >& atomToVisit,
    AtomId nextId,
    InvolvedPointer idOfDerivation,
    InvolvedType typeOfDerivation )
{
    Atom& next = program.getAtom( nextId );
    static unordered_set< AtomId > tempSet;
    assert( tempSet.empty() );
    Level currentLevel = backtrackingStack.getCurrentLevel();
    
    switch( typeOfDerivation )
    {
        case TYPEALL:
        {
            WaspRule* waspRule = idOfDerivation.waspRule;
            
            assert( waspRule != NULL );
            
            incrementLearnedWaspRuleActivity( waspRule );
            WaspRule& rule = *( waspRule );    
            
            unsigned int size = rule.size();
            for( unsigned int i = 0; i < size; ++i )
            {
                AtomId atomId = rule.getLiteral( i ).getAtomId();
                if( atomId != nextId && program.getAtom( atomId ).getRecursionLevel() == currentLevel && tempSet.insert( atomId ).second )
                    atomToVisit.push_back( atomId );        
            }
            break;
        }
        
        case TYPEVECTOR:
        {
            InvolvedVector* v = idOfDerivation.involvedVector;
            
            assert( v != NULL );
            unsigned int size = v->size();            
            //atomToVisit.reserve( size );
            for( unsigned int j = 0; j < size; ++j )
            {
                AtomId atomId = ( *v )[ j ];
                assert( atomId != 0 );
                assert( !interpretation.isUndefined( atomId ) );            
                
                if( atomId != nextId ) //&& atomId != conflictAtomId )
                {
                    if( program.getAtom( atomId ).getRecursionLevel() == currentLevel && tempSet.insert( atomId ).second )
                        atomToVisit.push_back( atomId );
                }
            }
            break;
        }

        case TYPESUPPORT:
        {
            const RuleContainer< WaspRule* >& waspRules = next.getNegativeHeadWaspRules();            
            unsigned int size = waspRules.size();
  
            for( unsigned int i = 0; i < size; i++ )
            {
                WaspRule& r = *( waspRules[ i ] );
                
                assert( !r.isActiveForAtom( nextId ) );
                AtomId atomId = getReasonForSupport( r, next );
                assert( atomId != 0 );

                assert( !interpretation.isUndefined( atomId ) );
                
                if( atomId != nextId ) //&& atomId != conflictAtomId )
                {
                    if( program.getAtom( atomId ).getRecursionLevel() == currentLevel && tempSet.insert( atomId ).second )
                        atomToVisit.push_back( atomId );
                }
            }
            
            const RuleContainer< ChoiceRule* >& choiceRules = next.getHeadChoiceRules();
            unsigned int choiceRulesSize = choiceRules.size();

            for( unsigned int i = 0; i < choiceRulesSize; i++ )
            {
                ChoiceRule& c = *( choiceRules[ i ] );
                
                assert( c.isBodyFalse() );
                AtomId reason = c.getIdFirstFalseLiteral();
                assert( reason != 0 );
                assert( !interpretation.isUndefined( reason ) );
                if( reason != nextId )
                {
                    if( program.getAtom( reason ).getRecursionLevel() == currentLevel && tempSet.insert( reason ).second )
                        atomToVisit.push_back( reason );
                }
            }
            break;
        }
        
        case TYPEDETERMINISTIC:
        {
            break;
        }

        default:
        {
            assert( 0 );
        }
    }
    
    tempSet.clear();
}

void
ModelGenerator::calculateAtomNeighboursLearning(
    vector< AtomId >& atomToVisitStack,
    AtomId nextId,
    InvolvedPointer idOfDerivation,
    InvolvedType typeOfDerivation )
{
    Atom& next = program.getAtom( nextId );
    unordered_set< AtomId > tempSet;

    switch( typeOfDerivation )
    {
        case TYPEALL:
        {
            WaspRule* waspRule = idOfDerivation.waspRule;
            
            assert( waspRule != NULL );
            
            incrementLearnedWaspRuleActivity( waspRule );
            WaspRule& rule = *( waspRule );    
            
            unsigned int size = rule.size();
            for( unsigned int i = 0; i < size; ++i )
            {
                AtomId atomId = rule.getLiteral( i ).getAtomId();
                assert( !interpretation.isUndefined( atomId ) );
                if( atomId != nextId && tempSet.insert( atomId ).second )
                    atomToVisitStack.push_back( atomId );    
            }
            break;
        }

        case TYPEVECTOR:
        {
            InvolvedVector* v = idOfDerivation.involvedVector;

            assert( v != NULL );
            unsigned int size = v->size();
            
            //atomToVisitStack.reserve( size );
            for( unsigned int j = 0; j < size; ++j )
            {
                AtomId atomId = ( *v )[ j ];
                
                assert( atomId != 0 );
                assert( !interpretation.isUndefined( atomId ) );

                if( atomId != conflictAtomId && atomId != nextId )
                {
                    if( tempSet.insert( atomId ).second )
                    {
                        atomToVisitStack.push_back( atomId );
                    }
                }
            }
            break;
        }

        case TYPESUPPORT:
        {
            const RuleContainer< WaspRule* >& waspRules = next.getNegativeHeadWaspRules();
            unsigned int waspRulesSize = waspRules.size();
            
            for( unsigned int i = 0; i < waspRulesSize; ++i )
            {
                WaspRule& r = *( waspRules[ i ] );
                assert( !r.isActiveForAtom( nextId ) );
                
                AtomId atomId = getReasonForSupport( r, next );
                assert( atomId != 0 );
                assert( !interpretation.isUndefined( atomId ) );
                if( atomId != conflictAtomId && atomId != nextId )
                {
                    if( tempSet.insert( atomId ).second )
                    {
                        atomToVisitStack.push_back( atomId );
                    }
                }
            }
            
            const RuleContainer< ChoiceRule* >& choiceRules = next.getHeadChoiceRules();
            unsigned int choiceRulesSize = choiceRules.size();
            
            for( unsigned int i = 0; i < choiceRulesSize; i++ )
            {
                ChoiceRule& c = *( choiceRules[ i ] );
                assert( c.isBodyFalse() );
                
                AtomId reason = c.getIdFirstFalseLiteral();
                assert( reason != 0 );
                assert( !interpretation.isUndefined( reason ) );
                if( reason != conflictAtomId && reason != nextId )
                {
                    if( tempSet.insert( reason ).second )
                    {
                        atomToVisitStack.push_back( reason );
                    }
                }           
            }
            break;
        }
        
        case TYPEDETERMINISTIC:
        {
            break;
        }

        default:
        {
            assert( 0 );
        }
    }
}

AtomId
ModelGenerator::calculateFirstUIP()
{
    trace( mg, 1, "Calculating firstUIP.\n" );
    startTimer( calculateFirstUip );
    calculateUIPdata.choiceFound = false;

    assert( conflictAtomId > 0 && conflictAtomId < program.getNumberOfAtoms() );

    if( conflictAtomId == currentChoice.first )
    {
        stopTimer( calculateFirstUip );
        return conflictAtomId;
    }
        
    AtomId res = 0;

    static vector< AtomId > implicants;
    assert( implicants.empty() );
    
    calculateAtomNeighbours( implicants, conflictAtomId, conflictImplicantIdType.first, conflictImplicantIdType.second );
    calculateAtomNeighbours( implicants, conflictAtomId, program.getAtom( conflictAtomId ).getIdOfDerivation(), program.getAtom( conflictAtomId ).getTypeOfDerivation() );
    
    assert( !implicants.empty() );
    
    calculateUIPdata.uip.push_back( MAXVALUE );
    calculateUIPdata.nextUip.push_back( calculateUIPdata.uip.size() ); 
    
    for( unsigned int i = 0; i < implicants.size(); ++i )
    {
        AtomId implicant = implicants[ i ];

        calculateUIPdata.forkBegin = 0;

        assert( interpretation.isTruePropagated( implicant ) || interpretation.isFalsePropagated( implicant ) || interpretation.isFalseUndefined( implicant ) );

        decisionHeuristic->onUIPComputationTraversal( implicant, interpretation.isTruePropagated( implicant ) );        

        
        {
            bool tmp = findConnection( implicant, currentChoice.first );
            assert( tmp );
        }

        // delete
        while( calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] < calculateUIPdata.connection[ implicant ] )
        {
            calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] = calculateUIPdata.nextUip[ calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] ];
        }
    }
    
    assert( calculateUIPdata.nextUip[ 0 ] > 0 );
    assert( calculateUIPdata.nextUip[ 0 ] < calculateUIPdata.uip.size() );
    res = calculateUIPdata.uip[ calculateUIPdata.nextUip[ 0 ] ];

    assert( res > 0 && res < program.getNumberOfAtoms() );        
    
//    if( findConnection( conflictAtomId ) )
//    {
//        assert( calculateUIPdata.nextUip[ 0 ] > 0 );
//        assert( calculateUIPdata.nextUip[ 0 ] < calculateUIPdata.uip.size() );
//        res = calculateUIPdata.uip[ calculateUIPdata.nextUip[ 0 ] ];
//
//        assert( res > 0 && res < program.getNumberOfAtoms() );        
//        trace( mg, 1, "Calculated firstUIP: atom %s.\n", program.getAtom( res ).getAtomName().c_str() );
//    }
//    else
//    {
//        trace( mg, 1, "UIP not found.\n" );
//    }

    calculateUIPdata.nextUip.clear();
    calculateUIPdata.uip.clear();
    calculateUIPdata.connection.clear();
    implicants.clear();
    
    stopTimer( calculateFirstUip );
        
    return res;
}

AtomId
ModelGenerator::calculateFirstUIPOfTwoNodes( 
    AtomId node1, 
    AtomId node2,
    AtomId choice )
{
    trace( mg, 1, "Calculating firstUIP.\n" );
    startTimer( calculateFirstUip );
    calculateUIPdata.choiceFound = false;

    assert( node1 != node2 );
    assert( program.getAtom( node1 ).getRecursionLevel() == program.getAtom( node2 ).getRecursionLevel() );
    assert( program.getAtom( node1 ).getRecursionLevel() == program.getAtom( choice ).getRecursionLevel() );
    
    if( node1 == choice )
    {
        stopTimer( calculateFirstUip );
        return node1;
    }
    else if( node2 == choice )
    {
        stopTimer( calculateFirstUip );
        return node2;
    }
        
    AtomId res = 0;

    vector< AtomId > implicants;
    calculateAtomNeighbours( implicants, node1, program.getAtom( node1 ).getIdOfDerivation(), program.getAtom( node1 ).getTypeOfDerivation() );
    calculateAtomNeighbours( implicants, node2, program.getAtom( node2 ).getIdOfDerivation(), program.getAtom( node2 ).getTypeOfDerivation() );
    
    assert( !implicants.empty() );
    
    calculateUIPdata.uip.push_back( MAXVALUE );
    calculateUIPdata.nextUip.push_back( calculateUIPdata.uip.size() ); 
    
    for( int i = 0; i < implicants.size(); i++ )
    {
        AtomId implicant = implicants[ i ];

        calculateUIPdata.forkBegin = 0;

        assert( interpretation.isTruePropagated( implicant ) || interpretation.isFalsePropagated( implicant ) || interpretation.isFalseUndefined( implicant ) );

        decisionHeuristic->onUIPComputationTraversal( implicant, interpretation.isTruePropagated( implicant ) );        

        bool tmp = findConnection( implicant, choice );
        assert( tmp );

        // delete
        while( calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] < calculateUIPdata.connection[ implicant ] )
        {
            calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] = calculateUIPdata.nextUip[ calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] ];
        }
    }
    
    assert( calculateUIPdata.nextUip[ 0 ] > 0 );
    assert( calculateUIPdata.nextUip[ 0 ] < calculateUIPdata.uip.size() );
    res = calculateUIPdata.uip[ calculateUIPdata.nextUip[ 0 ] ];

    assert( res > 0 && res < program.getNumberOfAtoms() );        
    
    calculateUIPdata.nextUip.clear();
    calculateUIPdata.uip.clear();
    calculateUIPdata.connection.clear();
    
    stopTimer( calculateFirstUip );
        
    return res;
}

bool
ModelGenerator::findConnection(
    AtomId node,
    AtomId choice )
{   
    if( !calculateUIPdata.choiceFound )
    {
        // node is a UIP candidate
        calculateUIPdata.connection[ node ] = calculateUIPdata.uip.size();
        calculateUIPdata.uip.push_back( node );
        calculateUIPdata.nextUip.push_back( calculateUIPdata.uip.size() );        

//        if( node == currentChoice.first )
        if( node == choice )
        {
            calculateUIPdata.forkBegin = calculateUIPdata.connection[ node ];
            calculateUIPdata.choiceFound = true;
            return true;
        }
        
        //assert( !program.getAtom( node ).isChoice() || node == conflictAtomId );

        vector< AtomId > implicants;
        
        calculateAtomNeighbours( implicants, node, program.getAtom( node ).getIdOfDerivation(), program.getAtom( node ).getTypeOfDerivation() );

        assert( !implicants.empty() );
        
        for( unsigned int i = 0; i < implicants.size(); ++i )
        {
            AtomId implicant = implicants[ i ];
            calculateUIPdata.forkBegin = calculateUIPdata.connection[ node ];
            assert( interpretation.isTruePropagated( implicant ) || interpretation.isFalsePropagated( implicant ) || interpretation.isFalseUndefined( implicant ) );

            decisionHeuristic->onUIPComputationTraversal( implicant, interpretation.isTruePropagated( implicant ) );
            
            bool tmp = findConnection( implicant, choice );
            assert( tmp );

            // delete
            while( calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] < calculateUIPdata.connection[ implicant ] )
            {
                calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] = calculateUIPdata.nextUip[ calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] ];
            }
        }
    }
    else
    {
        if( calculateUIPdata.connection.find( node ) == calculateUIPdata.connection.end() )
        {
            assert( !program.getAtom( node ).isChoice() );
            
            vector< AtomId > implicants;
            calculateAtomNeighbours( implicants, node, program.getAtom( node ).getIdOfDerivation(), program.getAtom( node ).getTypeOfDerivation() );
            
            assert( !implicants.empty() );
        
            for( unsigned int i = 0; i < implicants.size(); ++i )
            {
                AtomId implicant = implicants[ i ];
                assert( interpretation.isTruePropagated( implicant ) || interpretation.isFalsePropagated( implicant ) || interpretation.isFalseUndefined( implicant ) );

                decisionHeuristic->onUIPComputationTraversal( implicant, interpretation.isTruePropagated( implicant ) );
                
                bool tmp = findConnection( implicant, choice );
                assert( tmp );
                
                // set connection[node] equals to the connection of the first
                // processed implicant
                if( i == 0 ) //calculateUIPdata.connection.find( node ) == calculateUIPdata.connection.end() )
                {
                    calculateUIPdata.connection[ node ] = calculateUIPdata.connection[ implicant ];
                }

                // delete
                while( calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] < calculateUIPdata.connection[ implicant ] )
                    calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] = calculateUIPdata.nextUip[ calculateUIPdata.nextUip[ calculateUIPdata.forkBegin ] ];
            }
        }
    }
    return true;
}

void
ModelGenerator::calculateAtomsTypeAll(
    WaspRule* waspRule,
    AtomId currentAtom,
    vector< AtomId >& atomsInvolved )
{
    assert( waspRule != NULL );
    WaspRule& rule = *( waspRule );
    unsigned int size = rule.size();
    
    atomsInvolved.reserve( size );
    for( int i = 0; i < size; i++ )
    {
        AtomId atomId = rule.getLiteral( i ).getAtomId();
        if( atomId != currentAtom ) // && atomId != conflictAtomId )
            atomsInvolved.push_back( atomId );        
    }   
}

void
ModelGenerator::computeGUS()
{
    for( unsigned int i = 0; i < componentsInTheMap.size(); ++i )
    {
        ComponentId componentId = componentsInTheMap[ i ];
        computeGUS( mapComponentListOfAtomsLostSourcePointer[ componentId ] );
    }
    
    componentsInTheMap.clear();
}

void
ModelGenerator::computeGUS(
    List< Atom* >& atomsLostSourcePointer )
{
    if( atomsLostSourcePointer.empty() )
        return;
    startTimer( computeGUS );
    statistics( incrementNumComputeGUS() );
    trace( us, 1, "Calling unfounded set procedure.\n" );    
    
    #ifndef TRACE_OFF
    if( options.traceLevels.us >= 2 )
    {
        cerr << "List of atoms:";
        for( List< Atom* >::iterator it = atomsLostSourcePointer.begin(); it != atomsLostSourcePointer.end(); ++it )
        {
            cerr << " " << *( *it );
        }
        cerr << endl;
    }
    #endif
    
    trace( us, 1, "Computing unfounded set candidate.\n" );
    for( List< Atom* >::iterator it = atomsLostSourcePointer.begin(); it != atomsLostSourcePointer.end();  )
    {
        Atom* currentAtomPointer = *it;
        Atom& currentAtom = *currentAtomPointer;
        
        trace( us, 2, "Current Atom: %s.\n", currentAtom.getAtomName().c_str() );
        
        assert( !currentAtom.isFounded() );
        
        AtomId currentAtomId = currentAtom.getId();
        ComponentId componentOfAtom = currentAtom.getComponent();
        ComponentAtom componentAtom( componentOfAtom, currentAtomId );
        
        if( interpretation.isFalsePropagated( currentAtomId ) )
        {
            atomsLostSourcePointer.erase( it++ );
            continue;
        }
        
        vector< Rule* >& rulesInComponentWithPositiveBodyLiteral = program.getRulesInComponentWithPositiveBodyLiteral().find( componentAtom )->second;
        unsigned int rulesSize = rulesInComponentWithPositiveBodyLiteral.size();
        for( unsigned int i = 0; i < rulesSize; ++i )
        {
            Rule* rule = rulesInComponentWithPositiveBodyLiteral[ i ];

            ComponentRule componentRule( componentOfAtom, rule );
            vector< AtomId >& headAtomsInComponent = program.getHeadAtomInComponent().find( componentRule )->second;
            assert( !headAtomsInComponent.empty() );

            //TODO: DA RIMUOVERE QUANDO AGGIUSTIAMO LE CHOICE.
            assert( headAtomsInComponent.size() == 1 );
            AtomId headAtomId = headAtomsInComponent[ 0 ];
            Atom& headAtom = program.getAtom( headAtomId );
            
            if( interpretation.isFalsePropagated( headAtomId ) )
            {
                continue;
            }
            
            trace( us, 2, "Head Atom: %s.\n", headAtom.getAtomName().c_str() );
            
            assert( headAtom.isFounded() ^ find( atomsLostSourcePointer.begin(), atomsLostSourcePointer.end(), &headAtom ) != atomsLostSourcePointer.end() );
            
            if( headAtom.getSourcePointer() == rule && headAtom.isFounded() )
            {
                assert( find( atomsLostSourcePointer.begin(), atomsLostSourcePointer.end(), &headAtom ) == atomsLostSourcePointer.end() );
                trace( us, 2, "Added %s in the list.\n", headAtom.getAtomName().c_str() );
                atomsLostSourcePointer.push_back( &headAtom );

                backtrackingStack.push( &( headAtom.getFoundedAddress() ) );
                headAtom.setFounded( false );
            }
        }
        
        ++it;
    }
    
    trace( us, 1, "Finding new source pointers.\n" );
    //For each atom which lost a source pointer
    for( List< Atom* >::iterator it = atomsLostSourcePointer.begin(); it != atomsLostSourcePointer.end(); )
    {
        List< Atom* >::iterator temp_it = it++;
        
        Atom* currentAtomPointer = *temp_it;
        Atom& currentAtom = *currentAtomPointer;
        
        trace( us, 2, "Current Atom: %s.\n", currentAtom.getAtomName().c_str() );
        AtomId currentAtomId = currentAtom.getId();
        ComponentId componentOfAtom = currentAtom.getComponent();
        
        assert( !interpretation.isFalsePropagated( currentAtomId ) );
        if( currentAtom.isFounded() )
        {
            trace( us, 2, "Current Atom is founded.\n" );
            atomsLostSourcePointer.erase( temp_it );
            continue;
        }
        
        ComponentAtom componentAtom( componentOfAtom, currentAtomId );
        bool hasASourcePointer = false;
        
        trace( us, 1, "External search.\n" );
        vector< Rule* >& externalRules = program.getExternalRulesForAtomInComponent()[ componentAtom ];
        unsigned int externalRulesSize = externalRules.size();
        
        for( unsigned int i = 0; i < externalRulesSize; ++i )
        {
            Rule* rule = externalRules[ i ];
        
            if( rule->isRuleActiveForAtom( currentAtomId ) )
            {
                trace( us, 2, "Current Atom has a new source pointer(external).\n" );
                backtrackingStack.push( &( currentAtom.getFoundedAddress() ) );
                currentAtom.setFounded();
                currentAtom.setSourcePointer( rule );
                hasASourcePointer = true;
                
                atomsLostSourcePointer.erase( temp_it );
                break;
            }
        }
        
        if( !hasASourcePointer )
        {
            trace( us, 1, "Internal search.\n" );
            vector< Rule* >& internalRules = program.getInternalRulesForAtomInComponent()[ componentAtom ];
            unsigned int internalRulesSize = internalRules.size();
            for( unsigned int i = 0; i < internalRulesSize; ++i )
            {
                Rule* rule = internalRules[ i ];
                if( rule->isRuleActiveForAtom( currentAtomId ) )
                {
                    bool bodyFounded = true;
                    ComponentRule componentRule( componentOfAtom, rule );
                    vector< Atom* >& recursiveAtomsInRule = program.getRecursiveAtomsInRule()[ componentRule ];

                    for( unsigned int j = 0; j < recursiveAtomsInRule.size(); ++j )
                    {
                        if( !recursiveAtomsInRule[ j ]->isFounded() )
                        {
                            bodyFounded = false;
                            break;
                        }
                    }

                    if( bodyFounded )
                    {
                        trace( us, 2, "Current Atom has a new source pointer(internal).\n" );
                        backtrackingStack.push( &( currentAtom.getFoundedAddress() ) );
                        currentAtom.setFounded();
                        currentAtom.setSourcePointer( rule );
                        hasASourcePointer = true;

                        atomsLostSourcePointer.erase( temp_it );
                        break;
                    }
                }
            }        
        }
        
        if( hasASourcePointer )
        {
            stack< Atom* > founded;
            founded.push( currentAtomPointer );
            
            while( !founded.empty() )
            {
                Atom& current = *( founded.top() );
                founded.pop();
                ComponentId component = current.getComponent();
                AtomId currentAtomId = current.getId();
                
                ComponentAtom componentAtom( componentOfAtom, currentAtomId );
                trace( us, 1, "Propagating source pointers.\n" );
                vector< Rule* >& rulesInComponentWithPositiveBodyLiteral = program.getRulesInComponentWithPositiveBodyLiteral().find( componentAtom )->second;
                unsigned int rulesSize = rulesInComponentWithPositiveBodyLiteral.size();
                for( unsigned int i = 0; i < rulesSize; ++i )
                {
                    Rule* rule = rulesInComponentWithPositiveBodyLiteral[ i ];
                    ComponentRule componentRule( componentOfAtom, rule );

                    vector< AtomId >& headAtomsInComponent = program.getHeadAtomInComponent().find( componentRule )->second;
                    assert( !headAtomsInComponent.empty() );

                    AtomId headAtomId = headAtomsInComponent[ 0 ];
                    Atom& headAtom = program.getAtom( headAtomId );

                    if( headAtom.isFounded() || !rule->isRuleActiveForAtom( headAtomId ) )
                        continue;

                    bool bodyFounded = true;
                    vector< Atom* >& recursiveAtomsInRule = program.getRecursiveAtomsInRule()[ componentRule ];
                    for( unsigned int j = 0; j < recursiveAtomsInRule.size(); ++j )
                    {
                        if( !recursiveAtomsInRule[ j ]->isFounded() )
                        {
                            bodyFounded = false;
                            break;
                        }
                    }

                    if( bodyFounded )
                    {
                        trace( us, 2, "Atom %s has a new source pointer.\n", headAtom.getAtomName().c_str() );
                        backtrackingStack.push( &( headAtom.getFoundedAddress() ) );
                        headAtom.setFounded();
                        headAtom.setSourcePointer( rule );
                        founded.push( &headAtom );
                    }
                }
            }
        }
    }
    
    for( List< Atom* >::iterator it = atomsLostSourcePointer.begin(); it != atomsLostSourcePointer.end(); )
    {
        List< Atom* >::iterator temp_it = it++;
        Atom* currentAtom = *temp_it;
        if( currentAtom->isFounded() )
            atomsLostSourcePointer.erase( temp_it );        
    }
    
    trace( us, 1, "End of procedure. Calling computeGUSFinal.\n" );

    assert( isUnfoundedSet( atomsLostSourcePointer ) );
    computeGUSFinal( atomsLostSourcePointer );
    atomsLostSourcePointer.clear();
    stopTimer( computeGUS );
}

bool
ModelGenerator::isUnfoundedSet(
    List< Atom* >& atomsLostSourcePointer )
{
    if( atomsLostSourcePointer.size() == 0 )
        return true;
    
    UnorderedSet< AtomId > unfoundedSet;
    
    for( List< Atom* >::iterator it = atomsLostSourcePointer.begin(); it != atomsLostSourcePointer.end(); ++it )
    {
        unfoundedSet.insert( ( *it )->getId() );
    }
    
    bool positiveBodyCond = false, falseBodyCondition = false, multipleTrueInHeadCondition = false;

    //Iterate over the unfounded set
    for( unsigned int it = 0; it < unfoundedSet.size(); it++ )
    {
        AtomId unfounded = unfoundedSet[ it ];
        
        const RuleContainer< WaspRule* >& rules = program.getAtom( unfounded ).getNegativeHeadWaspRules();
        unsigned int rulesSize = rules.size();
        
        for( unsigned int rh = 0; rh != rulesSize; rh++ )
        {
            positiveBodyCond = false, falseBodyCondition = false, multipleTrueInHeadCondition = false;
            WaspRule& r = *( rules[ rh ] );

            unsigned int bodySize = r.getNumberOfLiteralsInTheBody();
            for( int at = 0; at != bodySize; at++ )
            {
                int val = r.getBodyLiteral( at ).getId();
                if( val > 0 && unfoundedSet.contains( val ) )//unfoundedSet.find( val ) != unfoundedSet.end() )
                {
                    positiveBodyCond = true;
                    break;
                }
            }

            //If the first condition does not hold check the other two
            if( !positiveBodyCond )
            {
                for( int at = 0; at != bodySize; at++ )
                {
                    if( interpretation.isFalsePropagated( r.getBodyLiteral( at ) ) )
                    {
                        falseBodyCondition = true;
                        break;
                    }
                }

                if( !falseBodyCondition )
                {                    
                    unsigned int headSize = r.getNumberOfAtomsInTheHead();
                    if( headSize > 1 )
                    {
                        for( unsigned int h2 = 0; h2 < headSize; h2++ )
                        {
                            AtomId atomId = r.getHeadAtom( h2 )->getId();
                            if( interpretation.isTruePropagated( atomId ) && !unfoundedSet.contains( atomId ) )// unfoundedSet.find( atomId ) == unfoundedSet.end() )
                            {
                                multipleTrueInHeadCondition = true;
                                break;
                            }
                        }
                    }
                }//Third condition

            }//second condition

            if( !falseBodyCondition && !positiveBodyCond && !multipleTrueInHeadCondition )
                return false;
        }//for each rule with it in the head
    }
    return true;
}

void
ModelGenerator::computeGUSFinal( 
    List< Atom* >& atomsLostSourcePointer )
{
    startTimer( computeGUSFinal );
    if( atomsLostSourcePointer.size() == 0 )
    {
        statistics( incrementNoDerivations() );
        stopTimer( computeGUSFinal );
        return;
    }
    
    statistics( incrementDerivations() );
    Atom* firstUnfoundedTrue = NULL;
    
    InvolvedVector* involvedRules = new InvolvedVector();
    
    ReasonPointer reasonPointer;
    reasonPointer.involvedVector = involvedRules;

    for( List< Atom* >::iterator it = atomsLostSourcePointer.begin(); it != atomsLostSourcePointer.end(); ++it )
    {
        Atom* atomPointer = *it;
        AtomId unfoundedAtom = atomPointer->getId();
        
        const RuleContainer< WaspRule* >& rules = program.getAtom( unfoundedAtom ).getNegativeHeadWaspRules();
        calculateReasons( rules, *involvedRules, unfoundedAtom );
        
        const RuleContainer< ChoiceRule* >& choiceRules = program.getAtom( unfoundedAtom ).getHeadChoiceRules();
        calculateReasonsForChoice( choiceRules, *involvedRules, unfoundedAtom );
        
        assert( !interpretation.isFalsePropagated( unfoundedAtom ) );
        if( interpretation.isUndefined( unfoundedAtom ) )
        {
            Atom& atom = *atomPointer;
            this->pushAtomInStack( atom, FALSE );
            addInvolvedRules( atom, reasonPointer, TYPEVECTOR );
        }
        else if( firstUnfoundedTrue == NULL )
        {
            firstUnfoundedTrue = atomPointer;
        }
    }
    
    if( firstUnfoundedTrue != NULL )
    {
        statistics( incrementGUSInconsistences() );
        Atom& atom = *firstUnfoundedTrue;
        this->pushAtomInStack( atom, FALSE );
        addInvolvedRules( atom, reasonPointer, TYPEVECTOR );
    }
    reasonsForGus.push_back( involvedRules );
    stopTimer( computeGUSFinal );
}

void
ModelGenerator::calculateReasons(
    const RuleContainer< WaspRule* >& waspRules,
    vector< unsigned int >& involvedRules,
    AtomId atomId )
{
    unsigned int waspRulesSize = waspRules.size();
    for( unsigned int i = 0; i < waspRulesSize; i++ )
    {
        WaspRule& r = *( waspRules[ i ] );

        AtomId reason = 0;        
        AtomId firstFalseLiteral = r.getIdFirstFalseLiteral();
        
        if( atomId == firstFalseLiteral )
        {
            if( r.getIdSecondFalseLiteral() != 0 )
                reason = r.getIdSecondFalseLiteral();
        }
        else
        {
            reason = firstFalseLiteral;
        }

        if( reason != 0 )
            involvedRules.push_back( reason );
    }
}

void
ModelGenerator::calculateReasonsForChoice(
    const RuleContainer< ChoiceRule* >& choiceRules,
    vector< unsigned int >& involvedRules,
    AtomId atomId )
{
    unsigned int choiceRulesSize = choiceRules.size();
    for( unsigned int i = 0; i < choiceRulesSize; i++ )
    {
        ChoiceRule& r = *( choiceRules[ i ] );

        AtomId reason = 0;
        AtomId firstFalseLiteral = r.getIdFirstFalseLiteral();
        
        if( atomId != firstFalseLiteral )
        {
            reason = firstFalseLiteral;
        }

        if( reason != 0 )
            involvedRules.push_back( reason );
    }
}

void
ModelGenerator::deleteLearnedConstraint()
{
    statistics( incrementDeleteLearnedRulesInvocationsNumber() );
    startTimer( deleteRules );
    List< LearnedWaspRule* >& learnedWaspRules = program.getLearnedWaspRules();
    
    List< List< LearnedWaspRule* >::iterator > tempList;
    
    unsigned int size = learnedWaspRules.size();
    int toDelete = size / 2;

    unsigned int threshold = constraintIncrement / size;
    Activity activitySum = 0;
    unsigned int activityCount = 0;
    
    for( List< LearnedWaspRule* >::iterator it = learnedWaspRules.begin(); it != learnedWaspRules.end(); )
    {
        assert( *it != NULL );
        LearnedWaspRule& learnedWaspRule = *( *it );
        
        if( !learnedWaspRule.isLocked() )
        {
            activitySum += learnedWaspRule.getActivity();
            ++activityCount;
            
            if( learnedWaspRule.getActivity() < threshold )
            {
                List< LearnedWaspRule* >::iterator temp_it = it++;
                
                assert( *learnedWaspRule.getIteratorFirstWatch() != NULL );
                assert( *learnedWaspRule.getIteratorSecondWatch() != NULL );

                assert( ( *learnedWaspRule.getIteratorFirstWatch() ) == ( *learnedWaspRule.getIteratorSecondWatch() ) );
                //deleteLearnedConstraint( learnedWaspRule );
                learnedWaspRule.getFirstWatchValue().eraseWatchedConstraint( learnedWaspRule.getIteratorFirstWatch() );
                learnedWaspRule.getSecondWatchValue().eraseWatchedConstraint( learnedWaspRule.getIteratorSecondWatch() );
                
                assert( ( *temp_it )->size() > 1 );
                
                delete *temp_it;
                learnedWaspRules.erase( temp_it );
                toDelete--;
            }
            else
            {
                tempList.push_back( it );
                ++it;
            }
        }
        else
        {
            ++it;
        }
    }
    
    if( activityCount == 0 )
    {
        stopTimer( deleteRules );
        return;
    }
    
    Activity activityAvg = activitySum / activityCount; 
    
    
    for( List< List< LearnedWaspRule* >::iterator >::iterator it = tempList.begin(); it != tempList.end() && toDelete > 0; ++it )
    {
        List< LearnedWaspRule* >::iterator temp_it = *it;//= it++;
        assert( *temp_it != NULL );
        LearnedWaspRule& learnedWaspRule = *( *temp_it );
        
        assert( !learnedWaspRule.isLocked() );
        
        if( learnedWaspRule.getActivity() < activityAvg )
        {
            assert( interpretation.isUndefined( learnedWaspRule.getFirstWatchValue() ) || 
                    learnedWaspRule.getFirstWatchValue().getAtom()->getIdOfDerivation().waspRule != &learnedWaspRule );
            
            assert( interpretation.isUndefined( learnedWaspRule.getSecondWatchValue() ) || 
                    learnedWaspRule.getSecondWatchValue().getAtom()->getIdOfDerivation().waspRule != &learnedWaspRule );
            
            assert( *learnedWaspRule.getIteratorFirstWatch() != NULL );
            assert( *learnedWaspRule.getIteratorSecondWatch() != NULL );
            
            assert( ( *learnedWaspRule.getIteratorFirstWatch() ) == ( *learnedWaspRule.getIteratorSecondWatch() ) );
            
            assert( ( *temp_it )->size() > 1 );
            //deleteLearnedConstraint( learnedWaspRule );
            learnedWaspRule.getFirstWatchValue().eraseWatchedConstraint( learnedWaspRule.getIteratorFirstWatch() );
            learnedWaspRule.getSecondWatchValue().eraseWatchedConstraint( learnedWaspRule.getIteratorSecondWatch() );
            delete *temp_it;
            learnedWaspRules.erase( temp_it );
            toDelete--;
        }
    }
    
    stopTimer( deleteRules );
}

void
ModelGenerator::incrementLearnedWaspRuleActivity(
    WaspRule* rule )
{
    assert( rule != NULL );

    if( !rule->isLearned() )
        return;
    
    LearnedWaspRule& r = *( static_cast< LearnedWaspRule* >( rule ) );
    
    r.incrementActivity( constraintIncrement );
    if( r.getActivity() > 1e20 )
    {
        List< LearnedWaspRule* >& learnedWaspRules = program.getLearnedWaspRules();        
        
        for( List< LearnedWaspRule* >::iterator i = learnedWaspRules.begin(); i != learnedWaspRules.end(); ++i )
        {
            assert( *i != NULL );
            
            Activity& activity = ( *i )->getActivityAddress();
            activity *= 1e-20;
            
        }
        constraintIncrement *= 1e-20;
    }  
}

LearnedWaspRule*
ModelGenerator::learnRule(
        AtomId firstUip,
        unsigned int& levelToBacktrack )
{
    statistics( incrementLearnRuleInvocationsNumber() );
    startTimer( learnRule );

    assert( firstUip > 0 && firstUip < program.getNumberOfAtoms() );

    Atom& conflictAtom = program.getAtom( conflictAtomId );

    //This set contains the atoms will add in the learned constraint.
    static unordered_set< AtomId > tempSet;
    static vector< AtomId > tempVector;
    assert( tempSet.empty() );
    assert( tempVector.empty() );

    //assert( interpretation.isTrue( firstUip ) || interpretation.isFalse( firstUip ) || interpretation.isFalseUndefined( firstUip ) );
    assert( interpretation.isTruePropagated( firstUip ) || interpretation.isFalsePropagated( firstUip ) );

    //FirstUip is always in the constraint.
    tempSet.insert( firstUip );
    tempVector.push_back( firstUip );

    levelToBacktrack = 0;
    unsigned int maxPos = 0;
    
    //If conflictAtom is a choice we have to add it in the constraint.
    if( conflictAtom.isChoice() )
    {
        assert( interpretation.isTruePropagated( conflictAtomId ) || interpretation.isFalsePropagated( conflictAtomId ) || interpretation.isFalseUndefined( conflictAtomId ) );
        if( tempSet.insert( conflictAtomId ).second )
            tempVector.push_back( conflictAtomId );
        
        if( conflictAtom.getRecursionLevel() != backtrackingStack.getCurrentLevel() )
        {
            levelToBacktrack = conflictAtom.getRecursionLevel() + 1;
            maxPos = tempVector.size() - 1;
        }                
    }

    trace( mg, 1, "First Uip: %s, ConflictAtom: %s, CurrentLevelChoice: %s.\n",
           program.getAtom( firstUip ).getAtomName().c_str(),
           conflictAtom.getAtomName().c_str(),
           program.getAtom( currentChoice.first ).getAtomName().c_str() );
    Level conflictLevel = backtrackingStack.getCurrentLevel();
    static vector< AtomId > implicants;
    static unordered_set< AtomId > visitedAtoms;
    assert( implicants.empty() );
    assert( visitedAtoms.empty() );

    calculateAtomNeighboursLearning( implicants, conflictAtomId, conflictAtom.getIdOfDerivation(), conflictAtom.getTypeOfDerivation() );
    calculateAtomNeighboursLearning( implicants, conflictAtomId, conflictImplicantIdType.first, conflictImplicantIdType.second );

    while( !implicants.empty() )
    {
        AtomId atomId = implicants.back();
        implicants.pop_back();

        if( atomId != firstUip && visitedAtoms.insert( atomId ).second )
        {
            Atom& currentAtom = program.getAtom( atomId );
            Level currentAtomLevel = currentAtom.getRecursionLevel();
            
            if( currentAtomLevel != conflictLevel )
            {
                if( currentAtomLevel != 0 && tempSet.insert( atomId ).second )
                {
                    tempVector.push_back( atomId );

                    if( currentAtom.getRecursionLevel() >= levelToBacktrack )
                    {
                        levelToBacktrack = currentAtom.getRecursionLevel() + 1;
                        maxPos = tempVector.size() - 1;
                    }                    
                }
            }
            else
            {
                calculateAtomNeighboursLearning( implicants, atomId, currentAtom.getIdOfDerivation(), currentAtom.getTypeOfDerivation() );
            }
        }
    }    
    
    if( tempVector.size() == 0 )
        levelToBacktrack = FOUND_INCOHERENCE;
    else if( tempVector.size() == 1 )
        levelToBacktrack = 1;
    else
    {
        AtomId tmp = tempVector.back();
        tempVector.back() = tempVector[ maxPos ];
        tempVector[ maxPos ] = tmp;
    }        
    
    LearnedWaspRule* w = insertConstraint( tempVector );
    
    tempSet.clear();
    tempVector.clear();
    implicants.clear();
    visitedAtoms.clear();
    
    stopTimer( learnRule );
    return w;
}

void
ModelGenerator::updateSupportCountersForHeadAtoms(
    WaspRule& r, 
    Atom* atom )
{
    unsigned int size = r.getNumberOfAtomsInTheHead();    
    for( unsigned int i = 0; i < size; i++ )
    {
        Atom* a = r.getHeadAtom( i );
        assert( a != NULL );
        assert( !a->isAggregate() );
        Atom& headAtom = *a;        
        
        if( !interpretation.isFalse( headAtom ) && a != atom )
        {     
            haveToCalculateGUS( headAtom, &r );
            
            if( !headAtom.isSupported() )
            {
                this->addAddressInBackTrackingStack( headAtom.supportCounter );
                headAtom.updateSupportCounter();
                if( !headAtom.canBeSupported() )
                {
                    statistics( incrementCountSupport() );
                    this->addInvolvedRules( headAtom, { NULL }, TYPESUPPORT );
                    pushAtomInStack( headAtom, FALSE );
                }
                else if( interpretation.isTruePropagated( headAtom ) && headAtom.onlyOneSupportingRuleLeft() )
                {
                    assert( !headAtom.isAggregate() );
                    lastSupportingRule( headAtom );
                }
            }
        }
    }
}

void
ModelGenerator::updateSupportCountersForOneAtom(
    WaspRule& r,
    Atom* a )
{
    assert( a != NULL );
    
    Atom& atom = *a;
    assert( !atom.isAggregate() );
    assert( interpretation.isTruePropagated( atom ) );
    
    haveToCalculateGUS( atom, &r );
    
    if( atom.isSupported() )
        return;
    
    this->addAddressInBackTrackingStack( atom.supportCounter );
    atom.updateSupportCounter();

    if( !atom.canBeSupported() )
    {
        statistics( incrementCountSupport() );
        this->addInvolvedRules( atom, { NULL }, TYPESUPPORT );
        pushAtomInStack( atom, FALSE );
    }
    else if( atom.onlyOneSupportingRuleLeft() )
    {
        assert( !atom.isAggregate() );
        lastSupportingRule( atom );
    }
}

unsigned int
ModelGenerator::updateFirstWatch(
    WaspRule& r )
{
    startTimer( updateLeftWatchWaspRule );
    trace( mg, 4, "Updating first watch.\n" );
    trace( mg, 4, "Initial value: %d.\n", r.getFirstWatch() );
    
    if( interpretation.isFalsePropagated( r.getSecondWatchValue() ) )
    {
        stopTimer( updateLeftWatchWaspRule );
        return NOTHING_TO_DO;
    }
    
    unsigned int oldPosition = r.getFirstWatch();
    while( r.updateFirstWatch() != oldPosition )
    {
        if( !interpretation.isTruePropagated( r.getFirstWatchValue() ) )
        {
            trace( mg, 4, "Value: %d.\n", r.getFirstWatch() );
            stopTimer( updateLeftWatchWaspRule );
            return WATCH_CHANGED;//MAXVALUE;
        }
    }
    
    //if we reach this point, it means that the rule is unit.
    assert( interpretation.isUndefined( r.getSecondWatchValue() ) );

    stopTimer( updateLeftWatchWaspRule );
    return r.getSecondWatch();    
}

unsigned int
ModelGenerator::updateSecondWatch(
    WaspRule& r )
{
    startTimer( updateRightWatchWaspRule );
    trace( mg, 4, "Updating first watch.\n" );
    trace( mg, 4, "Initial value: %d.\n", r.getSecondWatch() );
    
    if( interpretation.isFalsePropagated( r.getFirstWatchValue() )  )
    {
        stopTimer( updateRightWatchWaspRule );
        return NOTHING_TO_DO;
    }
    
    unsigned int oldPosition = r.getSecondWatch();
	while( r.updateSecondWatch() != oldPosition )
    {
        if( !interpretation.isTruePropagated( r.getSecondWatchValue() ) )
        {
            trace( mg, 4, "Value: %d.\n", r.getSecondWatch() );
            stopTimer( updateRightWatchWaspRule );
            return WATCH_CHANGED;
        }
    }

    //if we reach this point, it means that the rule is unit.
    assert( interpretation.isUndefined( r.getFirstWatchValue() ) );
    stopTimer( updateRightWatchWaspRule );
    return r.getFirstWatch();
}

void
ModelGenerator::checkForSatisfyWaspRulesForAtomsInHead(
    const RuleContainer< WaspRule* >& waspRuleVector,
    Atom& atom )
{
    unsigned int size = waspRuleVector.size();
    for( unsigned int i = 0; i < size; i++ )
    {        
        assert( waspRuleVector[ i ] != NULL );
        trace( mg, 2, "Performing waspRule %s", program.waspRuleToString( *waspRuleVector[ i ] ).c_str() );        
        WaspRule& r = *( waspRuleVector[ i ] );
        assert( !r.isLearned() );
                
        if( !r.isSatisfied() )
        {
            this->addAddressInBackTrackingStack( r.state );
            r.makeRuleCanSupport();

            this->addAddressInBackTrackingStack( r.idFirstFalseLiteral );
            r.setIdFirstFalseLiteral( atom.getId() );

            assert( !r.isConstraint() );            
            assert( !atom.isAggregate() );
            
            updateSupportCountersForHeadAtoms( r, &atom );
        }   
        else if( r.isSatisfiedCanSupport() )
        {
            assert( !r.isConstraint() );
            
            this->addAddressInBackTrackingStack( r.state );
            r.makeRuleSatisfied();
            
            this->addAddressInBackTrackingStack( r.idSecondFalseLiteral );
            r.setIdSecondFalseLiteral( atom.getId() );
            
            Atom* a = program.getAtomPointer( r.getIdFirstFalseLiteral() );
            
            assert( a != NULL );
            assert( !a->isAggregate() );       
            
            updateSupportCountersForOneAtom( r, a );
        }
    }
}

void
ModelGenerator::checkForSatisfyWaspRules( 
    const RuleContainer< WaspRule* >& waspRuleVector,
    Atom& atom )
{
    unsigned int size = waspRuleVector.size();
    for( unsigned int i = 0; i < size; ++i )
    {        
        assert( waspRuleVector[ i ] != NULL );
        trace( mg, 2, "Performing waspRule %s", program.waspRuleToString( *waspRuleVector[ i ] ).c_str() );        
        WaspRule& r = *( waspRuleVector[ i ] );
        assert( !r.isLearned() );
        
        if( !r.isSatisfied() )
        {
            this->addAddressInBackTrackingStack( r.state );
            //r.makeRuleSatisfied();
            r.makeRuleSatisfiedByBody();

            this->addAddressInBackTrackingStack( r.idFirstFalseLiteral );
            r.setIdFirstFalseLiteral( atom.getId() );

            if( !r.isConstraint() )
            {
                updateSupportCountersForHeadAtoms( r, NULL );
            }
        }
        else if( r.isSatisfiedCanSupport() )
        {
            assert( !r.isConstraint() );
            
            this->addAddressInBackTrackingStack( r.state );
            //r.makeRuleSatisfied();
            r.makeRuleSatisfiedByBody();
            
            this->addAddressInBackTrackingStack( r.idSecondFalseLiteral );
            r.setIdSecondFalseLiteral( atom.getId() );
            
            Atom* a = program.getAtomPointer( r.getIdFirstFalseLiteral() );
            
            assert( a != NULL );
            assert( !a->isAggregate() );            
            
            updateSupportCountersForOneAtom( r, a );
        }
    }
}

void
ModelGenerator::checkForSatisfyLearnedWaspRules( 
    const RuleContainer< WaspRule* >& waspRuleVector,
    Atom& atom )
{
    unsigned int size = waspRuleVector.size();
    for( unsigned int i = 0; i < size; ++i )
    {        
        assert( waspRuleVector[ i ] != NULL );
        trace( mg, 2, "Performing waspRule %s", program.waspRuleToString( *waspRuleVector[ i ] ).c_str() );        
        WaspRule& r = *( waspRuleVector[ i ] );
        assert( r.isLearned() );
        
        if( r.isSatisfied() )
            continue;

        this->addAddressInBackTrackingStack( r.state );
        r.makeRuleSatisfiedByBody();

        this->addAddressInBackTrackingStack( r.idFirstFalseLiteral );
        r.setIdFirstFalseLiteral( atom.getId() );
    }
}
    
void
ModelGenerator::checkForUpdateWatches( 
    WatchedList< WaspRule* >& waspRules,
    Atom& atom )
{
    //If the atom A is false, for each body where A appears negative, update the body watch.
    for( WatchedList< WaspRule* >::iterator it = waspRules.begin() ; it != waspRules.end();  )
    {
        WatchedList< WaspRule* >::iterator oldIt = it++;
        WaspRule& r = *( *oldIt );
        trace( mg, 2, "Performing waspRule %s", program.waspRuleToString( r ).c_str() );
        
        assert( !r.isConstraint() );

        if( r.isSatisfied() )
            continue;

        assert( !r.isSatisfied() );
        assert( !r.isViolated() );

        unsigned int value;
        if( r.getFirstWatchValue().getAtom() == &atom )
        {
            value = updateFirstWatch( r );
            if( value == WATCH_CHANGED )
            {
                waspRules.erase( oldIt );
                WatchedList< WaspRule* >::iterator new_iterator = r.getFirstWatchValue().addWatchedWaspRule( &r );
                r.setIteratorFirstWatch( new_iterator );
                continue;
            }
            else if( value == NOTHING_TO_DO )
                continue;
        }
        else
        {   
            assert( r.getSecondWatchValue().getAtom() == &atom );
            value = updateSecondWatch( r );
            
            if( value == WATCH_CHANGED )
            {
                waspRules.erase( oldIt );
                WatchedList< WaspRule* >::iterator new_iterator = r.getSecondWatchValue().addWatchedWaspRule( &r );
                r.setIteratorSecondWatch( new_iterator );
                continue;
            }
            else if( value == NOTHING_TO_DO )
                continue;
        }
        
        assert( !r.isSatisfied() );
        assert( !r.isViolated() );        
        
        Literal& literal = r.getLiteral( value );
        Atom& absNewAtom = *( literal.getAtom() );
        
        assert( !interpretation.isTruePropagated( literal ) );
        if( !interpretation.isFalse( literal ) )
        {
            statistics( incrementCountUnit() );
            addInvolvedRules( absNewAtom, { *oldIt }, TYPEALL );
            pushAtomInStack( absNewAtom, ( literal.isPositive() ? FALSE : TRUE ) );
            
            if( value < r.getNumberOfAtomsInTheHead() )
            {
                addAddressInBackTrackingStack( absNewAtom.supportCounter );
                absNewAtom.setSupported();
            }
        }
    }
}

void
ModelGenerator::checkForUpdateWatchesConstraints( 
    WatchedList< WaspRule* >& waspRules,
    Atom& atom )
{
    //If the atom A is false, for each body where A appears negative, update the body watch.
    for( WatchedList< WaspRule* >::iterator it = waspRules.begin() ; it != waspRules.end();  )
    {
        WatchedList< WaspRule* >::iterator oldIt = it++;
        WaspRule& r = *( *oldIt );
        trace( mg, 2, "Performing waspRule %s", program.waspRuleToString( r ).c_str() );
        
        assert( r.isConstraint() );

        if( r.isSatisfied() )
            continue;

        assert( !r.isSatisfied() );
        assert( !r.isViolated() );

        unsigned int value;
        if( r.getFirstWatchValue().getAtom() == &atom )
        {
            value = updateFirstWatch( r );
            if( value == WATCH_CHANGED )
            {
                waspRules.erase( oldIt );
                WatchedList< WaspRule* >::iterator new_iterator = r.getFirstWatchValue().addWatchedConstraint( &r );
                r.setIteratorFirstWatch( new_iterator );
                continue;
            }
            else if( value == NOTHING_TO_DO )
                continue;
        }
        else
        {   
            assert( r.getSecondWatchValue().getAtom() == &atom );
            value = updateSecondWatch( r );
            
            if( value == WATCH_CHANGED )
            {
                waspRules.erase( oldIt );
                WatchedList< WaspRule* >::iterator new_iterator = r.getSecondWatchValue().addWatchedConstraint( &r );
                r.setIteratorSecondWatch( new_iterator );
                continue;
            }
            else if( value == NOTHING_TO_DO )
                continue;
        }
        
        assert( !r.isSatisfied() );
        assert( !r.isViolated() );        
        
        Literal& literal = r.getLiteral( value );
        Atom& absNewAtom = *( literal.getAtom() );
        
        assert( !interpretation.isTruePropagated( literal ) );
        if( !interpretation.isFalse( literal ) )
        {
            if( r.isLearned() )
            {
                LearnedWaspRule& learnedRule = *( static_cast< LearnedWaspRule* >( *oldIt ) );
                
                backtrackingStack.push( &( learnedRule.getLockedAddress() ) );
                learnedRule.setLocked();
            }
            statistics( incrementCountUnit() );
            addInvolvedRules( absNewAtom, { *oldIt }, TYPEALL );
            pushAtomInStack( absNewAtom, ( literal.isPositive() ? FALSE : TRUE ) );
            
            assert( r.getNumberOfAtomsInTheHead() == 0 );
        }
    }
}

void
ModelGenerator::checkForUpdateChoiceRulesStatus( 
    const RuleContainer< ChoiceRule* >& choiceRuleVector,
    Atom& atom )
{
    unsigned int size = choiceRuleVector.size();
    //If the atom A is false, for each body where A appears negative, update the body watch.
    for( unsigned int i = 0; i < size; i++ )
    {
        assert( choiceRuleVector[ i ] != NULL );
        ChoiceRule& r = *( choiceRuleVector[ i ] );

        this->addAddressInBackTrackingStack( r.counter );
        r.updateCounter();
        
        if( r.isBodyTrue() )
        {
            setHeadAtomsChoiceRuleSupported( r );
        }
    }
}

void
ModelGenerator::checkForFalsifyChoiceRulesBody( 
    const RuleContainer< ChoiceRule* >& choiceRuleVector,
    Atom& atom )
{
    unsigned int size = choiceRuleVector.size();
    for( unsigned int i = 0; i < size; i++ )
    {        
        assert( choiceRuleVector[ i ] != NULL );
        ChoiceRule& r = *( choiceRuleVector[ i ] );
        
        assert( !r.isBodyTrue() );
        
        if( !r.isBodyFalse() )
        {
            this->addAddressInBackTrackingStack( r.idFirstFalseLiteral );
            r.setIdFirstFalseLiteral( atom.getId() );
            
            updateSupportCounterForHeadAtomsChoiceRule( r );        
        }        
    }
}

void 
ModelGenerator::setHeadAtomsChoiceRuleSupported( 
    ChoiceRule& r )
{
    unsigned int size = r.getNumberOfAtomsInTheHead();
    for( unsigned int i = 0; i < size; i++ )
    {
        assert( r.getHeadAtom( i ) != NULL );
        
        Atom& atom = *( r.getHeadAtom( i ) );
        
        if( !atom.isSupported() )
        {
            addAddressInBackTrackingStack( atom.supportCounter );
            atom.setSupported();
        }
    }
}

void 
ModelGenerator::updateSupportCounterForHeadAtomsChoiceRule( 
    ChoiceRule& r )
{
    unsigned int size = r.getNumberOfAtomsInTheHead();
    
    for( unsigned int i = 0; i < size; i++ )
    {
        Atom* a = r.getHeadAtom( i );
        assert( a != NULL );
        assert( !a->isAggregate() );
        Atom& headAtom = *a;        
        
        if( !interpretation.isFalse( headAtom ) )
        {     
            haveToCalculateGUS( headAtom, &r );
            if( !headAtom.isSupported() )
            {
                this->addAddressInBackTrackingStack( headAtom.supportCounter );
                headAtom.updateSupportCounter();
                
                if( !headAtom.canBeSupported() )
                {
                    statistics( incrementCountSupport() );
                    this->addInvolvedRules( headAtom, { NULL }, TYPESUPPORT );
                    pushAtomInStack( headAtom, FALSE );
                }
                else if( interpretation.isTruePropagated( headAtom ) && headAtom.onlyOneSupportingRuleLeft() )
                {
                    assert( !headAtom.isAggregate() );
                    lastSupportingRule( headAtom );
                }
            }
        }
    }
}

void
ModelGenerator::unitPropagation(
    Atom& atom )
{
    startTimer( unitPropagationTimer );

   // AtomId atomId = atom.getId();
    trace( mg, 2, "Evaluating waspRules where atom %s appears:\n" , atom.getAtomName().c_str() );

    assert( interpretation.isTruePropagated( atom ) || interpretation.isFalsePropagated( atom ) );

    const RuleContainer< ChoiceRule* >& positiveBodyChoiceRules = atom.getPositiveBodyChoiceRules();
    const RuleContainer< ChoiceRule* >& negativeBodyChoiceRules = atom.getNegativeBodyChoiceRules();

    if( interpretation.isFalsePropagated( atom ) )
    {
        const RuleContainer< WaspRule* >& positiveWaspRules = atom.getPositiveWaspRules();
        WatchedList< WaspRule* >& watchedNegativeConstraints = atom.getWatchedNegativeConstraints();
        WatchedList< WaspRule* >& watchedNegativeWaspRules = atom.getWatchedNegativeWaspRules();
    
        /* Satisfy positive waspRules */
        checkForSatisfyWaspRules( positiveWaspRules, atom );
        
        /* Update watches*/
        checkForUpdateWatches( watchedNegativeWaspRules, atom );
        checkForUpdateWatchesConstraints( watchedNegativeConstraints, atom );
        
        checkForFalsifyChoiceRulesBody( positiveBodyChoiceRules, atom );
        checkForUpdateChoiceRulesStatus( negativeBodyChoiceRules, atom );
    }
    else if( interpretation.isTruePropagated( atom ) )
    {
        const RuleContainer< WaspRule* >& negativeBodyWaspRules = atom.getNegativeBodyWaspRules();
        const RuleContainer< WaspRule* >& negativeHeadWaspRules = atom.getNegativeHeadWaspRules();

        WatchedList< WaspRule* >& watchedPositiveConstraints = atom.getWatchedPositiveConstraints();
        WatchedList< WaspRule* >& watchedPositiveWaspRules = atom.getWatchedPositiveWaspRules();
        
        /* Satisfy positive waspRules */
        checkForSatisfyWaspRules( negativeBodyWaspRules, atom );
        checkForSatisfyWaspRulesForAtomsInHead( negativeHeadWaspRules, atom );
        
        /* Update watches*/
        checkForUpdateWatches( watchedPositiveWaspRules, atom );
        checkForUpdateWatchesConstraints( watchedPositiveConstraints, atom );
        
        checkForFalsifyChoiceRulesBody( negativeBodyChoiceRules, atom );
        checkForUpdateChoiceRulesStatus( positiveBodyChoiceRules, atom );
        
    }
    stopTimer( unitPropagationTimer );
}

void
ModelGenerator::supportControl(
    Atom& atom )
{
    startTimer( supportControlTimer );
    //If the atom is true, we have to check if it is supported.
    
    trace( mg, 2, "Evaluating if atom %s is supported.\n", atom.getAtomName().c_str() );
    assert( interpretation.isTruePropagated( atom ) || interpretation.isFalsePropagated( atom ) );

    if( interpretation.isTruePropagated( atom ) )
    {
        if( atom.isAggregate() )
        {
            stopTimer( supportControlTimer );
            return;
        }

        trace( mg, 2, "Atom %s is true and it is not an aggregate.\n", atom.getAtomName().c_str() );
        if( !atom.isSupported() )
        {            
            trace( mg, 2, "Atom %s is not yet supported.\n", atom.getAtomName().c_str() );            
            
            if( atom.onlyOneSupportingRuleLeft() )
            {
                trace( mg, 2, "Only one rule can support the atom %s.\n", atom.getAtomName().c_str() );
                lastSupportingRule( atom );
            }
            else if( !atom.canBeSupported() )
            {
                trace( mg, 2, "Atom %s can't be supported.\n", atom.getAtomName().c_str() );
                statistics( incrementCountSupport() );
                this->addInvolvedRules( atom, { NULL }, TYPESUPPORT );
                pushAtomInStack( atom, FALSE );
            }
        } 
    }
    stopTimer( supportControlTimer );
    return;
}

void
ModelGenerator::lastSupportingRule(
    Atom& atom )
{
    trace( mg, 2, "Performing lastSupportingRule propagation for atom %s.\n", atom.getAtomName().c_str() );
    assert( interpretation.isTruePropagated( atom ) );
    AtomId atomId = atom.getId();
        
    InvolvedVector* involved = new InvolvedVector();
    ReasonPointer reasonPointer;
    reasonPointer.involvedVector = involved;
    
    involved->push_back( atomId );
   
    const RuleContainer< WaspRule* >& waspRules = atom.getNegativeHeadWaspRules();
    unsigned int size = waspRules.size();
    
    for( unsigned int i = 0; i < size; i++ )
    {
        WaspRule& r = *( waspRules[ i ] );
        if( r.isSatisfiedCannotSupport() )
        {
            AtomId reason = getReasonForSupport( r, atom );
            assert( reason != 0 );
            assert( !interpretation.isUndefined( reason ) );
            involved->push_back( reason );
        }
        else
        {
            assert( r.isSatisfiedCanSupport() );
            assert( r.getIdFirstFalseLiteral() == atomId );
            
            assert( !r.isViolated() );
        
            //SET WASPRULE LITERALS AS TRUE.
            unsigned int size = r.size();
            for( unsigned int j = 0; j < size; j++ )    
            {
                Literal& literal = r.getLiteral( j );
                if( !literal.isTrue() )
                {
                    Atom* atomPointer = literal.getAtom();
                    if( atomPointer != &atom )
                    {
                        statistics( incrementCountLastSupport() );    
                        Atom& a = *( atomPointer );
                        addInvolvedRules( a, reasonPointer, TYPEVECTOR );
                        this->pushAtomInStack( a, literal.isPositive() ? TRUE : FALSE );
                    }
                    else
                    {
                        assert( atomId == r.getIdFirstFalseLiteral() );
                    }
                }
            }
            //setWaspRuleLiteralsTrue( r, atom );
            this->addAddressInBackTrackingStack( atom.supportCounter );
            atom.setSupported();
        }
    }
    
    const RuleContainer< ChoiceRule* >& choiceRules = atom.getHeadChoiceRules();
    unsigned int choiceRulesSize = choiceRules.size();
    
    for( unsigned int i = 0; i < choiceRulesSize; i++ )
    {
        ChoiceRule& c = *( choiceRules[ i ] );
        if( c.isBodyFalse() )
        {
            AtomId reason = c.getIdFirstFalseLiteral();
            assert( reason != 0 );
            assert( !interpretation.isUndefined( reason ) );
            involved->push_back( reason );
        }
        else
        {
            assert( c.isBodyUndefined() || c.getNumberOfLiteralsInTheBody() == 0 );
            assert( !c.isBodyTrue() || c.getNumberOfLiteralsInTheBody() == 0 );
        
            unsigned int size = c.getNumberOfLiteralsInTheBody();
            for( unsigned int j = 0; j < size; j++ )    
            {
                Literal& literal = c.getBodyLiteral( j );
                if( !literal.isTrue() )
                {
                    statistics( incrementCountLastSupport() );
                    Atom& atom = *( literal.getAtom() );
                    addInvolvedRules( atom, reasonPointer, TYPEVECTOR );
                    this->pushAtomInStack( atom, literal.isPositive() ? TRUE : FALSE );            
                }
            }
            setHeadAtomsChoiceRuleSupported( c );
        }
    }
    
    reasonsForGus.push_back( involved );
}

AtomId
ModelGenerator::getReasonForSupport(
    WaspRule& r,
    Atom& atom )
{
    AtomId atomId = atom.getId();    
    AtomId firstFalseLiteral = r.getIdFirstFalseLiteral();
    
    if( atomId == firstFalseLiteral )
    {
        assert( r.getIdSecondFalseLiteral() != 0 );
        return r.getIdSecondFalseLiteral();
    }
    
    return firstFalseLiteral;
}

void
ModelGenerator::updateInterpretationWatch()
{
    startTimer( updateInterpretationWatches );
    if( interpretation.getLeftWatch() >= interpretation.getRightWatch() )
    {
        stopTimer( updateInterpretationWatches );
        return;
    }
    
    if( !interpretation.isUndefined( interpretation.getLeftWatch() ) )
    {
        backtrackingStack.push( &( interpretation.leftWatch ) );
        do{
            interpretation.updateLeftWatch();
            
            if( interpretation.getLeftWatch() >= interpretation.getRightWatch() )
            {
                stopTimer( updateInterpretationWatches );
                return;
            }
        } while( !interpretation.isUndefined( interpretation.getLeftWatch() ) );
    }
    
    assert( interpretation.isUndefined( interpretation.getLeftWatch() ) );
    assert( interpretation.getRightWatch() > interpretation.getLeftWatch() );
    
    if( !interpretation.isUndefined( interpretation.getRightWatch() ) )
    {
        backtrackingStack.push( &( interpretation.rightWatch ) );
        do{            
            interpretation.updateRightWatch(); 
            assert( interpretation.getRightWatch() >= interpretation.getLeftWatch() );           
        } while( !interpretation.isUndefined( interpretation.getRightWatch() ) );
    }
    stopTimer( updateInterpretationWatches );
}

//Used in testing the model checker
Program&
ModelGenerator::getProgram()
{
    return program;
}

LearnedWaspRule*
ModelGenerator::insertConstraint(
    vector< AtomId >& atomsToAddInConstraint )
{        
    unsigned int size = atomsToAddInConstraint.size();

    assert( size > 0 );
    assert( size < options.maxLearnedConstraintSize );
    
    statistics( incrementNumberOfLearnedWaspRules() );
    statistics( incrementSizeOfLearnedWaspRules( size ) );

    LearnedWaspRule* learnedWaspRule = new LearnedWaspRule( atomsToAddInConstraint.size() );
    
    for( vector< AtomId >::iterator i = atomsToAddInConstraint.begin(); i != atomsToAddInConstraint.end(); ++i )
    {
        AtomId atom = *i;
        assert( atom > 0 && atom < program.getNumberOfAtoms() );

        Atom* atomPointer = program.getAtomPointer( atom );
        Atom& a = *atomPointer;
        assert( interpretation.isTruePropagated( a ) || interpretation.isFalsePropagated( a ) || interpretation.isFalseUndefined( a ) );
        
        decisionHeuristic->onLearnedWaspRule( atom, interpretation.isTruePropagated( atom ) );
        
        if( interpretation.isTruePropagated( atom ) )
        {            
            learnedWaspRule->addLiteral( Literal( atomPointer ) );
            //a.addPositiveLearnedWaspRule( learnedWaspRule );
        }
        else
        {
            assert( interpretation.isFalsePropagated( a ) || interpretation.isFalseUndefined( a ) );
            learnedWaspRule->addLiteral( Literal( atomPointer, true ) );
            //a.addNegativeLearnedWaspRule( learnedWaspRule );
        }
    }
    
    WatchedList< WaspRule* >::iterator it1 = learnedWaspRule->getFirstWatchValue().addWatchedConstraint( learnedWaspRule );
    learnedWaspRule->setIteratorFirstWatch( it1 );
    WatchedList< WaspRule* >::iterator it2 = learnedWaspRule->getSecondWatchValue().addWatchedConstraint( learnedWaspRule );
    learnedWaspRule->setIteratorSecondWatch( it2 );

    //Add the learned constraint.
    program.addLearnedWaspRule( learnedWaspRule );
    
    backtrackingStack.push( &( learnedWaspRule->getLockedAddress() ) );
    learnedWaspRule->setLocked();

    incrementLearnedWaspRuleActivity( learnedWaspRule );       
    
    trace( mg, 1, "Learned WaspRule %s", program.waspRuleToString( *learnedWaspRule ).c_str() );
    return learnedWaspRule;
}

LearnedWaspRuleFromModel*
ModelGenerator::insertConstraintFromModel(
    vector< AtomId >& atomsToAddInConstraint )
{        
    unsigned int size = atomsToAddInConstraint.size();

    assert( size > 0 );
    assert( size < options.maxLearnedConstraintSize );
    
    statistics( incrementNumberOfLearnedWaspRules() );
    statistics( incrementSizeOfLearnedWaspRules( size ) );

    unsigned int learnedWaspRulesSize = program.getLearnedWaspRules().size();

    unordered_map< AtomId, unsigned int >* positions = new unordered_map< AtomId, unsigned int >();
    
    LearnedWaspRuleFromModel* learnedWaspRule = new LearnedWaspRuleFromModel( atomsToAddInConstraint.size() );

    for( vector< AtomId >::iterator i = atomsToAddInConstraint.begin(); i != atomsToAddInConstraint.end(); ++i )
    {
        AtomId atom = *i;
        assert( atom > 0 && atom < program.getNumberOfAtoms() );

        Atom* atomPointer = program.getAtomPointer( atom );
        Atom& a = *atomPointer;
        assert( interpretation.isTruePropagated( a ) || interpretation.isFalsePropagated( a ) || interpretation.isFalseUndefined( a ) );
        
        decisionHeuristic->onLearnedWaspRule( atom, interpretation.isTruePropagated( atom ) );
        
        if( interpretation.isTruePropagated( a ) )
        {            
            learnedWaspRule->addLiteral( Literal( atomPointer ) );
        }
        else
        {
            assert( interpretation.isFalsePropagated( a ) || interpretation.isFalseUndefined( a ) );
            learnedWaspRule->addLiteral( Literal( atomPointer, true ) );
        }
    }
    
    WatchedList< WaspRule* >::iterator it1 = learnedWaspRule->getFirstWatchValue().addWatchedConstraint( learnedWaspRule );
    learnedWaspRule->setIteratorFirstWatch( it1 );
    WatchedList< WaspRule* >::iterator it2 = learnedWaspRule->getSecondWatchValue().addWatchedConstraint( learnedWaspRule );
    learnedWaspRule->setIteratorSecondWatch( it2 );

    //Add the learned constraint.
    program.addLearnedWaspRule( learnedWaspRule );
    
    backtrackingStack.push( &( learnedWaspRule->getLockedAddress() ) );
    learnedWaspRule->setLocked();

    incrementLearnedWaspRuleActivity( learnedWaspRule );       
    
    trace( mg, 1, "%s", program.waspRuleToString( *learnedWaspRule ).c_str() );
    return learnedWaspRule;
}

bool
ModelGenerator::restart()
{
    trace( mg, 1, "Performing restart. \n" );
    unroll( 1 );    

    if( !propagationStack.empty() && !propagate() )
    {
        return false;
    }

    backtrackingAtomStack.clear();
    backtrackingLevelAtomStack.clear();
    propagationStack = stack< pair< unsigned int, TruthValue > >();
    if( options.geometricRestarts )
        nextRestartValue *= 1.5;
    else
        nextRestartValue = computeRestartNumber( countRestart ) * options.restartsNumbers;
    
    bool result = decisionHeuristic->onRestart( countRestart );
    if( !result )
        return false;

//    subsume();
    if( program.getLearnedWaspRules().size() > 500000 )
    {
        this->deleteLearnedConstraint();
    }    
    return true;
}

void
ModelGenerator::unroll(
    Level level )
{
	startTimer( unroll );
    conflictImplicantIdType.first = { NULL };
    conflictImplicantIdType.second = 0;
    backtrackingStack.unroll( level );
    stackInvolvedForGus.unroll( level );
    //atomsLostSourcePointer.clear();
    for( unsigned int i = 0; i < componentsInTheMap.size(); ++i )
    {
        ComponentId componentId = componentsInTheMap[ i ];
        mapComponentListOfAtomsLostSourcePointer[ componentId ].clear();
    }
    
    componentsInTheMap.clear();
	stopTimer( unroll );
}

void
ModelGenerator::unrollOne()
{
	startTimer( unrollOne );
    conflictImplicantIdType.first = { NULL };
    conflictImplicantIdType.second = 0;
    backtrackingStack.unrollOne();
    stackInvolvedForGus.unrollOne();
    //atomsLostSourcePointer.clear();
    for( unsigned int i = 0; i < componentsInTheMap.size(); ++i )
    {
        ComponentId componentId = componentsInTheMap[ i ];
        mapComponentListOfAtomsLostSourcePointer[ componentId ].clear();
    }
    
    componentsInTheMap.clear();
	stopTimer( unrollOne );
}

void
ModelGenerator::nextLevel()
{
    backtrackingStack.nextLevel();
    stackInvolvedForGus.nextLevel();
}

void
ModelGenerator::pushAtomInStack(
    Atom& atom,
    TruthValue truthValue )
{
    assert( truthValue == TRUE || truthValue == FALSE );
    trace( mg, 1, "Pushing atom %s in stack with truth value %s.\n", atom.getAtomName().c_str(), truthValue == TRUE ? "true" : "false" );

    AtomId atomId = atom.getId();

    //Insert the atomId in the stack with the truth value stored in the variable truthValue.
    if( truthValue == TRUE )
    {
        if( interpretation.isFalsePropagated( atomId ) )
        {
            if( conflictLiteral.first == 0 )
            {
                conflictLiteral.first = atomId;
                conflictLiteral.second = UNDEFINED;
                trace( mg, 1, "Setting %s as conflict Literal truth value undefined.\n", atom.getAtomName().c_str() );
            }
        }
        else if( interpretation.isFalseUndefined( atomId ) )
        {
            if( conflictLiteral.first == 0 )
            {
                conflictLiteral.first = atomId;
                conflictLiteral.second = FALSE;
                trace( mg, 1, "Setting %s as conflict Literal truth value false.\n", atom.getAtomName().c_str() );
            }
        }
        else
        {
            if( interpretation.isUndefined( atomId ) && !interpretation.isTrueUndefined( atomId ) )
            {
                backtrackingStack.push( &( interpretation.atoms[ atomId ] ) );
                interpretation.setAtomTruthValue( atomId, TRUE_UNDEFINED );
                propagationStack.push( AtomTruthValue( atomId, truthValue ) );
            }
        }
    }
    else
    {
        assert( truthValue == FALSE );
        if( interpretation.isTruePropagated( atomId ) )
        {
            if( conflictLiteral.first == 0 )
            {
                conflictLiteral.first = atomId;
                conflictLiteral.second = UNDEFINED;
                trace( mg, 1, "Setting %s as conflict Literal truth value undefined.\n", atom.getAtomName().c_str() );
            }
        }
        else if( interpretation.isTrueUndefined( atomId ) )
        {
            if( conflictLiteral.first == 0 )
            {
                conflictLiteral.first = atomId;
                conflictLiteral.second = TRUE;
                trace( mg, 1, "Setting %s as conflict Literal truth value true.\n", atom.getAtomName().c_str() );
            }
        }
        else{
            if( interpretation.isUndefined( atomId ) && !interpretation.isFalseUndefined( atomId ) )
            {
                backtrackingStack.push( &( interpretation.atoms[ atomId ] ) );
                interpretation.setAtomTruthValue( atomId, FALSE_UNDEFINED );
                propagationStack.push( AtomTruthValue( atomId, truthValue ) );
            }
        }
    }
}

void
ModelGenerator::aggregatesPropagation(
    Atom& a )
{
    ( this->*( a.getPropagationFunction() ) )( a );
}

void 
ModelGenerator::computeReasonsForAggregate( 
    InvolvedVector* involvedRules, 
    Aggregate& aggregate )
{
    involvedRules->push_back( aggregate.getId() );
    
    unsigned int size = aggregate.getSize();
    //Computing reasons.
    for( unsigned int i = 0; i < size; i++ )
    {
        Atom& atom = *( aggregate.getLiteral( i ).getAtom() );

        if( interpretation.isTruePropagated( atom ) ||
            interpretation.isFalsePropagated( atom ) )
        {
            involvedRules->push_back( atom.getId() );
        }
    }
}

void 
ModelGenerator::computeTrueAtomAsReasonsForAggregate( 
        InvolvedVector* involvedRules, 
        Aggregate& aggregate )
{
    unsigned int size = aggregate.getSize();
//    int aggregateValue = aggregate.getMinGuard();
    
    for( int i = size - 1; i >= 0; i-- )
    {
        Literal& literal = aggregate.getLiteral( i );
        if( interpretation.isTruePropagated( literal ) )
        {
            AtomId atomId = literal.getAtomId();
            involvedRules->push_back( atomId );
//            aggregateValue -= aggregate.getLiteralValue( i );
            
            //TODO: decommentare.
//            if( aggregateValue <= 0 )
//            {
//                return;
//            }
        }
    }
}

void 
ModelGenerator::computeFalseAtomAsReasonsForAggregate( 
        InvolvedVector* involvedRules, 
        Aggregate& aggregate )
{
    unsigned int size = aggregate.getSize();    
    for( int i = size - 1; i >= 0; i-- )
    {
        Literal& literal = aggregate.getLiteral( i );
        if( interpretation.isFalsePropagated( literal ) )
        {
            AtomId atomId = literal.getAtomId();        
            involvedRules->push_back( atomId );           
        }
    }    
}

void 
ModelGenerator::aggregateIsFalse(
    Aggregate& aggregate )
{
    bool haveToComputeReasons = true;
    InvolvedVector* involvedRules;
    ReasonPointer reasonPointer;
    unsigned int minGuard = aggregate.getMinGuard();
       
    assert( aggregate.getCurrentValue() < minGuard );
    if( ( aggregate.getRightWatchValue() + aggregate.getCurrentValue() ) >= minGuard )
    {
        unsigned int rightWatch = aggregate.getRightWatch();
        
//        backtrackingStack.push( &( aggregate.getRightWatchAddress() ) );
        assert( interpretation.isUndefined( aggregate.getRightWatchLiteral() ) );
//        while( ( aggregate.getRightWatchValue() + aggregate.getCurrentValue() ) >= minGuard )
        while( ( aggregate.getLiteralValue( rightWatch ) + aggregate.getCurrentValue() ) >= minGuard )
        {
//            Atom& currentAtom = *( aggregate.getRightWatchAtom() );
            Literal& currentLiteral = aggregate.getLiteral( rightWatch );
            if( interpretation.isUndefinedNotFalse( currentLiteral ) )
//            if( interpretation.isUndefined( currentAtom ) && !interpretation.isFalse( currentAtom ) )
            {
                if( haveToComputeReasons )
                {
                    involvedRules = new InvolvedVector();
                    reasonPointer.involvedVector = involvedRules;
                    computeReasonsForAggregate( involvedRules, aggregate );
                    haveToComputeReasons = false;
                }
                
                Atom& currentAtom = *( currentLiteral.getAtom() );
                TruthValue truth = currentLiteral.isPositive() ? FALSE : TRUE;
                addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
                this->pushAtomInStack( currentAtom, truth );
            }

            if( aggregate.getLeftWatch() == rightWatch )
                break;
            
            rightWatch--;
//            aggregate.updateRightWatch();
        }
    }

    if( !haveToComputeReasons )
    {
        assert( reasonPointer.involvedVector == involvedRules );        
        reasonsForGus.push_back( involvedRules );
    }
}

void
ModelGenerator::aggregateIsTrue(
    Aggregate& aggregate )
{
    bool haveToComputeReasons = true;
    InvolvedVector* involvedRules;
    ReasonPointer reasonPointer;
    unsigned int minGuard = aggregate.getMinGuard();
    
    assert( aggregate.getMaxPossibleValue() >= minGuard );
    if( ( aggregate.getMaxPossibleValue() - aggregate.getRightWatchValue() ) < minGuard )
    {
        unsigned int rightWatch = aggregate.getRightWatch();
        
//        backtrackingStack.push( &( aggregate.getRightWatchAddress() ) );
        assert( interpretation.isUndefined( aggregate.getRightWatchLiteral() ) );
        
//        while( ( aggregate.getMaxPossibleValue() - aggregate.getRightWatchValue() ) < minGuard )
        while( ( aggregate.getMaxPossibleValue() - aggregate.getLiteralValue( rightWatch ) ) < minGuard )
        {
//            Atom& currentAtom = *( aggregate.getRightWatchAtom() );
            Literal& currentLiteral = aggregate.getLiteral( rightWatch );
            
//            if( interpretation.isUndefined( currentAtom ) && !interpretation.isTrue( currentAtom ) )
            if( interpretation.isUndefinedNotTrue( currentLiteral ) )
            {
                if( haveToComputeReasons )
                {
                    involvedRules = new InvolvedVector();
                    reasonPointer.involvedVector = involvedRules;
                    computeReasonsForAggregate( involvedRules, aggregate );
                    haveToComputeReasons = false;
                }

//                addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
//                this->pushAtomInStack( currentAtom, TRUE );
                Atom& currentAtom = *( currentLiteral.getAtom() );
                TruthValue truth = currentLiteral.isPositive() ? TRUE : FALSE;
                addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
                this->pushAtomInStack( currentAtom, truth );
            }

            if( aggregate.getLeftWatch() == rightWatch )
                break;
            
            rightWatch--;
            
//            if( aggregate.getLeftWatch() == aggregate.getRightWatch() )
//                break;
//            aggregate.updateRightWatch();                
        }
    }

    if( !haveToComputeReasons )
    {
        assert( reasonPointer.involvedVector == involvedRules );
        reasonsForGus.push_back( involvedRules );                
    }
}

void 
ModelGenerator::literalInTheAggregateSetIsTrue( 
    Atom& atom, 
    Aggregate& aggregate,
    unsigned int position )
{
    this->backtrackingStack.push( &( aggregate.getCurrentValueAddress() ) );
    aggregate.atomSetTrue( position );

    if( aggregate.getCurrentValue() >= aggregate.getMinGuard() )
    {
        //This is the atom representing the aggregate.
        Atom& atom = program.getAtom( aggregate.getId() );

        if( !interpretation.isTrue( aggregate.getId() ) )
        {                
            //The vector containing the reasons.
            InvolvedVector* involvedRules = new InvolvedVector();
            ReasonPointer reasonPointer;
            reasonPointer.involvedVector = involvedRules;

            //Computing the reasons.
            computeTrueAtomAsReasonsForAggregate( involvedRules, aggregate );

            addInvolvedRules( atom, reasonPointer, TYPEVECTOR );

            //We infer that the aggregate is true.
            this->pushAtomInStack( atom, TRUE );
            reasonsForGus.push_back( involvedRules );
        }
    }

    if( aggregate.getLeftWatchLiteral().getAtom() == &atom )
    {
        backtrackingStack.push( &( aggregate.getLeftWatchAddress() ) );
        while( !interpretation.isUndefined( aggregate.getLeftWatchLiteral() ) && aggregate.getLeftWatch() != aggregate.getRightWatch() )
        {
            aggregate.updateLeftWatch();
        }
    }
    else if( aggregate.getRightWatchLiteral().getAtom() == &atom )
    {
        backtrackingStack.push( &( aggregate.getRightWatchAddress() ) );
        while( !interpretation.isUndefined( aggregate.getRightWatchLiteral() ) && aggregate.getLeftWatch() != aggregate.getRightWatch() )
        {
            aggregate.updateRightWatch();
        }
    }

    assert( aggregate.getLeftWatch() == aggregate.getRightWatch() || 
            ( interpretation.isUndefined( aggregate.getLeftWatchLiteral() ) && interpretation.isUndefined( aggregate.getRightWatchLiteral() ) ) );                        

    if( interpretation.isFalsePropagated( aggregate.getId() ) && aggregate.getMaxPossibleValue() >= aggregate.getMinGuard() )
    {
        aggregateIsFalse( aggregate );                
    }
}
    
void
ModelGenerator::literalInTheAggregateSetIsFalse( 
    Atom& atom, 
    Aggregate& aggregate,
    unsigned int position )
{
    this->backtrackingStack.push( &( aggregate.getMaxPossibleValueAddress() ) );            
    aggregate.atomSetFalse( position );

    if( aggregate.getMaxPossibleValue() < aggregate.getMinGuard() )
    {                
        if( !interpretation.isFalse( aggregate.getId() ) )
        {
            //This is the atom representing the aggregate.
            Atom& atom = program.getAtom( aggregate.getId() );

            //The vector containing the reasons. 
            InvolvedVector* involvedRules = new InvolvedVector();
            ReasonPointer reasonPointer;
            reasonPointer.involvedVector = involvedRules;

            //Computing the reasons.
            computeFalseAtomAsReasonsForAggregate( involvedRules, aggregate );

            addInvolvedRules( atom, reasonPointer, TYPEVECTOR );

            //We infer that the aggregate is true.
            this->pushAtomInStack( atom, FALSE );
            reasonsForGus.push_back( involvedRules );
        }
    }

    if( aggregate.getLeftWatchLiteral().getAtom() == &atom )
    {
        backtrackingStack.push( &( aggregate.getLeftWatchAddress() ) );
        while( !interpretation.isUndefined( aggregate.getLeftWatchLiteral() ) && aggregate.getLeftWatch() != aggregate.getRightWatch() )
        {
            aggregate.updateLeftWatch();
        }
    }
    else if( aggregate.getRightWatchLiteral() == &atom )
    {
        backtrackingStack.push( &( aggregate.getRightWatchAddress() ) );
        while( !interpretation.isUndefined( aggregate.getRightWatchLiteral() ) && aggregate.getLeftWatch() != aggregate.getRightWatch() )
        {
            aggregate.updateRightWatch();
        }
    }

    assert( aggregate.getLeftWatch() == aggregate.getRightWatch() || 
            ( interpretation.isUndefined( aggregate.getLeftWatchLiteral() ) && interpretation.isUndefined( aggregate.getRightWatchLiteral() ) ) );

    if( interpretation.isTruePropagated( aggregate.getId() ) && aggregate.getCurrentValue() < aggregate.getMinGuard() )
    {
        aggregateIsTrue( aggregate );
    }
}

void
ModelGenerator::aggregatesPropagationAtom(
    Atom& a )
{
    trace( mg, 2, "Evaluating aggregates where atom %s appears:\n", a.getAtomName().c_str() );
    
    AtomInAnAggregateSet& at = static_cast< AtomInAnAggregateSet& >( a );
    
    if( interpretation.isTruePropagated( a ) )
    {
        for( unsigned int i = 0; i < at.getAggregatesWhereAtomAppears().size(); i++ )
        {   
            Aggregate& aggregate = *( at.getAggregate( i ) );
            unsigned int position = at.getPositionInAggregatesWhereAtomAppears()[ i ];
            
            if( aggregate.getLiteral( position ).isPositive() )
                literalInTheAggregateSetIsTrue( a, aggregate, position );
            else
                literalInTheAggregateSetIsFalse( a, aggregate, position );
        }
    }
    else
    {
        assert( interpretation.isFalsePropagated( a ) );
                
        for( unsigned int i = 0; i < at.getAggregatesWhereAtomAppears().size(); i++ )
        {
            Aggregate& aggregate = *( at.getAggregate( i ) );
            unsigned int position = at.getPositionInAggregatesWhereAtomAppears()[ i ];
            
            if( aggregate.getLiteral( position ).isPositive() )
                literalInTheAggregateSetIsFalse( a, aggregate, position );
            else
                literalInTheAggregateSetIsTrue( a, aggregate, position );
        }
    }
}

void
ModelGenerator::aggregatesPropagationGreater(
    Atom& a )
{    
    trace( mg, 2, "Atom %s is an aggregate(greater) and we are evaluating its aggregate set.\n", a.getAtomName().c_str() );
    assert( a.isAggregate() );
    
    AggregateAtomGreater& aggregateAtom = static_cast< AggregateAtomGreater& >( a );
    
    assert( aggregateAtom.getAggregate() != NULL );
    Aggregate& aggregate = *( aggregateAtom.getAggregate() );    
    assert( a.getId() == aggregate.getId() );
    
    //The aggregate is true
    if( interpretation.isTruePropagated( a ) )
    {
        if( aggregate.getCurrentValue() < aggregate.getMinGuard() )
            aggregateIsTrue( aggregate );
    }//The aggregate is false.
    else
    {   
        assert( interpretation.isFalsePropagated( a ) );        
        if( aggregate.getMaxPossibleValue() >= aggregate.getMinGuard() )
            aggregateIsFalse( aggregate );        
    }
}

void
ModelGenerator::aggregatesPropagationDoubleGuards(
    Atom& a )
{
    assert( 0 );
    cerr << "Aggregates with double guards are not supported." << endl;
    exit( 0 );
//    trace( mg, 2, "Atom %s is an aggregate(double guards) and we are evaluating its aggregate set.\n", a.getAtomName().c_str() );
//    assert( a.isAggregate() );
//    
//    AggregateAtomDoubleGuards& aggregateAtom = static_cast< AggregateAtomDoubleGuards& >( a );
//    
//    Aggregate& aggregate = *( aggregateAtom.getAggregate() );
//    
//    unsigned int minGuard = aggregate.getMinGuard();
//    unsigned int maxGuard = aggregate.getMaxGuard();
//    assert( a.getId() == aggregate.getId() );
//  
//    //We have to compute reasons only for the first time that an atom is inferred.
//    bool haveToComputeReasons = true;
//
//    //The vector containing the reasons. 
//    InvolvedVector* involvedRules;
//    ReasonPointer reasonPointer;
//    
//    //The aggregate is true
//    if( interpretation.isTruePropagated( a ) )
//    {        
//        //In this case we have to infer all undefined atoms in the aggregate set as true.
//        //The reasons are the already defined atoms in the aggregate set.
//        if( aggregate.getMaxPossibleValue() == minGuard )
//        {
//            unsigned int size = aggregate.getSize();                        
//
//            for( int i = 0; i < size; i++ )
//            {
//                Atom& currentAtom = aggregate.getLiteral( i );
//                if( interpretation.isUndefined( currentAtom ) )
//                {
//                    if( haveToComputeReasons )
//                    {
//                        involvedRules = new InvolvedVector();
//                        reasonPointer.involvedVector = involvedRules;
//                        computeReasonsForAggregate( involvedRules, aggregate );
//                        haveToComputeReasons = false;
//                    }
//                    
//                    addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
//                    this->pushAtomInStack( currentAtom, TRUE );
//                }
//            }            
//        }
//        //In this case, if there exists an atom A such that the value of
//        //A makes the aggregate false then we have to infer A as false.
//        else if( aggregate.getCurrentValue() <= maxGuard )
//        {
//            unsigned int size = aggregate.getSize();
//            for( int i = 0; i < size; i++ )
//            {
//                Atom& currentAtom = aggregate.getLiteral( i );
//                if( interpretation.isUndefined( currentAtom ) )
//                {
//                    unsigned int value = aggregate.getLiteralValue( i );
//                    unsigned int currentValue = aggregate.getCurrentValue();
//
//                    if( ( value + currentValue) > maxGuard )
//                    {
//                        if( haveToComputeReasons )
//                        {
//                            involvedRules = new InvolvedVector();
//                            reasonPointer.involvedVector = involvedRules;
//                            //Maybe in this case we can consider as reason only true atoms in the aggregate set.
//                            computeReasonsForAggregate( involvedRules, aggregate );
//                            haveToComputeReasons = false;
//                        }
//                        
//                        addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
//                        this->pushAtomInStack( currentAtom, FALSE );
//                    }
//                }
//            }
//        }
//    }//The aggregate is false.
//    else
//    {
//        assert( interpretation.isFalsePropagated( a ) );
//        
//        if( aggregate.getMaxPossibleValue() <= maxGuard )
//        {
//            unsigned int size = aggregate.getSize();
//            for( int i = 0; i < size; i++ )
//            {
//                Atom& currentAtom = aggregate.getLiteral( i );
//                if( interpretation.isUndefined( currentAtom ) )
//                {
//                    unsigned int value = aggregate.getLiteralValue( i );
//                    unsigned int currentValue = aggregate.getCurrentValue();
//
//                    if( ( value + currentValue) >= minGuard )
//                    {
//                        if( haveToComputeReasons )
//                        {
//                            involvedRules = new InvolvedVector();
//                            reasonPointer.involvedVector = involvedRules;
//                            computeReasonsForAggregate( involvedRules, aggregate );
//                            haveToComputeReasons = false;
//                        }
//                        
//                        addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
//                        this->pushAtomInStack( currentAtom, FALSE );
//                    }
//                }
//            }
//        }
//        else if( aggregate.getCurrentValue() >= minGuard &&
//                 aggregate.getMaxPossibleValue() == ( maxGuard + 1 ) )
//        {
//            unsigned int size = aggregate.getSize();
//
//            for( int i = 0; i < size; i++ )
//            {
//                Atom& currentAtom = aggregate.getLiteral( i );
//                if( interpretation.isUndefined( currentAtom ) )
//                {
//                    if( haveToComputeReasons )
//                    {
//                        InvolvedVector* involvedRules = new InvolvedVector();
//                        reasonPointer.involvedVector = involvedRules;
//                        computeReasonsForAggregate( involvedRules, aggregate );
//                        haveToComputeReasons = false;
//                    }
//                    
//                    addInvolvedRules( currentAtom, reasonPointer, TYPEVECTOR );
//                    this->pushAtomInStack( currentAtom, TRUE );
//                }
//            }            
//        }                
//    }
//    
//    if( !haveToComputeReasons )
//        reasonsForGus.push_back( involvedRules );
}

void
ModelGenerator::onLearningAWaspRuleRestartBasedPolicy()
{    
    unsigned int value1 = ( programSize / 3.0 ) * pow( 1.1, countRestart );
    unsigned int value2 = programSize * 3;
    
    unsigned int min = value1 > value2 ? value1 : value2;    
    
    if( program.getLearnedWaspRules().size() > min )
    {
        deleteLearnedConstraint();
    }
}

void
ModelGenerator::onLearningAWaspRule()
{
    conservativeDeletionPolicy();
}

void
ModelGenerator::onFindingAModel()
{
    conservativeDeletionPolicy();
}

//void
//ModelGenerator::computePT(
//    vector< pair< AtomId, unsigned int > >& possTrue )
//{
//    unordered_set< AtomId > possTrueSet;
//
//    for( int i = interpretation.getLeftWatch(); i <= interpretation.getRightWatch(); i++ )
//    {
//        if( interpretation.isUndefined( i ) && possTrueSet.find( i ) == possTrueSet.end() )
//        {
//            if( isPossiblyTrue( i ) )
//            {
//                possTrueSet.insert( i );
//                Atom& a = *( program.getAtoms()[ i ] );
//                possTrue.push_back( pair< AtomId, unsigned int >( i, a.getTotalCounter() ) );
//            }
//        }
//    }
//    sort( possTrue.begin(), possTrue.end(), compareAtoms );
//}

//void
//ModelGenerator::computePTWithBerkmin(
//    vector< pair< AtomId, unsigned int > >& possTrue )
//{
//    bool found = false;
//
//    list< BodyId >& learnedConstraints = program.getLearnedWaspRulesOrderedVector();
//
//    for( list< BodyId >::reverse_iterator it = learnedConstraints.rbegin(); it != learnedConstraints.rend(); it++ )
//    {
//        BodyId i = *it;
//        Body& body = program.getLearnedConstraints()[ i ];
//
//        if( &body != NULL && !body.isFalse() )
//        {
//            ConjunctiveBody& b = *( static_cast< ConjunctiveBody* >( &body ) );
//            
//            assert( !b.isTrueStatic() );
//            assert( b.getSizeStatic() > 0 );
//            for( int j = b.getLeftWatch(); j <= b.getRightWatch(); j++ )
//            {
//                Atom& a = *( b.getLiteralsStatic()[ j ].getAtom() );
//                if( interpretation.isUndefined( a ) )
//                {
//                    possTrue.push_back( pair< AtomId, unsigned int >( a.getId(), a.getTotalCounter() ) );
//                    found = true;
//                }
//            }
//            if( found )
//                break;
//        }
//    }
//    sort( possTrue.begin(), possTrue.end(), compareAtoms );
//}

////NEW
//void
//ModelGenerator::computePT()
//{
//    Counter max = 0;
//    AtomId chosenAtom = 0;
//
//    Counter maxWithoutPT = 0;
//    AtomId chosenAtomWithoutPT = 0;
//
//    Counter mostOccurrencyMax = 0;
//    AtomId mostOccurrencyId = 0;
//
//    for( int i = interpretation.getLeftWatch(); i <= interpretation.getRightWatch(); i++ )
//    {
//        if( interpretation.isUndefined( i ) )
//        {
//            Atom& currentAtom = *( program.getAtoms()[ i ] );
//            if( max == 0 && maxWithoutPT == 0 && currentAtom.getOccurrences() > mostOccurrencyMax )
//            {
//                mostOccurrencyMax = currentAtom.getOccurrences();
//                mostOccurrencyId = i;
//            }
//
//            if( currentAtom.getTotalCounter() > max )
//            {
//                if( max == 0 )
//                {
//                    if( currentAtom.getTotalCounter() > maxWithoutPT )
//                    {
//                        maxWithoutPT = currentAtom.getTotalCounter();
//                        chosenAtomWithoutPT = i;
//                    }
//                }
//
//                if( isPossiblyTrue( i ) )
//                {
//                    if( currentAtom.getTotalCounter() > max )
//                    {
//                        max = currentAtom.getTotalCounter();
//                        chosenAtom = i;
//                    }
//                }                
//            }
//        }
//    }
//
//    if( max > 0 )
//    {
//        assert( chosenAtom != 0 );
//        currentChoice.first = chosenAtom;
//        choosePolarity( *( program.getAtoms()[ currentChoice.first ] ) );
//    }
//    else if( maxWithoutPT > 0 )
//    {        
//        assert( chosenAtomWithoutPT != 0 );
//        currentChoice.first = chosenAtomWithoutPT;
//        choosePolarity( *( program.getAtoms()[ currentChoice.first ] ) );
//    }
//    else
//    {
//        assert( mostOccurrencyMax > 0 );
//        assert( mostOccurrencyId > 0 );
//
//        currentChoice.first = mostOccurrencyId;
//        Atom& a = *( program.getAtoms()[ mostOccurrencyId ] );
//        currentChoice.second = a.isMoreNegativeOccurrences() ? FALSE : TRUE;
//    }
//}

//void
//ModelGenerator::computeUTs(
//    int& ut,
//    int& ut2,
//    int& ut3,
//    int& trueAtoms )
//{
//    startTimer( computeUTs );
//    for( int i = 1; i < interpretation.getSize(); i++ )
//    {
//        Atom& a = *( program.getAtoms()[ i ] );
//        if( interpretation.isTruePropagated( a ) && !a.isSupported() )
//        {
//            ut++;
//            int count = 0;
//           
//            for( int j = a.getBottomWatch(); j < a.getTopWatch() && count <= 3; j++ )
//            {
//                Rule& r = *( a.getRulesAtomInHead()[ j ] );
//
//                if( isRuleActive( r, i ) )
//                {
//                    count++;
//                }
//            }
//
//            if( count == 2 )
//                ut2++;
//            else if( count == 3 )
//                ut3++;
//        }
//        else if( interpretation.isTruePropagated( a ) )
//        {
//            trueAtoms++;
//        }
//    }
//    stopTimer( computeUTs );
//}

//void
//ModelGenerator::computeSatisfiedRules( 
//    int& count,
//    int& supportingRulesForTrueAtoms )
//{
//    for( int i = 1; i < program.getRules().size(); i++ )
//    {
//        Rule& r = *( program.getRules()[ i ] );
//        Body& b = *( r.getBody() );
//        Head& h = *( r.getHead() );
//
//        bool bodyFalse = b.isFalse();
//
//        if( bodyFalse )
//            count++;
//
//        if( h.getFirstTrueAtom() != 0 )
//        {
//            count++;
//            if( h.getSecondTrueAtom() == 0 && bodyFalse )
//            {
//                supportingRulesForTrueAtoms++;
//            }
//        }        
//    }
//}

//bool
//ModelGenerator::isPossiblyTrue(
//    AtomId atomId )
//{
//    startTimer( isPTFunction );
//    Atom& atom = *( program.getAtoms()[ atomId ] );
//
//    const vector< Rule* >& rulesInHead = atom.getRulesAtomInHead();
//
//    for( int i = 0; i < rulesInHead.size(); i++ )
//    {
//        Rule& r = *( rulesInHead[ i ] );
//        Head& head = *( r.getHead() );
//        
//        if( head.isValid() )
//        {
//            Body& body = *( r.getBody() );
//
//            if( body.isTrue() )
//            {
//                stopTimer( isPTFunction );
//                return true;
//            }
//        }
//    }
//
//    const vector< ConjunctiveBody* >& negativeBodies = atom.getNegativeConjunctiveBodies();
//
//    for( int i = 0; i < negativeBodies.size(); i++ )
//    {
//        ConjunctiveBody& b = *( negativeBodies[ i ] );
//
//        bool toContinue = true;
//
//        for( int j = 0; j < b.getSizeStatic(); j++ )
//        {
//            Literal& literal = b.getLiteralsStatic()[ j ];
//            if( literal.isPositive() && !interpretation.isTruePropagated( *( literal.getAtom() ) ) )
//            {
//                toContinue = false;
//                break;
//            }
//        }
//
//        if( toContinue )
//        {
//            vector< Rule* >& rulesBody = b.getRulesBodyStatic();
//            
//            for( int j = 0; j < rulesBody.size(); j++ )
//            {
//                Rule& r = *( rulesBody[ j ] );
//                Head& head = *( r.getHead() );
//
//                if( head.isValid() )
//                {
//                    Body& body = *( r.getBody() );
//                    if( body.isTrue() )
//                    {
//                        stopTimer( isPTFunction );
//                        return true;
//                    }
//                }
//            }
//        }
//    }
//
//    stopTimer( isPTFunction );
//    return false;
//}

//void ModelGenerator::findSubsumedFinal(
//	const int * atoms,
//	int size,
//	BodyId id,
//	const vector< BodyId >& bodyVector,
//	vector< BodyId >& result )
//{
//    unsigned int signatureBody = Util::computeSignature( atoms, size );
//    for( int i = 0; i < bodyVector.size(); i++ )
//    {
//        BodyId idBody = bodyVector[ i ] - program.getBodies().size();
//        assert( idBody >= 0 && idBody < program.getLearnedConstraints().size() );
//
//        Body& body = program.getLearnedConstraints()[ idBody ];
//
//        assert( &body != NULL );
//        ConjunctiveBody& b = *( static_cast< ConjunctiveBody* >( &body ) );
//
//        int* literals = new int[ b.getSizeStatic() ];
//        for( int j = 0; j < b.getSizeStatic(); j++ )
//        {
//            literals[ j ] = b.getLiteralsStatic()[ j ].getId();
//        }
//
//        if( id != idBody && size <= b.getSizeStatic() && subset( signatureBody, atoms, size, b.getSignatureStatic(), literals, b.getSizeStatic() ) )
//        {
//            result.push_back( idBody );
//        }
//
//        delete [] literals;
//    }
//}
//
//void ModelGenerator::findSubsumed(
//	BodyId bodyId,
//	const Literal* literals,
//	int size,
//	vector< BodyId >& result )
//{
//    assert( size > 0 );
//
//    int id = literals[ 0 ].getId();
//    unsigned int min = ( id > 0 ? program.getAtoms()[ id ]->getPositiveLearnedBodiesIds().size() :
//                                  program.getAtoms()[ -id ]->getNegativeLearnedBodiesIds().size() );
//    
//    //Looking for the atom which appears in the minimum number of bodies.
//
//    int* atoms = new int[ size ];
//
//    for( int i = 1; i < size; i++ )
//    {
//        int atom = literals[ i ].getId();
//        unsigned int numberOfBodies = ( atom > 0 ? program.getAtoms()[ atom ]->getPositiveLearnedBodiesIds().size() :
//                                                   program.getAtoms()[ -atom ]->getNegativeLearnedBodiesIds().size() );
//
//        if( numberOfBodies < min )
//        {
//            min = numberOfBodies;
//            id = atom;
//        }
//        atoms[ i ] = atom;
//    }
//
//    assert( id != 0 );
//
//    if( id > 0 )
//    	findSubsumedFinal( atoms, size, bodyId, program.getAtoms()[ id ]->getPositiveLearnedBodiesIds(), result );
//    else
//    	findSubsumedFinal( atoms, size, bodyId, program.getAtoms()[ -id ]->getNegativeLearnedBodiesIds(), result );
//
//    delete[] atoms;
//}
//
//bool ModelGenerator::subset(
//    unsigned int signatureBody,
//    const int * atoms,
//    int size,
//    unsigned int signatureBody2,
//    const int * atoms2,
//    int size2 )
//{    
//    if( ( ( signatureBody & 63 ) & ~( signatureBody2 & 63 ) ) != 0 )
//        return false;
//
//    int j = 0;
//    for( int i = 0; i < size; i++ )
//    {
//        bool found = false;
//        for( ; j < size2; j++ )
//        {
//            if( atoms[ i ] == atoms2[ j ] )
//            {
//                found = true;
//                break;
//            }
//        }
//        
//        if( !found )
//            return false;
//    }
//
//    return true;
//}
//
//void
//ModelGenerator::subsume()
//{
//    if( !options.doSubsumption )
//        return;
//    startTimer( subsumption );
//    
//    while( !constraintsCheckSubsumption.empty() )
//    {
//        BodyId i = constraintsCheckSubsumption.front();
//        constraintsCheckSubsumption.pop();
//        Body& body = program.getLearnedConstraints()[ i ];
//        if( &body != NULL )
//        {
//            ConjunctiveBody& b = *( static_cast< ConjunctiveBody* >( &body ) );
//            
//            vector< BodyId > result;            
//            findSubsumed( i, b.getLiteralsStatic(), b.getSizeStatic(), result );
//
//            for( int j = 0; j < result.size(); j++ )
//            {
//                statistics( incrementCountSubsumption() );
//                BodyId bodyId = result[ j ];
//
//                deleteLearnedConstraint( program.getLearnedConstraints(), bodyId );
//                program.getLearnedConstraints().remove( bodyId );
//            }
//
//            assert( this->selfSubsume != NULL );
//            ( this->*selfSubsume )( i, body );
//        }
//    }
//    stopTimer( subsumption );
//}
//
//void
//ModelGenerator::selfSubsumption(
//    BodyId bodyId,
//    Body & body )
//{
//    startTimer( selfsubsumption );
//
//    ConjunctiveBody* bb = static_cast< ConjunctiveBody* >( &body );
//    ConjunctiveBody& b = *bb;
//
//    int atomsSize = b.getSizeStatic();
//    int *atoms = new int[ atomsSize ];
//
//    for( int i = 0; i < atomsSize; i++ )
//    {
//        atoms[ i ] = b.getLiteralsStatic()[ i ].getId();
//    }
//
//    for( int i = 0; i < atomsSize; i++ )
//    {
//        int atom = atoms[ i ];
//
//        atoms[ i ] = -atom;
//        vector< BodyId > result;
//
//        if( atom < 0 )
//            findSubsumedFinal( atoms, atomsSize, bodyId, program.getAtoms()[ -atom ]->getPositiveLearnedBodiesIds(), result );
//        else
//            findSubsumedFinal( atoms, atomsSize, bodyId, program.getAtoms()[ atom ]->getNegativeLearnedBodiesIds(), result );
//
//        for( int j = 0; j < result.size(); j++ )
//        {
//            statistics( incrementCountSelfSubsumption() );
//            vector< int > atomsInConstraint;
//            BodyId bodyToDelete = result[ j ];
//            ConjunctiveBody& b = *( static_cast< ConjunctiveBody* >( &program.getLearnedConstraints()[ bodyToDelete ] ) );
//
//            for( int k = 0; k < b.getSizeStatic(); k++ )
//            {
//                int currentAtom = b.getLiteralsStatic()[ k ].getId();
//
//                if( currentAtom != -atom )
//                {
//                    atomsInConstraint.push_back( currentAtom );
//                }
//            }
//
//            deleteLearnedConstraint( program.getLearnedConstraints(), bodyToDelete );
//            program.getLearnedConstraints().remove( bodyToDelete );
//            list< BodyId >& bodiesToDelete = program.getLearnedConstraintsOrderedVector();
//
//            list< BodyId >::iterator it = bodiesToDelete.begin();
//            while( it != bodiesToDelete.end() )
//            {
//                BodyId bodyId = *it;
//                list< BodyId >::iterator tmp = it++;
//
//                if( bodyId == bodyToDelete )
//                {
//                    bodiesToDelete.erase( tmp );
//                    break;
//                }
//            }
//            
//            insertConstraintSubsumption( atomsInConstraint );
//        }
//        atoms[ i ] = atom;
//    }
//
//    delete [] atoms;
//    stopTimer( selfsubsumption );    
//}
//
//void
//ModelGenerator::selfSubsumptionDoNothing(
//    BodyId bodyId,
//    Body & b )
//{
//    return;
//}
//
//bool
//compareInteger( int i, int j )
//{
//    return ( abs( i ) < abs( j ) );
//}
//
//void
//ModelGenerator::insertConstraintSubsumption(
//    vector< int >& originalAtoms )
//{
//    sort( originalAtoms.begin(), originalAtoms.end(), compareInteger );
//
//    BodyId idBody = program.getBodies().size();
//    unsigned int firstFreePosition = program.getLearnedConstraints().getFirstFreePosition();
//
//    BodyId newId = idBody + firstFreePosition;
//    int size = originalAtoms.size();
//    Literal* literals = new Literal[ size ];
//
//    RuleId newRuleId = programSize + firstFreePosition;
//    unsigned int learnedConstraintsSize = program.getLearnedConstraints().size();    
//
//    unsigned int* positions = new unsigned int[ size ];
//    int j = 0;
//    
//    for( int i = 0; i < size; i++ )
//    {
//        int atom = originalAtoms[ i ];
//
//        assert( idBody != 0 );
//
//        if( atom > 0 )
//        {
//            Atom* atomPointer = program.getAtoms()[ atom ];
//            Atom& a = *atomPointer;
//            positions[ j ] = a.getPositiveLearnedBodiesIds().size();
//            literals[ j ].setAtom( atomPointer );
//            literals[ j ].setPositive();
//            a.addPositiveLearnedBodyId( newId );
//        }
//        else
//        {
//            Atom* atomPointer = program.getAtoms()[ -atom ];
//            Atom& a = *atomPointer;
//            positions[ j ] = a.getNegativeLearnedBodiesIds().size();
//            literals[ j ].setAtom( atomPointer );
//            literals[ j ].setNegative();
//            a.addNegativeLearnedBodyId( newId );
//        }
//        j++;
//    }
//
//    //Add the learned constraint.
//    if( firstFreePosition == learnedConstraintsSize )
//        program.addLearnedConstraints( size, literals, newId, positions );
//    else
//        program.getLearnedConstraints().push( firstFreePosition, size, literals, newId, positions );
//
//    ConjunctiveBody& b = program.getLearnedConstraints()[ firstFreePosition ];
//    
//    //Add rule in vector with negative id, because the body is "complex".
//    if( firstFreePosition == learnedConstraintsSize )
//        program.addRuleForLearnedConstraint( &b, newRuleId );
//    else
//        program.getRules()[ programSize + firstFreePosition ]->setBody( &b );
//
//    Rule* rule = program.getRules()[ programSize + firstFreePosition ];
//    assert( rule != NULL );
//
//    b.addRule( rule );
//    b.setLearnedStatic();
//    constraintsCheckSubsumption.push( firstFreePosition );
//
//    program.addLearnedConstraintsOrderedVector( firstFreePosition );
//
//    #ifndef TRACE_OFF
//    if( options.traceLevels.mg >= 1 )
//    {
//        cerr << "Subsumption constraint: ";
//        program.printRuleStdError( *( program.getRules()[ programSize + firstFreePosition ] ) );
//    }
//    #endif
//}
