/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   ModelGenerator.h
 * Author: Carmine Dodaro
 *
 * Created on 11 October 2010, 18.33
 */

#ifndef _MODELGENERATOR_H
#define	_MODELGENERATOR_H

#include <stack>
#include <iostream>
#include <cstdlib>
#include <set>
#include <algorithm>
#include <queue>
#include <unordered_set>
#include <unordered_map>
#include "stl/List.h"

#include "Constants.h"
#include "Interpretation.h"
#include "InputHandler.h"
#include "Program.h"
#include "stack.h"
#include "options.h"
#include "stack.h"
#include "timer.h"
#include "trace.h"
#include "ModelChecker.h"
#include "StackInvolvedForGUS.h"
#include "AtomInAnAggregateSet.h"
#include "DecisionHeuristic.h"
#include "LearnedWaspRuleFromModel.h"
#include "OutputBuilder.h"

using namespace std;
using namespace wasp;

/**
 * This class contains the model generator, i.e. 
 * it performs all operation about propagation, model generations and model checker.
 */
class ModelGenerator
{
    friend class DecisionHeuristic;
public:

    ModelGenerator( int heuristicId = BERKMIN_HEURISTIC );

    ModelGenerator( istream& input );
    
    ~ModelGenerator();

    void backtracking();
    
    inline bool checkModel();

    void computeGUS();   
    
    /**
     * This function compute the Greatest Unfounded Set for all Component.
     * 
     * @return false if an inconsistency is detected, true otherwise.
     */
    void computeGUS( List< Atom* >& atomsLostSourcePointer );   

    /**
     * This function returns the current interpretation.
     * @return the current interpretation.
     */
    inline Interpretation& getInterpretation();

    /**
     * This function performs basic operations.
     * @return false if an inconsistency is detected, true otherwise.
     */
    bool initialPropagation();

    /**
     * The model generator function.
     */
    void mg();
    
    /**
     * This function deterministically infers atoms truth values.
     *
     * @return false if an inconsistency is detected, true otherwise.
     */
    bool propagate();

    /**
     * This function deterministically infers atoms truth values.
     * 
     * @param the literal to propagate.
     * @return false if an inconsistency is detected, true otherwise.
     */
    inline int propagate( const AtomTruthValue& literal );

    /**
     * This function take the first atom from stack and propagate logical consequences of its truth assignment.
     *
     * @param atomId the atom id to propagate.
     * @return false if an inconsistency is detected, true otherwise.
     */
    inline void propagateAtom( Atom& atom );

    /**
     * This function print the value of interpretation.
     *
     */
    inline void printInterpretation();

    /**
     * This function print true and undefined atoms in the interpretation.
     *
     */
    void printInterpretationWithUndefined();

    Program& getProgram();

    static Interpretation* getCurrentInterpretation()
    {
        return currentInterpretation;
    }
    
    void aggregatesPropagationGreater( Atom& atom );
    
    void aggregatesPropagationDoubleGuards( Atom& atom );
    
    void aggregatesPropagationAtom( Atom& atom );
    
    inline void aggregatesPropagationNothing( Atom& atom );
    
private:    
    
    /**
     * The literals to propagate during the backtracking. In this stack when we add the literal to propagate in the next step.
     */
    vector< AtomTruthValue > backtrackingAtomStack;
    
    vector< Level > backtrackingLevelAtomStack;

    /**
     * This stack contains memory address of object added.
     * When a backtracking step is performed it's necessary to add the previous value of changed object, i.e. watches values
     * or atoms truth values.
     */
    Stack< unsigned int > backtrackingStack;
    
    struct
    {
        bool choiceFound;
        int forkBegin;
        vector< AtomId > uip;
        vector< unsigned int > nextUip;
        unordered_map< AtomId, unsigned int > connection;
    } calculateUIPdata;
    
    bool callComputeGUS;
    
    vector< ComponentId > componentsToComputeGus;
    
    /**
     * The atom that caused a conflict.
     */
    AtomId conflictAtomId;
    
    TruthValue conflictAtomIdTruthvalue;

    unsigned int conflictCount;

    unsigned int conflictCountForRestart;
    
    pair< InvolvedPointer, InvolvedType > conflictImplicantIdType;
    
    pair< AtomId, TruthValue > conflictLiteral;    
    
    Activity constraintDecrement;
    
    Activity constraintIncrement;
    
    DecisionHeuristic* decisionHeuristic;
    
//    queue< RuleId > constraintsCheckSubsumption;    
    
    unsigned int countRestart;
    
    /**
     * The current choice.
     */
    AtomTruthValue currentChoice;
    
    static Interpretation* currentInterpretation;
      
    /**
     * Interpretation for atoms in the program.
     */
    Interpretation interpretation;
    
    ModelChecker* mc;
    
    /**
     * The number of models have been found.
     */
    unsigned int models;

    unsigned int nextRestartValue;    
    
    /**
     * The program received in input.
     */
    Program program;
    
    OutputBuilder* outputBuilder;

    /*
     * The number of wasprules in the program.
     */
    unsigned int programSize;
    
    /**
     * The atoms stack. In this stack we add the atom for propagation.
     */
    stack< AtomTruthValue > propagationStack;
    
    vector< InvolvedVector* > reasonsForGus;

    StackInvolvedForGUS stackInvolvedForGus;
    
//    List< Atom* > atomsLostSourcePointer;
    
    unordered_map< ComponentId, List< Atom* > > mapComponentListOfAtomsLostSourcePointer;
    vector< ComponentId > componentsInTheMap;
    
    /**
     * This function adds the address of this variable in the backtracking stack.
     *
     * @param value the value to push in the stack.
     */
    inline void addAddressInBackTrackingStack( unsigned int & value );
    
    /**
     * This function sets the reason of an atom truth value assignment.
     *
     * @param atom the atom.
     * @param id the id of the derivation.
     * @param type the type of the reason.
     */
    inline void addInvolvedRules( Atom& atom, InvolvedPointer id, InvolvedType type );
    
    void aggregatesPropagation( Atom& atom );
    
    void aggregateIsFalse( Aggregate& aggregate );
    
    void aggregateIsTrue( Aggregate& aggregate );
    
    /**
     * This function adds in tempSet the atoms that appears in the rules that involved the atom a truth value.
     *
     * @param tempSet the set that contains atom id.
     * @param atomId the atom id.
     */
    void calculateAtomNeighbours( vector< AtomId >& atomToVisitStack, AtomId atom, InvolvedPointer idOfDerivation, InvolvedType typeOfDerivation );

    void calculateAtomNeighboursLearning( vector< AtomId >& atomToVisitStack, AtomId nextId, InvolvedPointer idOfDerivation, InvolvedType typeOfDerivation );
    
    void calculateAtomsTypeAll( WaspRule* rule, AtomId currentAtom, vector< AtomId >& atomsInvolved );
    
    /**
     * Calculate the first UIP.
     *
     */    
    AtomId calculateFirstUIP();
    
    AtomId calculateFirstUIPOfTwoNodes( AtomId node1, AtomId node2, AtomId choice );
    
    void calculateReasons( const RuleContainer< WaspRule* >& foundingRuleInComponent, vector< unsigned int >& involvedRules, AtomId atomId );
    
    void calculateReasonsForChoice( const RuleContainer< ChoiceRule* >& choiceRules, vector< unsigned int >& involvedRules, AtomId atomId );
        
    void checkForSatisfyWaspRules( const RuleContainer< WaspRule* >&, Atom& a );
    void checkForSatisfyLearnedWaspRules( const RuleContainer< WaspRule* >&, Atom& a );
    
    void checkForSatisfyWaspRulesForAtomsInHead( const RuleContainer< WaspRule* >&, Atom& a );
    
    void checkForUpdateWatches( WatchedList< WaspRule* >&, Atom& a );
    void checkForUpdateWatchesConstraints( WatchedList< WaspRule* >&, Atom& a );
    
    void checkForUpdateChoiceRulesStatus( const RuleContainer< ChoiceRule* >& choiceRuleVector, Atom& a );
    
    void checkForFalsifyChoiceRulesBody( const RuleContainer< ChoiceRule* >& choiceRuleVector, Atom& a );
    
    void computeDeterministicConsequences();
            
    void computeGUSFinal( List< Atom* >& atomsLostSourcePointer );
    
    bool isUnfoundedSet( List< Atom* >& atomsLostSourcePointer );
    
    void computeFalseAtomAsReasonsForAggregate( InvolvedVector* involvedRules, Aggregate& aggregate );
    
//    void computePT();

//    void computePT( vector< pair< AtomId, unsigned int > > & );

    void computeReasonsForAggregate( InvolvedVector* involvedRules, Aggregate& aggregate );
    
//    void computeReasonsForAtomInAggregate( InvolvedVector* involvedRules, Aggregate& aggregate );

//    void computeSatisfiedRules( int& count, int& supportingRulesForTrueAtom );
    
    void computeTrueAtomAsReasonsForAggregate( InvolvedVector* involvedRules, Aggregate& aggregate );
    
//    void computeUTs( int& ut, int& ut2, int& ut3, int& trueAtoms );

    inline void decrementConstraintActivity();
    
    inline void conservativeDeletionPolicy();
    
    void deleteLearnedConstraint();

    void deleteLearnedConstraint( LearnedWaspRule& learnedConstraint );  
    
    void onLearningAWaspRule();
    
    void onFindingAModel();
    
    void onLearningAWaspRuleRestartBasedPolicy();
    
    inline void onModelFound();
    
    bool findConnection( AtomId node, AtomId choice );    
    
    unsigned int estimateShortWaspRulesPropagation( const Literal& ) const;
    
//    void findSubsumed( BodyId bodyId, Body& b, vector< BodyId >& result );

//    void findSubsumed( BodyId bodyId, const Literal* atoms, int atomsSize, vector< BodyId >& result );

//    void findSubsumedFinal( const int * atoms, int size, BodyId id, const vector< BodyId >& bodyVector, vector< BodyId >& result );

    AtomId getReasonForSupport( WaspRule& rule, Atom& atom );
    
    void haveToCalculateGUS( Atom& atom, Rule* rule );        
    
    void incrementLearnedWaspRuleActivity( WaspRule* r );
    
    void init( int heuristicId = BERKMIN_HEURISTIC );
    
    LearnedWaspRule* insertConstraint( vector< AtomId >& atomToAddInConstraint );
    
    LearnedWaspRuleFromModel* insertConstraintFromModel( vector< AtomId >& atomToAddInConstraint );
    
    LearnedWaspRule* insertConstraintFromLoopFormula( vector< pair< AtomId, TruthValue > >& loopFormula );
    
//    void insertConstraintSubsumption( vector< int >& originalAtoms );
    
//    bool isPossiblyTrue( AtomId atomId );
    
    void lastSupportingRule( Atom& atom );
   
    /**
     * This function adds a constraint C in the program. C contains the literals that caused the conflict.
     *
     * @param firstUip the first UIP.
     */
    LearnedWaspRule* learnRule( AtomId firstUip, unsigned int& levelToBacktrack );
    
    void learnRuleModelFound();
    
    bool learnRuleModelCheckerFailure();
    
    bool updateAggregateRepresentingWeakConstraints();
    
//    void lookaheadHeuristic();

    void makeAChoice();
        
    void nextLevel();

    /**
     * Push an atom in the stack and set its truth value in the interpretation.
     *
     * @param atom the atom.
     * @param truthValue the atom truth value.
     * 
     */
    void pushAtomInStack( Atom& atom, TruthValue truthValue );

    void setWaspRuleLiteralsTrue( WaspRule& r, Atom& atom );
    
    void setWaspRuleLiteralsTrueChoiceRule( ChoiceRule& r );
    
    void setHeadAtomsChoiceRuleSupported( ChoiceRule& r );

    /**
     * This function checks if an atom is supported.
     *
     * @param atom the atom.     
     */
    void supportControl( Atom& atom );
    
    void unitPropagation( Atom& atom );
    
    void unroll( Level level );

    void unrollOne();
    
    // TODO: DOCS
    void updateInterpretationWatch();
    
    /**
     * This function is called to update the left watch of a wasprule.
     *
     * @param r the waspRule.
     */
    unsigned int updateFirstWatch( WaspRule& r );
    
    /**
     * This function is called to update the right watch of a wasprule.
     *
     * @param r the waspRule.
     */
    unsigned int updateSecondWatch( WaspRule& r );
        
    void updateSupportCountersForHeadAtoms( WaspRule& r, Atom* atom );
    
    void updateSupportCounterForHeadAtomsChoiceRule( ChoiceRule& r );
    
    void updateSupportCountersForOneAtom( WaspRule& r, Atom* atom );

    void literalInTheAggregateSetIsTrue( Atom&, Aggregate&, unsigned int );
    
    void literalInTheAggregateSetIsFalse( Atom&, Aggregate&, unsigned int );
    
    /*
     * Stop the computation and backtrack to the level 0.
     * 
     * @return false if an inconsistency is detected, true otherwise.
     */
    bool restart();
    
//    void selfSubsumption( BodyId bodyId, Body & b );

//    void selfSubsumptionDoNothing( BodyId bodyId, Body & b );
    
//    bool subset( const int * atoms, int size, const int * atoms2, int size2 );

//    bool subset( unsigned int signatureBody, const int * atoms, int size, unsigned int signatureBody2, const int * atoms2, int size2 );

//    void subsume();
    
    /**
     * This function search the first undefined in the interpretation and set ModelGenerator field firstUndefined with its value.
     */
    inline void setCurrentChoiceAccordingToAHeuristic();

//    void ( ModelGenerator::*selfSubsume )( BodyId bodyId, Body & b );
    
    #ifndef TRACE_OFF
    void printInterpretationForDebug()
    {
        unsigned int size = program.getNumberOfAtoms();
        for( unsigned int i = 1; i < size; i++ )
        {
            Atom& atom = program.getAtom( i );
            cerr << "[" << i << "]" << atom << ":(";

            if( interpretation.isTruePropagated( atom ) )
                cerr << "T,";
            else if( interpretation.isFalsePropagated( atom ) )
                cerr << "F,";
            else
                cerr << "U,";

            if( atom.getRecursionLevel() != MAXVALUE )
                cerr << atom.getRecursionLevel() << ",";
            else
                cerr << "ND,";

            if ( atom.isChoice() )
                cerr << "C);";
            else
                cerr << "NC);";
            cerr << " ";
            if( i % 3 == 0 )
                cerr << endl;
        }
        cerr << endl;
    }

    void printInterpretationForDebug( AtomId first, AtomId last )
    {
        for( unsigned int i = first; i < last; i++ )
        {
            Atom& atom = program.getAtom( i );
            cerr << "[" << i << "]" << atom << ":(";

            if( interpretation.isTruePropagated( atom ) )
                cerr << "T,";
            else if( interpretation.isFalsePropagated( atom ) )
                cerr << "F,";
            else
                cerr << "U,";

            if( atom.getRecursionLevel() != MAXVALUE )
                cerr << atom.getRecursionLevel() << ",";
            else
                cerr << "ND,";

            if ( atom.isChoice() )
                cerr << "C);";
            else
                cerr << "NC);";
            cerr << " ";
            if( i % 3 == 0 )
                cerr << endl;
        }
        cerr << endl;
    }
    
    void printLearnedWaspRulesDebug()
    {
        List< LearnedWaspRule* >& learnedWaspRules = program.getLearnedWaspRules();
        
        for( List< LearnedWaspRule* >::iterator i = learnedWaspRules.begin(); i != learnedWaspRules.end(); ++i )
        {
            if( *i != NULL )
            {
                cerr << *i << endl;
            }
        }
    }
    
    #endif
};

void
ModelGenerator::addAddressInBackTrackingStack(
    unsigned int& value )
{
    this->backtrackingStack.push( &value );
}

void
ModelGenerator::addInvolvedRules(
    Atom& atom,
    InvolvedPointer pointer,
    InvolvedType type )
{
    if( atom.getTypeOfDerivation() == 0 )
    {
        backtrackingStack.push( &atom.typeOfDerivation );
        atom.addInvolvedRules( pointer, type );
    }
    else if( conflictImplicantIdType.second == 0 )
    {
        conflictImplicantIdType.first = pointer;
        conflictImplicantIdType.second = type;
    }
    else
    {
    }
}

void
ModelGenerator::aggregatesPropagationNothing(
    Atom& )
{
}

bool
ModelGenerator::checkModel()
{
    startTimer( modelChecker );
    if( program.isHeadCycleFree() )
    {
        assert( program.hasChoiceRules() || mc->isStable( interpretation ) );
        stopTimer( modelChecker );
        return true;
    }
    else
    {    
        assert( !program.hasChoiceRules() );
        bool isStable = mc->isStable( interpretation );
        stopTimer( modelChecker );
        return isStable;
    }
}

void
ModelGenerator::decrementConstraintActivity()
{
    constraintIncrement *= constraintDecrement;
}

void
ModelGenerator::conservativeDeletionPolicy()
{
    if( program.getLearnedWaspRules().size() > programSize / 3 )
    {
        deleteLearnedConstraint();
    }
}

Interpretation&
ModelGenerator::getInterpretation()
{
    return interpretation;
}

// Return:
// * -1 if the propagation of literal return false;
// * +1 if the propagation of literal return true.
int
ModelGenerator::propagate(
    const AtomTruthValue & literal )
{
    assert( interpretation.isUndefined( literal.first ) );
    this->propagationStack.push( literal );
    return propagate() ? 1 : -1;
}

void
ModelGenerator::propagateAtom(
    Atom& atom )
{
    unitPropagation( atom );
    supportControl( atom );
    aggregatesPropagation ( atom );
}

void
ModelGenerator::setCurrentChoiceAccordingToAHeuristic()
{
    assert( interpretation.thereAreUndefinedAtoms() );
    
    decisionHeuristic->makeAChoice();
    currentChoice.first = decisionHeuristic->getHeuristicAtom();
    
    assert( currentChoice.first > 0 && currentChoice.first < program.getNumberOfAtoms() );
    assert( interpretation.isUndefined( currentChoice.first ) );
    
    currentChoice.second = decisionHeuristic->getHeuristicPolarity();
    assert( currentChoice.second == FALSE || currentChoice.second == TRUE );
}

void
ModelGenerator::onModelFound()
{
    outputBuilder->onAnswerSetFound( interpretation, program );
    if( program.hasWeakConstraints() )
        outputBuilder->onAnswerSetFoundWithWeakConstraint( interpretation, program );
}

void
ModelGenerator::printInterpretation()
{
    onModelFound();    
}

#endif	/* _MODELGENERATOR_H */
