/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "MostOccurrencesForAggregates.h"
#include <cassert>

MostOccurrencesForAggregates::MostOccurrencesForAggregates()
{
}

pair< bool, unsigned int >
MostOccurrencesForAggregates::findInBinCache(
    unsigned int n,
    unsigned int k )
{
    unordered_map< int, unordered_map< int, unsigned int > >::const_iterator it1 = binCache.find( n );
    if( it1 == binCache.end() )
        return pair< bool, unsigned int >( false, 0 );

    unordered_map< int, unsigned int >::const_iterator it = it1->second.find( k );
    if( it == it1->second.end() )
        return pair< bool, unsigned int >( false, 0 );
    else
        return pair< bool, unsigned int >( true, it->second );
}

void
MostOccurrencesForAggregates::insertIntoBinCache(
    unsigned int n,
    unsigned int k,
    unsigned int bin )
{
    unordered_map< int, unordered_map< int, unsigned int > >::const_iterator it1 = binCache.find( n );
    if( it1 == binCache.end() )
        binCache[ n ] = unordered_map< int, unsigned int >();

    binCache[ n ][ k ] = bin;
}

pair< bool, unsigned int >
MostOccurrencesForAggregates::findInBinSumCache(
    unsigned int n,
    unsigned int k )
{
    unordered_map< int, unordered_map< int, unsigned int > >::const_iterator it1 = binSumCache.find( n );
    if( it1 == binSumCache.end() )
        return pair< bool, unsigned int >( false, 0 );

    unordered_map< int, unsigned int >::const_iterator it = it1->second.find( k );
    if( it == it1->second.end() )
        return pair< bool, unsigned int >( false, 0 );
    else
        return pair< bool, unsigned int >( true, it->second );
}

void
MostOccurrencesForAggregates::insertIntoBinSumCache(
    unsigned int n,
    unsigned int k,
    unsigned int bin )
{
    unordered_map< int, unordered_map< int, unsigned int > >::const_iterator it1 = binSumCache.find( n );
    if( it1 == binSumCache.end() )
        binSumCache[ n ] = unordered_map< int, unsigned int >();

    binSumCache[ n ][ k ] = bin;
}

unsigned int
MostOccurrencesForAggregates::binomialCoefficient(
    unsigned int n,
    unsigned int k )
{
    if( k > n )
        return 0;

    if( k > n / 2 )
        k = n - k; // faster

    pair< bool, unsigned int > cached = findInBinCache( n, k );
    if( cached.first )
        return cached.second;

    long double accum = 1;
    for( unsigned int i = 1; i <= k; i++ )
            accum = accum * ( n - k + i ) / i;

    unsigned int ris = accum + 0.5; // avoid rounding error

    insertIntoBinCache( n, k, ris );

    return ris;
}

unsigned int 
MostOccurrencesForAggregates::sumOfBinomialCoefficients(
    int from,
    int toInclusive,
    int n )
{
    pair< bool, unsigned int > cached = findInBinSumCache( from, toInclusive );
    if( cached.first )
        return cached.second;

    unsigned int ris = 0;
    for( int i = from; i <= toInclusive; i++ )
        ris += binomialCoefficient( n, i );

    insertIntoBinCache( from, toInclusive, ris );

    return ris;
}

void 
MostOccurrencesForAggregates::computeCombinatorialHeursticForCount(
    int normalizedUpperGuard,
    int normalizedLowerGuard,
    unsigned int& aggSize,
    unsigned int& aggOcc,
    unsigned int& posConjOcc,
    unsigned int& negConjOcc )
{    
    bool OptionAggBinomialSumVSIDSHeur = true;
    // consider Lg < #count < Ug as: #count > Lg AND #count < Ug
    // NOTE: This function is not supposed to work with negative integers inside the aggregate.
    if( OptionAggBinomialSumVSIDSHeur )
    {
        if( normalizedLowerGuard < 0 || normalizedLowerGuard > aggSize )
        {
            // #count > -1 always satisfied, equivalent program is a single fact h.
            // #count > Lg never satisfied if Lg > |A|, equivalent program is :- h.
            aggOcc++;
        }
        else
        {
            assert( aggSize - 1 >= normalizedLowerGuard );
            aggOcc+= sumOfBinomialCoefficients( 0, aggSize - normalizedLowerGuard - 1, aggSize );
            posConjOcc = sumOfBinomialCoefficients( normalizedLowerGuard, aggSize - 1, aggSize - 1 );
        }

        if( normalizedUpperGuard <= 0 || aggSize < static_cast< unsigned int >( normalizedUpperGuard ) )
        {
            // #count < 0 never satisfied, equivalent program is :- h.
            // #count < Ug always satisfied if |A| < Ug OR Ug<0, equivalent program is a single fact h.
            aggOcc++;
        }
        else
        {
            aggOcc += sumOfBinomialCoefficients( 0, normalizedUpperGuard - 1, aggSize );
            negConjOcc = sumOfBinomialCoefficients( 0, normalizedUpperGuard - 1, aggSize - 1 );
        }
    }
    else
    {
        if( normalizedLowerGuard < 0 || normalizedLowerGuard > aggSize )
        {
            // #count > -1 always satisfied, equivalent program is a single fact h.
            // #count > Lg never satisfied if Lg > |A|, equivalent program is :- h.
            aggOcc++;
        }
        else
        {
            assert( aggSize - 1 >= normalizedLowerGuard );
            aggOcc += binomialCoefficient( aggSize, aggSize - normalizedLowerGuard - 1 );
            posConjOcc = binomialCoefficient( aggSize - 1, normalizedLowerGuard );
        }

        if( normalizedUpperGuard <= 0 || aggSize < static_cast< unsigned int >( normalizedUpperGuard ) )
        {
            // #count < 0 never satisfied, equivalent program is :- h.
            // #count < Ug always satisfied if |A| < Ug OR Ug<0, equivalent program is a single fact h.
            aggOcc++;
        }
        else
        {
            aggOcc += binomialCoefficient( aggSize, normalizedUpperGuard - 1 );
            negConjOcc = binomialCoefficient( aggSize - 1, normalizedUpperGuard - 1 );
        }
    }
}

void
MostOccurrencesForAggregates::computeCounters(
    Program* program )
{
 
    vector< Aggregate* >& aggregates = program->getAggregates();
    
    unsigned int size = aggregates.size();
    for( unsigned int i = 0; i < size; i++ )
    {
        Aggregate& aggregate = *( aggregates[ i ] );
        
        // To avoid getting crazy, guards are normalized
        // in such way that they are *not* inclusive i.e. Lg < #AGG < Ug
        unsigned int normalizedUpperGuard = normalizeMaxGuard( aggregate.getMaxGuard() );
        unsigned int normalizedLowerGuard = normalizeMinGuard( aggregate.getMinGuard() );

        unsigned int aggSizeSimulated = 0;
        unsigned int aggOcc = 0;
        unsigned int posConjOcc = 0;
        unsigned int negConjOcc = 0;

        unsigned int aggregateSize = aggregate.getSize();
        //Da migliorare, calcolare solo per il primo nell'aggregate set.
        
        // we basically approximate sum by count, each v:conj
        // is considered like v different v,x:conj
        for( unsigned int j = 0; j < aggregateSize; j++ )
        {
            unsigned int value = aggregate.getLiteralValue( j );
            if( value > 0 )
                aggSizeSimulated += value;
            else
                aggSizeSimulated += 1;
        }

        assert( aggSizeSimulated >= aggregate.getSize() );

        computeCombinatorialHeursticForCount( normalizedUpperGuard, normalizedLowerGuard,
                        aggSizeSimulated, aggOcc, posConjOcc, negConjOcc );

        AtomId aggregateId = aggregate.getId();
        AggregateAtomGreater& aggregateAtom = *( static_cast< AggregateAtomGreater* >( program->getAtomPointer( aggregateId ) ) );
        
        aggregateAtom.incrementPositiveOccurrencesForAggregates( aggOcc );
        
        if( posConjOcc || negConjOcc )
        {
            for( unsigned int j = 0; j < aggregateSize; j++ )
            {
                unsigned int value = aggregate.getLiteralValue( j );
                if( value > 0 )
                    aggSizeSimulated += value;
                else
                    aggSizeSimulated += 1;
            }
            
            for( unsigned int j = 0; j < aggregateSize; j++ )
            {
//                Atom* a = aggregate.getAtomPointer( j );
                Atom* a = aggregate.getLiteral( j ).getAtom();
                AtomInAnAggregateSet& atom = *( static_cast< AtomInAnAggregateSet* >( a ) );
                //Atom& atom = program->getAtoms()[ atomId ];
                atom.incrementPositiveOccurrencesForAggregates( posConjOcc );
                atom.incrementNegativeOccurrencesForAggregates( negConjOcc );
            }
        }
    }
}
