/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include <vector>
#include <algorithm>

#include "Program.h"
#include "Interpretation.h"
#include "options.h"
#include "unordered_map.h"
#include "MostOccurrencesForAggregates.h"
using namespace std;

void
Program::aggregatesInit()
{
    if( wasp::options.mostOccurrencesForAggregates )
    {
        MostOccurrencesForAggregates m;
        m.computeCounters( this );
    }    
}

void
Program::buildDependencyGraph()
{
    //For each atom, if it appears in the head add an edge to each atom in the body.
    for( unsigned int i = 0; i < atoms.size(); i++ )
    {
        //This is a fake edge. It needs to add all atom in the graph
        //and it inserts atom 0 in first component.
        dependencyGraph.addEdge( i, 0 );
        
        const RuleContainer< WaspRule* >& negativeHeadWaspRules = atoms[ i ]->getNegativeHeadWaspRules();
        unsigned int negativeHeadWaspRulesSize = negativeHeadWaspRules.size();
        //For each rule where the atom appears in the head.
        for( unsigned int j = 0; j < negativeHeadWaspRulesSize; j++ )
        {
            assert( negativeHeadWaspRules[ j ] != NULL );
            
            //get current rule
            WaspRule& r = *( negativeHeadWaspRules[ j ] );
            
            unsigned int size = r.getNumberOfLiteralsInTheBody();

            for( unsigned int k = 0; k < size; k++ )
            {
                Literal& literal = r.getBodyLiteral( k );
                int id = literal.getId();
                assert( id != 0 );
                if( id > 0 )
                {
                    dependencyGraph.addEdge( i, id );
                }
            }
        }
        
        const RuleContainer< ChoiceRule* >& headChoiceRules = atoms[ i ]->getHeadChoiceRules();
        unsigned int headChoiceRulesSize = headChoiceRules.size();
        //For each rule where the atom appears in the head.
        for( unsigned int j = 0; j < headChoiceRulesSize; j++ )
        {
            assert( headChoiceRules[ j ] != NULL );
            
            //get current rule
            ChoiceRule& r = *( headChoiceRules[ j ] );
            
            unsigned int size = r.getNumberOfLiteralsInTheBody();
            for( unsigned int k = 0; k < size; k++ )
            {
                Literal& literal = r.getBodyLiteral( k );
                int id = literal.getId();
                assert( id != 0 );
                if( id > 0 )
                {
                    dependencyGraph.addEdge( i, id );
                }
            }
        }
    }    
}

void
Program::componentsStrongConnected()
{
    //Computing the strong connected components.
    components = dependencyGraph.componentsStrongConnected();
    componentsSize = dependencyGraph.getComponentSize();
    
    //First component ( 0 ) contains only atom 0.
    //For each component C
    for( unsigned int i = 1; i < dependencyGraph.getComponentSize(); i++ )
    {
        //For each atom A in C, set that A occurs in C.
        vector< AtomId >& atomsInComponent = components[ i ].getAtomsInComponent();
        for ( unsigned int j = 0; j < atomsInComponent.size(); j++ )
        {
            assert( atomsInComponent[ j ] > 0 );
            atoms [ atomsInComponent[ j ] ]->setComponent( i );
        }
    }
    
    //For each component C
    for( unsigned int i = 1; i < dependencyGraph.getComponentSize(); i++ )
    {
        //We are interested only on cyclic and head cycle free component.
        if( isCyclic( i ) )
        {
            tight = false;
            if( isHeadCycleFree( i ) )
            {
                cyclicAndHeadCycleFreeComponents.push_back( i );

                //Get all atoms appear in C.
                vector< AtomId >& atomsInComponent = components[ i ].getAtomsInComponent();            

                //For each atom A
                for( unsigned int j = 0; j < atomsInComponent.size(); j++ )
                {
                    vector< Rule* > externalRules;
                    vector< Rule* > internalRules;
                    
                    assert( atomsInComponent[ j ] > 0 );

                    AtomId atomId = atomsInComponent[ j ];
                    Atom& atom = *( atoms[ atomId ] );                

                    if( atom.isTrueSupported() )
                    {
                        atom.setFounded();
                        continue;
                    }

                    //Get all rules where A appears in the head.
                    const RuleContainer< WaspRule* >& negativeHeadWaspRules = atom.getNegativeHeadWaspRules();
                    
                    unsigned int negativeHeadWaspRulesSize = negativeHeadWaspRules.size();
                    //For each rule R where A appears in the head.
                    for( unsigned int k = 0; k < negativeHeadWaspRulesSize; k++ )
                    {
                        WaspRule* wPointer = negativeHeadWaspRules[ k ]; 
                        WaspRule& w = *( wPointer );

                        //Create a pair component, rule.
                        ComponentRule componentRule( i, wPointer );
                        
                        //Associate the head atom to the pair component, rule.
                        headAtomInComponent[ componentRule ].push_back( atomId );

                        //This counter needs to count all rules with an atom in the positive body that appears in the component.
                        unsigned int rulesCount = 0;
                        vector< Atom* > positiveBodyLiterals;

                        unsigned int bodySize = w.getNumberOfLiteralsInTheBody();
                        for( unsigned int t = 0; t < bodySize; t++ )
                        {
                            Literal& literal = w.getBodyLiteral( t );
                            if( literal.isPositive() )
                            {
                                Atom* atom = literal.getAtom();
                                if( atom->getComponent() == i )
                                {
                                    rulesCount++;
                                    positiveBodyLiterals.push_back( atom );
                                }
                            }
                        }

                        //Associate the pair C,R with the rulesCount.
                        recursiveAtomsInRule[ componentRule ] = positiveBodyLiterals;

                        //If rulesCount is equal to 0, can add this rule in founding rule list.
                        if( rulesCount == 0 )
                        {
                            externalRules.push_back( wPointer );
                        }
                        else
                        {
                            internalRules.push_back( wPointer );
                        }
                    }
                    
                    //Get all rules where A appears in the head.
                    const RuleContainer< ChoiceRule* >& headChoiceRules = atom.getHeadChoiceRules();
                    unsigned int headChoiceRulesSize = headChoiceRules.size();
                    //For each choice rule R where A appears in the head.
                    for( unsigned int k = 0; k < headChoiceRulesSize; k++ )
                    {
                        ChoiceRule* rPointer = headChoiceRules[ k ]; 
                        ChoiceRule& r = *( rPointer );

                        //Associate the pair C,R with the atom A.
                        ComponentRule componentRule;
                        componentRule.first = i;
                        componentRule.second = rPointer;
                        headAtomInComponent[ componentRule ].push_back( atomId );

                        //This counter needs to count all rules with an atom in the positive body that appears in the component.
                        unsigned int rulesCount = 0;
                        vector< Atom* > positiveBodyLiterals;

                        unsigned int bodySize = r.getNumberOfLiteralsInTheBody();
                        for( unsigned int t = 0; t < bodySize; t++ )
                        {
                            Literal& literal = r.getBodyLiteral( t );
                            if( literal.isPositive() )
                            {
                                Atom* atom = literal.getAtom();
                                if( atom->getComponent() == i )
                                {
                                    rulesCount++;
                                    positiveBodyLiterals.push_back( atom );
                                }
                            }
                        }
                        
                        //Associate the pair C,R with the rulesCount.
                        recursiveAtomsInRule[ componentRule ] = positiveBodyLiterals;
                        //If rulesCount is equal to 0, can add this rule in founding rule list.
                        if( rulesCount == 0 )
                        {
                            externalRules.push_back( rPointer );
                        }
                        else
                        {
                            internalRules.push_back( rPointer );
                        }
                    }
                    
                    ComponentAtom componentAtom( i, atomId );
                    externalRulesForAtomInComponent[ componentAtom ] = externalRules;
                    internalRulesForAtomInComponent[ componentAtom ] = internalRules;
                }

                for( unsigned int j = 0; j < atomsInComponent.size(); j++ )
                {
                    vector< Rule* > listRules2;
                    assert( atomsInComponent[ j ] > 0 );

                    AtomId atomId = atomsInComponent[ j ];
                    Atom& atom = *( atoms[ atomId ] );
                    
                    const RuleContainer< WaspRule* >& positiveWaspRules = atom.getPositiveWaspRules();
                    unsigned int positiveWaspRulesSize = positiveWaspRules.size(); 
                    for( unsigned int k = 0; k < positiveWaspRulesSize; k++ )
                    {
                        WaspRule* wPointer = positiveWaspRules[ k ];
                        
                        ComponentRule componentRule;
                        componentRule.first = i;
                        componentRule.second = wPointer;
                        if( headAtomInComponent.find( componentRule ) != headAtomInComponent.end() )
                        {
                            listRules2.push_back( wPointer );
                        }
                    }
                    
                    const RuleContainer< ChoiceRule* >& positiveBodyChoiceRules = atom.getPositiveBodyChoiceRules();
                    unsigned int positiveBodyChoiceRulesSize = positiveBodyChoiceRules.size(); 

                    for( unsigned int k = 0; k < positiveBodyChoiceRulesSize; k++ )
                    {
                        ChoiceRule* cPointer = positiveBodyChoiceRules[ k ];
                        
                        ComponentRule componentRule;
                        componentRule.first = i;
                        componentRule.second = cPointer;
                        if( headAtomInComponent.find( componentRule ) != headAtomInComponent.end() )
                        {
                            listRules2.push_back( cPointer );
                        }
                    }

                    ComponentAtom componentAtom;
                    componentAtom.first = i;
                    componentAtom.second = atomId;
                    rulesInComponentWithPositiveBodyLiteral[ componentAtom ] = listRules2;
                }
            }
            else
            {
                headCycleFree = false;
            }
        }
    }
}

bool
Program::isCyclic(
    ComponentId componentId )
{
    //Get the atoms appear in the component.
    vector< AtomId >& atomsInComponent = components[ componentId ].getAtomsInComponent();

    assert( atomsInComponent.size() > 0 );

    if( atomsInComponent.size() == 1 )
    {
        //There is only one atom in the component.
        AtomId atomId = atomsInComponent[ 0 ];
        
        //If there is not a self loop than the component is not cyclic. 
        if( !dependencyGraph.existSelfLoop( atomId ) )
        {
            return false;
        }
    }
    
    //If two or more atoms are in the component C, then C is cyclic.
    components[ componentId ].setCyclic();
    return true;
}

bool
Program::isHeadCycleFree(
    ComponentId componentId )
{
    //A component C is not head cycle free if there exists a rule with two atoms a1,a2 in the head such that a1 and a2 appears in C.
    //Create a rule set.
    unordered_set< WaspRule* > ruleSet;

    //Get the atoms appear in the component.
    vector< AtomId >& atomsInComponent = components[ componentId ].getAtomsInComponent();

    //For each atom A
    for( unsigned int i = 0; i < atomsInComponent.size(); i++ )
    {
        AtomId atomId = atomsInComponent[ i ];
        
        const RuleContainer< WaspRule* >& negativeHeadWaspRules = atoms[ atomId ]->getNegativeHeadWaspRules();        
        unsigned int negativeHeadWaspRulesSize = negativeHeadWaspRules.size();
        
        for( unsigned int j = 0; j < negativeHeadWaspRulesSize; j++ )
        {
            assert( negativeHeadWaspRules[ j ] != NULL );
            WaspRule* w = negativeHeadWaspRules[ j ];
            //Try to add the waspRule in the rule set. If the rule is in the set, the component have a head cycle.
            if( !ruleSet.insert( w ).second )
            {
                return false;
            }
        }
    }
    
    //If function doesn't return, the component is head cycle free.
    components[ componentId ].setHeadCycleFree();
    return true;
}

void
Program::printWaspRule(
    WaspRule& rule )
{
    printWaspRule( rule, cout );
}

void
Program::printWaspRule(
    WaspRule& waspRule,
    ostream & o )
{
    o << waspRule << endl;
}

void
Program::printWaspRuleStdError(
    WaspRule& rule )
{
    printWaspRule( rule, cerr );
}

string
Program::waspRuleToString(
    WaspRule& rule )
{
    stringstream ruleStream;

    printWaspRule( rule, ruleStream );

    return ruleStream.str();
}

void
Program::reduceProgram()
{
    /*for( unsigned int i = 0; i < waspRules.size(); )
    {
        WaspRule* rule = waspRules[ i ];
        WaspRule& r = *rule;
        if( r.isSatisfiedCannotSupport() )
        {
            for( unsigned int j = 0; j < r.getNumberOfAtomsInTheHead(); j++ )
            {
                Atom& headAtom = *( r.getHeadAtom( j ) );
                headAtom.removeNegativeHeadWaspRule( rule );
            }
            
            for( unsigned int j = 0; j < r.getNumberOfLiteralsInTheBody(); j++ )
            {
                Literal& l = r.getBodyLiteral( j );
                Atom& atom = *( l.getAtom() );
                if( l.isPositive() )
                {
                    atom.removePositiveWaspRule( rule );
                }
                else
                {
                    atom.removeNegativeBodyWaspRule( rule );
                }
            }
            
            
            
            waspRules[ i ] = waspRules.back();
            waspRules.pop_back();
            
            delete rule;
        }
        else
        {
            i++;
        }
    }
    
    if( !isTight() )
    {
        dependencyGraph.clear();
        cyclicAndHeadCycleFreeComponents.clear();
        foundingRulesInComponent.clear();
        foundingAtomsInComponent.clear();
        headAtomInComponent.clear();
        positiveLiteralsInComponent.clear();
        positiveLiteralsInComponentVector.clear();
        rulesInComponentWithPositiveBodyLiteral.clear();
        buildDependencyGraph();
        componentsStrongConnected();
    }
    
    statistics( setRulesNumberAfterReduction( waspRules.size() ) );
    */
}

void
Program::computeLinearCostsForWeakConstraints()
{
    sort( this->levelsOfWeakConstraints.begin(), this->levelsOfWeakConstraints.end() );
    maxCostOfLevelOfWeakConstraints.push_back( 1 );
    int currentMaxCost = 1;
    for( unsigned int i = 0; i < this->levelsOfWeakConstraints.size(); ++i )
    {
        int currLevel = this->levelsOfWeakConstraints[ i ];
        List< WeakConstraint* >& currList = this->weakConstraintsPerLevel[ currLevel ];
        
        for( List< WeakConstraint* >::iterator it = currList.begin(); it != currList.end(); ++it )
        {
            WeakConstraint& wc = **it;
            assert( wc.getLevel() == currLevel );
            int weight = wc.getWeight() * maxCostOfLevelOfWeakConstraints.back();
            currentMaxCost += weight;
            wc.setLinearCost( weight );
        }
        assert( currentMaxCost >= maxCostOfLevelOfWeakConstraints.back() );
        maxCostOfLevelOfWeakConstraints.push_back( currentMaxCost );
    }
}

int
Program::getCostOfLevel(
    int levelIdx,
    int totalCost )
{
    assert( 0 <= levelIdx && levelIdx < levelsOfWeakConstraints.size() );
    assert( levelsOfWeakConstraints.size() + 1 == maxCostOfLevelOfWeakConstraints.size() );

    return ( totalCost % maxCostOfLevelOfWeakConstraints[ levelIdx + 1 ] ) / maxCostOfLevelOfWeakConstraints[ levelIdx ];
}