/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   Program.h
 * Author: Carmine Dodaro
 *
 * Created on 7 October 2010, 9.49
 */

#ifndef _PROGRAM_H
#define	_PROGRAM_H

#include <cassert>
#include <iostream>
#include "stl/List.h"
#include <sstream>
#include <string>
#include <unordered_set>
#include <vector>
using namespace std;

#include "Aggregate.h"
#include "AggregateAtomGreater.h"
#include "AggregateAtomDoubleGuards.h"
#include "Atom.h"
#include "AtomInAnAggregateSet.h"
#include "DependencyGraph.h"
#include "options.h"
#include "Statistics.h"
#include "unordered_map.h"
#include "WaspRule.h"
#include "LearnedWaspRule.h"
#include "ChoiceRule.h"
#include "WeakConstraint.h"

class Interpretation;

/**
 * This is a class storing all informations about a program, i.e.
 * atoms, heads, bodies and rules appear in the program.
 */
class Program
{    
public:

    unordered_set< Rule* > foundingRulesSet;

    inline Program();

    ~Program()
    {        
        for( unsigned int i = 0; i < atoms.size(); i++ )
        {
            delete atoms[ i ];
        }

        for( unsigned int i = 0; i < waspRules.size(); i++ )
        {
            delete waspRules[ i ];
        }        
        
        for( unsigned int i = 0; i < choiceRules.size(); i++ )
        {
            delete choiceRules[ i ];
        }
        
        for( unsigned int i = 0; i < aggregates.size(); i++ )
        {
            assert( aggregates[ i ] != NULL );
            delete aggregates[ i ];
        }        
        
        for( List< LearnedWaspRule* >::iterator it = learnedWaspRules.begin(); it != learnedWaspRules.end(); ++it )
        {
            delete *it;
        }
        
        learnedWaspRules.clear();
    }

    /**
     * It adds an aggregate in the aggregate vector.
     *
     * @param Aggregate* a pointer to the aggregate.
     * 
     */
    inline void addAggregate( Aggregate* aggregate );

    inline void addAggregateAtomDoubleGuards( AtomId id, string atomName );
    
    inline void addAggregateAtomGreater( AtomId id, string atomName );
    
    /**
     * It adds the atom in the atoms vector.
     *
     * @param id The id of the atom.
     * @param atomName The atom name.
     */
    inline void addAtom( AtomId id, string atomName );
    
    inline void addAtomInAnAggregateSet( AtomId id, string atomName );
    
    inline void addFalseAtom( Atom* a );
        
    inline void addLearnedWaspRule( LearnedWaspRule* w );//, unordered_map< AtomId, unsigned int >* positions );
    
    inline void addTrueSupportedAtom( Atom* a );
    
    inline void addTrueUnsupportedAtom( Atom* a );
    
    inline void addChoiceRule( ChoiceRule* c );
    
    /**
     * It adds a wasprule in the wasprule vector.
     * 
     * @param the wasprule.
     */
    inline void addWaspRule( WaspRule* );
    
    inline void addWeakConstraint( WeakConstraint* );
    
    /**
     * This function converts an atom to a string.
     *
     * @return the string representing the atom.
     */
    inline string atomToString( AtomId atomId );

    /**
     * It performs the inital operations for the aggregates.
     *
     */
    void aggregatesInit();

    /**
     * this function creates the dependency graph.
     *
     */
    void buildDependencyGraph();

    /**
     * This function computes and stores the strong connected components.
     * 
     */
    void componentsStrongConnected();

    /**
     * Returns a vector, which contain Aggregates.
     *
     * @return aggregates appearing in the program.
     */
    inline vector< Aggregate* >& getAggregates();

    /**
     * Returns a vector, which contain Atoms.
     *
     * @return atoms appearing in the program.
     */
    inline vector< Atom* >& getAtoms();
    
    /**
     * Returns a pointer to the atom with the id.
     * @param the id of the atom.
     * 
     * @return a pointer to the atom.
     */
    inline Atom* getAtomPointer( unsigned int atomId );
    
    /**
     * Returns the atom with the id.
     * @param the id of the atom.
     * 
     * @return a reference to the atom.
     */
    inline Atom& getAtom( unsigned int atomId );
    
    /**
     * Returns the number of atoms in the program.
     * 
     * @return the number of atoms in the program.
     */
    inline unsigned int getNumberOfAtoms() const;
    
    /**
     * Returns a vector containing the cyclic and HCF components.
     * 
     * @return a const vector containing the cyclic and HCF components.
     */
    inline const vector< ComponentId >& getCyclicAndHeadCycleFreeComponents() const;
    
    /**
     * Returns a dinamic array of Component.
     *
     * @return the components of the program.
     */
    inline Component* getComponents() const;

    /**
     * Returns the component array size.
     *
     * @return the components size.
     */
    inline unsigned int getComponentsSize() const;
    
    /**
     * This function returns the private field headAtomInComponent.
     * @return a map of ComponentRule in AtomId
     */
    inline std::unordered_map< ComponentRule, vector< AtomId >, RuleIntegerPairHasher >& getHeadAtomInComponent();

    inline List< LearnedWaspRule* >& getLearnedWaspRules();
    
    inline const vector< Atom* >& getFalseAtoms() const;
    
    /**
     * This function returns the private field rulesInComponentWithPositiveBodyLiteral.
     * @return a map of ComponentAtom in vector of a RuleId.
     */
    inline std::unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher >&  getRulesInComponentWithPositiveBodyLiteral();
    
    inline const vector< Atom* >& getTrueSupportedAtoms() const;
    
    inline const vector< Atom* >& getTrueUnsupportedAtoms() const;

    inline QueryType getType();

    inline QueryValue getValue();
    
    inline ChoiceRule* getChoiceRule( RuleId ruleId );
    
    inline vector< int >& getLevelsOfWeakConstraints();
    
    int getCostOfLevel( int levelIdx, int totalCost );
    
    inline int getInitialMaxValueWeakConstraints() const;
    
    /**
     * Levels of weak constraints can be ignored after calling this method
     * because there are only a finite number of them and thus cost@level can
     * be transformed in level*c, where c is a constant depending on the
     * costs of the previous levels.
     * 
     */
    void computeLinearCostsForWeakConstraints();
    
    /**
     * Returns the number of waspRules in the original program.
     *
     * @return the number of waspRules in the original program.
     */
    inline unsigned int getWaspRulesSize() const;
    
    inline unsigned int getChoiceRulesSize() const;
    
    inline bool hasWeakConstraints() const;
        
    /**
     * Returns if grounder failed.
     *
     * @return true if there was an error in the grounding step, false otherwise.
     */
    inline bool isErrorFromGrounding();
    
    /**
     * This function determines if the program is head-cycle-free.
     *
     * @return if the program is head-cycle-free
     */
    inline bool isHeadCycleFree();
    
    /**
     * Returns if there is a message from grounder.
     *
     * @return true if there is a message from the grounding step, false otherwise.
     */
    inline bool isMessageFromGrounding();

    /**
     * It prints the program.     
     */
    inline void printProgram();

    /**
     * This function print a rule on the standard output.
     *
     * @param ruleId the id of the rule that is getting to be printed.
     */
    void printWaspRule( RuleId ruleId );
    
    /**
     * This function print a rule on the standard output.
     *
     * @param rule the rule that is getting to be printed.
     */
    void printWaspRule( WaspRule& rule );

    /**
     * This function print a rule on the standard error.
     *
     * @param rule the rule that is getting to be printed.
     */    
    void printWaspRuleStdError( WaspRule& rule );
    
    inline void setErrorFromGrounding();

    inline void setMessageFromGrounding();

    inline void setType( QueryType );

    inline void setValue( QueryValue );

    inline bool isTight();
    
    inline bool hasChoiceRules();
    
    inline unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher >& getExternalRulesForAtomInComponent();
    
    inline unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher >& getInternalRulesForAtomInComponent();
    
    inline unordered_map< ComponentRule, vector< Atom* >, RuleIntegerPairHasher >& getRecursiveAtomsInRule();
    
    void reduceProgram();
    
    inline void setAggregateRepresentingWeakConstraints( Aggregate* aggr );
    
    inline Aggregate* getAggregateRepresentingWeakConstraints();
    
    inline bool hasQuery();
    
    /**
     * This function transforms the rule into a string.
     *
     * @param the rule.
     * 
     * @return a string representing the rule.
     */
    string waspRuleToString( WaspRule& );

    /**
     * This function transforms the rule into a string.
     * 
     * @param the id of the rule.
     * 
     * @return a string representing the rule.
     */
    string waspRuleToString( RuleId ruleId );
    
    #ifndef STATS_OFF
    void computeStatistics()
    {
        //Computing statistics
        for( unsigned int i = 0; i < waspRules.size(); i++ )
        {
            WaspRule* w = waspRules[ i ];

            if( w->size() == 2 )
                statistics( incrementBinaryRules() );
            else if( w->size() == 3 )
                statistics( incrementTernaryRules() );

            if( w->getNumberOfAtomsInTheHead() > 1 )
            {
                statistics( incrementNumberOfDisjunctiveRules() );
                if( w->getNumberOfLiteralsInTheBody() == 0 )
                    statistics( incrementNumberOfDisjunctiveFacts() );
            }
            else if( w->isConstraint() )
            {
                statistics( incrementNumberOfConstraints() );
            }

            int countNegative = 0;
            for( unsigned int j = 0; j < w->size() && countNegative < 2; j++ )
            {
                Literal& l = w->getLiteral( j );
                if( !l.isPositive() )
                    countNegative++;
            }
            if( countNegative == 1 )
                statistics( incrementNumberOfHornRules() );
        }

        for( unsigned int i = 0; i < cyclicAndHeadCycleFreeComponents.size(); i++ )
        {
            ComponentId c_id = cyclicAndHeadCycleFreeComponents[ i ];
            Component& c = components[ c_id ];
            
            statistics( addComponentSize( c.size() ) );
        }
    }
    #endif    
    
private:

    /**
     * Aggregates contained by this program are stored in a vector.
     */
    vector< Aggregate* > aggregates;

     /**
     * All atoms of the program are stored in this vector.
     */
    vector< Atom* > atoms;

    /**
     * This vector contains only cyclic and head cycle free components.
     */
    vector< ComponentId > cyclicAndHeadCycleFreeComponents;
    
    /**
     * The array of strong connected components.
     */
    Component* components;

    /**
     * The size of components array.
     */
    unsigned int componentsSize;

    /**
     * This field contains the dependency graph.
     */
    DependencyGraph dependencyGraph;
    
    /*
     * A flag representing if there is an error from the grounding step.
     */
    bool errorFromGrounding;
    
    vector< Atom* > falseAtoms;
        
    /**
     * This field is a map of a ComponentRule (a pair of a ComponentId and a RuleId) to AtomId.
     * The AtomId represents the only atom in a component that is in the head of a specific rule.
     * When a component is head cycle free there is only an atom in the head of a specific rule for that component.
     */
    unordered_map< ComponentRule, vector< AtomId >, RuleIntegerPairHasher > headAtomInComponent;
    
    /*
     * Ids of constraints learned when a conflict is encountered.
     */
    List< LearnedWaspRule* > learnedWaspRules;
    
    /*
     * A flag representing if there is a message from the grounding step.
     */
    bool messageFromGrounding;
    
    /**
     * This field is a map of a ComponentAtom (a pair of a ComponentId and an AtomId) to a vector of RuleId.
     */
    unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher > rulesInComponentWithPositiveBodyLiteral;
    
    unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher > externalRulesForAtomInComponent;
    
    unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher > internalRulesForAtomInComponent;
    
    unordered_map< ComponentRule, vector< Atom* >, RuleIntegerPairHasher > recursiveAtomsInRule;
    
    bool headCycleFree;
    
    bool tight;

    vector< Atom* > trueSupportedAtoms;
    
    vector< Atom* > trueUnsupportedAtoms;
    
    QueryType type;

    QueryValue value;    

    /**
     * This vector contains all the WaspRules of the program.
     */
    vector< WaspRule* > waspRules;
    
    vector< ChoiceRule* > choiceRules;
    
    vector< int > levelsOfWeakConstraints;
    
    vector< int > maxCostOfLevelOfWeakConstraints;
    
    unordered_map< int, List< WeakConstraint* > > weakConstraintsPerLevel;
    
    Aggregate* aggregateRepresentingWeakConstraints;

    /**
     * Returns if a component is cyclic.
     *
     * @param componentId The id of the component.
     * 
     * @return true if it is cyclic, false otherwise.
     */
    bool isCyclic( ComponentId componentId );
    
    /**
     * Returns if a component is head cycle free.
     *
     * @param componentId the id of the component.
     * 
     * @return true if it is head cycle free, false otherwise.
     */
    bool isHeadCycleFree( ComponentId componentId );
    
    /**
     * This function prints a waspRule on the ostream o.
     *
     * @param waspRule The waspRule.
     * @param o The ostream.
     * 
     */    
    void printWaspRule( WaspRule& waspRule, ostream& o );
    
    /**
     * Copy constructor. By default, it is disabled.
     */
    Program( const Program & )
    {
        assert( 0 );
    }
};

Program::Program() : errorFromGrounding( false ), messageFromGrounding( false ), tight( true ), headCycleFree( true ), type( MAXVALUE ), value( MAXVALUE ), aggregateRepresentingWeakConstraints( NULL )
{
}

void
Program::addAggregate(
    Aggregate* aggregate )
{
    this->aggregates.push_back( aggregate );
}

void
Program::addAggregateAtomDoubleGuards(
    AtomId id,
    string atomName )
{
    assert( id == atoms.size() );
    AggregateAtomDoubleGuards* atom = new AggregateAtomDoubleGuards( id, atomName );
    this->atoms.push_back( atom );
}
    
void
Program::addAggregateAtomGreater(
    AtomId id,
    string atomName )
{
    assert( id == atoms.size() );
    AggregateAtomGreater* atom = new AggregateAtomGreater( id, atomName );
    this->atoms.push_back( atom );
}

void
Program::addAtom(
    AtomId id,
    string atomName )
{
    assert( id == atoms.size() );
    Atom* atom = new Atom( id, atomName );
    this->atoms.push_back( atom );
}

void
Program::addAtomInAnAggregateSet(
    AtomId id,
    string atomName )
{
    assert( id == atoms.size() );
    AtomInAnAggregateSet* atom = new AtomInAnAggregateSet( id, atomName );
    this->atoms.push_back( atom );
}

void
Program::addChoiceRule(
    ChoiceRule* c )
{
    assert( c != NULL );
    choiceRules.push_back( c );
}

void
Program::addFalseAtom(
    Atom* a )
{
    assert( a != NULL );
    falseAtoms.push_back( a );
}

void
Program::addLearnedWaspRule(
    LearnedWaspRule* w )
{
    assert( w != NULL );    
    learnedWaspRules.push_back( w );
}

void
Program::addTrueSupportedAtom(
    Atom* a )
{
    assert( a != NULL );
    trueSupportedAtoms.push_back( a );
    a->setTrueSupported();
}

void
Program::addTrueUnsupportedAtom(
    Atom* a )
{
    assert( a != NULL );
    trueUnsupportedAtoms.push_back( a );
}

void
Program::addWaspRule(
    WaspRule* w )
{
    assert( w != NULL );
    this->waspRules.push_back( w );
}

void
Program::addWeakConstraint(
    WeakConstraint* w )
{
    pair< unordered_map< int, List< WeakConstraint* > >::iterator, bool > res =
            this->weakConstraintsPerLevel.insert( unordered_map< int, List< WeakConstraint* > >::value_type( w->getLevel(), List< WeakConstraint* >() ) );
    if( res.second )
    {
        this->levelsOfWeakConstraints.push_back( w->getLevel() );
    }
    res.first->second.push_back( w );
}

vector< Aggregate* >&
Program::getAggregates()
{
    return aggregates;
}

vector< Atom* >&
Program::getAtoms()
{
    return atoms;
}

Atom* 
Program::getAtomPointer( 
    unsigned int position )
{
    assert( position < atoms.size() );
    return atoms[ position ];
}

Atom& 
Program::getAtom( 
    unsigned int position )
{
    assert( position < atoms.size() );
    assert( atoms[ position ] != NULL );
    return *( atoms[ position ] );
}

unsigned int 
Program::getNumberOfAtoms() const
{
    return atoms.size();
}

Component*
Program::getComponents() const
{
    return components;
}

unsigned int
Program::getComponentsSize() const
{
    return componentsSize;
}

const vector< ComponentId >&
Program::getCyclicAndHeadCycleFreeComponents() const
{
    return cyclicAndHeadCycleFreeComponents;
}

const vector< Atom* >&
Program::getFalseAtoms() const
{
    return falseAtoms;
}

std::unordered_map< ComponentRule, vector< AtomId >, RuleIntegerPairHasher >&
Program::getHeadAtomInComponent()
{
    return headAtomInComponent;
}

List< LearnedWaspRule* >&
Program::getLearnedWaspRules()
{
    return learnedWaspRules;
}

std::unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher >&
Program::getRulesInComponentWithPositiveBodyLiteral()
{
    return rulesInComponentWithPositiveBodyLiteral;
}

const vector< Atom* >&
Program::getTrueSupportedAtoms() const
{
    return trueSupportedAtoms;
}

const vector< Atom* >&
Program::getTrueUnsupportedAtoms() const
{
    return trueUnsupportedAtoms;
}

QueryType
Program::getType()
{
    return type;
}

QueryValue
Program::getValue()
{
    return value;
}

ChoiceRule*
Program::getChoiceRule( 
    RuleId ruleId )
{
    assert( ruleId < choiceRules.size() );
    return choiceRules[ ruleId ];
}

unsigned int 
Program::getWaspRulesSize() const
{
    return waspRules.size();
}

unsigned int 
Program::getChoiceRulesSize() const
{
    return choiceRules.size();
}

bool
Program::isErrorFromGrounding()
{
    return errorFromGrounding;
}

bool
Program::isMessageFromGrounding()
{
    return messageFromGrounding;
}

bool
Program::isTight()
{
    return tight;
}

bool
Program::isHeadCycleFree()
{
    return headCycleFree;
}

void
Program::printProgram()
{
    for( unsigned int i = 0; i < waspRules.size(); i++ )
    {        
        printWaspRule( *( waspRules[ i ] ) );
    }
    
    for( unsigned int i = 0; i < choiceRules.size(); i++ )
    {        
        cout << ( *( choiceRules[ i ] ) ) << endl;
    }    
    
    for( unsigned int i = 0; i < trueSupportedAtoms.size(); i++ )
    {
        cout << *( trueSupportedAtoms[ i ] ) << "." << endl;
    }
    
    for( unsigned int i = 0; i < falseAtoms.size(); i++ )
    {
        cout << ":- " << *( falseAtoms[ i ] ) << "." << endl;
    }
    
    for( unsigned int i = 0; i < trueUnsupportedAtoms.size(); i++ )
    {
        cout << ":- not " << *( trueUnsupportedAtoms[ i ] ) << "." << endl;
    }
    
//    for( unsigned int i = 0; i < aggregates.size(); i++ )
//    {
//        cout << *aggregates[ i ] << endl;
//    }    
}

void
Program::setErrorFromGrounding()
{
    errorFromGrounding = true;
}

void
Program::setMessageFromGrounding()
{
    messageFromGrounding = true;
}

void
Program::setType(
    QueryType t )
{
    type = t;
}

void
Program::setValue(
    QueryValue v )
{
    value = v;
}

string
Program::atomToString(
    AtomId atomId )
{    
    return getAtom( atomId ).getAtomName();
}

bool 
Program::hasChoiceRules()
{
    return !choiceRules.empty();
}

unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher >& 
Program::getExternalRulesForAtomInComponent()
{
    return externalRulesForAtomInComponent;
}

unordered_map< ComponentAtom, vector< Rule* >, IntegerPairHasher >& 
Program::getInternalRulesForAtomInComponent()
{
    return internalRulesForAtomInComponent;
}

unordered_map< ComponentRule, vector< Atom* >, RuleIntegerPairHasher >& 
Program::getRecursiveAtomsInRule()
{
    return recursiveAtomsInRule;
}

vector< int >& 
Program::getLevelsOfWeakConstraints()
{ 
    return levelsOfWeakConstraints;
}

bool 
Program::hasWeakConstraints() const
{
    return !levelsOfWeakConstraints.empty();
}

int 
Program::getInitialMaxValueWeakConstraints() const
{
    return maxCostOfLevelOfWeakConstraints.back();
}

void 
Program::setAggregateRepresentingWeakConstraints( 
    Aggregate* aggr )
{
    aggregateRepresentingWeakConstraints = aggr;
}

Aggregate* 
Program::getAggregateRepresentingWeakConstraints()
{
    return aggregateRepresentingWeakConstraints;
}

bool
Program::hasQuery()
{
    return type != MAXVALUE;
}

#endif	/* _PROGRAM_H */