/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/*
 * File:   StackInvolvedForGUS.h
 * Author: Carmine Dodaro
 *
 * Created on 19 October 2011, 17.46
 */

#ifndef _STACKINVOLVEDFORGUS_H
#define	_STACKINVOLVEDFORGUS_H

#include <vector>
#include <cassert>
#include <ostream>
#include "trace.h"
#include "options.h"
#include "Constants.h"

using namespace std;
using namespace wasp;

class StackInvolvedForGUS
//
// This structure is used by our linear time propagation machinery.
//
// Whenever we change one of those many counters there, we save its
// address and its old value on this stack so that we can easily
// restore the counter during backtracking.
//
{
//    friend ostream& operator<<( ostream&, const StackVector& );
private:
    vector< unsigned int > sizes;
    vector< InvolvedVector* >& gusVectors;

    // disabled methods
    inline StackInvolvedForGUS( const StackInvolvedForGUS& );
    inline void operator=( const StackInvolvedForGUS& );

public:
    inline StackInvolvedForGUS( vector< InvolvedVector* >& v );

    inline unsigned int getCurrentLevel() const;

    inline void clear();
    
    inline void nextLevel();
    
    void unroll( const unsigned level )
    {
        trace( mg, 4, "Unrolling level(s) from %u to %u.\n", static_cast<unsigned int>(getCurrentLevel()), level );

        // Do nothing if we already are at the desired level or even below.
        // (This is the case when a user of StackVector operates at higher levels
        // without pushing anything on the STACK.)
        if( level >= sizes.size() )
            return;

        // Restore stored values.
        while( sizes.size() - 1 > level )
        {
            sizes.pop_back();
        }

        assert( !sizes.empty() );
        assert( sizes.size() - 1 == level );
        while( gusVectors.size() > sizes.back() )
        {
            delete gusVectors.back();
            gusVectors.pop_back();
        }
        sizes.pop_back();
    }

    inline void unrollOne();
};

StackInvolvedForGUS::StackInvolvedForGUS(
    const StackInvolvedForGUS& init )
    : gusVectors( init.gusVectors )
{
    assert( 0 );
}

void
StackInvolvedForGUS::operator=(
    const StackInvolvedForGUS& )
{
    assert( 0 );
}

StackInvolvedForGUS::StackInvolvedForGUS(
    vector< InvolvedVector* >& v )
    : sizes(), gusVectors( v )
{
    nextLevel();
}

unsigned int
StackInvolvedForGUS::getCurrentLevel() const
{
    assert( !sizes.empty() );

    return sizes.size() - 1;
}

void
StackInvolvedForGUS::clear()
{
    sizes.clear();
}

void
StackInvolvedForGUS::nextLevel()
{
    trace( mg, 4, "Pushing level %u.\n", static_cast<unsigned int>(sizes.size()) );
    sizes.push_back( gusVectors.size() );
}

void
StackInvolvedForGUS::unrollOne()
{
    unroll( getCurrentLevel() );
}

#endif	/* _STACKINVOLVEDFORGUS_H */
