/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   Statistics.h
 * Author: Carmine Dodaro
 *
 * Created on 3 February 2011, 12.36
 */

#ifndef _STATISTICS_H
#define	_STATISTICS_H
#include <cassert>
#include <vector>
#include <iostream>
using namespace std;

#ifdef STATS_OFF
    #define statistics( METHOD )
#else
    #define statistics( METHOD ) Statistics::inst().METHOD

class Statistics {
    
    public:

        Statistics();

        Statistics( const Statistics& orig );

        virtual ~Statistics();

        inline void addReachedLevel( unsigned int level );
        
        inline void addComponentSize( unsigned int size );
        
        inline void incrementNumModels();

        inline void incrementNumPropagations();

        inline void incrementNumComputeGUS();

        inline void incrementNumDeletedRules();

        inline void incrementNumCheckModelFailures();

        inline void setMaxBacktrackingLevel( unsigned int number );

        inline void setRulesNumber( unsigned int number );

        inline void setAtomsNumber( unsigned int number );

        inline void setInitialPropagationNumber( unsigned int number );

        inline void incrementBerkminInvocationsNumber();

        inline void incrementHigherGlobalCounterInvocationsNumber();

        inline void incrementLearnRuleInvocationsNumber();

        inline void incrementCalculateFirstUipInvocationsNumber();

        inline void incrementDeleteLearnedRulesInvocationsNumber();

        inline void incrementRestartNumber();

        inline void incrementDerivations();

        inline void incrementNoDerivations();

        inline void incrementGUSInconsistences();

        inline void incrementCountUnit();

        inline void incrementCountSupport();

        inline void incrementCountLastSupport();

        inline void incrementCountSubsumption();

        inline void incrementCountSelfSubsumption();
        
        inline void incrementBinaryRules();
        
        inline void incrementTernaryRules();

        inline void incrementNumberOfHornRules();

        inline void incrementNumberOfDisjunctiveFacts();

        inline void incrementNumberOfDisjunctiveRules();

        inline void incrementNumberOfConstraints();

        inline void incrementNumberOfLearnedWaspRules();
        
        inline void incrementSizeOfLearnedWaspRules( unsigned int s );

        inline void incrementMostOccurrencesInvocationsNumber();

        inline void incrementChoosePolarityInvocationsNumber();
        
        inline void setRulesNumberAfterReduction( unsigned int value );
        
        inline void incrementNumberOfPartialModelChecks();
        
        inline void incrementOperatorRCalls();
        
        inline void incrementFixOperatorRCalls();
        
        inline void incrementNumberOfChoices();
        
        void printStatistics();

        inline static Statistics& inst();

    private:
        
        static Statistics statistics;

        unsigned int numModels;

        unsigned int numPropagations;

        unsigned int numComputeGUS;

        unsigned int numDeletedRules;

        unsigned int numCheckModelFailures;

        unsigned int maxBackTrackingLevel;

        unsigned int rulesNumber;

        unsigned int atomsNumber;

        unsigned int initialPropagationNumber;

        unsigned int berkminInvocationsNumber;

        unsigned int higherGlobalCounterInvocationsNumber;
        
        unsigned int mostOccurrencesInvocationsNumber;
        
        unsigned int choosePolarityInvocationsNumber;

        unsigned int learnRuleInvocationsNumber;

        unsigned int calculateFirstUipInvocationsNumber;

        unsigned int deleteLearnedRulesInvocationsNumber;
        
        unsigned int restartsNumber;

        unsigned int numberOfDerivationsGUS;

        unsigned int numberOfNoDerivationsGUS;

        unsigned int gusInconsistences;

        unsigned int minConstraintSize;

        unsigned int maxConstraintSize;

        unsigned int countUnitPropagation;

        unsigned int countSupport;

        unsigned int countLastSupport;

        unsigned int countSubsumption;

        unsigned int countSelfSubsumption;
        
        unsigned int binaryRules;
        
        unsigned int ternaryRules;
        
        unsigned int numberOfHornRules;
        
        unsigned int numberOfDisjunctiveFacts;
        
        unsigned int numberOfDisjunctiveRules;
        
        unsigned int numberOfConstraints;
        
        unsigned int numberOfLearnedWaspRules;
        
        unsigned int sizeOfLearnedWaspRules;
        
        unsigned int rulesNumberAfterReduction;
        
        unsigned int numberOfPartialModelChecks;
        
        unsigned int operatorRCalls;
        
        unsigned int fixPointOperatorRCalls;
        
        unsigned int numberOfChoices;
        
        vector< unsigned int > sizeOfComponents;

        vector< unsigned int > reachedLevel;
};

Statistics&
Statistics::inst()
{
    return statistics;
}

void
Statistics::incrementNumModels()
{
    numModels++;
}

void
Statistics::incrementNumPropagations()
{
    numPropagations++;
}

void
Statistics::incrementNumCheckModelFailures()
{
    numCheckModelFailures++;
}

void
Statistics::incrementNumComputeGUS()
{
    numComputeGUS++;
}

void
Statistics::incrementNumDeletedRules()
{
    numDeletedRules++;
}

void
Statistics::setMaxBacktrackingLevel(
    unsigned int number )
{
    if( number > maxBackTrackingLevel )
        maxBackTrackingLevel = number;
}

void
Statistics::setAtomsNumber(
    unsigned int number )
{
    atomsNumber = number;
}

void
Statistics::setRulesNumber(
    unsigned int number )
{
    rulesNumber = number;
}

void
Statistics::setInitialPropagationNumber(
    unsigned int number )
{
    initialPropagationNumber = number;
}

void
Statistics::incrementBerkminInvocationsNumber()
{
    berkminInvocationsNumber++;
}

void
Statistics::incrementHigherGlobalCounterInvocationsNumber()
{
    higherGlobalCounterInvocationsNumber++;
}

void
Statistics::incrementMostOccurrencesInvocationsNumber()
{
    mostOccurrencesInvocationsNumber++;
}

void
Statistics::incrementChoosePolarityInvocationsNumber()
{
    choosePolarityInvocationsNumber++;
}

void
Statistics::incrementLearnRuleInvocationsNumber()
{
    learnRuleInvocationsNumber++;
}

void
Statistics::incrementCalculateFirstUipInvocationsNumber()
{
    calculateFirstUipInvocationsNumber++;
}

void
Statistics::incrementDeleteLearnedRulesInvocationsNumber()
{
    deleteLearnedRulesInvocationsNumber++;
}

void
Statistics::incrementRestartNumber()
{
    restartsNumber++;
}

void
Statistics::addReachedLevel(
    unsigned int level )
{
    if( level >= reachedLevel.size() )
        reachedLevel.push_back( 1 );
    else
        reachedLevel[ level ]++;
}

void
Statistics::incrementNoDerivations()
{
    numberOfNoDerivationsGUS++;
}

void
Statistics::incrementDerivations()
{
    numberOfDerivationsGUS++;
}

void
Statistics::incrementGUSInconsistences()
{
    gusInconsistences++;
}

void
Statistics::incrementCountUnit()
{
    countUnitPropagation++;
}

void
Statistics::incrementCountSupport()
{
    countSupport++;
}

void
Statistics::incrementCountLastSupport()
{
    countLastSupport++;
}

void
Statistics::incrementCountSubsumption()
{
    countSubsumption++;
}

void
Statistics::incrementCountSelfSubsumption()
{
    countSelfSubsumption++;
}

void
Statistics::incrementBinaryRules()
{
    binaryRules++;
}

void
Statistics::incrementTernaryRules()
{
    ternaryRules++;
}

void
Statistics::incrementNumberOfHornRules()
{
    numberOfHornRules++;
}

void
Statistics::incrementNumberOfDisjunctiveFacts()
{
    numberOfDisjunctiveFacts++;
}

void
Statistics::incrementNumberOfDisjunctiveRules()
{
    numberOfDisjunctiveRules++;
}

void
Statistics::incrementNumberOfConstraints()
{
    numberOfConstraints++;
}

void
Statistics::addComponentSize(
    unsigned int size )
{
    sizeOfComponents.push_back( size );
}

void
Statistics::incrementNumberOfLearnedWaspRules()
{
    numberOfLearnedWaspRules++;
}
        
void
Statistics::incrementSizeOfLearnedWaspRules( 
    unsigned int size )
{
    sizeOfLearnedWaspRules += size;    
    if( minConstraintSize > size )
        minConstraintSize = size;
    if( maxConstraintSize < size )
        maxConstraintSize = size;
}

void
Statistics::setRulesNumberAfterReduction(
    unsigned int value )
{
    rulesNumberAfterReduction = value;
}

void
Statistics::incrementNumberOfPartialModelChecks()
{
    numberOfPartialModelChecks++;
}

void 
Statistics::incrementOperatorRCalls()
{
    operatorRCalls++;
}

void 
Statistics::incrementFixOperatorRCalls()
{
    fixPointOperatorRCalls++;
}

void
Statistics::incrementNumberOfChoices()
{
    numberOfChoices++;
}

#endif

#endif	/* _STATISTICS_H */
