/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   Statistics.cpp
 * Author: Mario Alviano
 * 
 * Created on 18 January 2013, 12.18
 */

#ifndef _UNORDERED_SET_H_
#define _UNORDERED_SET_H_

#include <cassert>
#include <iostream>
#include <vector>
#include <unordered_map>

using namespace std;

#define NOT_FOUND -1

template< class T >
class UnorderedSet
{
    template< class TT >
    friend ostream& operator<<( ostream& out, const UnorderedSet< TT >& set );
    public:
        UnorderedSet(){}
        ~UnorderedSet(){}

        inline T& operator[]( unsigned idx );
        inline const T& operator[]( unsigned idx ) const;
        inline T& at( unsigned idx );
        inline const T& at( unsigned idx ) const;

        inline unsigned size() const;
        inline bool empty() const;
        inline void clear();

        bool insert( const T& value );
        bool erase( const T& value );

        inline int find( const T& value ) const;
        inline bool contains( const T& value ) const;

    private:
        vector< T > data;
        unordered_map< T, int > index;
        UnorderedSet( const UnorderedSet& init );
};

template< class T >
ostream&
operator<<(
    ostream& out,
    const UnorderedSet< T >& s )
{
    for( unsigned i = 0; i < s.size(); ++i )
        out << s.data[ i ] << " ";
    return out;
}

template< class T >
UnorderedSet< T >::UnorderedSet(
    const UnorderedSet< T >& init )
: data( init.data )
{
    assert( 0 );
    for( unsigned i = 0; i < data.size(); i++ )
        index[ data[ i ] ] = i;
}
    
template< class T >
T&
UnorderedSet< T >::operator[](
    unsigned idx )
{
    assert( idx < data.size() );
    return data[ idx ];
}

template< class T >
const T&
UnorderedSet< T >::operator[](
    unsigned idx ) const
{
    assert( idx < data.size() );
    return data[ idx ];
}

template< class T >
T&
UnorderedSet< T >::at(
    unsigned idx )
{
    return operator[]( idx );
}

template< class T >
const T&
UnorderedSet< T >::at(
    unsigned idx ) const
{
    return operator[]( idx );
}

template< class T >
unsigned
UnorderedSet< T >::size() const
{
    return data.size();
}

template< class T >
bool
UnorderedSet< T >::empty() const
{
    return data.empty();
}

template< class T >
void
UnorderedSet< T >::clear()
{
    data.clear();
    index.clear();
}

template< class T >
bool
UnorderedSet< T >::insert(
    const T& value )
{
    if( index.find( value ) != index.end() )
        return false;
    
    index.insert( typename unordered_map< T, int >::value_type( value, data.size() ) );
    data.push_back( value );
    return true;
}

template< class T >
bool
UnorderedSet< T >::erase(
    const T& value )
{
    typename unordered_map< T, int >::iterator it = index.find( value );
    if( it == index.end() || it->second == NOT_FOUND )
        return false;
    
    data[ it->second ] = data.back();
    index[ data.back() ] = it->second;
    data.pop_back();
    it->second = NOT_FOUND;
    return true;
}

template< class T >
int
UnorderedSet< T >::find(
    const T& value ) const
{
    typename unordered_map< T, int >::const_iterator it = index.find( value );
    return ( it != index.end() ) ? it->second : NOT_FOUND;
}

template< class T >
bool
UnorderedSet< T >::contains(
    const T& value ) const
{
    return find( value ) != NOT_FOUND;
}

#endif /* _UNORDERED_SET_H_ */