/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "WaspNumericFormat.h"
#include "WaspRule.h"

#include <cstdlib>
#include <cstring>
#include <vector>
#include <ios>

WaspNumericFormat::WaspNumericFormat( 
    ProgramBuilder* p ) : programBuilder( p )
{
}

void
WaspNumericFormat::parse()
{
    this->parse( cin );
}

void
WaspNumericFormat::parse(
    istream& input )
{
    programBuilder->startProgram();
    // Read version
    int number;
   
    input >> number;
    
    if( number != VERSION_ID )
    {
        cerr << ERRORPARSING << endl;
        exit( 0 );
    }
    
    int major;
    int minor;
    input >> major >> minor;
    
    switch( major )
    {
    case 1:
        switch( minor )
        {
        case 0:
            cerr << VERSION_NOT_SUPPORTED << endl;
            exit( 0 );
            break;

        case 1:
            cerr << VERSION_NOT_SUPPORTED << endl;
            exit( 0 );
            break;
        
        default:
            cerr << ERRORPARSING << endl;
            exit( 0 );
            break;
        }
        break;
        
    case 2:
        switch( minor )
        {
        case 0:
            parse_2_0( input );
            break;
            
        default:
            cerr << ERRORPARSING << endl;
            exit( 0 );
            break; 
        }
        break;
        
    default:
        cerr << ERRORPARSING << endl;
        exit( 0 );
        break;
    }
    programBuilder->endProgram();
}

void
WaspNumericFormat::parse_2_0(
    istream& input )
{
    int number;
    
    while( input >> number )
    {
        switch ( number )
        {
        case COMMENT_ID:
            readComment( input );
            break;

        case RULE_ID:
            readWaspRule( input );
            break;
            
        case WEAK_CONSTRAINT_ID:
            readWeakConstraint( input );
            break;

        case INPUT_FALSE_ATOMS:
            readFalseAtoms( input );
            break;
        
        case INPUT_TRUE_SUPPORTED_ATOMS:
            readTrueSupportedAtoms( input );
            break;

        case INPUT_TRUE_UNSUPPORTED_ATOMS:
            readTrueUnsupportedAtoms( input );
            break;
                
        case SUM_ID:
            readSum( input );
            break;

        case COUNT_ID:
            readCount( input );
            break;
            
        case ATOM_TABLE_ID:
            readAtomTable( input );
            break;
            
        case ATOM_TABLE_ID_WITH_AGGREGATES:
            readAtomTableWithAggregates( input );
            break;

        case GROUNDER_MESSAGE:
            readMessage( input );
            break;

        case HAS_QUERY:
            readQuery( input );
            break;
            
        default:
            cerr << ERRORPARSING << endl;
            exit( 0 );
            break;
        }
    }
    if( !input.good() && !input.eof() )
    {
        cerr << ERRORPARSING << endl;
        exit( 0 );
    }  
}

void
WaspNumericFormat::readComment( 
    istream & input )
{
    char buf[ 10240 ];

    input.getline( buf, 10240 );
    assert( input.gcount() < 10239 );
    
    programBuilder->readComment( string( buf ) );
}

void
WaspNumericFormat::readWeakConstraint(
    istream& input )
{
    int weight;
    int level;
    unsigned int size;
    
    input >> weight >> level >> size;
    
    unordered_set< int > tempSet;
    
    bool trivialRule = false;
    vector< int > wRule;
    wRule.reserve( size );
    
    int id;
    for( unsigned int i = 0; i < size; i++ )
    {
        input >> id;
        
        if( tempSet.insert( id ).second )
        {
            wRule.push_back( id );
            if( tempSet.find( -id ) != tempSet.end() )
            {
                trivialRule = true;
            }
        }
    }
    
    if( trivialRule )
        return;
    
    AtomId auxAtomId = programBuilder->addAuxAtomForWeakConstraint();
    WeakConstraint* w = programBuilder->startWeakConstraint( wRule.size() + 1, weight, level );
    
    programBuilder->addAuxAtomInWeakConstraint( w, auxAtomId );
    
    for( unsigned int i = 0; i < wRule.size(); i++ )
    {
        int id = wRule[ i ];
        if( id < 0 )
            programBuilder->addLiteralInWeakConstraint( w, -id, false );
        else
            programBuilder->addLiteralInWeakConstraint( w, id, true );
    }
    programBuilder->endWeakConstraint( w );
}

void
WaspNumericFormat::readWaspRule(
    istream& input )
{
    unsigned int size;
    unsigned int headAtoms;
    input >> size >> headAtoms;
    
    if( headAtoms > size )
    {
        cerr << ERRORPARSING << endl;
        exit( 0 );
    }
    
    unordered_set< int > tempSet;
    
    bool trivialRule = false;
    vector< int > wRule;
    wRule.reserve( size );
    
    unsigned int numberOfAtomsInTheHead = 0;
    
    int id;
    for( unsigned int i = 0; i < headAtoms; i++ )
    {
        input >> id;
        
        if( tempSet.insert( id ).second )
        {
            wRule.push_back( id );
            numberOfAtomsInTheHead++;
        }
    }
    
    for( unsigned int i = headAtoms; i < size; i++ )
    {
        input >> id;
        
        if( tempSet.insert( id ).second )
        {
            wRule.push_back( id );
            if( tempSet.find( -id ) != tempSet.end() )
            {
                trivialRule = true;
            }
        }
    }
    
    if( trivialRule )
        return;
    
    WaspRule* w = programBuilder->startWaspRule( wRule.size(), numberOfAtomsInTheHead );
    
    for( unsigned int i = 0; i < numberOfAtomsInTheHead; i++ )
    {
        programBuilder->addHeadAtom( w, -wRule[ i ] );
    }
    
    if( numberOfAtomsInTheHead > 0 )
    {
        for( unsigned int i = numberOfAtomsInTheHead; i < wRule.size(); i++ )
        {
            int id = wRule[ i ];
            if( id < 0 )
            {
                programBuilder->addBodyLiteral( w, -id, false );
            }
            else
            {
                programBuilder->addBodyLiteral( w, id, true );
            }
        }
    }
    else
    {
        for( unsigned int i = numberOfAtomsInTheHead; i < wRule.size(); i++ )
        {
            int id = wRule[ i ];
            if( id < 0 )
            {
                programBuilder->addBodyLiteralInConstraint( w, -id, false );
            }
            else
            {
                programBuilder->addBodyLiteralInConstraint( w, id, true );
            }
        }
    }
    
    programBuilder->endWaspRule( w );
}

void
WaspNumericFormat::readAtomTable(
    istream& input )
{
    int id;
    char name[ 10240 ];

    int oldId = 0;
    
    while( input >> id )
    {
        if( id == ATOM_SIZE )        
        {
            input >> id;
            while( id > ++oldId )
            {
                programBuilder->addAtom( oldId, "" );
            }
            return;
        }
        
        while( id > ++oldId )
        {
            programBuilder->addAtom( oldId, "" );
        }
        
        switch( id )
        {
        case -1:
            readComment( input );
            break;
            
        default:
            assert( id >= 0 );
            
            input.getline( name, 10240 );
            assert( input.gcount() < 10230 );           

            programBuilder->addAtom( id, name + 2 );     // there are two TABs
            break;
        }

        oldId = id;
    }
}

void
WaspNumericFormat::readAtomTableWithAggregates(
    istream& input )
{
    int id;
    char name[ 10240 ];

    int oldId = 0;
    
    while( input >> id )
    {           
        if( id == ATOM_SIZE )
        {
            input >> id;
            while( id > ++oldId )
            {
                programBuilder->addAtom( oldId, "" );
            }
            return;
        }
        
        while( id > ++oldId )
        {
            programBuilder->addAtom( oldId, "" );
        }                
        
        switch( id )
        {
        case COMMENT_ID:
            readComment( input );
            break;
            
        default:
            assert( id >= 0 );
                        
            int value;
            input >> value;
            
            assert( value == ATOM_NO_AGGREGATE || value == ATOM_IN_AGGREGATE_VALUE ||
                    value == AGGREGATE_GREATER_VALUE || value == AGGREGATE_DOUBLEGUARDS_VALUE );
            
            input.getline( name, 10240 );
            assert( input.gcount() < 10230 );
            switch( value )
            {
                case ATOM_NO_AGGREGATE:
                    programBuilder->addAtom( id, name + 2 );     // there are two TABs
                    break;
                case ATOM_IN_AGGREGATE_VALUE:
                    if( name[ 2 ] == '#' )
                    {
                        programBuilder->addAtomInAnAggregateSet( id, "" );
                    }            
                    else
                    {
                        programBuilder->addAtomInAnAggregateSet( id, name + 2 );                        
                    }
                    break;
                case AGGREGATE_GREATER_VALUE:                    
                    //there are two TABs
                    assert( name[ 2 ] == '#' );
                    programBuilder->addAggregateAtomGreater( id, "" );
                    break;
                case AGGREGATE_DOUBLEGUARDS_VALUE:
                    //there are two TABs
                    assert( name[ 2 ] == '#' );
                    programBuilder->addAggregateAtomDoubleGuards( id, "" );
                    break;
                default:
                    assert( 0 );
                    break;
            }            
            
            break;
        }

        oldId = id;        
    }
}

void
WaspNumericFormat::readCount(
    istream& input )
{
    unsigned int id;
    unsigned int size;

    unsigned int minValue;
    unsigned int maxValue;

    input >> id >> minValue >> maxValue >> size;
    
    Aggregate* aggregate = programBuilder->startAggregateCount( id, minValue, maxValue, size );
    
    AtomId atom;
    for( unsigned int i = 0; i < size; i++ )
    {
        input >> atom;
        programBuilder->addAtomInTheAggregate( atom, 1, aggregate, i );
    }

    programBuilder->endAggregateCount( aggregate );
    
}

bool compare( pair< int, int >& v1, pair< int, int >& v2 )
{
    return v1.second < v2.second;
}

void
WaspNumericFormat::readSum(
    istream& input )
{
    unsigned int id;
    unsigned int size;    

    unsigned int minValue;
    unsigned int maxValue;
    
    input >> id >> minValue >> maxValue >> size;
    
    list< pair< int, int > > tempList;
    int atom;
    int value;
    for( unsigned int i = 0; i < size; i++ )
    {
        input >> atom;
        input >> value;
        
        tempList.push_back( pair< int, int >( atom, value ) );        
    }
    tempList.sort( compare );
    
    Aggregate* aggregate = programBuilder->startAggregateSum( id, minValue, maxValue, size );
    
    unsigned int count = 0;
    for( list< pair< int, int > >::iterator it = tempList.begin(); it != tempList.end(); ++it )
    {
        programBuilder->addAtomInTheAggregate( it->first, it->second, aggregate, count );
        count++;
    }    
    
    programBuilder->endAggregateSum( aggregate );
}

void
WaspNumericFormat::readMessage(
    istream& input )
{
    unsigned int exitCode;
    unsigned int type;
    unsigned int value;

    input >> exitCode;

    programBuilder->startMessageFromGrounding( exitCode );
    
    assert( exitCode == GROUNDER_ERROR || exitCode == GROUNDER_OK );
    
    if( exitCode == GROUNDER_ERROR )
    {
        input >> value;

        switch( value )
        {
            case 0:
                cerr << QUERYNOTSUPPORTED << endl;
                break;

            case 1:
                cerr << ERRORPARSING << endl;
                break;

            default:
                break;
        }
        
        programBuilder->setErrorFromGrounding( value );
    }
    else if( exitCode == GROUNDER_OK )
    {
        input >> type >> value;
        assert( type == QUERY_BRAVE || type == QUERY_CAUTIOUS );        
        assert( value == QUERY_FALSE || value == QUERY_TRUE );

        programBuilder->setMessageFromGrounding( type, value );
    }
    
    programBuilder->endMessageFromGrounding();
}

void
WaspNumericFormat::readQuery(
    istream& input )
{
    unsigned int type;
    input >> type;
    
    assert( type == QUERY_BRAVE || type == QUERY_CAUTIOUS );
    
    programBuilder->setQueryType( type );
}

void 
WaspNumericFormat::readTrueSupportedAtoms( 
    istream& input )
{
    unsigned int size;
    input >> size;
    
    for( int i = 0; i < size; i++ )
    {
        unsigned int id;
        input >> id;
        
        programBuilder->addTrueSupportedAtom( id );
    }
}

void
WaspNumericFormat::readTrueUnsupportedAtoms( 
    istream& input )
{
    unsigned int size;
    input >> size;
    
    for( int i = 0; i < size; i++ )
    {
        unsigned int id;
        input >> id;
        
        programBuilder->addTrueUnsupportedAtom( id );
    }
}

void 
WaspNumericFormat::readFalseAtoms( 
    istream& input )
{
    unsigned int size;
    input >> size;
    
    for( int i = 0; i < size; i++ )
    {
        unsigned int id;
        input >> id;
        
        programBuilder->addFalseAtom( id );
    }
}