/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "WaspProgramBuilder.h"
#include "Program.h"

#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>

void 
WaspProgramBuilder::addAtom( 
    AtomId atomId,
    string atomName )
{
    program->addAtom( atomId, atomName );
}

void 
WaspProgramBuilder::addAtomInAnAggregateSet( 
    AtomId atomId, 
    string atomName )
{
    program->addAtomInAnAggregateSet( atomId, atomName );
}

void 
WaspProgramBuilder::addAggregateAtomGreater( 
    AtomId atomId, 
    string atomName )
{
    program->addAggregateAtomGreater( atomId, atomName );
}

void 
WaspProgramBuilder::addAggregateAtomDoubleGuards( 
    AtomId atomId, 
    string atomName )
{
    program->addAggregateAtomDoubleGuards( atomId, atomName );
}

AtomId
WaspProgramBuilder::addAuxAtom()
{
    AtomId id = program->getNumberOfAtoms();
    program->addAtom( id, "" );
    return id;
}

AtomId
WaspProgramBuilder::addAuxAtomForWeakConstraint()
{
    AtomId id = program->getNumberOfAtoms();
    program->addAtomInAnAggregateSet( id, "" );
    return id;
}

AtomId
WaspProgramBuilder::addAuxAggregateAtomGreater(
    )
{
    AtomId id = program->getNumberOfAtoms();
    program->addAggregateAtomGreater( id, "" );
    return id;
}

void 
WaspProgramBuilder::addHeadAtom( 
    WaspRule* w, 
    AtomId atomId )
{
    Atom* a = program->getAtomPointer( atomId );    
    assert( a != NULL );
    
    Literal literal( a, true );
    w->addLiteral( literal );
    a->addNegativeHeadWaspRule( w );
}

void
WaspProgramBuilder::addBodyLiteral( 
    WaspRule* w, 
    AtomId atomId, 
    bool positive )
{
    Atom* a = program->getAtomPointer( atomId );
    assert( a != NULL );
    
    if( positive )
    {
        Literal literal( a );
        w->addLiteral( literal );
        a->addPositiveWaspRule( w );
    }
    else
    {
        Literal literal( a, true );
        w->addLiteral( literal );
        a->addNegativeBodyWaspRule( w );
    }
}

void 
WaspProgramBuilder::addBodyLiteralInConstraint( 
    WaspRule* w, 
    AtomId atomId, 
    bool positive )
{
    assert( w->getNumberOfAtomsInTheHead() == 0 );
    
    Atom* a = program->getAtomPointer( atomId );
    assert( a != NULL );
    
    if( positive )
    {
        Literal literal( a );
        w->addLiteral( literal );
    }
    else
    {
        Literal literal( a, true );
        w->addLiteral( literal );
    }
}

void 
WaspProgramBuilder::addHeadAtomInChoiceRule( 
    ChoiceRule* c, 
    AtomId atomId )
{
    Atom* a = program->getAtomPointer( atomId );    
    assert( a != NULL );
    
    c->addHeadAtom( a );
    a->addHeadChoiceRule( c );
}

void
WaspProgramBuilder::addBodyLiteralInChoiceRule( 
    ChoiceRule* c, 
    AtomId atomId, 
    bool positive )
{
    Atom* a = program->getAtomPointer( atomId );
    assert( a != NULL );
    
    if( positive )
    {
        Literal literal( a );
        c->addBodyLiteral( literal );
        a->addPositiveBodyChoiceRule( c );
    }
    else
    {
        Literal literal( a, true );
        c->addBodyLiteral( literal );
        a->addNegativeBodyChoiceRule( c );
    }
}

void 
WaspProgramBuilder::startProgram()
{
}

bool compare( pair< AtomId, int >& v1, pair< AtomId, int >& v2 )
{
    return v1.second < v2.second;
}

void 
WaspProgramBuilder::endProgram()
{
    program->computeLinearCostsForWeakConstraints();
    if( !atomsInWeakConstraints.empty() )
    {
        AtomId auxAtom = this->addAuxAggregateAtomGreater();
        
        this->addFalseAtom( auxAtom );
        
        unsigned int aggregateSize = atomsInWeakConstraints.size();
        int maxValue = program->getInitialMaxValueWeakConstraints();
        
        list< pair< AtomId, int > > listOfCosts;
        for( unsigned int i = 0; i < aggregateSize; ++i )
        {
            assert( atomsInWeakConstraints[ i ].first != NULL );
            assert( atomsInWeakConstraints[ i ].second != NULL );
            listOfCosts.push_back( pair< AtomId, int >( atomsInWeakConstraints[ i ].first->getId(), atomsInWeakConstraints[ i ].second->getLinearCost() ) );
        }
        
        listOfCosts.sort( compare );
        
        Aggregate* aggregate = this->startAggregateSum( auxAtom, maxValue, MAXAGGREGATE, aggregateSize );
        unsigned int count = 0;
        for( list< pair< AtomId, int > >::iterator it = listOfCosts.begin(); it != listOfCosts.end(); ++it )
        {
            this->addAtomInTheAggregate( it->first, it->second, aggregate , count );
            count++;
        }
        this->endAggregateSum( aggregate );
        program->setAggregateRepresentingWeakConstraints( aggregate );        
    }
    
    program->aggregatesInit();
    program->buildDependencyGraph();
    program->componentsStrongConnected();
    #ifndef STATS_OFF
    program->computeStatistics();
    #endif
}

void 
WaspProgramBuilder::addFalseAtom( 
    AtomId atomId )
{
    assert( program->getAtomPointer( atomId ) != NULL );
    program->addFalseAtom( program->getAtomPointer( atomId ) );
}

void
WaspProgramBuilder::addTrueUnsupportedAtom( 
    AtomId atomId )
{
    assert( program->getAtomPointer( atomId ) != NULL );
    program->addTrueUnsupportedAtom( program->getAtomPointer( atomId ) );
}

void
WaspProgramBuilder::addTrueSupportedAtom( 
    AtomId atomId )
{
    assert( program->getAtomPointer( atomId ) != NULL );
    program->addTrueSupportedAtom( program->getAtomPointer( atomId ) );
}

void
WaspProgramBuilder::readComment(
    string )
{
}

void
WaspProgramBuilder::startMessageFromGrounding(
    unsigned int )
{
}

void
WaspProgramBuilder::endMessageFromGrounding()
{
}

Aggregate*
WaspProgramBuilder::startAggregateCount( AtomId id, unsigned int minGuard, unsigned int maxGuard, unsigned int size )
{
    Aggregate* aggregatePointer = new Aggregate( id, minGuard, maxGuard, size, COUNT_ID );
    AggregateAtomGreater& aggregateAtom = *( static_cast< AggregateAtomGreater* >( program->getAtomPointer( id ) ) );
    aggregateAtom.setAggregate( aggregatePointer );
    return aggregatePointer;
}

void
WaspProgramBuilder::endAggregateCount(
    Aggregate* aggregate )
{
    program->addAggregate( aggregate );
}

Aggregate*
WaspProgramBuilder::startAggregateSum( AtomId id, unsigned int minGuard, unsigned int maxGuard, unsigned int size )
{
    Aggregate* aggregatePointer = new Aggregate( id, minGuard, maxGuard, size, SUM_ID );
    AggregateAtomGreater& aggregateAtom = *( static_cast< AggregateAtomGreater* >( program->getAtomPointer( id ) ) );
    aggregateAtom.setAggregate( aggregatePointer );
    return aggregatePointer;
}

void
WaspProgramBuilder::endAggregateSum(
    Aggregate* aggregate )
{
    program->addAggregate( aggregate );
}

WaspRule*
WaspProgramBuilder::startWaspRule( 
    unsigned int size, 
    unsigned int numberOfHeadAtoms )
{
    return new WaspRule( size, numberOfHeadAtoms );
}

void
WaspProgramBuilder::endWaspRule(
    WaspRule* w )
{
    if( w->getNumberOfAtomsInTheHead() > 0 )
    {
        WatchedList< WaspRule* >::iterator iteratorFirst = w->getFirstWatchValue().addWatchedWaspRule( w );
        w->setIteratorFirstWatch( iteratorFirst );
        WatchedList< WaspRule* >::iterator iteratorSecond = w->getSecondWatchValue().addWatchedWaspRule( w );
        w->setIteratorSecondWatch( iteratorSecond );
    }
    else
    {
        WatchedList< WaspRule* >::iterator iteratorFirst = w->getFirstWatchValue().addWatchedConstraint( w );
        w->setIteratorFirstWatch( iteratorFirst );
        WatchedList< WaspRule* >::iterator iteratorSecond = w->getSecondWatchValue().addWatchedConstraint( w );
        w->setIteratorSecondWatch( iteratorSecond );
    }
        
    program->addWaspRule( w );
}

ChoiceRule*
WaspProgramBuilder::startChoiceRule( 
    unsigned int numberOfBodyLiterals, 
    unsigned int numberOfHeadAtoms )
{
    return new ChoiceRule( program->getChoiceRulesSize(), numberOfBodyLiterals, numberOfHeadAtoms );
}

void
WaspProgramBuilder::endChoiceRule(
    ChoiceRule* c )
{
    program->addChoiceRule( c );
}

WeakConstraint*
WaspProgramBuilder::startWeakConstraint( 
    unsigned int size,
    int weight,
    int level )
{
    return new WeakConstraint( size, weight, level );
}

void
WaspProgramBuilder::endWeakConstraint(
    WeakConstraint* wc )
{
    assert( wc->getNumberOfAtomsInTheHead() == 1 );
    
    WatchedList< WaspRule* >::iterator iteratorFirst = wc->getFirstWatchValue().addWatchedWaspRule( wc );
    wc->setIteratorFirstWatch( iteratorFirst );
    WatchedList< WaspRule* >::iterator iteratorSecond = wc->getSecondWatchValue().addWatchedWaspRule( wc );
    wc->setIteratorSecondWatch( iteratorSecond );
    program->addWeakConstraint( wc );    
}

void
WaspProgramBuilder::addAuxAtomInWeakConstraint( 
    WeakConstraint* wc, 
    AtomId atomId )
{
    Atom* a = program->getAtomPointer( atomId );
    assert( a != NULL );

    //Negative Literal
    Literal literal( a, false );
    wc->addLiteral( literal );
    
    a->addNegativeHeadWaspRule( wc );
    
    atomsInWeakConstraints.push_back( pair< Atom*, WeakConstraint* > ( a, wc ) );
}

void
WaspProgramBuilder::addLiteralInWeakConstraint( 
    WeakConstraint* wc, 
    AtomId atomId, 
    bool positive )
{
    Atom* a = program->getAtomPointer( atomId );
    assert( a != NULL );
    
    if( positive )
    {
        Literal literal( a );
        wc->addLiteral( literal );
        a->addPositiveWaspRule( wc );
    }
    else
    {
        Literal literal( a, false );
        wc->addLiteral( literal );
        a->addNegativeBodyWaspRule( wc );
    }    
}

void 
WaspProgramBuilder::addAtomInTheAggregate( 
    int atom, 
    int value,
    Aggregate* aggregate,
    unsigned int position )
{
    Atom* a = program->getAtomPointer( atom );
    assert( a != NULL );
    
    assert( aggregate != NULL );
    
    if( value < 0 )
    {
        //Negative literal
        Literal l( a, false );
        aggregate->addLiteralInTheAggregateSet( l, -value );
    }
    else
    {
        //Positive literal
        Literal l( a );
        aggregate->addLiteralInTheAggregateSet( l, value );
    }
    
    
    AtomInAnAggregateSet& at = *( static_cast< AtomInAnAggregateSet* >( a ) );
    at.addAggregate( aggregate, position );
}

void 
WaspProgramBuilder::setMessageFromGrounding( 
    unsigned int type, 
    unsigned int value )
{
    program->setType( type );
    program->setValue( value );
    program->setMessageFromGrounding();
}

void 
WaspProgramBuilder::setErrorFromGrounding( 
    unsigned int )
{
    program->setErrorFromGrounding();
}

void 
WaspProgramBuilder::setQueryType( 
    QueryType queryType )
{
    program->setType( queryType );
}