/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   WaspRule.h
 * Author: Carmine Dodaro
 *
 * Created on 4 September 2012, 17.00
 */
#ifndef _WASPRULE_H
#define	_WASPRULE_H

#include <iostream>
#include <cassert>
#include "Constants.h"
#include "Literal.h"
#include "Rule.h"
using namespace std;

class WaspRule;
template< class T >
class WatchedList;
//template< class T1 >
//typedef typename WatchedList< T1 >::iterator iterator;

class WaspRule : public Rule
{
    friend class ModelGenerator;
    friend ostream& operator<<( ostream& o, WaspRule& w )
    {   
        unsigned int headAtoms = w.numberOfAtomsInTheHead;
        unsigned int bodyLiterals = w.getNumberOfLiteralsInTheBody();
        
        if( headAtoms > 0 )
        {
            Literal l = w.literals[ 0 ];
            l.setPositive();
            o << l;
            
            for( unsigned int i = 1; i < headAtoms; ++i )
            {
                Literal l = w.literals[ i ];
                l.setPositive();
                o << " | " << l;
            }
        }
        
        if( bodyLiterals > 0 )
        {
            o << " :- ";
            
            o << w.getBodyLiteral( 0 );
            
            for( unsigned int i = 1; i < bodyLiterals; ++i )
            {
                o << ", " << w.getBodyLiteral( i );
            }
        }
        o << ".";
        
        return o;
    }
    
    public:
        inline WaspRule( unsigned int size, unsigned int headAtoms );
        inline WaspRule( unsigned int size, unsigned int headAtoms, unsigned int fWatch, unsigned int sWatch );
        
        virtual ~WaspRule();
        inline void addLiteral( const Literal& literal );
        inline unsigned int size() const;
        inline void setNumberOfAtomsInTheHead( unsigned int headAtoms );
        inline unsigned int getNumberOfAtomsInTheHead() const;
        inline unsigned int getNumberOfLiteralsInTheBody() const;
        inline Literal& getLiteral( unsigned int i );
        inline const Literal& getLiteral( unsigned int i ) const;
        inline unsigned int getIdFirstFalseLiteral() const;
        inline void setIdFirstFalseLiteral( unsigned int i );
        inline unsigned int getIdSecondFalseLiteral() const;
        inline void setIdSecondFalseLiteral( unsigned int i );
        
        inline bool isUndefined() const;
        inline bool isSatisfied() const;
        inline bool isSatisfiedCannotSupport() const;
        inline bool isSatisfiedCanSupport() const;
        inline bool isConstraint() const;
        inline bool isBodyFalse() const;
        
        #ifndef NDEBUG
        inline bool isViolated() const;
        #endif
        
        inline unsigned int getFirstWatch() const;
        inline Literal& getFirstWatchValue();
        inline const Literal& getFirstWatchValue() const;
        
        inline unsigned int getSecondWatch() const;
        inline Literal& getSecondWatchValue();
        inline const Literal& getSecondWatchValue() const;
        
        inline void makeRuleSatisfied();
        inline void makeRuleCanSupport();
        inline void makeRuleSatisfiedByBody();
        
        inline unsigned updateFirstWatch();
        inline unsigned updateSecondWatch();
        
        inline Atom* getHeadAtom( unsigned int position );
        inline Literal& getBodyLiteral( unsigned int position );
        
        inline WatchedList< WaspRule* >::iterator getIteratorFirstWatch();
        inline WatchedList< WaspRule* >::iterator getIteratorSecondWatch();
        
        inline void setIteratorFirstWatch( WatchedList< WaspRule* >::iterator );
        inline void setIteratorSecondWatch( WatchedList< WaspRule* >::iterator );
        
        virtual bool isLearned() const;
        
        bool isActiveForAtom( AtomId ) const;
        
        virtual bool isRuleActiveForAtom( AtomId atomId ) const;
        virtual bool isWaspRule() const;
        virtual void print();
        virtual unsigned int getSize() const;
        
        virtual Literal& getRuleBodyLiteral( unsigned int position );

    protected:
        inline WaspRule( const WaspRule& );
        
    private:        
        unsigned int numberOfAtomsInTheHead;
        vector< Literal > literals;        
        
        unsigned int firstWatch;
        unsigned int secondWatch;
        
        WatchedList< WaspRule* >::iterator iterator_firstWatch;
        WatchedList< WaspRule* >::iterator iterator_secondWatch;
        
        unsigned int state;
        unsigned int idFirstFalseLiteral;
        unsigned int idSecondFalseLiteral;
};

WaspRule::WaspRule(
    unsigned int size,
    unsigned int headAtoms ) : 
        numberOfAtomsInTheHead( headAtoms ), state( RULE_UNDEFINED ), 
        firstWatch( rand() % size ), secondWatch ( rand() % ( size - 1 ) ), 
        idFirstFalseLiteral( 0 ), idSecondFalseLiteral( 0 )
{
    assert( size > 1 );
    literals.reserve( size );
    if( secondWatch >= firstWatch )
        secondWatch++;
}

WaspRule::WaspRule(
    unsigned int size,
    unsigned int headAtoms,
    unsigned int fWatch,
    unsigned int sWatch ) : 
        numberOfAtomsInTheHead( headAtoms ), state( RULE_UNDEFINED ), 
        firstWatch( fWatch ), secondWatch ( sWatch ), 
        idFirstFalseLiteral( 0 ), idSecondFalseLiteral( 0 )
{
    assert( size > 0 );
    literals.reserve( size );
    assert( fWatch != sWatch || size == 1 );
}

WaspRule::WaspRule(
    const WaspRule& )
{
    assert( 0 );
}

void
WaspRule::addLiteral(
    const Literal& literal )
{
    literals.push_back( literal );
}

unsigned int 
WaspRule::size() const
{
    return literals.size(); 
}

void
WaspRule::setNumberOfAtomsInTheHead(
    unsigned int headAtoms )
{
    numberOfAtomsInTheHead = headAtoms;
}

unsigned int 
WaspRule::getNumberOfAtomsInTheHead() const
{
    return numberOfAtomsInTheHead;
}

unsigned int
WaspRule::getNumberOfLiteralsInTheBody() const
{
    return literals.size() - numberOfAtomsInTheHead;
}

unsigned int
WaspRule::getIdFirstFalseLiteral() const
{
    return idFirstFalseLiteral;
}

void
WaspRule::setIdFirstFalseLiteral(
    unsigned int i )
{
    idFirstFalseLiteral = i;
}

unsigned int
WaspRule::getIdSecondFalseLiteral() const
{
    return idSecondFalseLiteral;
}

void
WaspRule::setIdSecondFalseLiteral(
    unsigned int i )
{
    idSecondFalseLiteral = i;
}

Literal& 
WaspRule::getLiteral( 
    unsigned int i )
{
    assert( i < literals.size() );
    return literals[ i ];
}

const Literal&
WaspRule::getLiteral(
    unsigned int i ) const
{
    assert( i < literals.size() );
    return literals[ i ];
}

Atom*
WaspRule::getHeadAtom(
    unsigned int i )
{
    assert( i < numberOfAtomsInTheHead );
    return literals[ i ].getAtom();
}

Literal&
WaspRule::getBodyLiteral(
    unsigned int i )
{
    assert( i < getNumberOfLiteralsInTheBody() );
    assert( ( i + numberOfAtomsInTheHead ) < literals.size() );
    return literals[ i + numberOfAtomsInTheHead ];
}

#ifndef NDEBUG
bool 
WaspRule::isViolated() const
{
    for( unsigned int i = 0; i < literals.size(); i++ )
    {
        if( !literals[ i ].isTruePropagated() )
            return false;
    }
    
    return true;
}
#endif


bool 
WaspRule::isUndefined() const
{
    return state == RULE_UNDEFINED;
}

bool 
WaspRule::isSatisfied() const
{
    return ( state & 1 );
//    return ( ( state & 1 ) == 1 );
        //return true;
    //return false;
}

bool 
WaspRule::isSatisfiedCannotSupport() const
{
    return ( ( state & 3 ) == 3 ); 
    //return state == RULE_SATISFIED;
}

bool
WaspRule::isBodyFalse() const
{
    return state == RULE_SATISFIED_BY_BODY;
}

bool 
WaspRule::isSatisfiedCanSupport() const
{
    return state == RULE_CAN_SUPPORT;
}

unsigned int 
WaspRule::getFirstWatch() const
{
    return firstWatch;
}

Literal& 
WaspRule::getFirstWatchValue()
{
    assert( firstWatch < size() );
    return literals[ firstWatch ];
}

const Literal& 
WaspRule::getFirstWatchValue() const
{
    assert( firstWatch < size() );
    return literals[ firstWatch ];
}

unsigned int 
WaspRule::getSecondWatch() const
{
    return secondWatch;
}

Literal& 
WaspRule::getSecondWatchValue()
{
    assert( secondWatch < size() );
    return literals[ secondWatch ];
}

const Literal& 
WaspRule::getSecondWatchValue() const
{
    assert( secondWatch < size() );
    return literals[ secondWatch ];
}

void 
WaspRule::makeRuleSatisfied()
{
    state = RULE_SATISFIED;
}

void
WaspRule::makeRuleCanSupport()
{
    state = RULE_CAN_SUPPORT;
}

void
WaspRule::makeRuleSatisfiedByBody()
{
    state = RULE_SATISFIED_BY_BODY;
}

unsigned int 
WaspRule::updateFirstWatch()
{
    assert( size() > 1 );
    assert( firstWatch < size() );
    firstWatch = ( firstWatch + 1 ) % size();
    
    if( firstWatch == secondWatch )
    {
        firstWatch = ( firstWatch + 1 ) % size();
    }
    
    assert( firstWatch < size() );
    return firstWatch;
}

unsigned int 
WaspRule::updateSecondWatch()
{
    assert( size() > 1 );
    assert( secondWatch < size() );
    secondWatch = ( secondWatch + 1 ) % size();
    
    if( firstWatch == secondWatch )
    {
        secondWatch = ( secondWatch + 1 ) % size();
    }
    
    assert( secondWatch < size() );
    return secondWatch;
}

//Literal*
//WaspRule::onlyOneUndefined()
//{
//    assert( secondWatch != firstWatch || secondWatch < literals.size() );
//    return secondWatch == firstWatch ? &( literals[ secondWatch ] ) : NULL;
//}

//bool 
//WaspRule::watchingTheSameLiteral() const
//{
//    return leftWatch == rightWatch;
//}

bool 
WaspRule::isConstraint() const
{
    return numberOfAtomsInTheHead == 0;
}

WatchedList< WaspRule* >::iterator 
WaspRule::getIteratorFirstWatch()
{
    return iterator_firstWatch;
}

WatchedList< WaspRule* >::iterator 
WaspRule::getIteratorSecondWatch()
{
    return iterator_secondWatch;
}

void 
WaspRule::setIteratorFirstWatch( 
    WatchedList< WaspRule* >::iterator it )
{
    iterator_firstWatch = it;
}

void 
WaspRule::setIteratorSecondWatch( 
    WatchedList< WaspRule* >::iterator it )
{
    iterator_secondWatch = it;
}

#endif	/* _WASPRULE_H */
