/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include <vector>

#include "BerkminHeuristic.h"
#include "../Interpretation.h"
#include "../Program.h"
#include "../timer.h"

BerkminHeuristic::~BerkminHeuristic()
{
}

BerkminHeuristic::BerkminHeuristic( 
    ModelGenerator* mg, 
    unsigned int as ) : DecisionHeuristic( mg ), atomsSize( as ), lastUsedRule( 0 )
{
    atomsNegativeGlobalCounters.reserve( atomsSize );
    atomsPositiveGlobalCounters.reserve( atomsSize );
    atomsNegativeCounters.reserve( atomsSize );
    atomsPositiveCounters.reserve( atomsSize );
    atomsNegativeOccurrences.reserve( atomsSize );
    atomsPositiveOccurrences.reserve( atomsSize );
    
    pushElement();
    
    for( int i = 1; i < atomsSize; i++ )
    {
        pushElement();
        computeOccurrencesForAtom( i );
    }
}

void
BerkminHeuristic::onProgramReduction()
{
    for( int i = 1; i < atomsSize; i++ )
    {
        computeOccurrencesForAtom( i );
    }
}

void
BerkminHeuristic::pushElement()
{
    HEURISTIC_TYPE initialValue = 0;
    atomsNegativeGlobalCounters.push_back( initialValue );
    atomsPositiveGlobalCounters.push_back( initialValue );
    atomsNegativeCounters.push_back( initialValue );
    atomsPositiveCounters.push_back( initialValue );
    atomsNegativeOccurrences.push_back( 0 );
    atomsPositiveOccurrences.push_back( 0 );
}

void
BerkminHeuristic::makeAChoice()
{    
    startTimer( berkmin );    
    if( !pickLiteralFromLearnedConstraints() )
    {
        if( !pickLiteralWithActivity() )
        {
            pickLiteralWithNoActivity();
        }
    }    
    stopTimer( berkmin );
    assert( chosenAtom != 0 );
}

bool
BerkminHeuristic::pickLiteralFromLearnedConstraints()
{
    return topMostUndefinedLearnedConstraint();
}

bool
BerkminHeuristic::topMostUndefinedLearnedConstraint()
{
    bool found = false;

    Interpretation& interpretation = getInterpretation();
    Program& program = getProgram();
    vector< AtomId > maxVector;

    unsigned int count = 0;
    
    List< LearnedWaspRule* >& learnedWaspRules = program.getLearnedWaspRules();
    for( List< LearnedWaspRule* >::reverse_iterator it = learnedWaspRules.rbegin(); it != learnedWaspRules.rend(); ++it )
    {
        if( count++ > options.maxBerkminNumber )
            break;
        
        assert( *it != NULL );
        LearnedWaspRule& currentRule = *( *it );
        
        if( interpretation.isFalsePropagated( currentRule.getFirstWatchValue() ) || interpretation.isFalsePropagated( currentRule.getSecondWatchValue() ) || currentRule.getNumberOfLiteralsInTheBody() == 1 )
            continue;
        
//            if( lastUsedRule == i )
//            {
//                if( useCache() )
//                    return true;
//            }
//            else
//            {
//                atomsCache.clear();
//                lastUsedRule = i;
//            }
            
        
        assert( !currentRule.isViolated() );

        AtomId atomId = 0;
        HEURISTIC_TYPE max = 0;
        
        unsigned int j = 0;
        for( ; j < currentRule.size(); j++ )
        {
            AtomId currentAtomId = currentRule.getLiteral( j ).getAtomId();
            if( interpretation.isUndefined( currentAtomId ) )
            {
                atomId = currentAtomId;
                max = getTotalCounter( currentAtomId );
                break;
            }
        }

        found = true;
        assert( atomId != 0 );
        
        maxVector.push_back( atomId );
        for( ; j < currentRule.size(); j++ )
        {
            AtomId currentAtomId = currentRule.getLiteral( j ).getAtomId();
            if( interpretation.isUndefined( currentAtomId ) )
            {
                HEURISTIC_TYPE currentAtomTotalCounter = getTotalCounter( currentAtomId );
                if( currentAtomTotalCounter > max )
                {
                    maxVector.clear();
                    
                    max = getTotalCounter( currentAtomId );
                    atomId = currentAtomId;
                    
                    maxVector.push_back( currentAtomId );
                }
                else if( currentAtomTotalCounter == max )
                {
                    maxVector.push_back( currentAtomId );
                }
            }
        }

        assert( atomId != 0 );

        assert( maxVector.size() > 0 );
        if( maxVector.size() > 1 )
        {
            if( tieBreak( maxVector ) )
                return true;
        }
        chosenAtom = atomId;
        break;
    }

    if( !found )    
    {
        return false;
    }
    else
    {
        if( getGlobalPositiveCounter( chosenAtom ) > getGlobalNegativeCounter( chosenAtom ) )
            chosenPolarity = TRUE;
        else
            chosenPolarity = FALSE;
    }
    
    statistics( incrementBerkminInvocationsNumber() );
    return true;
}

bool
BerkminHeuristic::pickLiteralWithActivity()
{
    if( higherGlobalCounter() )
    {
        choosePolarity( getAtom( chosenAtom ) );
        return true;
    }
    return false;    
}

void
BerkminHeuristic::pickLiteralWithNoActivity()
{
    doMostOccurrences();
}

bool
BerkminHeuristic::higherGlobalCounter()
{
    startTimer( higherGlobalCounterTimer );

    statistics( incrementHigherGlobalCounterInvocationsNumber() );
    HEURISTIC_TYPE max = 0;
    AtomId id = 0;    

    unsigned int mostOccurrencesMax = 0;
    AtomId mostOccurrencesId = 0;
    
    Interpretation& interpretation = getInterpretation();

    unsigned int i = interpretation.getLeftWatch();
    for( ; i <= interpretation.getRightWatch(); i++ )
    {
        if( interpretation.isUndefined( i ) )
        {
            if( getTotalCounter( i ) > max )
            {
                max = getTotalCounter( i );
                id = i;
                break;
            }            

            if( getOccurrences( i ) > mostOccurrencesMax )
            {
                mostOccurrencesMax = getOccurrences( i );
                mostOccurrencesId = i;
            }
        }
    }
    
    i++;
    for( ; i <= interpretation.getRightWatch(); i++ )
    {
        if( interpretation.isUndefined( i ) )
        {
            if( getTotalCounter( i ) > max )
            {
                max = getTotalCounter( i );
                id = i;
            }            
        }
    }

    stopTimer( higherGlobalCounterTimer );

    if( id != 0 )
    {
        chosenAtom = id;
        return true;
    }
    else
    {
        assert( mostOccurrencesId != 0 );
        chosenAtom = mostOccurrencesId;
        return false;        
    }
}

void
BerkminHeuristic::doMostOccurrences()
{
    statistics( incrementMostOccurrencesInvocationsNumber() );
    chosenPolarity = isMoreNegativeOccurrences( chosenAtom ) ? FALSE : TRUE;    
}

void
BerkminHeuristic::choosePolarity(
    Atom& a )
{
    startTimer( lookahead );
    statistics( incrementChoosePolarityInvocationsNumber() );
    AtomId atomId = a.getId();
    
    TruthValue atomTruthValue;
    TruthValue otherAtomTruthValue;
    
    if( isMoreNegativeOccurrences( atomId ) )
    {
        atomTruthValue = TRUE;
        otherAtomTruthValue = FALSE;
    }
    else
    {
        atomTruthValue = FALSE;
        otherAtomTruthValue = TRUE;
    }

    AtomTruthValue literal;
    literal.first = atomId;    
    literal.second = atomTruthValue;

    nextLevel();

    Stack< unsigned int >& backtrackingStack = getBacktrackingStack();
    stack< AtomTruthValue >& propagationStack = getPropagationStack();
    backtrackingStack.push( &( a.getChoiceAddress() ) );
    a.setChoice();
    propagationStack.push( literal );
    if( !propagate() )
    {
        unrollOne();
//        currentChoice.second = otherAtomTruthValue;
        chosenPolarity = atomTruthValue;
    }
    else
    {
        unsigned int diff = backtrackingStack.size();
        unrollOne();
        nextLevel();

        backtrackingStack.push( &( a.getChoiceAddress() ) );
        a.setChoice();
        
        literal.second = otherAtomTruthValue;
        propagationStack.push( literal );
        if( !propagate() )
        {
            unrollOne();            
//            currentChoice.second = atomTruthValue;
            chosenPolarity = otherAtomTruthValue;
        }
        else
        {
            unrollOne();
            if( diff <= backtrackingStack.size() )
                chosenPolarity = otherAtomTruthValue;
            else
                chosenPolarity = atomTruthValue;
        }
    }
    
    stopTimer( lookahead );
}

void
BerkminHeuristic::onUIPComputationTraversal( 
    AtomId atomId, 
    bool isTrue )
{
    incrementHeuristicsCounters( atomId, isTrue );
}

void
BerkminHeuristic::onLearnedWaspRule( 
    AtomId atomId, 
    bool isTrue )
{
    incrementHeuristicsCounters( atomId, isTrue );
}

void
BerkminHeuristic::onReachedANumberOfConflicts()
{
    unsigned int size = getAtomsSize();
    for( unsigned int i = 1; i < size; i++ )
    {
        divideNegativeCounter( i );
        dividePositiveCounter( i );
    }
}

bool
compareCache( 
    CacheForHeuristic& c1, 
    CacheForHeuristic& c2 )
{
    unsigned int value1 = c1.countPositive > c1.countNegative ? c1.countPositive : c1.countNegative;
    unsigned int value2 = c2.countPositive > c2.countNegative ? c2.countPositive : c2.countNegative;
    
    return value1 > value2;
}

bool
BerkminHeuristic::tieBreak(
    vector< AtomId >& maxVector )
{
    unsigned int maxVectorSize = maxVector.size();
    unsigned int maxNegative = 0;
    unsigned int maxPositive = 0;
    AtomId idNegative = 0;
    AtomId idPositive = 0;    

    for( unsigned int j = 0; j < maxVectorSize; j++ )
    {
        unsigned int countNegative = 0;
        unsigned int countPositive = 0;

        AtomId currentAtom = maxVector[ j ];
        Atom& a = getAtom( currentAtom );

        unsigned int negativeBodySize = a.getNegativeBodyWaspRules().size();
        for( unsigned int t = 0; t < negativeBodySize; t++ )
        {
            WaspRule& w = *( a.getNegativeBodyWaspRules()[ t ] );
            if( !w.isSatisfiedCannotSupport() )
            {
                countNegative++;
            }
        }

        unsigned int negativeHeadSize = a.getNegativeHeadWaspRules().size();
        for( unsigned int t = 0; t < negativeHeadSize; t++ )
        {
            WaspRule& w = *( a.getNegativeHeadWaspRules()[ t ] );
            if( w.isSatisfiedCanSupport() )
            {
                countNegative++;
            }
        }

        unsigned int positiveBodySize = a.getPositiveWaspRules().size();
        for( unsigned int t = 0; t < positiveBodySize; t++ )
        {
            WaspRule& w = *( a.getPositiveWaspRules()[ t ] );
            if( !w.isSatisfiedCannotSupport() )
            {
                countPositive++;
            }
        }

        if( countPositive > maxPositive )
        {
            maxPositive = countPositive;
            idPositive = currentAtom;
        }

        if( countNegative > maxNegative )
        {
            maxNegative = countNegative;
            idNegative = currentAtom;
        }
        
        //atomsCache.push_back( CacheForHeuristic( currentAtom, countPositive, countNegative ) );
    }

    //atomsCache.sort( compareCache );
    
    if( idPositive != 0 || idNegative != 0 )
    {
        if( maxPositive > maxNegative )
        {
            chosenAtom = idPositive;
            chosenPolarity = FALSE;
//            chosenPolarity = TRUE;
        }
        else
        {
            chosenAtom = idNegative;
            chosenPolarity = TRUE;
//            chosenPolarity = FALSE;
        }
        
        return true;
    }
    
    return false;
}

void
BerkminHeuristic::computeOccurrencesForAtom(
    AtomId atomId )
{
    Atom& a = getAtom( atomId );
    assert( atomId > 0 );
    assert( atomId < atomsPositiveOccurrences.size() );
    atomsPositiveOccurrences[ atomId ] = a.getPositiveWaspRules().size() + a.getPositiveOccurrencesAggregates() + a.getWatchedPositiveConstraints().size();
    assert( atomId < atomsNegativeOccurrences.size() );    
    atomsNegativeOccurrences[ atomId ] = a.getNegativeBodyWaspRules().size() + a.getNegativeHeadWaspRules().size() + a.getNegativeOccurrencesAggregates() + a.getWatchedNegativeConstraints().size();
}

bool
BerkminHeuristic::useCache()
{
    if( atomsCache.empty() )
        return false;    
    
    Interpretation& interpretation = getInterpretation();
    
    while( !atomsCache.empty() )
    {
        CacheForHeuristic& c = atomsCache.front();    
        atomsCache.pop_front();
    
        if( interpretation.isUndefined( c.atomId ) )
        {
            chosenAtom = c.atomId;
            chosenPolarity = c.getTruthValue();
            
            return true;
        }
    }
    
    return false;
}
