/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   BerkminHeuristic.h
 * Author: Carmine Dodaro
 *
 * Created on 04 December 2012, 17.58
 */

#ifndef _BERKMINHEURISTIC_H
#define	_BERKMINHEURISTIC_H

#include <cassert>
#include "../stl/List.h"
#include "../Constants.h"
#include "../DecisionHeuristic.h"

typedef unsigned int HEURISTIC_TYPE;

class Atom;

class CacheForHeuristic
{
    public:
        CacheForHeuristic( 
            AtomId a,
            unsigned int countPos, 
            unsigned int countNeg ) : 
                atomId( a ), countPositive( countPos ), countNegative( countNeg )
        {
        }
    
        TruthValue 
        getTruthValue()
        {
            return ( countPositive > countNegative ) ? FALSE : TRUE;            
        }
                
        AtomId atomId;
        unsigned int countPositive;
        unsigned int countNegative;
};

class BerkminHeuristic : public DecisionHeuristic
{
    public:
        BerkminHeuristic( ModelGenerator* mg, unsigned int atomsSize );
        virtual ~BerkminHeuristic();
                
        virtual void makeAChoice();
        
        virtual void onUIPComputationTraversal( AtomId atomId, bool isTrue );
        virtual void onLearnedWaspRule( AtomId atomId, bool isTrue );
        virtual void onReachedANumberOfConflicts();
        virtual void onProgramReduction();
        
    protected:
        virtual bool pickLiteralFromLearnedConstraints();
        virtual bool pickLiteralWithActivity();
        virtual void pickLiteralWithNoActivity();
        virtual bool tieBreak( vector< AtomId >& );
        
        virtual void computeOccurrencesForAtom( AtomId atomId );
        inline bool isMoreNegativeOccurrences( AtomId atomId );
        inline unsigned int getNegativeOccurrences( AtomId atomId );
        inline unsigned int getPositiveOccurrences( AtomId atomId );
        inline unsigned int getOccurrences( AtomId atomId );
        
        
        bool topMostUndefinedLearnedConstraint();
        bool higherGlobalCounter();        
        void doMostOccurrences();
        virtual void choosePolarity( Atom& a );                
        
        vector< HEURISTIC_TYPE > atomsPositiveCounters;
        vector< HEURISTIC_TYPE > atomsNegativeCounters;
        vector< HEURISTIC_TYPE > atomsPositiveGlobalCounters;
        vector< HEURISTIC_TYPE > atomsNegativeGlobalCounters;
        vector< unsigned int > atomsPositiveOccurrences;
        vector< unsigned int > atomsNegativeOccurrences;
        
        List< CacheForHeuristic > atomsCache;
        RuleId lastUsedRule;
        
        inline void divideNegativeCounter( AtomId atomId );
        
        inline void dividePositiveCounter( AtomId atomId );

        inline unsigned int getAtomsSize();
        
        inline HEURISTIC_TYPE getGlobalNegativeCounter( AtomId atomId ) const;

        inline HEURISTIC_TYPE getGlobalPositiveCounter( AtomId atomId ) const;
        
        inline HEURISTIC_TYPE getNegativeCounter( AtomId atomId ) const;
        
        inline HEURISTIC_TYPE getPositiveCounter( AtomId atomId ) const;
        
        inline HEURISTIC_TYPE getTotalCounter( AtomId atomId ) const;
        
        inline void incrementHeuristicsCounters( AtomId atomId, bool isTrue );
        
        inline void incrementNegativeCounter( AtomId atomId );

        inline void incrementPositiveCounter( AtomId atomId );
        
        inline bool useCache();
        
    private:
        unsigned int atomsSize;
        inline bool checkAtomId( AtomId atomId ) const;
        
        void pushElement();
};

unsigned int 
BerkminHeuristic::getAtomsSize()
{
    return atomsSize;
}

bool 
BerkminHeuristic::checkAtomId(
    AtomId atomId ) const
{
    assert( atomsPositiveCounters.size() == atomsSize );
    assert( atomsNegativeCounters.size() == atomsSize );
    assert( atomsPositiveGlobalCounters.size() == atomsSize );
    assert( atomsNegativeGlobalCounters.size() == atomsSize );
    assert( atomsPositiveOccurrences.size() == atomsSize );
    assert( atomsNegativeOccurrences.size() == atomsSize );
    
    return( atomId > 0 && atomId < atomsSize );
}

void
BerkminHeuristic::incrementHeuristicsCounters(
    AtomId atomId,
    bool isTrue )
{
    assert( checkAtomId( atomId ) );
    if( isTrue )
        incrementPositiveCounter( atomId );
    else
        incrementNegativeCounter( atomId );
}

void
BerkminHeuristic::incrementNegativeCounter( 
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsNegativeCounters[ atomId ] = atomsNegativeCounters[ atomId ] + 1;
    atomsNegativeGlobalCounters[ atomId ] = atomsNegativeGlobalCounters[ atomId ] + 1;
}

void
BerkminHeuristic::incrementPositiveCounter(
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsPositiveCounters[ atomId ] = atomsPositiveCounters[ atomId ] + 1;
    atomsPositiveGlobalCounters[ atomId ] = atomsPositiveGlobalCounters[ atomId ] + 1;
}

HEURISTIC_TYPE
BerkminHeuristic::getPositiveCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsPositiveCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminHeuristic::getTotalCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsPositiveCounters[ atomId ] + atomsNegativeCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminHeuristic::getNegativeCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsNegativeCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminHeuristic::getGlobalNegativeCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsNegativeGlobalCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminHeuristic::getGlobalPositiveCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsPositiveGlobalCounters[ atomId ];
}

void
BerkminHeuristic::divideNegativeCounter(
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsNegativeCounters[ atomId ] = atomsNegativeCounters[ atomId ] / 4;
}

void
BerkminHeuristic::dividePositiveCounter(
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsPositiveCounters[ atomId ] = atomsPositiveCounters[ atomId ] / 4;
}

bool
BerkminHeuristic::isMoreNegativeOccurrences(
    AtomId atomId )
{
    checkAtomId( atomId );
    return atomsNegativeOccurrences[ atomId ] > atomsPositiveOccurrences[ atomId ];
}

unsigned int
BerkminHeuristic::getNegativeOccurrences(
    AtomId atomId )
{
    checkAtomId( atomId );
    return atomsNegativeOccurrences[ atomId ];
}

unsigned int
BerkminHeuristic::getPositiveOccurrences(
    AtomId atomId )
{
    checkAtomId( atomId );
    return atomsPositiveOccurrences[ atomId ];
}

unsigned int
BerkminHeuristic::getOccurrences(
    AtomId atomId )
{
    checkAtomId( atomId );
    return getNegativeOccurrences( atomId ) + getPositiveOccurrences( atomId );
}

#endif /* _BERKMINHEURISTIC_H */
