/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include <vector>

#include "BerkminHeuristicMoms.h"
#include "../Interpretation.h"
#include "../Program.h"
#include "../timer.h"

#define USE_RANDOM 0
#define USE_MOSTOCCURRENCES 1
#define USE_MOMS 2

void
BerkminHeuristicMoms::makeAChoice()
{    
    startTimer( berkmin );    
    if( !pickLiteralFromLearnedConstraints() )
    {
        if( !pickLiteralWithActivity() )
        {
            doMostOccurrences();
        }
    }    
    stopTimer( berkmin );
    assert( chosenAtom != 0 );    
    assert( getInterpretation().isUndefined( chosenAtom ) );
}

unsigned int 
BerkminHeuristicMoms::calculateMoms( 
    const Literal& literal )
{    
    return estimateShortWaspRulesPropagation( literal );
}

bool
BerkminHeuristicMoms::tieBreak( 
    vector< AtomId >& maxVector )
{
    unsigned int maxVectorSize = maxVector.size();
    Program& program = getProgram();
    
    unsigned int max = 0;
    TruthValue truth = FALSE;
    
    AtomId id = 0;

    for( int j = 0; j < maxVectorSize; j++ )
    {                        
        AtomId currentAtom = maxVector[ j ];
        Atom& a = program.getAtom( currentAtom );

        Literal litNegative( &a, true );
        unsigned int momsNeg = calculateMoms( litNegative );

        Literal litPositive( &a );
        unsigned int momsPos = calculateMoms( litPositive );

        unsigned int value = ( ( momsNeg * momsPos ) << 10 ) + ( momsNeg + momsPos );

        if( value > max )
        {
            if( momsPos > momsNeg )
                truth = TRUE;
            else
                truth = FALSE;

            max = value;
            id = currentAtom;
        }                   
    }
    
    if( id != 0 )
    {
        chosenAtom = id;
        chosenPolarity = truth;
        return true;
    }
    
    return false;
}

void
BerkminHeuristicMoms::choosePolarity(
    Atom& a )
{
    Atom* atomPointer = &a;
    Literal litNegative( atomPointer, true );
    unsigned int momsNeg = calculateMoms( litNegative );

    Literal litPositive( atomPointer );
    unsigned int momsPos = calculateMoms( litPositive );

    if( momsPos > momsNeg )
        chosenPolarity = TRUE;
    else
        chosenPolarity = FALSE;
    
    countMoms++;
}

bool
BerkminHeuristicMoms::pickLiteralWithActivity()
{
    unsigned int value = higherGlobalCounterRandom();
    
    switch( value )
    {
        case USE_MOMS:
            return true;
    
        case USE_RANDOM:
            return true;
    
        case USE_MOSTOCCURRENCES:
            return false;
        
        default:
            assert( 0 );
            cerr << "Error while picking a literal." << endl;
            exit( 0 );
            return false;
    }
}

unsigned int
BerkminHeuristicMoms::higherGlobalCounterRandom()
{
    startTimer( higherGlobalCounterTimer );

    statistics( incrementHigherGlobalCounterInvocationsNumber() );
    HEURISTIC_TYPE max = 0;
    AtomId id = 0;
    
    Interpretation& interpretation = getInterpretation();

    vector< AtomId > maxVector;
    
    int i = interpretation.getLeftWatch();
    for( ; i <= interpretation.getRightWatch(); i++ )
    {
        if( interpretation.isUndefined( i ) )
        {
            if( getTotalCounter( i ) > max )
            {
                max = getTotalCounter( i );
                id = i;

                maxVector.clear();
                maxVector.push_back( i );
            }
            else if( getTotalCounter( i ) == max )
            {
                maxVector.push_back( i );
            }
        }
    }
    
    stopTimer( higherGlobalCounterTimer );

    unsigned int size = maxVector.size();

    if( id != 0 )
    {
        unsigned int idx = rand() % size;

        chosenAtom = maxVector[ idx ];

        if( size > 1 || countMoms > 5 )
        {
            chosenPolarity = FALSE;
            return USE_RANDOM;
        }
            
        return USE_MOMS;
    }
    else
    {
        assert( max == 0 );
        for( int i = 0; i < size; i++ )
        {
            AtomId atomId = maxVector[ i ];
            assert( interpretation.isUndefined( atomId ) );

            if( getOccurrences( atomId ) > max )
            {
                max = getOccurrences( atomId );
                id = atomId;
            }
        }

        assert( id != 0 );
        chosenAtom = id;        
        return USE_MOSTOCCURRENCES;
    }        
}

bool 
BerkminHeuristicMoms::onRestart( 
    unsigned int )
{
    countMoms = 0;
    return true;
}
