/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "BerkminHeuristicTwo.h"
#include "../Interpretation.h"
#include "../Program.h"
#include "../WaspRule.h"
#include "../timer.h"
#include "../Atom.h"

BerkminHeuristicTwo::BerkminHeuristicTwo( 
    ModelGenerator* mg, 
    unsigned int as ) : BerkminHeuristic( mg, as )
{
    atomsTwoNegativeValues.reserve( as );
    atomsTwoPositiveValues.reserve( as );
    
    for( int i = 0; i < as; i++ )
    {
        atomsTwoNegativeValues.push_back( 0 );
        atomsTwoPositiveValues.push_back( 0 );
    }
}

void
BerkminHeuristicTwo::choosePolarity(
    Atom& a )
{ 
    calculateTwo( a.getId() );
}

void
BerkminHeuristicTwo::calculateTwo(
    AtomId atomId )
{    
    startTimer( calculateTwo );
    assert( atomId == chosenAtom );
        
    if( reachedMaxTwoValue( atomId, false ) )
    {
        chosenPolarity = FALSE;
        stopTimer( calculateTwo );
        return;
    }
    else if( reachedMaxTwoValue( atomId, true ) )
    {
        chosenPolarity = TRUE;
        stopTimer( calculateTwo );
        return;
    }

    Interpretation& interpretation = getInterpretation();   
    for( int i = interpretation.getLeftWatch(); i <= interpretation.getRightWatch(); i++ )
    {
        if( interpretation.isUndefined( i ) )
        {            
            calculateTwoInit( i );
        }
    }
    calculateTwoFinal( atomId );
    chosenPolarity = getTwoPositive( atomId ) > getTwoNegative( atomId ) ? TRUE : FALSE;
    stopTimer( calculateTwo );
}

void
BerkminHeuristicTwo::calculateTwoFinal(
    AtomId atomId )
{   
    Atom& a = getAtom( atomId );
    computeTwoValuesFinal( atomId, a.getPositiveWaspRules(), true );
    if( getTwoPositive( atomId ) < 100 )
        computeTwoValuesFinal( atomId, a.getWatchedPositiveConstraints(), true );

    computeTwoValuesFinal( atomId, a.getNegativeBodyWaspRules(), false );
    if( getTwoNegative( atomId ) < 100 )        
        computeTwoValuesFinal( atomId, a.getNegativeHeadWaspRules(), false );
    if( getTwoNegative( atomId ) < 100 )        
        computeTwoValuesFinal( atomId, a.getWatchedNegativeConstraints(), false );
}

void 
BerkminHeuristicTwo::computeTwoValues( 
    AtomId atomId, 
    const RuleContainer< WaspRule* >& waspRules, 
    bool positive )
{
    Interpretation& interpretation = getInterpretation();
    unsigned int size = waspRules.size();
    for( unsigned int i = 0; i < size; i++ )
    {        
        if( reachedMaxTwoValue( atomId, positive ) )
            break;        
        
        assert( waspRules[ i ] != NULL );
        
        WaspRule& r = *waspRules[ i ];
        
        if( r.isUndefined() )
        {
            if( r.size() > 2 )
            {
                unsigned int count = 2;
                for( unsigned int j = 0; j < r.size(); j++ )
                {
                    if( interpretation.isUndefined( r.getLiteral( j ) ) )
                    {
                        count++;
                        break;
                    }
                }
                
                if( count == 2 )
                    incrementTwo( atomId, 1, positive );
            }
            else
            {
                incrementTwo( atomId, 1, positive );
            }
        }
    }
}

void 
BerkminHeuristicTwo::computeTwoValues( 
    AtomId atomId, 
    const WatchedList< WaspRule* >& waspRules, 
    bool positive )
{
    Interpretation& interpretation = getInterpretation();
    for( WatchedList< WaspRule* >::const_iterator it = waspRules.begin(); it != waspRules.end(); ++it )
    {        
        if( reachedMaxTwoValue( atomId, positive ) )
            break;
        
        assert( ( *it ) != NULL );
        
        WaspRule& r = *( *it );
        
        if( interpretation.isUndefined( r.getFirstWatchValue () ) && interpretation.isUndefined( r.getSecondWatchValue() ) )
        {
            if( r.size() > 2 )
            {
                unsigned int count = 2;
                for( unsigned int j = 0; j < r.size(); j++ )
                {
                    if( interpretation.isUndefined( r.getLiteral( j ) ) )
                    {
                        count++;
                        break;
                    }
                }
                
                if( count == 2 )
                    incrementTwo( atomId, 1, positive );
            }
            else
            {
                incrementTwo( atomId, 1, positive );
            }
        }
    }
}

void
BerkminHeuristicTwo::computeTwoValuesFinal(
    AtomId atomId, 
    const RuleContainer< WaspRule* >& waspRules, 
    bool isPositive )
{
    Interpretation& interpretation = getInterpretation();
    unsigned int size = waspRules.size();
    for( unsigned int i = 0; i < size; i++ )
    {
        if( reachedMaxTwoValue( atomId, isPositive ) )
            break;        
        
        assert( waspRules[ i ] != NULL );
        
        WaspRule& r = *waspRules[ i ];
        
        if( r.isUndefined() )
        {
            if( r.size() > 2 )
            {
                unsigned int count = 2;
                for( unsigned int j = 0; j < r.size(); j++ )
                {
                    if( interpretation.isUndefined( r.getLiteral( j ) ) )
                    {
                        count++;
                        break;
                    }
                }
                
                if( count == 2 )
                {
                    AtomId otherAtomId = 0;
                    bool positive = false;
                    
                    if( r.getFirstWatchValue().getAtomId() == atomId )
                    {
                        positive = r.getSecondWatchValue().isPositive();
                        otherAtomId = r.getSecondWatchValue().getAtomId();
                    }
                    else
                    {
                        assert( r.getSecondWatchValue().getAtomId() == atomId );
                        positive = r.getFirstWatchValue().isPositive();
                        otherAtomId = r.getFirstWatchValue().getAtomId();
                    }
                    
                    assert( otherAtomId != 0 );                                        
                    if( positive )
                        incrementTwo( atomId, getTwoNegative( otherAtomId ), isPositive );
                    else
                        incrementTwo( atomId, getTwoPositive( otherAtomId ), isPositive );
                }
            }
            else
            {
                AtomId otherAtomId = 0;
                bool positive = false;

                if( r.getFirstWatchValue().getAtomId() == atomId )
                {
                    positive = r.getSecondWatchValue().isPositive();
                    otherAtomId = r.getSecondWatchValue().getAtomId();
                }
                else
                {
                    assert( r.getSecondWatchValue().getAtomId() == atomId );
                    positive = r.getFirstWatchValue().isPositive();
                    otherAtomId = r.getFirstWatchValue().getAtomId();
                }

                assert( otherAtomId != 0 );                                        
                if( positive )
                    incrementTwo( atomId, getTwoNegative( otherAtomId ), isPositive );
                else
                    incrementTwo( atomId, getTwoPositive( otherAtomId ), isPositive );
            }
        }
    }
}

void
BerkminHeuristicTwo::computeTwoValuesFinal(
    AtomId atomId, 
    const WatchedList< WaspRule* >& waspRules, 
    bool isPositive )
{
    Interpretation& interpretation = getInterpretation();
    for( WatchedList< WaspRule* >::const_iterator it = waspRules.begin(); it != waspRules.end(); ++it )
    {
        if( reachedMaxTwoValue( atomId, isPositive ) )
            break;        
        
        assert( ( *it ) != NULL );
        
        WaspRule& r = *( *it );
        
        if( interpretation.isUndefined( r.getFirstWatchValue() ) && interpretation.isUndefined( r.getSecondWatchValue() ) )
        {
            if( r.size() > 2 )
            {
                unsigned int count = 2;
                for( unsigned int j = 0; j < r.size(); j++ )
                {
                    if( interpretation.isUndefined( r.getLiteral( j ) ) )
                    {
                        count++;
                        break;
                    }
                }
                
                if( count == 2 )
                {
                    AtomId otherAtomId = 0;
                    bool positive = false;
                    
                    if( r.getFirstWatchValue().getAtomId() == atomId )
                    {
                        positive = r.getSecondWatchValue().isPositive();
                        otherAtomId = r.getSecondWatchValue().getAtomId();
                    }
                    else
                    {
                        assert( r.getSecondWatchValue().getAtomId() == atomId );
                        positive = r.getFirstWatchValue().isPositive();
                        otherAtomId = r.getFirstWatchValue().getAtomId();
                    }
                    
                    assert( otherAtomId != 0 );                                        
                    if( positive )
                        incrementTwo( atomId, getTwoNegative( otherAtomId ), isPositive );
                    else
                        incrementTwo( atomId, getTwoPositive( otherAtomId ), isPositive );
                }
            }
            else
            {
                AtomId otherAtomId = 0;
                bool positive = false;

                if( r.getFirstWatchValue().getAtomId() == atomId )
                {
                    positive = r.getSecondWatchValue().isPositive();
                    otherAtomId = r.getSecondWatchValue().getAtomId();
                }
                else
                {
                    assert( r.getSecondWatchValue().getAtomId() == atomId );
                    positive = r.getFirstWatchValue().isPositive();
                    otherAtomId = r.getFirstWatchValue().getAtomId();
                }

                assert( otherAtomId != 0 );                                        
                if( positive )
                    incrementTwo( atomId, getTwoNegative( otherAtomId ), isPositive );
                else
                    incrementTwo( atomId, getTwoPositive( otherAtomId ), isPositive );
            }
        }
    }
}

void
BerkminHeuristicTwo::calculateTwoInit(
    AtomId atomId )
{
    Atom& a = getAtom( atomId );
    resetTwoCounters( atomId );

    computeTwoValues( atomId, a.getPositiveWaspRules(), true );
    if( getTwoPositive( atomId ) < 100 )
        computeTwoValues( atomId, a.getWatchedPositiveConstraints(), true );
    
    computeTwoValues( atomId, a.getNegativeBodyWaspRules(), false );
    if( getTwoNegative( atomId ) < 100 )        
        computeTwoValues( atomId, a.getNegativeHeadWaspRules(), false );
    if( getTwoNegative( atomId ) < 100 )        
        computeTwoValues( atomId, a.getWatchedNegativeConstraints(), false );        
}
