/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include <vector>

#include "BerkminRandomHeuristic.h"
#include "../Interpretation.h"
#include "../Program.h"
#include "../timer.h"

BerkminRandomHeuristic::~BerkminRandomHeuristic()
{
}

BerkminRandomHeuristic::BerkminRandomHeuristic( 
    ModelGenerator* mg, 
    unsigned int as ) : DecisionHeuristic( mg ), atomsSize( as )
{
    atomsNegativeGlobalCounters.reserve( atomsSize );
    atomsPositiveGlobalCounters.reserve( atomsSize );
    atomsNegativeCounters.reserve( atomsSize );
    atomsPositiveCounters.reserve( atomsSize );
    
    pushElement();
    
    for( int i = 1; i < atomsSize; i++ )
    {
        pushElement();
    }
}

void
BerkminRandomHeuristic::onProgramReduction()
{
}

void
BerkminRandomHeuristic::pushElement()
{
    HEURISTIC_TYPE initialValue = 0;
    atomsNegativeGlobalCounters.push_back( initialValue );
    atomsPositiveGlobalCounters.push_back( initialValue );
    atomsNegativeCounters.push_back( initialValue );
    atomsPositiveCounters.push_back( initialValue );
}

void
BerkminRandomHeuristic::makeAChoice()
{    
    startTimer( berkmin );    
    if( !pickLiteralFromLearnedConstraints() )
    {
        chosenAtom = getInterpretation().getLeftWatch();
        chosenPolarity = FALSE;
    }    
    stopTimer( berkmin );
    assert( chosenAtom != 0 );
}

bool
BerkminRandomHeuristic::pickLiteralFromLearnedConstraints()
{
    return topMostUndefinedLearnedConstraint();
}

bool
BerkminRandomHeuristic::topMostUndefinedLearnedConstraint()
{
    Interpretation& interpretation = getInterpretation();
    Program& program = getProgram();
    
    unsigned int count = 0;
    List< LearnedWaspRule* >& learnedWaspRules = program.getLearnedWaspRules();
    for( List< LearnedWaspRule* >::reverse_iterator it = learnedWaspRules.rbegin(); it != learnedWaspRules.rend(); ++it )
    {
        if( count++ > options.maxBerkminNumber )
            break;
        assert( *it != NULL );
        LearnedWaspRule& currentRule = *( *it );

        if( interpretation.isFalsePropagated( currentRule.getFirstWatchValue() ) || interpretation.isFalsePropagated( currentRule.getSecondWatchValue() ) || currentRule.getNumberOfLiteralsInTheBody() == 1 )
            continue;
        
        assert( !currentRule.isViolated() );

        AtomId atomId = 0;
        HEURISTIC_TYPE max = 0;
        
        unsigned int j = 0;
        for( ; j < currentRule.size(); j++ )
        {
            AtomId currentAtomId = currentRule.getLiteral( j ).getAtomId();
            if( interpretation.isUndefined( currentAtomId ) )
            {
                atomId = currentAtomId;
                max = getTotalCounter( currentAtomId );
                break;
            }
        }

        assert( atomId != 0 );
        
        for( ; j < currentRule.size(); j++ )
        {
            AtomId currentAtomId = currentRule.getLiteral( j ).getAtomId();
            if( interpretation.isUndefined( currentAtomId ) )
            {
                HEURISTIC_TYPE currentAtomTotalCounter = getTotalCounter( currentAtomId );
                if( currentAtomTotalCounter > max )
                {
                    max = getTotalCounter( currentAtomId );
                    atomId = currentAtomId;
                }
            }
        }

        assert( atomId != 0 );
        
        chosenAtom = atomId;
        
        if( getGlobalPositiveCounter( chosenAtom ) > getGlobalNegativeCounter( chosenAtom ) )
            chosenPolarity = TRUE;
        else
            chosenPolarity = FALSE;    

        statistics( incrementBerkminInvocationsNumber() );
        return true;
    }

    return false;
}

void
BerkminRandomHeuristic::onUIPComputationTraversal( 
    AtomId atomId, 
    bool isTrue )
{
    incrementHeuristicsCounters( atomId, isTrue );
}

void
BerkminRandomHeuristic::onLearnedWaspRule( 
    AtomId atomId, 
    bool isTrue )
{
    incrementHeuristicsCounters( atomId, isTrue );
}

void
BerkminRandomHeuristic::onReachedANumberOfConflicts()
{
    unsigned int size = getAtomsSize();
    for( unsigned int i = 1; i < size; i++ )
    {
        divideNegativeCounter( i );
        dividePositiveCounter( i );
    }
}

