/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

/* 
 * File:   BerkminRandomHeuristic.h
 * Author: Carmine Dodaro
 *
 * Created on 19 February 2013, 16.58
 */

#ifndef _BERKMINRANDOMHEURISTIC_H
#define	_BERKMINRANDOMHEURISTIC_H

#include <cassert>
#include <list>
#include "../Constants.h"
#include "../DecisionHeuristic.h"

typedef unsigned int HEURISTIC_TYPE;

class Atom;

class BerkminRandomHeuristic : public DecisionHeuristic
{
    public:
        BerkminRandomHeuristic( ModelGenerator* mg, unsigned int atomsSize );
        virtual ~BerkminRandomHeuristic();
                
        virtual void makeAChoice();
        
        virtual void onUIPComputationTraversal( AtomId atomId, bool isTrue );
        virtual void onLearnedWaspRule( AtomId atomId, bool isTrue );
        virtual void onReachedANumberOfConflicts();
        virtual void onProgramReduction();
        
    protected:
        virtual bool pickLiteralFromLearnedConstraints();
        
        bool topMostUndefinedLearnedConstraint();
        
        vector< HEURISTIC_TYPE > atomsPositiveCounters;
        vector< HEURISTIC_TYPE > atomsNegativeCounters;
        vector< HEURISTIC_TYPE > atomsPositiveGlobalCounters;
        vector< HEURISTIC_TYPE > atomsNegativeGlobalCounters;
        
        inline void divideNegativeCounter( AtomId atomId );
        
        inline void dividePositiveCounter( AtomId atomId );

        inline unsigned int getAtomsSize();
        
        inline HEURISTIC_TYPE getGlobalNegativeCounter( AtomId atomId ) const;

        inline HEURISTIC_TYPE getGlobalPositiveCounter( AtomId atomId ) const;
        
        inline HEURISTIC_TYPE getNegativeCounter( AtomId atomId ) const;
        
        inline HEURISTIC_TYPE getPositiveCounter( AtomId atomId ) const;
        
        inline HEURISTIC_TYPE getTotalCounter( AtomId atomId ) const;
        
        inline void incrementHeuristicsCounters( AtomId atomId, bool isTrue );
        
        inline void incrementNegativeCounter( AtomId atomId );

        inline void incrementPositiveCounter( AtomId atomId );
        
    private:
        unsigned int atomsSize;
        inline bool checkAtomId( AtomId atomId ) const;
        
        void pushElement();
};

unsigned int 
BerkminRandomHeuristic::getAtomsSize()
{
    return atomsSize;
}

bool 
BerkminRandomHeuristic::checkAtomId(
    AtomId atomId ) const
{
    assert( atomsPositiveCounters.size() == atomsSize );
    assert( atomsNegativeCounters.size() == atomsSize );
    assert( atomsPositiveGlobalCounters.size() == atomsSize );
    assert( atomsNegativeGlobalCounters.size() == atomsSize );
    
    return( atomId > 0 && atomId < atomsSize );
}

void
BerkminRandomHeuristic::incrementHeuristicsCounters(
    AtomId atomId,
    bool isTrue )
{
    assert( checkAtomId( atomId ) );
    if( isTrue )
        incrementPositiveCounter( atomId );
    else
        incrementNegativeCounter( atomId );
}

void
BerkminRandomHeuristic::incrementNegativeCounter( 
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsNegativeCounters[ atomId ] = atomsNegativeCounters[ atomId ] + 1;
    atomsNegativeGlobalCounters[ atomId ] = atomsNegativeGlobalCounters[ atomId ] + 1;
}

void
BerkminRandomHeuristic::incrementPositiveCounter(
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsPositiveCounters[ atomId ] = atomsPositiveCounters[ atomId ] + 1;
    atomsPositiveGlobalCounters[ atomId ] = atomsPositiveGlobalCounters[ atomId ] + 1;
}

HEURISTIC_TYPE
BerkminRandomHeuristic::getPositiveCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsPositiveCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminRandomHeuristic::getTotalCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsPositiveCounters[ atomId ] + atomsNegativeCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminRandomHeuristic::getNegativeCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsNegativeCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminRandomHeuristic::getGlobalNegativeCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsNegativeGlobalCounters[ atomId ];
}

HEURISTIC_TYPE
BerkminRandomHeuristic::getGlobalPositiveCounter(
    AtomId atomId ) const
{
    assert( checkAtomId( atomId ) );
    return atomsPositiveGlobalCounters[ atomId ];
}

void
BerkminRandomHeuristic::divideNegativeCounter(
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsNegativeCounters[ atomId ] = atomsNegativeCounters[ atomId ] / 4;
}

void
BerkminRandomHeuristic::dividePositiveCounter(
    AtomId atomId )
{
    assert( checkAtomId( atomId ) );
    atomsPositiveCounters[ atomId ] = atomsPositiveCounters[ atomId ] / 4;
}

#endif /* _BerkminRandomHeuristic_H */
