/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "PrelookaheadHeuristic.h"
#include "../Interpretation.h"
#include "../Program.h"

void
PrelookaheadHeuristic::pickLiteralWithNoActivity()
{
    doHeuristicPrelookahead();
}

bool
PrelookaheadHeuristic::onInit()
{
    trace( mg, 1, "Starting prelookahead.\n" );
    setCallComputeGUS( false );
    
    if( !lookaheadInitial() )
    {
        trace( mg, 1, "Found conflict in lookahead.\n");
        return false;
    }
    
    setCallComputeGUS( true );

    computeGUS();
    
    if( conflictLiteralIsSetted() )
        return false;
    
    stack< AtomTruthValue>& propagationStack = getPropagationStack();
    if( !propagationStack.empty() )
    {
        unsigned int size = propagationStack.size();
        if( !propagate() )
            return false;

        if( size > 100 && size > getProgram().getNumberOfAtoms() * 9/ 10 )
        {
            setCallComputeGUS( false );
            if( !lookaheadInitial() )
                return false;
            setCallComputeGUS( true );
            computeGUS();
            if( conflictLiteralIsSetted() )
                return false;
            
            if( !propagationStack.empty() && !propagate() )
            {
                trace( mg, 1, "Found conflict in lookahead after GUS step.\n");
                return false;
            }
        }
    }
    
    trace( mg, 1, "Finished lookahead.\n");
    return true;
}

/*
 * Returns 0 for inconsistence.
 * Returns -1 if derived something.
 * Returns 1 if derived nothing.
 */
int
PrelookaheadHeuristic::performLookahead(
    AtomId atomId,
    Program& program,    
    stack< AtomTruthValue >& propagationStack,
    Stack< unsigned int >& backtrackingStack )
{
    Atom& a = program.getAtom( atomId );
    trace( mg, 1, "Performing atom %s in look-ahead pre-computation.\n", a.getAtomName().c_str() );

    TruthValue atomTruthValue = isMoreNegativeOccurrences( atomId ) ? TRUE : FALSE;
    TruthValue otherAtomTruthValue = isMoreNegativeOccurrences( atomId ) ? FALSE : TRUE;

    AtomTruthValue literal;
    literal.first = atomId;
    literal.second = atomTruthValue;

    nextLevel();

    backtrackingStack.push( &( a.getChoiceAddress() ) );
    a.setChoice();
    propagationStack.push( literal );
    if( !propagate() )
    {
        unrollOne();

        if( otherAtomTruthValue == TRUE )
            pushAtomInStack( a, TRUE );
        else
            pushAtomInStack( a, FALSE );
    }
    else
    {
        unsigned int diff = backtrackingStack.size();
        unrollOne();
        nextLevel();

        backtrackingStack.push( &( a.getChoiceAddress() ) );
        a.setChoice();

        literal.second = otherAtomTruthValue;
        propagationStack.push( literal );

        if( !propagate() )
        {
            unrollOne();

            if( atomTruthValue == TRUE )
                pushAtomInStack( a, TRUE );
            else
                pushAtomInStack( a, FALSE );
        }
        else
        {
            if( diff >= backtrackingStack.size() )
            {                            
                literal.second = atomTruthValue;

            }
            preferredAtoms.push_back( pair< AtomTruthValue, unsigned int >( literal, diff + backtrackingStack.size() ) );
            unrollOne();
        }
    }
    
    if( !propagationStack.empty() )
    {
        if( !propagate() )
            return 0;
        else
            return -1;
    }
    
    return 1;
}

bool
PrelookaheadHeuristic::lookaheadInitial()
{    
    bool derivedSomething;
    unsigned int count = 0;

    Interpretation& interpretation = getInterpretation();
    Program& program = getProgram();
    
    stack< AtomTruthValue>& propagationStack = getPropagationStack();
    Stack< unsigned int >& backtrackingStack = getBacktrackingStack();
    
    unsigned int lastIndex = interpretation.getRightWatch();
    unsigned int last;
    do
    {
        trace( mg, 2, "Prelookahead step %d.\n", count );
        derivedSomething = false;
        
        last = lastIndex;
        for( unsigned int i = interpretation.getLeftWatch(); i <= interpretation.getRightWatch() && i <= last; i++ )
        {
            AtomId atomId = i;
            if( interpretation.isUndefined( atomId ) )
            {
                trace( mg, 2, "Processing atom %s.\n", getAtom( atomId ).getAtomName().c_str() );
                int exitValue = performLookahead( atomId, program, propagationStack, backtrackingStack );                
                if( exitValue == 0 )
                {
                    trace( mg, 2, "Conflict during propagation.\n" );
                    return false;
                }
                else if( exitValue == -1 )
                {
                    trace( mg, 2, "Derived deterministic consequence.\n" );
                    derivedSomething = true;
                    lastIndex = i;
                }
            }
        }
    } while( derivedSomething && ++count < options.depthlookahead );
    
    return true;
}

void
PrelookaheadHeuristic::doHeuristicPrelookahead()
{
    unsigned int max = 0;
    unsigned int id = 0;

    Interpretation& interpretation = getInterpretation();
    for( unsigned int i = 0; i < preferredAtoms.size(); i++ )
    {
        if( interpretation.isUndefined( preferredAtoms[ i ].first.first ) && preferredAtoms[ i ].second + 1 > max )
        {
            max = preferredAtoms[ i ].second + 1;
            id = i;
        }
    }
    if( id == 0 )
    {   
        assert( chosenAtom != 0 );
        chosenPolarity = isMoreNegativeOccurrences( chosenAtom ) ? FALSE : TRUE;
    }
    else
    {
        chosenAtom = preferredAtoms[ id ].first.first;
        chosenPolarity = preferredAtoms[ id ].first.second;
    }
}