/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "PrelookaheadHeuristicLight.h"
#include "../Interpretation.h"
#include "../Program.h"
#include <algorithm>
#include <queue>

bool compareOccurrences( const pair< AtomId, unsigned int >& i, const pair< AtomId, unsigned int >& j )
{
    return ( i.second < j.second );
}

PrelookaheadHeuristicLight::PrelookaheadHeuristicLight(
    ModelGenerator* mg,
    unsigned int atomsSize ) : PrelookaheadHeuristic( mg, atomsSize )
{    
}

bool
PrelookaheadHeuristicLight::onInit()
{
    Interpretation& interpretation = getInterpretation();
    for( unsigned int i = interpretation.getLeftWatch(); i <= interpretation.getRightWatch(); i++ )
    {        
        if( interpretation.isUndefined( i ) )
        {
            Atom& atom = getAtom( i );
            unsigned int occurrences = getOccurrences( i );
            if( occurrences > 30 || atom.getBinaryNegativeWaspRules().size() > 0 || atom.getBinaryPositiveWaspRules().size() > 0 )
                atomsOrderedByMO.push_back( pair< AtomId, unsigned int >( i, occurrences ) );
        }
    }
    
    atomsOrderedByMO.sort( compareOccurrences );
    
    return PrelookaheadHeuristic::onInit();
}

bool
PrelookaheadHeuristicLight::lookaheadInitial()
{    
    bool derivedSomething;
    unsigned int count = 0;

    Interpretation& interpretation = getInterpretation();
    Program& program = getProgram();
    
    stack< AtomTruthValue>& propagationStack = getPropagationStack();
    Stack< unsigned int >& backtrackingStack = getBacktrackingStack();
    
    do
    {
        trace( mg, 2, "Prelookahead step %d.\n", count );
        derivedSomething = false;

        for( list< pair< AtomId, unsigned int > >::iterator i = atomsOrderedByMO.begin(); i != atomsOrderedByMO.end() ; ++i )
        {
            AtomId atomId = i->first;
            if( interpretation.isUndefined( atomId ) )
            {
                trace( mg, 2, "Processing atom %s.\n", getAtom( atomId ).getAtomName().c_str() );
                int exitValue = performLookahead( atomId, program, propagationStack, backtrackingStack );                
                if( exitValue == 0 )
                {
                    trace( mg, 2, "Conflict during propagation." );
                    return false;
                }
                else if( exitValue == -1 )
                {
                    trace( mg, 2, "Derived deterministic consequence." );
                    derivedSomething = true;
                }
            }
        }
    } while( derivedSomething && count < options.depthlookahead );
    
    return true;
}