/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#include "options.h"
#include "Constants.h"
#include "Help.h"

#include <getopt.h>
#include <cstdlib>
#include <boost/algorithm/string.hpp>

namespace wasp
{

struct Options options;

// Options with a short alias must use ascii codes.
// Here are all the short/long options ordered by code.
#define OPTIONID_help 'h'
#define OPTIONID_models 'n'

// Options without a short alias must use non-ascii codes.
// Here are all the "only-long" options ordered by code.
#define OPTIONID_trace_std ( 'z' + 1 )
#define OPTIONID_trace_parser ( 'z' + 2 )
#define OPTIONID_trace_checker ( 'z' + 3 )
#define OPTIONID_stdin ( 'z' + 4 )
#define OPTIONID_trace_mg ( 'z' + 5 )
#define OPTIONID_trace_us ( 'z' + 7 )
#define OPTIONID_polarity ( 'z' + 8 )
#define OPTIONID_det ( 'z' + 9 )
#define OPTIONID_wf ( 'z' + 10 )
#define OPTIONID_silent ( 'z' + 11 )
#define OPTIONID_competition_output ( 'z' + 12 )
#define OPTIONID_partial_mc_off ( 'z' + 13 )
#define OPTIONID_max_lc_size ( 'z' + 14 )
#define OPTIONID_self_subsumption ( 'z' + 15 )
#define OPTIONID_no_subsumption ( 'z' + 16 )
#define OPTIONID_nprint ( 'z' + 17 )
#define OPTIONID_toprint ( 'z' + 18 )
#define OPTIONID_stats ( 'z' + 19 )
#define OPTIONID_stats_exit ( 'z' + 20 )
#define OPTIONID_delete ( 'z' + 21 )
#define OPTIONID_simplifyprogram ( 'z' + 22 )
#define OPTIONID_most_occurrences_for_aggregates ( 'z' + 23 )
#define OPTIONID_fuheuristic ( 'z' + 24 )
#define OPTIONID_berkmintwo ( 'z' + 25 )
#define OPTIONID_berkmintwolight ( 'z' + 26 )
#define OPTIONID_berkminheuristic ( 'z' + 27 )
#define OPTIONID_berkmin_moms ( 'z' + 28 )
#define OPTIONID_lookaheadheuristic ( 'z' + 29 )
#define OPTIONID_lookaheadberkminheuristic ( 'z' + 30 )
#define OPTIONID_prelookahead ( 'z' + 31 )
#define OPTIONID_prelookaheadlight ( 'z' + 32 )
#define OPTIONID_printprogram ( 'z' + 33 )
#define OPTIONID_geometric_restarts ( 'z' + 34 )
#define OPTIONID_restarts_numbers ( 'z' + 35 )
#define OPTIONID_randomheuristic ( 'z' + 36 )
#define OPTIONID_mixedheuristic ( 'z' + 37 )
#define OPTIONID_mixedheuristicprelookahead ( 'z' + 38 )
#define OPTIONID_mixedheuristicrestartlookahead ( 'z' + 39 )
#define OPTIONID_balancedMO ( 'z' + 40 )
#define OPTIONID_gringo ( 'z' + 41 )
#define OPTIONID_berkminrandomheuristic ( 'z' + 42 )
#define OPTIONID_berkminlimitedactivity ( 'z' + 43 )
#define OPTIONID_third_competition_output ( 'z' + 44 )

void
Options::parse(
    int argc,
    char* const* argv )
{
    // It will store the option code.
    int code;

    polarity = 0;
    numModels = MAXVALUE;
    silentMode = false;
    onlyDeterministicConsequences = false;
    onlyWellFounded = false;
    stats = false;
    statsExit = false;
//    heuristic = MIXED_HEURISTIC_PRELOOKAHEAD;
    heuristic = BERKMIN_HEURISTIC;
    deleteMode = DELETE_WHEN_LEARN_A_NEW_CONSTRAINT;
    printProgram = false;
    partialMCoff = false;
    maxLearnedConstraintSize = MAXVALUE;
    statsTimer = MAXVALUE;
    competitionOutput = false;
    thirdCompetitionOutput = false;
    depthlookahead = MAXVALUE;
    doSubsumption = false;
    restartsNumbers = 32;    
    mostOccurrencesForAggregates = false;
    geometricRestarts = false;
    modelSeparator = "\n";
    randomHeuristicSeed = 37;
    prelookaheadRestartsNumber = 40;
    momsHeuristicSeed = 1231023;
    gringoNumericFormat = false;
    simplifyProgram = false;
    maxBerkminNumber = MAXVALUE;
    
    do{
        // For each option there is an instance of option.
        // Each option specifies
        // - the long name;
        // - if it requires an argument (required_argument), accept an argument
        //   (optional_argument), or does not allow arguments (no_argument);
        // - the address of an integer flag to be setted followed by the value
        //   to use, or NULL followed by the code of the option.
        //
        // Here are all the long options ordered by long name.
        static struct option longOptions[] =
            {
                { "help", no_argument, NULL, OPTIONID_help },
                { "trace-parser", required_argument, NULL, OPTIONID_trace_parser },
                { "trace-std", required_argument, NULL, OPTIONID_trace_std },
                { "trace-mg", required_argument, NULL, OPTIONID_trace_mg },
                { "trace-mc", required_argument, NULL, OPTIONID_trace_checker },
                { "trace-us", required_argument, NULL, OPTIONID_trace_us },
                { "stdin", no_argument, NULL, OPTIONID_stdin },
                { "det", no_argument, NULL, OPTIONID_det },
                { "wf", no_argument, NULL, OPTIONID_wf },
                { "polarity", required_argument, NULL, OPTIONID_polarity},                
                { "silent", no_argument, NULL, OPTIONID_silent },
                { "stats", optional_argument, NULL, OPTIONID_stats },
                { "stats-exit", optional_argument, NULL, OPTIONID_stats_exit },                
                { "delete", required_argument, NULL, OPTIONID_delete },
                { "heuristic-lookahead", no_argument, NULL, OPTIONID_lookaheadheuristic },
                { "heuristic-lookahead-berkmin", no_argument, NULL, OPTIONID_lookaheadberkminheuristic },
                { "heuristic-berkmin", optional_argument, NULL, OPTIONID_berkminheuristic },
                { "heuristic-berkmintwo", optional_argument, NULL, OPTIONID_berkmintwo },
                { "heuristic-berkmintwolight", optional_argument, NULL, OPTIONID_berkmintwolight },                                
                { "heuristic-berkminmoms", optional_argument, NULL, OPTIONID_berkmin_moms },
                { "heuristic-berkminbalancedmo", optional_argument, NULL, OPTIONID_balancedMO },
                { "heuristic-firstundefined", no_argument, NULL, OPTIONID_fuheuristic },
                { "heuristic-random", optional_argument, NULL, OPTIONID_randomheuristic },
                { "heuristic-prelookahead", optional_argument, NULL, OPTIONID_prelookahead },
                { "heuristic-prelookaheadlight", optional_argument, NULL, OPTIONID_prelookaheadlight },
                { "heuristic-mixed", no_argument, NULL, OPTIONID_mixedheuristic },
                { "heuristic-mixedprelookahead", no_argument, NULL, OPTIONID_mixedheuristicprelookahead },
                { "heuristic-mixedrestartlookahead", optional_argument, NULL, OPTIONID_mixedheuristicrestartlookahead },
                { "heuristic-berkminrandom", optional_argument, NULL, OPTIONID_berkminrandomheuristic },
                { "heuristic-berkminlimitedactivity", optional_argument, NULL, OPTIONID_berkminlimitedactivity },
                { "printprogram", no_argument, NULL, OPTIONID_printprogram },
                { "neg-filter", required_argument, NULL, OPTIONID_nprint },
                { "filter", required_argument, NULL, OPTIONID_toprint },
                { "partial-mc-off", no_argument, NULL, OPTIONID_partial_mc_off },
                { "max-learnedconstraints-size", required_argument, NULL, OPTIONID_max_lc_size },
                { "competition-output", no_argument, NULL, OPTIONID_competition_output },
                { "third-competition-output", no_argument, NULL, OPTIONID_third_competition_output }, 
                { "subsumption", no_argument, NULL, OPTIONID_self_subsumption },
                { "no-subsumption", no_argument, NULL, OPTIONID_no_subsumption },
				{ "restarts", required_argument, NULL, OPTIONID_restarts_numbers },
                { "geometric-restarts", optional_argument, NULL, OPTIONID_geometric_restarts },
                { "no-heuristic-aggregates", no_argument, NULL, OPTIONID_most_occurrences_for_aggregates },
                { "gringo", no_argument, NULL, OPTIONID_gringo },
                { "simplify-program", no_argument, NULL, OPTIONID_simplifyprogram },

                // The NULL-option indicates the end of the array.
                { NULL, 0, NULL, 0 }
            };

        // The function getopt_long() stores the option index here.
        int optionIndex = 0;

        // The third argument specifies all the short options.
        code = getopt_long( argc, argv, "hn:01", longOptions, &optionIndex );

        switch( code )
        {
        case -1:
            // All the command line was parsed.
            break;

        case OPTIONID_help:
            Help::printHelp();
            exit( 0 );
            break;

        case OPTIONID_models:
            numModels = atoi( optarg );
            if( numModels == 0 )
                numModels = MAXVALUE;
            break;

        case OPTIONID_trace_parser:
            setTraceLevel( parser, atoi( optarg ) );
            break;

        case OPTIONID_trace_std:
            setTraceLevel( std, atoi( optarg ) );
            break;

        case OPTIONID_trace_checker:
            setTraceLevel( mc, atoi( optarg ) );
            break;

        case OPTIONID_trace_mg:
            setTraceLevel( mg, atoi( optarg ) );
            break;
            
        case OPTIONID_trace_us:
            setTraceLevel( us, atoi( optarg ) );
            break;

        case OPTIONID_det:
            onlyDeterministicConsequences = true;
            break;

        case OPTIONID_wf:
            onlyWellFounded = true;
            break;
        case OPTIONID_polarity:
            polarity = atoi( optarg );
            break;            

        case OPTIONID_silent:
            silentMode = true;
            modelSeparator = "";
            break;

        case OPTIONID_stats:
            stats = true;
            if( optarg )
                statsTimer = atoi( optarg );
            break;
            
        case OPTIONID_stats_exit:
            statsExit = true;
            if( optarg )
                statsTimer = atoi( optarg );
            break;

        case OPTIONID_fuheuristic:
            heuristic = FIRST_UNDEFINED_HEURISTIC;
            break;
           
        case OPTIONID_delete:
            if( optarg )
                deleteMode = atoi( optarg );
            
            if( deleteMode != DELETE_WHEN_MODEL_IS_FOUND && 
                deleteMode != NO_DELETE && 
                deleteMode != RESTART_BASED_POLICY )
            {
                deleteMode = DELETE_WHEN_LEARN_A_NEW_CONSTRAINT;
            }
            break;
        
        case OPTIONID_lookaheadheuristic:
            heuristic = LOOKAHEAD_HEURISTIC;
            break;

        case OPTIONID_lookaheadberkminheuristic:
            heuristic = LOOKAHEADBERKMIN_HEURISTIC;
            break;

        case OPTIONID_berkminheuristic:
            if( optarg )
                maxBerkminNumber = atoi( optarg );
            heuristic = BERKMIN_HEURISTIC;
            break;
            
        case OPTIONID_berkminlimitedactivity:
            if( optarg )
                maxBerkminNumber = atoi( optarg );
            heuristic = BERKMIN_LIMITED_ACTIVITY;
            break;
            
        case OPTIONID_berkmintwo:
            if( optarg )
                maxBerkminNumber = atoi( optarg );
            heuristic = BERKMIN_TWO;
            break;
            
        case OPTIONID_berkmintwolight:
            if( optarg )
                maxBerkminNumber = atoi( optarg );
            heuristic = BERKMIN_TWOLIGHT;
            break;
                       
        case OPTIONID_berkmin_moms:            
            heuristic = BERKMIN_MOMS;
            if( optarg )
                momsHeuristicSeed = atoi( optarg );
            break;
                       
        case OPTIONID_berkminrandomheuristic:
            if( optarg )
                maxBerkminNumber = atoi( optarg );
            heuristic = BERKMIN_RANDOM;
            break;
            
        case OPTIONID_randomheuristic:
            heuristic = RANDOM_HEURISTIC;
            if( optarg )
                randomHeuristicSeed = atoi( optarg );
            break;
            
        case OPTIONID_prelookahead:
            heuristic = PRELOOKAHEAD_HEURISTIC;
            if( optarg )
                depthlookahead = atoi( optarg );            
            break;
            
        case OPTIONID_prelookaheadlight:
            heuristic = PRELOOKAHEADLIGHT_HEURISTIC;
            if( optarg )
                depthlookahead = atoi( optarg );            
            break;            
            
        case OPTIONID_mixedheuristic:            
            heuristic = MIXED_HEURISTIC;
            break;
            
        case OPTIONID_mixedheuristicprelookahead:
            heuristic = MIXED_HEURISTIC_PRELOOKAHEAD;            
            break;
            
        case OPTIONID_mixedheuristicrestartlookahead:            
            heuristic = MIXED_HEURISTIC_RESTARTLOOKAHEAD;
            if( optarg )
                prelookaheadRestartsNumber = atoi( optarg );
            break;
            
        case OPTIONID_balancedMO:
            heuristic = BALANCED_MO;
            break;
        
        case OPTIONID_printprogram:
            printProgram = true;
            break;
            

        case OPTIONID_nprint:
        {
            vector< string > splittedString;
            boost::split( splittedString, optarg, boost::is_any_of(",") );
            for( int i = 0; i < splittedString.size(); i++ )
                doNotPrint.insert( splittedString[ i ] );
        }
            break;

        case OPTIONID_toprint:
        {
            vector< string > splittedString;
            boost::split( splittedString, optarg, boost::is_any_of(",") );
            for( int i = 0; i < splittedString.size(); i++ )
                toPrint.insert( splittedString[ i ] );
        }
            break;

        case OPTIONID_partial_mc_off:
            partialMCoff = true;
            break;

        case OPTIONID_max_lc_size:
            maxLearnedConstraintSize = atoi( optarg );
            if( maxLearnedConstraintSize < 32 )
                maxLearnedConstraintSize = 32;
            break;

        case OPTIONID_competition_output:
            competitionOutput = true;
            silentMode = true;
            modelSeparator = "";
            numModels = 1;
            break;
            
        case OPTIONID_third_competition_output:
            thirdCompetitionOutput = true;
            silentMode = true;
            modelSeparator = "";
            numModels = 1;
            break;

        case OPTIONID_self_subsumption:
            doSubsumption = true;
            break;

        case OPTIONID_no_subsumption:
            doSubsumption = false;
            break;

        case OPTIONID_stdin:
            inputFiles.push_back( "--" );            
            break;

		case OPTIONID_restarts_numbers:
            if( optarg )
                restartsNumbers = atoi( optarg );

            if( restartsNumbers < 32 )
                restartsNumbers = 32;
            break;
            
        case OPTIONID_geometric_restarts:
            geometricRestarts = true;
            if( optarg )                
                restartsNumbers = atoi( optarg );
            
            if( restartsNumbers < 100 )
                restartsNumbers = 100;
            break;
        
        case OPTIONID_most_occurrences_for_aggregates:
            mostOccurrencesForAggregates = false;
            break;
            
        case OPTIONID_gringo:
            gringoNumericFormat = true;
            break;
            
        case OPTIONID_simplifyprogram:
            simplifyProgram = true;
            break;
            
        default:
            // TODO: Exit with error.
            cerr << "error" << endl;
            break;
        }
    } while( code != -1 );

    // TODO: Manage the reminder options (probably, they are all input
    //       files).
    for( int i = optind; i < argc; i++ )
    {
        inputFiles.push_back( argv[ i ] );
    }
}

};
