/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#ifndef STACK_H
#define STACK_H

#include <vector>
#include <cassert>
#include <ostream>
#include "trace.h"
#include "options.h"

using namespace std;
using namespace wasp;

template< class T >
class Stack
//
// This structure is used by our linear time propagation machinery.
//
// Whenever we change one of those many counters there, we save its
// address and its old value on this stack so that we can easily
// restore the counter during backtracking.
//
{
    template< class U >
    friend ostream& operator<<( ostream&, const Stack< U >& );

private:
    vector< pair< T*, T > > data;
    vector< unsigned int > levels;
    
    // disabled methods
    inline Stack( const Stack& );
    inline void operator=( const Stack& );

public:
    inline Stack();
    
    inline unsigned int getCurrentLevel() const;

    inline unsigned int size() const;
    inline bool empty() const;
    inline void clear();
    
    inline void push( T* p );
    inline void nextLevel();
    void unroll( const unsigned level );
    inline void unrollOne();
    };

template< class T >
Stack< T >::Stack( 
    const Stack< T >& )
{
    assert( 0 );
}

template< class T >
void
Stack< T >::operator=( 
    const Stack< T >& )
{
    assert( 0 );
}

template< class T >
Stack< T >::Stack()
    : data(), levels()
{
    nextLevel();
}

template< class T >
unsigned int
Stack< T >::getCurrentLevel() const
{
    assert( ! levels.empty() );
    
    return levels.size() - 1;
}

template< class T >
unsigned int
Stack< T >::size() const
{
    return data.size();
}

template< class T >
bool
Stack< T >::empty() const
{
    return data.empty();
}

template< class T >
void
Stack< T >::clear()
{
    data.clear();
    levels.clear();
}

template< class T >
void
Stack< T >::push(
    T* value )
{
    trace( mg, 4, "Pushing element (%p,%u)\n", value, static_cast<unsigned int>(*value) );

    assert( ! levels.empty() );
    
    data.push_back( pair< T*, T >( value, *value ) );
}

template< class T >
void
Stack< T >::nextLevel()
{
    trace( mg, 4, "Pushing level %u starting at position %u.\n", static_cast<unsigned int>(levels.size()), static_cast<unsigned int>(data.size()) );
    levels.push_back( data.size() );
}

template< class T >
void
Stack< T >::unroll(
    const unsigned level )
{
    trace( mg, 4, "Unrolling level(s) from %u to %u.\n", static_cast<unsigned int>(getCurrentLevel()), level );

    // Do nothing if we already are at the desired level or even below.
    // (This is the case when a user of STACK operates at higher levels
    // without pushing anything on the STACK.)
    if( level >= levels.size() )
        return;

    assert( levels[ level ] <= data.size() );
    
    // Restore stored values.
    for( unsigned int n = data.size(); n > levels[level]; )
    {
        *( data[--n].first ) = data[n].second;
        //*( data.back().first ) = data.back().second;
        //data.pop_back();
    }
    data.resize(levels[level]);
    
    // Remove restored levels.
    for( unsigned int n = levels.size() - level; n > 0; --n )
        levels.pop_back();
}

template< class T >
void
Stack< T >::unrollOne()
{
    unroll( getCurrentLevel() );
}

template<class T>
ostream& 
operator<<(
    ostream& out, 
    const Stack< T >& s )
{
    out << s.data.size() << " element(s) starting at " << &s << ": ";

    for( unsigned int i = 0, j = 0; i < s.levels.size(); i++ )
    {
        out << "[level " << i << "] ";
        unsigned int last = ( i == s.levels.size() - 1 ? s.data.size() : s.levels[ i+1 ] );
        for( ; j < last; j++ )
            out << "[0x" << s.data[ j ].first << "," << s.data[ j ].second << "] ";
    }

    return out;
}

#endif
