/*
*
*  Copyright 2013 Mario Alviano, Carmine Dodaro, Wolfgang Faber, Nicola Leone, Francesco Ricca, and Marco Sirianni.
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*
*/

#ifndef TIMER_H
#define TIMER_H

/*
 * The timer macros are expanded only if TIMER_OFF is not defined.
 * The timer structures are defined only if TIMER_OFF is not defined.
 */
#ifdef TIMER_OFF
#   define startTimer( id )
#   define stopTimer( id )
#   define resetTimer( id )
#   define printTimer( ostream, prefix, id, suffix )
namespace wasp {
}
#else
#   define startTimer( id ) \
        waspTimers.id.start()
#   define stopTimer( id ) \
        waspTimers.id.stop()
#   define resetTimer( id ) \
        waspTimers.id.reset()
#   define printTimer( ostream, prefix, id, suffix ) \
        ostream << prefix << waspTimers.id << suffix


#include <cassert>
#include <ctime>
#include <iostream>

using namespace std;

namespace wasp
{

/**
 * A class for measuring the elapsed time.
 *
 * All the timer in wasp must be instantiated as fields of the Timers structure.
 *
 * Since the Timer class is not present if the macro TIMER_OFF is defined,
 * direct method invocation must be avoided; to use the corresponding macro is
 * the right way.
 */
class Timer
{
    friend struct Timers;
    friend ostream& operator<<( ostream& out, const Timer& timer );

public:
    /**
     * Call this method to start the timer.
     *
     * Note that the timer must be stopped.
     */
    void start();

    /**
     * Call this method to stop (or pause) the timer.
     *
     * Note that the timer must be started.
     */
    void stop();

    /**
     * Call this method to reset the timer; in other words, this method set to
     * zero the elapsed time.
     *
     * Note that the timer must be stopped.
     */
    void reset();

    /**
     * This method return the elapsed time in seconds.
     *
     * Note that the timer must be stopped.
     *
     * @return The elapsed time expressed in seconds.
     */
    double getElapsedTime() const;

    /**
     * Check if the timer is stopped.
     *
     * @return True if the timer is stopped, false otherwise.
     */
    bool isStopped() const;

private:
    /**
     * Construct a timer. Initially, the elapsed time is zero and the timer is
     * stopped.
     */
    Timer();

    /**
     * It stores the elapsed time as number of cpu clocks.
     */
    clock_t elapsedTime;

    /**
     * It stores the number of cpu clocks when the timer was started.
     * If the timer is stopped, then it stores -1.
     */
    clock_t startTime;
};


/**
 * In this structure are defined all the timer of wasp.
 *
 * Since the Timers structure is not present if the macro TIMER_OFF is defined,
 * direct access to it must be avoided; to use the corresponding macro is the
 * right way.
 */
struct Timers
{
    /* Program Timers */
    Timer main;
    Timer startup;
    
    /* Heuristics Timers */
    Timer berkmin;
    Timer calculateTwo;
    Timer lookahead;
    Timer higherGlobalCounterTimer;
    Timer divideCountersTimer;
    Timer pt;
    Timer mo;
    Timer chosen;
    Timer computeUTs;
    Timer look2;
    Timer isPTFunction;    
    
    /* Learning Timers */
    Timer calculateFirstUip;
    Timer learnRule;
    Timer deleteRules;    
    
    /* Propagation Timers */
    Timer initialPropagation;    
    Timer propagation;
    Timer scanNormalWaspRulesTimer;
    Timer scanLearnedWaspRulesTimer;
    Timer unitPropagationTimer;
    Timer supportControlTimer;
    
    /* GUS Timers */
    Timer computeGUS;
    Timer computeGUSInitial;
    Timer computeGUSCentral;
    Timer computeGUSFinal;
    
    /* Utility functions Timers */
    Timer selfsubsumption;
    Timer subsumption;
    Timer updateInterpretationWatches;
    Timer updateLeftWatchWaspRule;
    Timer updateRightWatchWaspRule;
    Timer unroll;
    Timer unrollOne;
    
    /* Model Checker */
    Timer modelChecker;
    Timer polynomialOperator;
    Timer satSolverInvocation;
    Timer buildFormula;
};

extern struct Timers waspTimers;

};

#endif

#endif
